// -*- c -*- 
// (setq c-basic-offset 4)
// "clsConfItem.h"

#include "clsConfItem.h"

#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#define ENDL wxT("\n")

#ifndef _ASSERT
#define _ASSERT assert
#endif

#ifdef WIN32
#define ENDLWX wxT("\r\n")
#else
#define ENDLWX wxT("\n")
#endif

ConfItem_cls::
ConfItem_cls(){
    msHeader = L"NONE";
    msBody   = L"NONE";
}

void ConfItem_cls::
ReplaceSeparator(){
    for (unsigned int i = 0; i < msBody.size(); i++){
	if (msBody[i] == '\\') msBody[i] = '/';
	if (msBody[i] == ' '){
	    msBody.insert(i, L"\\"); 
	    i++; }
    }
}

bool ConfItem_cls::
ReadStrLineBase(std::wstring strLine){
    std::wstring::size_type loc = strLine.find(msHeader, 0);
    if (loc != std::string::npos){
	msBody = wxConvFile.cWC2WX(strLine.c_str() + loc + wcslen(msHeader.c_str()));
	return TRUE;
    }
    return FALSE;
}

//----------------------------------

unsigned int ConfArray_cls::
CountDimension(wxString strBody){
    unsigned int dimension = 0;
    std::wstring::size_type pos = 0;
    while (pos != std::wstring::npos){
	pos = strBody.find(L" ", pos + 1);
	dimension++;
    }
    _ASSERT(dimension < _CONFARRAY_MAX);
    return dimension;
}

//----------------------------------

ConfItemSTR_cls::
ConfItemSTR_cls(){
}

ConfItemSTR_cls::
ConfItemSTR_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemSTR_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
}

bool ConfItemSTR_cls::
ReadStrLine(std::wstring strLine){
    return ReadStrLineBase(strLine);
}

wxString ConfItemSTR_cls::
Serialize(){
    if (msBody == wxT("!!SKIP!!")) return wxT("");
    return msHeader + msBody + ENDLWX;
}

wxString ConfItemSTR_cls::
GetItemStr(){
    return msBody;
}

//------------------------------------

ConfItemINT_cls::
ConfItemINT_cls():
ConfItem_cls(){}

ConfItemINT_cls::
ConfItemINT_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemINT_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
#ifdef WIN32
    mBody = _wtoi(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
}

bool ConfItemINT_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtoi(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemINT_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mBody << ENDLWX;
    return sOut;
}

wxString ConfItemINT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// -------------

void ConfItemINTARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;   
    msBody = strBody;
    mDimension = CountDimension(strBody);
    std::wstringstream ssTmp;
    ssTmp << msHeader.c_str() << strBody.c_str();
    ReadStrLine(ssTmp.str());
}

bool ConfItemINTARRAY_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(msBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemINTARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDLWX;
    return sOut;
}

wxString ConfItemINTARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++) 
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

//------------------------------------

ConfItemUINT_cls::
ConfItemUINT_cls(){
}

ConfItemUINT_cls::
ConfItemUINT_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemUINT_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody   = strBody;
#ifdef WIN32
    mBody = _wtoi(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
}

bool ConfItemUINT_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtoi(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemUINT_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mBody << ENDLWX;
    return sOut;
}

wxString ConfItemUINT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

//-------------------------------------

void ConfItemUINTARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
    mDimension = CountDimension(strBody);
    std::wstringstream ssTmp;
    ssTmp << msHeader.c_str() << strBody.c_str();
    ReadStrLine(ssTmp.str());
}

bool ConfItemUINTARRAY_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(msBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemUINTARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDLWX;
    return sOut;
}

wxString ConfItemUINTARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

//-------------------------------------

ConfItemFLOAT_cls::
ConfItemFLOAT_cls():
ConfItem_cls(){}

ConfItemFLOAT_cls::
ConfItemFLOAT_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemFLOAT_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody   = strBody;
#ifdef WIN32
    mBody = _wtof(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
}

bool ConfItemFLOAT_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtof(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemFLOAT_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mBody << ENDLWX;
    return sOut;
}

wxString ConfItemFLOAT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// ------------------------------------

ConfItemFLOATARRAY_cls::
ConfItemFLOATARRAY_cls(){
}

ConfItemFLOATARRAY_cls::
ConfItemFLOATARRAY_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemFLOATARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
    mDimension = CountDimension(strBody);
    std::wstringstream ssTmp;
    ssTmp << msHeader.c_str() << strBody.c_str();
    ReadStrLine(ssTmp.str());
}

bool ConfItemFLOATARRAY_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(msBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemFLOATARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDLWX;
    return sOut;
}

wxString ConfItemFLOATARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

// ------------------------------------

void ConfItemDOUBLE_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody   = strBody;
#ifdef WIN32
    mBody = _wtof(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
}

bool ConfItemDOUBLE_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtof(msBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemDOUBLE_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mBody << ENDLWX;
    return sOut;
}

wxString ConfItemDOUBLE_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// ------------------------------------

ConfItemBOOL_cls::
ConfItemBOOL_cls():
ConfItem_cls(){}

ConfItemBOOL_cls::
ConfItemBOOL_cls(wxString strHeader, wxString strBody){
    SetItem(strHeader, strBody);
}

void ConfItemBOOL_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
#ifdef WIN32
    mbBody = ( _wtoi(msBody.c_str()) != 0);
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mbBody = ( atoi(wxBody.ToAscii()) != 0);
#endif
}

bool ConfItemBOOL_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mbBody = ( _wtoi(msBody.c_str()) != 0);
#else
    wxString wxBody = wxConvFile.cWC2WX(msBody.c_str());
    mbBody = ( atoi(wxBody.ToAscii()) != 0);
#endif
    return TRUE;
}

wxString ConfItemBOOL_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mbBody << ENDLWX;
    return sOut;
}

wxString ConfItemBOOL_cls::
GetItemStr(){
    wxString sOut;
    sOut << mbBody ;
    return sOut;
}

//-------------------------------------

void ConfItemBOOLARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    msHeader = strHeader;
    if (strBody == wxT("")) return;
    msBody = strBody;
    mDimension = CountDimension(strBody);
    std::wstringstream ssTmp;
    ssTmp << msHeader.c_str() << strBody.c_str();
    ReadStrLine(ssTmp.str());
}

bool ConfItemBOOLARRAY_cls::
ReadStrLine(std::wstring strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(msBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpbBody[i];
    return TRUE;
}

wxString ConfItemBOOLARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << msHeader << mpbBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpbBody[i];
    sOut << ENDLWX;
    return sOut;
}

wxString ConfItemBOOLARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpbBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpbBody[i];
    return sOut;
}

// --------------------------------

ConfItemList_cls::
ConfItemList_cls(){
    mbSetWithInitValue = TRUE;
}

void ConfItemList_cls::
Clear(){
    mVpCList.clear();
}

void ConfItemList_cls::
Add(ConfItem_cls* pC, wxString strHeader, wxString strBody){
    if (mbSetWithInitValue == TRUE)
	pC->SetItem(strHeader, strBody);
    else
	pC->SetItem(strHeader, wxT(""));
    //pC->SetItem(strHeader, strBody);
    mVpCList.push_back(pC);
}

wxString ConfItemList_cls::
SerializeAll(){
    wxString sOut;
    for (unsigned int i = 0; i < mVpCList.size(); i++)
	sOut << mVpCList[i]->Serialize();
    return sOut;
}

bool ConfItemList_cls::
ReadStrLineAll(std::wstring strLine){
    bool bRet = FALSE;
    for (unsigned int i = 0; i < mVpCList.size(); i++)
	if (mVpCList[i]->ReadStrLine(strLine) == TRUE) bRet = TRUE;
    return bRet;
}
