// (setq c-basic-offset 4)

#include "clsReadBodyImgSeq.h"
#include "../clsDF3.h"
#include "../clsDF3Sequence.h"
#include "../GL/clsColoring.h"
extern Coloring_cls GCColoring;

#include "../../COMMON/clsFilename.h"
#include "../../MAINCANVAS/clsPreference.h"
#include "../../MYAPI/MsgBox.h"

#include <wx/wx.h>
#include <wx/bitmap.h>
#include <wx/stopwatch.h>
#include <fstream>
#include <time.h>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2

// ----------------

ThreadReaderOneImage::
ThreadReaderOneImage(ReadBodyImgSeq_cls *pCParent,
		     wxWindow *pCMsgParent, int initz, int finz, int offset,
		     FileSequenceWX_cls CFileImgSeq) : wxThread(wxTHREAD_JOINABLE)
{
    mpCParent = pCParent;
    mpCMsgParent = pCMsgParent;
    mInitZ = initz;
    mFinZ = finz;
    mOffset = offset;
    mCFileImgSeq = CFileImgSeq;
    if ( Create() != wxTHREAD_NO_ERROR )
	MsgBoxExit(mpCMsgParent, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadReaderOneImage::
Entry(){
    for (int z = mInitZ; z < mFinZ; z++){
	wxImage CImage;
	if (CImage.LoadFile(mCFileImgSeq.GetModifiedWXStr(z + mOffset), wxBITMAP_TYPE_PNG) == FALSE){
	    MsgBox(mpCMsgParent, _("read error thread"),
		   mpCParent->mpC->GetFilePathWXStr() + _("can not open"));
	    
	    return 0;
	}
	
	for (unsigned int addr = 0; addr < mpCParent->mpC->GetCurrSizeSxT(Z); addr++)
	    mpCParent->mpC->mpMainXYZ[Z][mpCParent->mpC-> GetCurrSizeSxT(Z)*z + addr]
		= CImage.GetData()[addr * 3 + mpCParent->mChannel];
    }
    
    return 0;
}

// ----------------

void ReadBodyImgSeq_cls::
InitializeMember(DF3_cls *pCDF3){
    mpC = pCDF3;
}

void ReadBodyImgSeq_cls::
InitializeStaticMembers(Preference_cls *pCPreference){
    mspCPreferenceLink = pCPreference;
}

int ReadBodyImgSeq_cls::
GetBitmapType(wxString filename){
    SplitPathWX_cls CPath(filename);
    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".bmp") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".BMP"))
	return wxBITMAP_TYPE_BMP;

    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".png") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".PNG"))
	return wxBITMAP_TYPE_PNG;
    
    if (CPath.GetPathWXStr(_FN_EXT) == wxT(".tif") || CPath.GetPathWXStr(_FN_EXT) == wxT(".tiff") ||
	CPath.GetPathWXStr(_FN_EXT) == wxT(".TIF") || CPath.GetPathWXStr(_FN_EXT) == wxT(".TIFF"))
	return wxBITMAP_TYPE_TIF;
    return -1;
}

bool ReadBodyImgSeq_cls::
MainProcess(bool type, wxWindow* pC){

    static unsigned int coreNumber = wxThread::GetCPUCount();
    if (mspCPreferenceLink->mCThreadNumImgRead.mBody > coreNumber)
	mspCPreferenceLink->mCThreadNumImgRead.mBody = coreNumber;
    
    Initialize(mpC->GetCurrSize(Z));

    mpC->mOldPointXYZ = 10;
    assert(mpC->IsAllocated() == TRUE);

    FileSequenceWX_cls CFileImgSeq(/*mpC->GetFilePathWXStr()*/
				   mpC->GetFilePathFormat());
    
    wxStopWatch CSW;
   
    int offset = mpC->mpCParent->mDefaultZSliceInit;

    if (mpC->mpDownSizeSrc[X] == 1 &&
	mpC->mpDownSizeSrc[Y] == 1 && 
	mpC->mpDownSizeSrc[Z] == 1)
	for (unsigned int z = mNext; z < mpC->mCurrSize[Z]; z++){

	    if (mspCPreferenceLink->mCThreadNumImgRead.mBody > 2 && 
		mspCPreferenceLink->mCbMultiThreadImgRead.mbBody == TRUE){
		ThreadReaderOneImage Thread1
		    (this, pC, 0, mpC->mCurrSize[Z]/2, offset, CFileImgSeq);
		ThreadReaderOneImage Thread2
		    (this, pC, mpC->mCurrSize[Z]/3, 2*mpC->mCurrSize[Z]/3, offset, CFileImgSeq);
		ThreadReaderOneImage Thread3
		    (this, pC, 2*mpC->mCurrSize[Z]/3, mpC->mCurrSize[Z], offset, CFileImgSeq);

		GbGLLockForMultithread = TRUE;
		Thread1.Run(); Thread2.Run(); Thread3.Run();
		Thread1.Wait(); Thread2.Wait(); Thread3.Wait();
		GbGLLockForMultithread = FALSE;

		mNext = mpC->mCurrSize[Z];
		break;
	    }
	    else if (mspCPreferenceLink->mCThreadNumImgRead.mBody > 1 && 
		     mspCPreferenceLink->mCbMultiThreadImgRead.mbBody == TRUE){
		ThreadReaderOneImage Thread1
		    (this, pC, 0, mpC->mCurrSize[Z]/2, offset, CFileImgSeq);
		ThreadReaderOneImage Thread2
		    (this, pC, mpC->mCurrSize[Z]/2, mpC->mCurrSize[Z], offset, CFileImgSeq);
		GbGLLockForMultithread = TRUE;
		Thread1.Run(); Thread2.Run();
		Thread1.Wait(); Thread2.Wait();
		GbGLLockForMultithread = FALSE;

		mNext = mpC->mCurrSize[Z];
		break;
	    }
	    
	    int bitmapType = 
		GetBitmapType(CFileImgSeq.GetModifiedWXStr(z + offset));
	    if (bitmapType == -1) return FALSE;
	    wxImage CImage;
	    wxString strFilePath = CFileImgSeq.GetModifiedWXStr(z + offset);
	    if (CImage.LoadFile(strFilePath, bitmapType) == FALSE){
		MsgBox(NULL, strFilePath + _("can not open"),
		       _("read error thread"));
		return FALSE;
	    }
	    
	    for (unsigned int addr = 0; addr < mpC->GetCurrSizeSxT(Z); addr++)
		mpC->mpMainXYZ[Z][mpC-> GetCurrSizeSxT(Z)*z + addr]
		    = CImage.GetData()[addr * 3 + mChannel];

	    mNext = z + 1;
	    if (IsResumeConditionSatisfied(type) == TRUE) break;
	}
    else{
	// ---- READ with DownSize ----
    }
    
    if (IsFinishConditionSatisfied() == FALSE) return TRUE;
    
    Finalize();
    mpC->mbDirty = TRUE;

    //MsgBoxf(pC, "!!", "%d ms", CSW.Time());

    return TRUE;
}

bool ReadBodyImgSeq_cls::
ReadHeaderInfo(wxWindow *pCWindow){
    
    //FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathWXStr());
    FileSequenceWX_cls CFileImgSeq(mpC->GetFilePathFormat());
    
    int index = 0;
    int offset = mpC->mpCParent->mDefaultZSliceInit;
    while (TRUE){
	wxString strFilePath = CFileImgSeq.GetModifiedWXStr(index + offset);
	FILE *fp;
#ifdef WIN32
	fp = fopen(wxFNCONV(strFilePath.ToAscii()), "rb");
#else
	fp = fopen(strFilePath.ToAscii(), "rb");
#endif
	if (fp == NULL) break;
	index++;
	fclose(fp);
    }
    
    mpC->mTrueSize[Z] = index;

    int bitmapType = GetBitmapType(CFileImgSeq.GetModifiedWXStr(0));
    wxImage CImage;
    //int zInit = mpC->mpCParent->mDefaultZSliceInit;
    if (CImage.LoadFile(/*CFileImgSeq.GetModifiedWXStr(zInit)*/mpC->GetFilePathWXStr(), bitmapType) == FALSE){
	MsgBox(pCWindow, /*CFileImgSeq.GetModifiedWXStr(zInit)*/mpC->GetFilePathWXStr() + _("can not open"),
	       _("read error Header Info Image"));
	return FALSE; }
    
    mpC->mTrueSize[X] = CImage.GetWidth();
    mpC->mTrueSize[Y] = CImage.GetHeight();

    return TRUE;
}

Preference_cls *ReadBodyImgSeq_cls::mspCPreferenceLink;
//unsigned int ReadBodyImgSeq_cls::msReadChannel = 0;
