// -*- c -*- 
// "clsColoring.cpp"

#ifndef _COLORING_H_
#define _COLORING_H_

#define _CLR_ALPHA 3
#define _CLR_CUSTOM 10
#define _CLR_CUSTOM_NUM_MAX 16

#define _TRANSMIT_TRANSMIT 1
#define _TRANSMIT_FILTER 2

#include "../../CONFITEM/clsConfItem.h"
#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../../WIDGETCTRL/clsDialogEXKey.h"
#include "../../WIDGETCTRL/clsDialogController.h"
#include "../../PARAMETER/clsMultiParameters.h"
#include "../../POVRAY/clsPovSerialize.h"

#include <vector>
#include <string>

class GraphPanel_cls; 
class PanelColorBar_cls; 
class FrameMain_cls; 
class Coloring_cls;
class CustomColor_cls; 
class SliderInt_cls; 
class wxButtonEX; 

#define _CLR_NOSHADE FALSE
#define _CLR_SHADE TRUE

class CustomColorGUI_cls : public MultiParametersGUI_cls
{
 public:
    CustomColorGUI_cls(CustomColor_cls* pCParent, wxWindow *pC,
		       wxPoint pos, wxSize size);
    ~CustomColorGUI_cls();

    static CustomColorGUI_cls *mspCThis;	
    CustomColor_cls *mpCParentColor;
    PanelColorBar_cls *mpCPanelColorBar, *mpCPanelColorBar2;
    GraphPanel_cls *mpCGraphColor;
    SliderInt_cls *mpCSlider;

    int mIndexIndex, mIndexR, mIndexG, mIndexB;
    
    wxButtonEX *mpCButtonAddKey, *mpCButtonRemKey, *mpCButtonColor;
    int mTemplate;

    void SetMNumParam(wxBoxSizer *pBoxSizer);
    void RemakeIconList();

    void EnableColorCtrl(bool bEnable);
    void UpdateValues(bool bRefresh = FALSE);
    void UpdateCurrentIndex(bool bRefresh = FALSE);
    void PageUp();
    void PageDown();
    
    void OnSize            (wxSizeEvent &event);
    void OnButtonAdd       (wxCommandEvent &event);
    void OnButtonRemove    (wxCommandEvent &event);
    void OnSlider          (wxCommandEvent &event);
    void OnSliderPageUp    (wxCommandEvent &event);
    void OnSliderPageDown  (wxCommandEvent &event);

    void OnButtonAddKey    (wxCommandEvent &event);
    void OnButtonRemKey    (wxCommandEvent &event);
    void OnButtonColor     (wxCommandEvent &event);
    void OnButtonInvert    (wxCommandEvent &event);
    void OnButtonChannelShift(wxCommandEvent &event);
    void OnKeyDown(wxKeyEvent &event);

    void UpdateActiveOnTL();
    void UpdateComboBoxPropagate();

    static double GetCustomColorR(double value, void *pClue = NULL);
    static double GetCustomColorG(double value, void *pClue = NULL);
    static double GetCustomColorB(double value, void *pClue = NULL);
    static double GetCustomColorRxD(double value, void *pClue = NULL);
    static double GetCustomColorGxD(double value, void *pClue = NULL);
    static double GetCustomColorBxD(double value, void *pClue = NULL);
    static void FuncHookSyncFromPanel(void *pArg, void *pEvent);
    static void FuncHookCombo(void *pArg, void *pEvent);
    static void FuncHookComboText(void *pArg, void *pEvent);

 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class CustomColor_cls : 
public DialogControl_cls, public PovSerialize_cls, public MultiParameters_cls
{
 public:
    CustomColor_cls(wxWindow *pC, Coloring_cls *pCColoring, wxBitmap *pBitmap);
    ~CustomColor_cls();

    CustomColorGUI_cls *mpCGUI;
    Coloring_cls *mpCColoringLink;
    wxBitmap     *mpBitmapLink;

    int  mIndexInt;
    int  mCacheIndexInt;
    int  mCacheColor[3];
    int  mCacheColorTrim[3];
    double GetDoubleIndex();
    bool mbMultiplyLevel[_CLR_CUSTOM_NUM_MAX];
    bool mbCacheMultiplyLevel;
    
    void SetToDefault();
    void SetColorTemplate(int index);
    void InitGUI(wxWindow *pC);
    void UpdateItems(bool bForce = FALSE);
    bool IsRefreshed(int refreshId);
    bool IsShortClrMapPossible(int index);
    bool IsMultiplyLevel(int index);
    int  mRefreshId; // XV++ē̕⏕

    void AddItemPropagate();
    void RemoveItemPropagate();
    void ReadStrLinePropagate();

    bool IsKeyPoint();
    double GetCustomColor(double value, unsigned int index, bool bTrim = TRUE,
			  int customIndex = -1);
    std::vector < float > GetV4CustomColor(double value, bool bTrim, int customIndex, float factor = 1);
    wxString GetShortClrMap
	(int index, int df3index, bool bShade, float shadeFactor,
	 bool bMultiplyLevel, float transpalency,
	 wxString strRGBA, bool bHalfMapping, bool bLevelAdjust/* = TRUE*/,
	 bool bShadeDiminish);
    void   UpdateCacheColor();
    void   SyncFromCache();
    void   UpdateColorBar();
};

// ---------------------

class Coloring_cls : public PovSerialize_cls {
 public:
    Coloring_cls();
    ~Coloring_cls();
    
 private:
    static Coloring_cls *mpThis;
    
 public:
    CustomColor_cls *mpCCustomColor;
    void PushBack4(std::vector < float > &Vec, float r, float g, float b, float a);
    void CopyRGB(float pRGB[3], int cindex, int colorType);
    float GetAlphaFactor(int cindex, int colorType);

    unsigned int GetPVVSize(int colorType);
    std::vector < std::vector < float > >* GetPVV(int colorType);
    std::vector < std::vector < std::vector < float > > > mVVVAColors;

    void GetColor(float pRGB[3], float value, int clrindex, bool bTrim = TRUE);

    wxString GetPovStrColorMapMacro
	(int df3index, int clrindex = -1,  bool bShade = FALSE,
	 float shadeFactor = 0, float monoClr[3] = NULL, 
	 bool bMultiplyLevel = TRUE, unsigned int shadePattern = 0);
    wxString GetPovStrPigment(int clrindex, float value, float alpha,
			      int transmitType = _TRANSMIT_FILTER);
    wxString GetClrMapItem
	(int clrindex, int df3index, float clrMap, 
	 bool bShade, float shadeFactor,
	 bool bMultiplyLevel, float transpalency, wxString strRGBA,
	 bool bHalfMapping, bool bLevelAdjust, bool bShadeDiminish,
	 float monoClr[3] = NULL);
    wxString GetPovStrColorMap
	(int df3index, int clrindex = -1,  bool bShade = FALSE,
	 float shadeFactor = 0, float monoClr[3] = NULL, 
	 bool bMultiplyLevel = TRUE, unsigned int shadePattern = 0,
	 float transpalency = 0, int transmitType = _TRANSMIT_FILTER,
	 bool bHalfMapping = FALSE, bool bLevelAdjust = TRUE, bool bShortEnable = TRUE);
    wxString GetPovStrClr
	(int df3index, int clrindex = -1, float level = 0, float transpalency = 0, int transmitType = _TRANSMIT_FILTER);
};

#endif
