// "clsDraw3DTexBoxA.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
extern PFNGLTEXIMAGE3DPROC glTexImage3D;
extern PFNGLBLENDCOLORPROC glBlendColor;
extern PFNGLBLENDEQUATIONEXTPROC glBlendEquationEXT;
//extern PFNGLCOLORTABLEEXTPROC glColorTableEXT;
#else
#define GL_GLEXT_PROTOTYPES 
//#include <GL/glext.h>
#include <GL/glu.h>
#endif

#include "clsDF3GL.h"
#include "clsDraw3DTexBoxA.h"
#include "clsPreviewProperty.h"
#include "../clsDF3SeqCtrl.h"
#include "../../FILTER/clsFilterCtrl.h"

#include <algorithm>

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

#define S 0
#define T 1

void DrawDataBoxBase_cls::
SetBlendFunction(){
    switch (mpCPropertyLink->mCRenderType.mBody){
    case 0:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //At@
	break;
    case 1:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE); // Z
	break;
    }
}

Draw3DTexBoxA_cls::
Draw3DTexBoxA_cls(DF3SeqCtrl_cls *pCDF3SeqCtrl, 
		  PreviewProperty_cls *pCProperty){
    mpCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mpCPropertyLink = pCProperty;
}

void Draw3DTexBoxA_cls::
CalcAllSlice(PositionAndDirection_cls CCam,
	     double clipBoxMin[3], double clipBoxMax[3]){
    
    static DF3GLInfo_cls sCInfoPrev;
    DF3GLInfo_cls CInfoCurrent(CCam, clipBoxMin, clipBoxMax);
    
    double scaleZdash = sqrt(CCam.mTgt[X]*CCam.mTgt[X]+
			     CCam.mTgt[Y]*CCam.mTgt[Y]+
			     CCam.mTgt[Z]*CCam.mTgt[Z]);
    
    CCam.NormalizeAndOrthogonalize();
    
    int NQuality = mpCPropertyLink->GetNQuality();
    double interval = 2.0 / NQuality / scaleZdash;

    if ((!(CInfoCurrent == sCInfoPrev)) ||
	mVVCIntersect.size() == 0 || TRUE){
	
	mVVCIntersect.clear();
	for (int i = +NQuality * scaleZdash * 0.87 ; i >= -NQuality * scaleZdash * 0.87; i--){
	    double basePos[3] = 
		{CCam.mTgt[X] * interval * i,
		 CCam.mTgt[Y] * interval * i,
		 CCam.mTgt[Z] * interval * i};
	    
	    CalcSingleSlice(CCam, basePos, clipBoxMin, clipBoxMax); 
	}
	
	sCInfoPrev = CInfoCurrent;
    }
}

void Draw3DTexBoxA_cls::
CalcSingleSlice(PositionAndDirection_cls CCam, double basePos[3],
		double clipBoxMin[3], double clipBoxMax[3]){

    double APos[4][2];
    
    std::vector < Position_cls > VCIntersect;

    APos[0][S] = clipBoxMin[X]-0.5;    APos[0][T] = clipBoxMin[Y]-0.5;
    APos[1][S] = clipBoxMax[X]-0.5;    APos[1][T] = clipBoxMin[Y]-0.5;
    APos[2][S] = clipBoxMax[X]-0.5;    APos[2][T] = clipBoxMax[Y]-0.5;
    APos[3][S] = clipBoxMin[X]-0.5;    APos[3][T] = clipBoxMax[Y]-0.5;

    if (CCam.mTgt[Z] != 0)
	for (int i = 0; i < 4; i++){
	    float z = 
		(CCam.mTgt[X] * (basePos[X]-APos[i][S]) +
		 CCam.mTgt[Y] * (basePos[Y]-APos[i][T])) / CCam.mTgt[Z] +
		basePos[Z];
	    if (z > clipBoxMin[Z] - 0.5 && z < clipBoxMax[Z] - 0.5)
		VCIntersect.push_back
		    (Position_cls(APos[i][S], APos[i][T], z));
	}

    APos[0][S] = clipBoxMin[Y]-0.5;    APos[0][T] = clipBoxMin[Z]-0.5;
    APos[1][S] = clipBoxMax[Y]-0.5;    APos[1][T] = clipBoxMin[Z]-0.5;
    APos[2][S] = clipBoxMax[Y]-0.5;    APos[2][T] = clipBoxMax[Z]-0.5;
    APos[3][S] = clipBoxMin[Y]-0.5;    APos[3][T] = clipBoxMax[Z]-0.5;

    if (CCam.mTgt[X] != 0)
	for (int i = 0; i < 4; i++){
	    float x = 
		(CCam.mTgt[Y] * (basePos[Y]-APos[i][S]) +
		 CCam.mTgt[Z] * (basePos[Z]-APos[i][T])) / CCam.mTgt[X] +
		basePos[X];
	    if (x > clipBoxMin[X] - 0.5 && x < clipBoxMax[X] - 0.5)
		VCIntersect.push_back
		    (Position_cls(x, APos[i][S], APos[i][T]));
	}

    APos[0][S] = clipBoxMin[Z]-0.5;    APos[0][T] = clipBoxMin[X]-0.5;
    APos[1][S] = clipBoxMax[Z]-0.5;    APos[1][T] = clipBoxMin[X]-0.5;
    APos[2][S] = clipBoxMax[Z]-0.5;    APos[2][T] = clipBoxMax[X]-0.5;
    APos[3][S] = clipBoxMin[Z]-0.5;    APos[3][T] = clipBoxMax[X]-0.5;

    if (CCam.mTgt[Y] != 0)
	for (int i = 0; i < 4; i++){
	    float y =
		(CCam.mTgt[Z] * (basePos[Z]-APos[i][S]) +
		 CCam.mTgt[X] * (basePos[X]-APos[i][T])) / CCam.mTgt[Y] +
		basePos[Y];
	    if (y > clipBoxMin[Y] - 0.5 && y < clipBoxMax[Y] - 0.5)
		VCIntersect.push_back
		    (Position_cls(APos[i][T], y, APos[i][S]));
	}
    
    if (VCIntersect.size() == 0) return;
    
    std::vector < PositionEX_cls > VCProjected;
    for (unsigned int i = 0; i < VCIntersect.size(); i++){
	float x = 
	    CCam.mRgt[X] * VCIntersect[i].mPos[X] + 
	    CCam.mRgt[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mRgt[Z] * VCIntersect[i].mPos[Z];
	float y = 
	    CCam.mTop[X] * VCIntersect[i].mPos[X] + 
	    CCam.mTop[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mTop[Z] * VCIntersect[i].mPos[Z];
	float z = 
	    CCam.mTgt[X] * VCIntersect[i].mPos[X] + 
	    CCam.mTgt[Y] * VCIntersect[i].mPos[Y] + 
	    CCam.mTgt[Z] * VCIntersect[i].mPos[Z];
	VCProjected.push_back
	    (PositionEX_cls(x, y, z, i));
    }

    int baseIndex = 0; //\[gp̃x[X
    double xMin = VCProjected[0].mPos[X];
    for (unsigned int i = 0; i < VCProjected.size(); i++)
	if (VCProjected[i].mPos[X] < xMin){
	    xMin = VCProjected[i].mPos[X];
	    baseIndex = i;
	}

    for (unsigned int i = 0; i < VCProjected.size(); i++){
	if (i != baseIndex){
	    VCProjected[i].mPos[X] -= VCProjected[baseIndex].mPos[X] ;
	    VCProjected[i].mPos[Y] -= VCProjected[baseIndex].mPos[Y] ;
	}
    }

    VCProjected[baseIndex].mPos[X] = 0;
    VCProjected[baseIndex].mPos[Y] = 0;
    // [baseIndex]_ɂ xyʏɓ_zuꂽ    
    
    // e_̊px
    for (unsigned int i = 0; i < VCProjected.size(); i++){
	if (i == baseIndex)
	    VCProjected[i].mAngleTan = -9.99e9;
	else {
	    if (VCProjected[i].mPos[X] == 0){
		if (VCProjected[i].mPos[Y] < 0 ){
		    VCProjected[i].mAngleTan = -9.99e9;
		    VCProjected[baseIndex].mAngleTan = +9.99e9;
		    // baseႢ_base͍Ō
		}
		else
		    VCProjected[i].mAngleTan = +9.90e9;
	    }
	    else 
		VCProjected[i].mAngleTan = VCProjected[i].mPos[Y] / VCProjected[i].mPos[X];
	}
    }
    
    std::sort(VCProjected.begin(), VCProjected.end());

    std::vector < Position_cls > VCIntrsectSorted;

    for (unsigned int i = 0; i < VCProjected.size(); i++)
	VCIntrsectSorted.push_back(VCIntersect[VCProjected[i].mIndex]);

    mVVCIntersect.push_back(VCIntrsectSorted);
}

void Draw3DTexBoxA_cls::
DrawAllSlice(double alpha, float clr[3], float clrinv[3], double fogAlpha, 
	     bool bDrawVolume, double *pShadeOffset, bool bFrame){
    double offset[3] = {0, 0 ,0}, offset0[3] = {0, 0, 0};
    if (pShadeOffset != NULL)
	LOOP_XYZ offset[xyz] = pShadeOffset[xyz];

    float clrshade[3];
    LOOP_RGB clrshade[rgb] = clr[rgb] + mpCPropertyLink->mCAbsorptionComplement.mBody * (clrinv[rgb] - clr[rgb]);

    // COLOR_TABLE ɂ`͂ǂobsolute
 
    int layerNum = mVVCIntersect.size();
    for (unsigned int index = 0; index < layerNum; index++){
	if (mpCPropertyLink->mCShadingType.mBody == 1 && bDrawVolume){
	    //Z
	    glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	    if (mpCPropertyLink->mCAbsorption.mBody != 0){
		glColor4f(clrshade[0], clrshade[1], clrshade[2], 
			  alpha * mpCPropertyLink->mCAbsorption.mBody);
		DrawSingleSlice(index, offset); 
	    }
	}

	if (TRUE && (index % 4 == 0)){ 
	    // Fogɂ錸Z
	    // `͏Ȃ߂ŗǂ
	    double fogAbsorption = mpCDF3SeqCtrlLink->GetPFilterCtrl(0)->GetFogAbsorption();
	    double fogEmission = mpCDF3SeqCtrlLink->GetPFilterCtrl(0)->GetFogEmission();
	    float fogColor[3]; mpCDF3SeqCtrlLink->GetPFilterCtrl(0)->GetFogDistAndClr(fogColor);
	    if (fogAbsorption != 0){
		glBlendEquationEXT(GL_FUNC_ADD_EXT);
		glBlendFunc(GL_ZERO, GL_CONSTANT_COLOR);
		glDisable(GL_TEXTURE_3D);
		
		glBlendColor(1 - fogColor[0] * fogAbsorption * fogAlpha,
			     1 - fogColor[1] * fogAbsorption * fogAlpha,
			     1 - fogColor[2] * fogAbsorption * fogAlpha,
			     1);
		DrawSingleSlice(index, offset);
		glEnable(GL_TEXTURE_3D);
	    }

	    if (fogEmission != 0){
		glBlendEquationEXT(GL_FUNC_ADD_EXT);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glDisable(GL_TEXTURE_3D);
		glDisable(GL_TEXTURE_2D);
		
		glColor4f(fogColor[0], fogColor[1], fogColor[2], fogEmission * fogAlpha);
		DrawSingleSlice(index, offset);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_TEXTURE_3D);
	    }
	}
	
	if (bDrawVolume){
	    SetBlendFunction();
	    glColor4f(clr[0], clr[1], clr[2], alpha);
	    glBlendEquationEXT(GL_FUNC_ADD_EXT);
	    DrawSingleSlice(index, offset0); 
	}
	
	if (bFrame == TRUE){
	    //glBindTexture(GL_TEXTURE_3D, 0);
	    glBegin(GL_LINE_LOOP);
	    size_t iSize = mVVCIntersect[index].size();
	    for (size_t i = 0; i < iSize; i++){
		glVertex3d(mVVCIntersect[index][i].mPos[X],
			   mVVCIntersect[index][i].mPos[Y],
			   mVVCIntersect[index][i].mPos[Z]);
	    }
	    glEnd();
	    //glBindTexture(GL_TEXTURE_3D, mTexture3D);
	}
    }
}

void Draw3DTexBoxA_cls::
DrawAllSliceWithOtherGL(double *Aalpha, int indexNum, int texFilter,
			float (*clr)[3],
			float (*clrinv)[3], 
			int *Atexture3D,
			double fogAlpha, bool *AbDrawVolume,
			double *pShadeOffset, bool bFrame){
    double offset[3] = {0, 0 ,0}, offset0[3] = {0, 0, 0};
    if (pShadeOffset != NULL)
	LOOP_XYZ offset[xyz] = pShadeOffset[xyz];

    if (mpCPropertyLink->mCTextureType.mBody == _PREVIEW_RGBA)
	for (int df3 = 0; df3 < indexNum; df3++)
	    if (mpCDF3SeqCtrlLink->GetPFilterCtrl(df3)->GetVolumeColorType() !=
		_POV_VOLUME_MONOCOLOR)
		LOOP_RGB clr[df3][rgb] = 1;

    float clrshade[_DF3_MAX_SLOT][3];
    for (int df3 = 0; df3 < indexNum; df3++){
	LOOP_RGB clrshade[df3][rgb] = clr[df3][rgb] + mpCPropertyLink->mCAbsorptionComplement.mBody * (clrinv[df3][rgb] - clr[df3][rgb]);
    }
 
    double fogAbsorption[_DF3_MAX_SLOT];
    double fogEmission[_DF3_MAX_SLOT];
    float fogColor[_DF3_MAX_SLOT][3];
    for (int df3 = 0; df3 < indexNum; df3++){
	fogAbsorption[df3] = 
	    mpCDF3SeqCtrlLink->GetPFilterCtrl(df3)->GetFogAbsorption();
	fogEmission[df3] = 
	    mpCDF3SeqCtrlLink->GetPFilterCtrl(df3)->GetFogEmission();
	mpCDF3SeqCtrlLink->GetPFilterCtrl(df3)->GetFogDistAndClr(fogColor[df3]);
    }


    size_t indexSize = mVVCIntersect.size();
    for (size_t index = 0; index < indexSize; index++)
	for (unsigned int df3 = 0; df3 < indexNum; df3++){
	    
	    // Fog
	    if (TRUE && (index % 4 == 0)){ 
		// Fog`͏Ȃ߂ŗǂ
		if (fogAbsorption[df3] != 0){
		    glBlendFunc(GL_ZERO, GL_CONSTANT_COLOR);
		    glDisable(GL_TEXTURE_3D);
		    
		    glBlendColor
			(1 - fogColor[df3][0] * fogAbsorption[df3] * fogAlpha,
			 1 - fogColor[df3][1] * fogAbsorption[df3] * fogAlpha,
			 1 - fogColor[df3][2] * fogAbsorption[df3] * fogAlpha,
			 1);
		    DrawSingleSlice((int)index, offset);
		    glEnable(GL_TEXTURE_3D);
		}
		
		if (fogEmission[df3] != 0){
		    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		    glDisable(GL_TEXTURE_3D);
		    glDisable(GL_TEXTURE_2D);
		    
		    glColor4f(fogColor[df3][0], fogColor[df3][1], fogColor[df3][2], fogEmission[df3] * fogAlpha);
		    DrawSingleSlice((int)index, offset);
		    glEnable(GL_TEXTURE_2D);
		    glEnable(GL_TEXTURE_3D);
		}
		
	    }

	    // COLOR_TABLE ͂ǂobsolute
       
	    // Volume
	    if (AbDrawVolume[df3]){
		glBindTexture(GL_TEXTURE_3D, Atexture3D[df3]);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, 
				texFilter);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, 
				texFilter);
			    
		if (mpCPropertyLink->mCShadingType.mBody == 1 && 
		    glBlendEquationEXT != NULL){
		    //Z
		    glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
		    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		    
		    if (mpCPropertyLink->mCAbsorption.mBody != 0){
			glColor4f(clrshade[df3][0], clrshade[df3][1], clrshade[df3][2], 
				  Aalpha[df3] * mpCPropertyLink->mCAbsorption.mBody);
			DrawSingleSlice(index, offset); 
		    }
		}

		if (glBlendEquationEXT != NULL)
		    glBlendEquationEXT(GL_FUNC_ADD_EXT);
		
		SetBlendFunction();
		glColor4f(clr[df3][0], clr[df3][1], clr[df3][2], Aalpha[df3]);
		DrawSingleSlice(index, offset0); 
	    }
	}
}

void Draw3DTexBoxA_cls::
DrawSingleSlice(int index, double *pShadeOffset){
    int vertexNum = mVVCIntersect[index].size();
    glBegin(GL_TRIANGLE_FAN);
    for (unsigned int i = 0; i < vertexNum; i++){
	glTexCoord3d(mVVCIntersect[index][i].mPos[X] + 0.5 - pShadeOffset[X],
		     mVVCIntersect[index][i].mPos[Y] + 0.5 - pShadeOffset[Y],
		     mVVCIntersect[index][i].mPos[Z] + 0.5 - pShadeOffset[Z]);
	glVertex3dv(mVVCIntersect[index][i].mPos);
    }
    glEnd();  
}

std::vector < std::vector < Position_cls > > Draw3DTexBoxA_cls::mVVCIntersect; 
