// -*- c -*- 
// "clsPanelAA.h"

#include "clsPanelAA.h"
#include "clsDF3Sequence.h"

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#define X 0
#define Y 1
#define Z 2

BEGIN_EVENT_TABLE(PanelAA_cls, wxPanel)
     EVT_PAINT       (PanelAA_cls::OnPaint)
     EVT_MOUSE_EVENTS(PanelAA_cls::OnMouse)
     EVT_CHAR        (PanelAA_cls::OnChar)
END_EVENT_TABLE()

PanelAA_cls::
PanelAA_cls(DF3Sequence_cls *pParent, wxWindow *parent, wxWindowID id, wxPoint pos, wxSize size):
    wxPanel(parent, id, pos, size)
{
    mpCParent = pParent;
    this->SetBackgroundColour(wxColour(232, 232, 232));
    mOffset = _MEM_PANEL_OFFSET; 
    mFrameCurrent = 0;
    mPosCurrent = 0;
    mbCaptureMouse = 0;
}

PanelAA_cls::
~PanelAA_cls(){
}

void PanelAA_cls::
SetMemberSize(wxSize size){
    mWidth = size.GetWidth();
    mHeight = size.GetHeight();
}

int  PanelAA_cls::
GetDeviceX(float x){
    return (int)(x * (mWidth - 2. * mOffset) + mOffset);
}

void PanelAA_cls::
UpdateDisplay(){
    wxPaintDC dc(this);
    
    SetMemberSize(dc.GetSize());

    unsigned int frame    = mpCParent->mFrame;
    unsigned int frameMax = mpCParent->mCMaxFrame.mBody;
    unsigned int minFrameWA = mpCParent->mCWAMin.mBody;
    unsigned int maxFrameWA = mpCParent->mCWAMax.mBody;
    unsigned int AAmin = mpCParent->mCAAMin.mBody;
    unsigned int AAmax = mpCParent->mCAAMax.mBody;
    unsigned int step     = mpCParent->mCStep.mBody;

    wxBrush PlayModeBrush(wxColour(191,216,216), wxSOLID);

    dc.SetPen(*wxTRANSPARENT_PEN);
    // WorkingArea
    dc.SetBrush(*wxWHITE_BRUSH);
    CalcCoordinateX0X1(minFrameWA, maxFrameWA-minFrameWA, 0, frameMax);
    dc.DrawRectangle(mX0, 0, mX1-mX0, mHeight - 0);

    // ActiveArea
    dc.SetBrush(PlayModeBrush);
    CalcCoordinateX0X1(AAmin, AAmax-AAmin, 0, frameMax);
    dc.DrawRectangle(mX0, 3, mX1-mX0, mHeight - 6);

    dc.SetBrush(wxBrush(wxT("AQUAMARINE"), wxSOLID));
    dc.SetPen(*wxBLACK_PEN);
    dc.DrawRectangle(mX0-7, 0, 7, mHeight);
    dc.DrawRectangle(mX1, 0, 7, mHeight);

    // Current
    CalcCoordinateX0X1(frame, 0, 0, frameMax);
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.SetPen(*wxGREY_PEN);
    dc.DrawRectangle(mX0-1, 0, 3, mHeight);

    dc.SetBrush(*wxTRANSPARENT_BRUSH);
    dc.SetPen(*wxRED_PEN);
    dc.DrawRectangle(mX0, 0, 1, mHeight);
    
    int offset = 0;
    if (mPosCurrent > mWidth * 0.7) offset = -60;

    if (mbCaptureMouse == TRUE && 
	(mSelected == _AA_SEL_MIN || mSelected == _AA_SEL_MAX)){
	wxString Mes;
	dc.SetTextForeground(*wxRED);
	Mes << mFrameCurrent; 
	dc.DrawText(Mes, mPosCurrent + 15 + offset, -3);
    }
}

void PanelAA_cls::
OnPaint(wxPaintEvent& event){
    event.Skip();
    UpdateDisplay();
}

void PanelAA_cls::
CalcCoordinateX0X1(unsigned int frame, unsigned int step,
		   unsigned int minFrame, unsigned int maxFrame){
    float x0, x1;

    if (maxFrame == minFrame){
	x0 = 0; x1 = 1; 
    }
    else{
	float diff  = (float)((int)frame    - (int)minFrame);
	float width = (float)((int)maxFrame - (int)minFrame);
	x0 = diff / width;
	x1 = (diff + step) / width;
    }
    
    mX0 = GetDeviceX(x0); mX1 = GetDeviceX(x1);
    
    if (mX0 == mX1) mX1 = mX0+1;
}

int PanelAA_cls::
CalcFrameFromPosition(int pos, unsigned int minFrame, unsigned int maxFrame){
    if (maxFrame == minFrame) return 0;
    
    double dIndex = 
	(double)(pos - mOffset) / (mWidth - 2.0 * mOffset);
    int frame = 
	(int)(dIndex * (maxFrame - minFrame) + minFrame + 0.5);
    if (frame < 0) frame = 0;
    if (frame > (int)maxFrame) frame = maxFrame;
    return frame;
}

void PanelAA_cls::
OnMouse(wxMouseEvent& event){
    
    unsigned int initMin = mpCParent->mCAAMin.mBody;
    unsigned int initMax = mpCParent->mCAAMax.mBody;

    int frame = CalcFrameFromPosition(event.GetX(), 
				      0, mpCParent->mCMaxFrame.mBody);

    CalcCoordinateX0X1(initMin, initMax - initMin,
		       0, mpCParent->mCMaxFrame.mBody);
    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	mSelected = 0;
	if (event.GetX() < mX0 && event.GetX() > mX0 - 7) mSelected = _AA_SEL_MIN;
	if (event.GetX() < mX1 + 7 && event.GetX() > mX1) mSelected = _AA_SEL_MAX;
	if (event.GetX() < mX1 && event.GetX() > mX0) mSelected = _AA_SEL_REGION;
	if (mSelected != 0){
	    CaptureMouse();
	    mbCaptureMouse = TRUE;
	    if (mSelected == _AA_SEL_MIN || mSelected == _AA_SEL_MAX)
		SetCursor(wxCURSOR_SIZEWE);
	    if (mSelected == _AA_SEL_REGION){
		SetCursor(wxCURSOR_HAND);
		mFrameAtClick = frame;
		mMinAtClick = mpCParent->mCAAMin.mBody;
		mMaxAtClick = mpCParent->mCAAMax.mBody;
	    }
	}
    }

    if (event.Dragging() == TRUE)
	switch (mSelected){
	case _AA_SEL_MIN:
	    mpCParent->mCAAMin.mBody = frame;
	    break;
	case _AA_SEL_MAX:
	    mpCParent->mCAAMax.mBody = frame;
	    break;
	case _AA_SEL_REGION:
	    int min = mMinAtClick + (frame - mFrameAtClick);
	    int max = mMaxAtClick + (frame - mFrameAtClick);
	    if (min < 0) min = 0;
	    if (max > (int)mpCParent->mCMaxFrame.mBody)
		max = mpCParent->mCMaxFrame.mBody;
	    mpCParent->mCAAMin.mBody = min;
	    mpCParent->mCAAMax.mBody = max;
	    break;
	}

    if (event.ButtonUp(wxMOUSE_BTN_LEFT))
	if (mbCaptureMouse == TRUE){
	    ReleaseMouse();
	    mbCaptureMouse = FALSE;
	    SetCursor(wxNullCursor);
	}
    
    mPosCurrent = event.GetX();
    mFrameCurrent = frame;

    if ((mpCParent->mCAAMin.mBody != initMin) ||
	(mpCParent->mCAAMax.mBody != initMax))
	GetParent()->AddPendingEvent(event);// $B?F$,(BRefresh
    //event.Skip(); 
    else
	Refresh(TRUE); //$B<+J,$@$1(BRefresh
}

void PanelAA_cls::
OnChar(wxKeyEvent& event){
    GetParent()->AddPendingEvent(event);
}
