// -*- c++ -*- 
// "clsDF3SeqCtrl.h"
// (setq c-basic-offset 4)

#include "clsDF3SeqCtrl.h"

#include "../../clsCameraPath.h"
#include "../../clsDrawingBox.h"
#include "../../clsTimeCtrl.h"

#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/GL/clsDF3GL.h"
#include "../DF3/GL/clsPreviewProperty.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../POVRAY/clsPovVolume.h"
#include "../POVRAY/clsPovIsosurface.h"
#include "../POVRAY/clsPovCrossSection.h"
#include "../POVRAY/clsPovThreeView.h"
#include "../POVRAY/clsTreeBoxRule.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

extern bool GbGLLockForMultithread;

#include <stdio.h>

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

DF3SeqCtrl_cls::
DF3SeqCtrl_cls(wxWindow *pCParentWindow,
	       CameraPath_cls *pCPath, TimeCtrl_cls *pCTime) : 
    mCBaseAlpha(1, L"BaseAlpha"),
    mVCCheckTreeFile(_DF3_SEQ_MAX)
{
    mpCParentWindow = pCParentWindow;
    mspCThis = this;
    mspCTimeLink = pCTime;
    mCConfActiveSlot.SetItem(wxT("ActiveDataSlot="), wxT("0"));//mActiveDataSlot = 0;
    
    double alpha=32;
    mCBaseAlpha.ExecCommandInitSet(&alpha);

    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i] = new DF3Sequence_cls(i, this, pCTime);

    mpCGUI = new DF3SeqGUI_cls(pCPath, this,
			       pCTime, pCParentWindow, 
			       wxDefaultPosition, wxSize(-1, -1));

    mCConfListRender.Add(&mCCritTexMemoryMB, L"DF3CritTexMemoryMB=", L"100");
    mCConfListRender.Add(&mCCritXYMemoryMB,  L"DF3CritXYMemoryMB=",  L"200");
    mCConfListRender.Add(&mCCritXYZMemoryMB, L"DF3CritXYZMemoryMB=", L"400");

    mCConfListScene.Add(&mCAppendType, L"DF3SequenceAppendType=", L"0");
    
    mbPreFetch = FALSE;
    mbQuickShow = FALSE;
    mTexMemoryMB = mXYMemoryMB = mXYZMemoryMB = 0;
    mPrefetchMin = mPrefetchMax = 0;

    LOOP_XYZ mAbShow[xyz] = TRUE;
}

DF3SeqCtrl_cls::
~DF3SeqCtrl_cls(){
    mVCCheckTreeFile.resize(0);
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	delete mpCDF3Seq[i];
}

CheckTreeFile_cls* DF3SeqCtrl_cls::
GetPCheckTreeFile(){
    return &mVCCheckTreeFile[GetRemappedIndex(-1)];
}

DF3Sequence_cls* DF3SeqCtrl_cls::
GetPDF3Seq(unsigned int index){
    return mpCDF3Seq[index];
}

DF3Sequence_cls* DF3SeqCtrl_cls::
GetPDF3SeqActive(){
    return mpCDF3Seq[mCConfActiveSlot.mBody];
}

FilterCtrl_cls* DF3SeqCtrl_cls::
GetPFilterCtrl(unsigned int index){
    return mpCDF3Seq[index]->mpCFilterCtrl;
}

FilterCtrl_cls* DF3SeqCtrl_cls::
GetPFilterCtrlActive(){
    return mpCDF3Seq[mCConfActiveSlot.mBody]->mpCFilterCtrl;
}

FilterBase_cls* DF3SeqCtrl_cls::
GetPFilterFromTag(wxString strTag){
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	FilterBase_cls *pFilter = mpCDF3Seq[i]->mpCFilterCtrl->GetPFromTag(strTag);
	if (pFilter != NULL) return pFilter; }
    return NULL;
}

DF3_cls* DF3SeqCtrl_cls::
GetPDF3Active(){
    return mpCDF3Seq[mCConfActiveSlot.mBody]->GetPDF3();
}

void DF3SeqCtrl_cls::
GetClippingInformation(double scaleMin[3], double scaleMax[3],
		       wxString targetStr, wxString strTagActivated){
    std::vector < std::vector < double > > VDummyMin, VDummyMax;
    std::vector < bool > VDummyActive;
    std::vector < bool > VDummyVisible;

    if (targetStr == wxT("Volume"))
	for (int i = 0; i < _DF3_MAX_SLOT; i++){
	    bool bFlagSkip = FALSE;
	    GetPDF3Seq(i)->mpCFilterCtrl->
		GetClippingInformation(targetStr, scaleMin, scaleMax,
				       &VDummyMin, &VDummyMax,
				       &VDummyActive, &VDummyVisible,
				       (i==0), 
				       strTagActivated,
				       &bFlagSkip);
	    if (bFlagSkip == TRUE) break;
	}
    else
	GetPFilterCtrlActive()->
	    GetClippingInformation(targetStr, scaleMin, scaleMax,
				   &VDummyMin, &VDummyMax,
				   &VDummyActive, &VDummyVisible,
				   TRUE,
				   strTagActivated); 
}

int  DF3SeqCtrl_cls::
GetMaxActiveSlot(){
    for (int i = _DF3_MAX_SLOT - 1; i > 0 ;i--){
	if (mpCDF3Seq[i]->mVCFileSequence.size() != 0)
	    return i;
    }
    return 0;
}

void DF3SeqCtrl_cls::
ClearFilterAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->ClearFilter();
}

void DF3SeqCtrl_cls::
ClearAlreadyOutputtedTagAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	for (std::list<FilterBase_cls*>::iterator itr = 
		 mpCDF3Seq[i]->mpCFilterCtrl->mVpCFilter.begin()
		 ;itr != mpCDF3Seq[i]->mpCFilterCtrl->mVpCFilter.end(); itr++){
	    (*itr)->mVSAlreadyOutputedTag.clear();
	    (*itr)->mVSAlreadyOutputedTagDProfile.clear();
	    (*itr)->mVSAlreadyOutputedTagDFile.clear();
	}
}

void DF3SeqCtrl_cls::
ReleaseStepDF3(int df3index, int index, int flagBit, bool bFakePermitted){
    if (df3index == _DF3INDEX_ALL)
	for (int i = 0; i < _DF3_MAX_SLOT; i++)
	    mpCDF3Seq[i]->ReleaseStep(index, flagBit, bFakePermitted);
    else
	mpCDF3Seq[df3index]->ReleaseStep(index, flagBit, bFakePermitted);
    //mVCCheckTreeFile[index].ClearTree();
}

void DF3SeqCtrl_cls::
ClearTreeInfo(int timeIndex){
    if (timeIndex == -1)
	mVCCheckTreeFile[GetRemappedIndex(-1)];
    else
	mVCCheckTreeFile[timeIndex].ClearTree();
}

void DF3SeqCtrl_cls::
ReleaseAllDF3Seq(int flagBit, bool bFakePermitted){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->ReleaseAll(flagBit, bFakePermitted);
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mVCCheckTreeFile[i].ClearTree();
}

bool DF3SeqCtrl_cls::
IsReprocessRequired(int df3index, int type){
    bool bRequired = FALSE;
    FilterBase_cls *pFilterAtThis = NULL;
    if (type == _CHECK_VOLUME_ONLY)
	pFilterAtThis = mpCDF3Seq[df3index]->mpCFilterCtrl->
	    GetPFilterBase(_FILTER_VOLUME);
    bRequired |= mpCDF3Seq[df3index]->mpCFilterCtrl->
	IsReprocessRequired(GetRemappedIndex(df3index), 
			    GetRemappedIndexDouble(df3index),
			    pFilterAtThis);
    return bRequired;
}

bool DF3SeqCtrl_cls::
ProcessXYZDataAllChannel(){
    LOOP_XYZ FilterBase_cls::msABoxScale[xyz] = 0.5 * 
	(DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz+3] -
	 DrawingBlock_cls::mspCBoxSize->mpInterpolatedValue[xyz]);
    
    bool bRetval = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bRetval |= mpCDF3Seq[i]->GetPDF3()->ProcessXYZData();
    
    FilterObject_cls::msbRetex = FALSE;
    return bRetval;
}

/*
  void DF3SeqCtrl_cls::
  UpdateFilterHideAndShowAll(){
  for (int i = 0; i < _DF3_MAX_SLOT; i++)
  mpCDF3Seq[i]->mpCFilterCtrl->UpdateHideAndShows();
  }
*/

void DF3SeqCtrl_cls::
UpdateFilterParametersAll(double doubleIndex){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->UpdateParameters(doubleIndex);
}

void DF3SeqCtrl_cls::
UpdateFilterPanelListAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->UpdateList();
}

void DF3SeqCtrl_cls::
UpdateFilterTargetsAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->UpdateTarget();
}

void DF3SeqCtrl_cls::
UpdateFilterColorsAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->UpdateColors();
}

void DF3SeqCtrl_cls::
UpdateFilterColorBarAll(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->UpdateColorBar();
}

/*
  void DF3SeqCtrl_cls::
  UpdateTargetFromListAll(){
  for (int i = 0; i < _DF3_MAX_SLOT; i++)
  mpCDF3Seq[i]->mpCFilterCtrl->UpdateTargetFromList();
  }
*/

unsigned int DF3SeqCtrl_cls::
GetCurrentSizeOf1Data(int axis, int index){
    return mpCDF3Seq[index]->GetPDF3()->mCurrSize[axis];
}

bool DF3SeqCtrl_cls::
IsAllDataDF3(){
    bool bRet = TRUE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	//f[^͂邯DF3Ȃ
	bRet &= !(mpCDF3Seq[i]->GetPDF3()->mFileType != _DF3_FILETYPE_DF3 && 
		  mpCDF3Seq[i]->GetPDF3()->IsAllocated() == TRUE); 
    return bRet;
}

bool DF3SeqCtrl_cls::
DoesFilterExistInAllDF3(int type){
    bool bRet = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bRet |= mpCDF3Seq[i]->mpCFilterCtrl->DoesFilterExist(type);
    return bRet;
}

bool DF3SeqCtrl_cls::
DoesSameFilterNameExistInAllDF3(wxString strTag, FilterBase_cls *pCException){
    bool bExist = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bExist |= mpCDF3Seq[i]->mpCFilterCtrl->DoesSameFilterNameExist(strTag, pCException);
    return bExist;
}

bool DF3SeqCtrl_cls::
DoesVolumeBoxRequiredAll(){
    bool bRet = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bRet |= mpCDF3Seq[i]->mpCFilterCtrl->DoesVolumeBoxRequired();
    return bRet;
}

double DF3SeqCtrl_cls::
GetFogInfoAll(float fogColor[3], wxString strTag){
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	double distance = 
	    mpCDF3Seq[i]->mpCFilterCtrl->GetFogDistAndClr(fogColor, strTag);
	if (distance != -1) return distance;
    }
    return -1;
}

FilterData_cls* DF3SeqCtrl_cls::
GetFirstPFilterDataInAllDF3(int type){
    FilterData_cls *pFilterData;
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	pFilterData = mpCDF3Seq[i]->mpCFilterCtrl->GetPFilterData(type);
	if (pFilterData != NULL) return pFilterData;
    }
    return NULL;
}

// ------------------

unsigned int DF3SeqCtrl_cls::
GetCurrentFrame(){
    return mspCTimeLink->mFrame;
}

unsigned int DF3SeqCtrl_cls::
GetMaxFrame(){
    return mspCTimeLink->mCMaxFrame.mBody;
}

unsigned int DF3SeqCtrl_cls::
GetRemappedIndex(int df3index, int frame){
    int remappedIndex = mspCTimeLink->GetRemappedIndex(frame);
    
    if (df3index >= 0)
	if (remappedIndex > (int)mpCDF3Seq[df3index]->mMaxIndex) 
	    remappedIndex = (int)mpCDF3Seq[df3index]->mMaxIndex;
    
    return (unsigned int)remappedIndex;
}

double DF3SeqCtrl_cls::
GetRemappedIndexDouble(int df3index, int frame){
    double remappedIndexDouble = mspCTimeLink->GetRemappedIndexDouble(frame);

    if (remappedIndexDouble > mpCDF3Seq[df3index]->mMaxIndex) 
	remappedIndexDouble = mpCDF3Seq[df3index]->mMaxIndex;

    return remappedIndexDouble;
}

long unsigned DF3SeqCtrl_cls::
GetCurrentTotalSizeAllChannel(){
    long unsigned int totalSize = 0;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	totalSize += mpCDF3Seq[i]->GetPDF3()->GetCurrentTotalSize();
    return totalSize;
}

std::vector < wxString > DF3SeqCtrl_cls::
GetCurrentFWXStrAllChannel(){
    std::vector < wxString > retObj;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	retObj.push_back(mpCDF3Seq[i]->GetCurrentFWXStr());
    return retObj;
}

unsigned int DF3SeqCtrl_cls::
GetTexMemoryFor1StepAllChannelMB(){
    unsigned int texMemoryFor1StepByte;
    unsigned int texMemoryFor1StepMB;
    if (DF3GL_cls::mspCProperty->mbUse3DTexture == FALSE)
	texMemoryFor1StepByte = 
	    GetCurrentTotalSizeAllChannel() * 3;
    else
	texMemoryFor1StepByte = 
	    GetCurrentTotalSizeAllChannel();

    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_LUMINANCE_ALPHA)
	texMemoryFor1StepByte *= 2;
    if (DF3GL_cls::mspCProperty->mCTextureType.mBody == _PREVIEW_RGBA){
	texMemoryFor1StepByte *= 4;
	if (DF3GL_cls::mspCPreferenceLink->mCTexCompress.mBody == 1)
	    texMemoryFor1StepByte /= 4;
    }
    
    texMemoryFor1StepMB = texMemoryFor1StepByte / 1024 / 1024;
    
    return texMemoryFor1StepMB;
}

void DF3SeqCtrl_cls::
CalcPrefetchRange(unsigned int memoryFor1StepMB){
    int forwardMax = mCCritTexMemoryMB.mBody / memoryFor1StepMB - 1;
    if (forwardMax > 200) forwardMax = 200;
    if (forwardMax < 0)  forwardMax = 0;
    int backwardMax = 0;
    while (backwardMax < forwardMax / 3){
	forwardMax--; backwardMax++;
    }
    mPrefetchMin = GetRemappedIndex(0) - 
	backwardMax * mspCTimeLink->mCStep.mBody;
    mPrefetchMax = GetRemappedIndex(0) +
	forwardMax  * mspCTimeLink->mCStep.mBody;
}

void DF3SeqCtrl_cls::
CalcCurrentUsedMemory(){
    mTexMemoryMB = 0;
    mXYMemoryMB = 0;
    mXYZMemoryMB = 0;
    for (int index = 0; index < _DF3_MAX_SLOT; index++){
	for (unsigned int i = 0; i <= mpCDF3Seq[index]->mMaxIndex; i++){
	    mTexMemoryMB += mpCDF3Seq[index]->mVDF3GL[i].GetAllocatedTextureKB();
	    mXYMemoryMB  += mpCDF3Seq[index]->mVDF3GL[i].GetAllocatedXYKB();
	    mXYZMemoryMB += mpCDF3Seq[index]->mVDF3GL[i].GetAllocatedXYZKB();
	}
    }
    
    mTexMemoryMB /= 1024;
    mXYMemoryMB  /= 1024;
    mXYZMemoryMB /= 1024;
}

bool DF3SeqCtrl_cls::
IsPlayTexConditionSatisfied(unsigned int step){
    bool bState = TRUE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bState &= mpCDF3Seq[i]->IsPlayTexConditionSatisfied(step);
    return bState;
}

bool DF3SeqCtrl_cls::
PlayTex(){
    unsigned int nextStep = 
	mspCTimeLink->CalcNextStep(mspCTimeLink->mPlayDirection);

    if (nextStep == mspCTimeLink->mFrame){
	StopPlayTex();
	return FALSE;
    }

    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	if (mpCDF3Seq[i]->IsPlayTexConditionSatisfied(nextStep) == FALSE){
	    StopPlayTex();
	    return FALSE;
	}
    
    if (mspCTimeLink->mPlayDirection == 1)
	mspCTimeLink->FrameStep(+1);
    else
	mspCTimeLink->FrameStep(-1);
    
    // Ɏ̃Xebv̕\RfBV
    nextStep = mspCTimeLink->CalcNextStep(mspCTimeLink->mPlayDirection);
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	if (mpCDF3Seq[i]->IsPlayTexConditionSatisfied(nextStep) == FALSE)
	    return FALSE;
    return TRUE;
}

void DF3SeqCtrl_cls::
StopPlayTex(){
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemPlayStop);
    mpCParentWindow->AddPendingEvent(com);
    mspCTimeLink->mbPlayTex = FALSE;
    mspCTimeLink->UpdateDF3SeqGUI(TRUE);
}

int DF3SeqCtrl_cls::
CheckUpdate(){
    bool bRetval = TRUE;
    for (int df3 = 0; df3 < _DF3_MAX_SLOT; df3++)
	bRetval &= IdleProcessesMainEach(df3, (int)GetRemappedIndex(df3), _DF3_FORCE_FINISH);
    /*
      if (mspCTimeLink->mCbUseTimeRemap.mbBody == TRUE &&	
      mpCDF3Seq[_DF3_SLOT_0_FIX_ME]->GetRemapInterpolateFactor() != 0){
      for (int df3 = 0; df3 < _DF3_MAX_SLOT; df3++){
      int nextStep = GetRemappedIndex(df3) + 1;
      if (nextStep > (int)mpCDF3Seq[df3]->mMaxIndex) 
      nextStep = (int)mpCDF3Seq[df3]->mMaxIndex;
      bRetval &= IdleProcessesMainEach
      (df3, nextStep, _DF3_FORCE_FINISH,
      _DF3SEQ_SKIPTEXPROCESS);
      }
      }
    */
    if (mspCTimeLink->mCbUseTimeRemap.mbBody == TRUE)
	for (int df3 = 0; df3 < _DF3_MAX_SLOT; df3++)
	    if (mpCDF3Seq[df3]->GetRemapInterpolateFactor() != 0){
		int nextStep = GetRemappedIndex(df3) + 1;
		if (nextStep > (int)mpCDF3Seq[df3]->mMaxIndex) 
		    nextStep = (int)mpCDF3Seq[df3]->mMaxIndex;
		bRetval &= IdleProcessesMainEach
		    (df3, nextStep, _DF3_FORCE_FINISH,
		     _DF3SEQ_SKIPTEXPROCESS);
	    }
    
    IdleProcessesRelease(_DF3INDEX_ALL, _DF3_FORCE_FINISH);
    return bRetval;
}

bool DF3SeqCtrl_cls::
IdleProcessesMainEach(int df3index, unsigned int timeIndex,
		      bool type, bool bSkipTextureProcess){
    bool bReadyForDrawing = TRUE;
    int IdleProcess = mpCDF3Seq[df3index]->IdleProcessesMain
	(timeIndex, type, bSkipTextureProcess);
    if ((IdleProcess & 2) == 0)
	// {[ȂΏłĂȂĂH _CHECK_ME!!
	if (mpCDF3Seq[df3index]-> mpCFilterCtrl->GetVolumeValidity() == TRUE)
	    bReadyForDrawing = FALSE; // łĂȂf[^FALSE
    return bReadyForDrawing;
}

bool DF3SeqCtrl_cls::
IdleProcessesMainAll(int timeIndex, bool type, bool bSkipTextureProcess){
    bool bReadyForDrawing = TRUE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bReadyForDrawing &= IdleProcessesMainEach
	    (i, timeIndex, type, bSkipTextureProcess);
    return bReadyForDrawing;
}

bool DF3SeqCtrl_cls::
IdleProcessesFetch(){
    // ǂݔ͈͂ƃgpʌvZ
    unsigned int texMemoryFor1StepMB = GetTexMemoryFor1StepAllChannelMB();
    if (texMemoryFor1StepMB == 0) return FALSE;
    CalcPrefetchRange(texMemoryFor1StepMB);
    CalcCurrentUsedMemory();
    
    bool bState = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bState |= mpCDF3Seq[i]->IdleProcessesFetch();
    return bState;
}

bool DF3SeqCtrl_cls::
IdleProcessesRelease(int df3index, bool type){
    // ǂݔ͈͂ƃgpʌvZ
    unsigned int texMemoryFor1StepMB = GetTexMemoryFor1StepAllChannelMB();
    if (texMemoryFor1StepMB == 0) return FALSE;
    CalcPrefetchRange(texMemoryFor1StepMB);
    CalcCurrentUsedMemory();

    bool bState = FALSE;
    if (df3index == _DF3INDEX_ALL)
	for (int i = 0; i < _DF3_MAX_SLOT; i++)
	    bState |= mpCDF3Seq[i]->IdleProcessesRelease(type);
    else
	bState |= mpCDF3Seq[df3index]->IdleProcessesRelease(type);

    return bState;
}

void DF3SeqCtrl_cls::
SetFileChangedTagAllChannel(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->SetFileChangedTagAll();
}

// ------

void DF3SeqCtrl_cls::
SetDF3GLVector(int timeIndex, int df3index){
    DF3GL_cls::msVpDF3GL.clear(); 
    DF3GL_cls::msVAlpha.clear();

    if (DF3GL_cls::mspCProperty->mCHiding.mBody == _PREVIEW_MUTUAL_HIDE)
	for (int i = 0; i < _DF3_MAX_SLOT; i++){
	    int timeIndexEach;
	    if (mbQuickShow == TRUE)
		timeIndexEach = mpCDF3Seq[i]->GetQuickIndex();
	    else
		timeIndexEach = GetRemappedIndex(i);
	    
	    DF3GL_cls::msVpDF3GL.push_back(&(mpCDF3Seq[i]->mVDF3GL[timeIndexEach]));
	    DF3GL_cls::msVAlpha.push_back
		(GetPFilterCtrl(i)->GetVolumeParameter(_VOLUME_EMISSION) + 
		 0.5 * GetPFilterCtrl(i)->GetVolumeParameter(_VOLUME_SCATTER));
	}
    else{
	DF3GL_cls::msVpDF3GL.push_back(&(mpCDF3Seq[df3index]->mVDF3GL[timeIndex]));
	DF3GL_cls::msVAlpha.push_back
	    (GetPFilterCtrl(df3index)->GetVolumeParameter(_VOLUME_EMISSION)+
	     0.5 * GetPFilterCtrl(df3index)->GetVolumeParameter(_VOLUME_SCATTER));
    }
}

// ------

void DF3SeqCtrl_cls::
DrawDataInitSetting(){
    DF3GL_cls::DrawDataInitSetting();
}

void DF3SeqCtrl_cls::
DrawDataEndSetting(){
    DF3GL_cls::DrawDataEndSetting();
}

void DF3SeqCtrl_cls::
DrawDataSlot(PositionAndDirection_cls CCam, 
	     double boxScaleX, double boxScaleY, double boxScaleZ,
	     unsigned int axis, bool bMirror[3],
	     double clipBoxMin[3], double clipBoxMax[3],
	     double *pShadeOffset){
    
    if (DF3GL_cls::mspCProperty->mCHiding.mBody == _PREVIEW_NO_HIDE){
	for (int i = 0; i < _DF3_MAX_SLOT; i++)
	    if (mpCDF3Seq[i]-> mpCFilterCtrl->GetVolumeValidity() == TRUE)
		mpCDF3Seq[i]->DrawData(CCam, boxScaleX, boxScaleY, boxScaleZ,
				       axis, bMirror, 
				       clipBoxMin, clipBoxMax,
				       pShadeOffset);
    }
    else if (DF3GL_cls::mspCProperty->mCHiding.mBody == _PREVIEW_INVERSE_HIDE){
	for (int i = _DF3_MAX_SLOT-1; i >= 0; i--)
	    if (mpCDF3Seq[i]-> mpCFilterCtrl->GetVolumeValidity() == TRUE)
		mpCDF3Seq[i]->DrawData(CCam, boxScaleX, boxScaleY, boxScaleZ,
				       axis, bMirror, 
				       clipBoxMin, clipBoxMax,
				       pShadeOffset);
    }
    else { 
	//HidingĂ1DrawőS`
	//̂̂DF3SeqȂƕ`łȂ̂
	/*
	  bool bpDrawVolume[_DF3_MAX_SLOT];
	  for (int i = 0; i < _DF3_MAX_SLOT; i++)
	  bpDrawVolume[i] = mpCDF3Seq[i]->mpCFilterCtrl->GetVolumeValidity();
	*/
	for (int i = 0; i < _DF3_MAX_SLOT; i++)
	    if (mpCDF3Seq[i]->DrawData
		(CCam, boxScaleX, boxScaleY, boxScaleZ,
		 axis, bMirror, clipBoxMin, clipBoxMax,
		 pShadeOffset) == TRUE)
		return;
    }
}

void DF3SeqCtrl_cls::
DrawItems(double center[3], double scale[3]){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->DrawItems(center, scale);
}

void DF3SeqCtrl_cls::
DrawObject(double scale, bool bPositiveObject, bool bNegativeObject){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->mpCFilterCtrl->DrawObject
	    (scale, bPositiveObject, bNegativeObject);
}

int  DF3SeqCtrl_cls::
GetMaxColorLayerNum(){
    int layerNum = 1;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	if (layerNum < mpCDF3Seq[i]->GetColorLayerNum())
	    layerNum = mpCDF3Seq[i]->GetColorLayerNum();
    return layerNum;
}

void DF3SeqCtrl_cls::
SetRGBAtoGLAllChannel(unsigned int cindex){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mpCDF3Seq[i]->SetRGBAtoGL(cindex);
}

// ------

void DF3SeqCtrl_cls::
DeselectAllFilter(){
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	mspCThis->mpCDF3Seq[i]->mpCFilterCtrl->DeselectFilter(wxT("all"));
}

FilterGUI_cls *DF3SeqCtrl_cls::
GetPFilterGUI(void *pArg){
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	FilterGUI_cls *pCFilterGUI = mspCThis->mpCDF3Seq[i]->mpCFilterCtrl->GetPFilterGUI(pArg);
	if (pCFilterGUI != NULL) return pCFilterGUI;
    }
    return NULL;
}

// ------

wxString DF3SeqCtrl_cls::
GetPovStrColorMapsAllDF3(){
    int orgActive = mCConfActiveSlot.mBody;
    wxString strColorMaps;
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	mCConfActiveSlot.mBody = i;
	strColorMaps << mpCDF3Seq[i]->mpCFilterCtrl->GetColorMaps();
    }
    mCConfActiveSlot.mBody = orgActive;
    return strColorMaps;
}

wxString DF3SeqCtrl_cls::
GetPovStrFog(){
    wxString strPovFog;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	strPovFog << mpCDF3Seq[i]->mpCFilterCtrl->GetPovStrFog();
    return strPovFog;
}

wxString DF3SeqCtrl_cls::
GetPovStrGlobalObject(wxString strTag,
		      PositionAndDirection_cls CCamGlobalCoord,
		      std::vector < wxString > VstrFilePath){
    wxString strPovGlobalObject;

    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	strPovGlobalObject << GetPFilterCtrl(i)->GetPovStrGlobalObject
	    (strTag, CCamGlobalCoord, VstrFilePath);
    }
    
    return strPovGlobalObject;
}

void DF3SeqCtrl_cls::
GetPovStrDensityBox(std::wfstream &ofstream,
		    wxString strTag, std::vector < wxString > VstrFilePath,
		    wxString strGeometry, DrawingBlock_cls *pCDrawingBlock,
		    PositionAndDirection_cls CCamLocalCoord,
		    int frame){

    int orgActive = mCConfActiveSlot.mBody; // ߂ǉŗv񂾂H
    
    double baseAlpha = mCBaseAlpha.mpInterpolatedValue[0];
    double doubleIndex = mspCTimeLink->GetDoubleIndex();

    std::vector < wxString > VstrInteriorTag;

    //!! strPathƂĂ̂ TreeRuleɂConvertedꂽ\f[^̖O
    //!! ۂɕKvȃf[^͂œׂ
    //!! TreeRuleɂ Convert ͌

    // Tree\p^C}bvpDF3֐ƂĒ`
    if (mspCTimeLink->mCbUseTimeRemap.mbBody == TRUE ||
	pCDrawingBlock->IsNestOnDensity() ||
	(pCDrawingBlock->mpCParentBlock != NULL &&
	 pCDrawingBlock->mpCParentBlock->IsNestOnDensity())){
	for (int i = 0; i < _DF3_MAX_SLOT; i++){
	    mCConfActiveSlot.mBody = i;
	    ofstream << wxConvFile.cWX2WC
		(GetPFilterCtrl(i)->GetPovStrDensityFile
		 (strTag, VstrFilePath, 
		  mspCTimeLink->mCbUseTimeRemap.mbBody, 
		  mspCTimeLink->GetRemappedIndexDouble(_TIME_CTRL_CURRENT) - GetRemappedIndex(i)));
	}
    }
    
    // e`悷ꍇ͂܂
    if (pCDrawingBlock->mpCParentBlock != NULL &&
	pCDrawingBlock->mpCParentBlock->IsNestOnDensity() == TRUE){
	mCConfActiveSlot.mBody = orgActive; 
	return;
    }

    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	mCConfActiveSlot.mBody = i;
	GetPFilterCtrl(i)->GetPovStrDProfile
	    (ofstream, strTag, VstrFilePath, 
	     pCDrawingBlock, mspCTimeLink->mCbUseTimeRemap.mbBody,
	     mspCTimeLink->GetRemappedIndexDouble(_TIME_CTRL_CURRENT) - GetRemappedIndex(i));
    }

    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	mCConfActiveSlot.mBody = i;
	GetPFilterCtrl(i)->GetPovStrDensityBox
	    (ofstream, baseAlpha, doubleIndex, strTag, VstrFilePath[i],
	     strGeometry, pCDrawingBlock,
	     &VstrInteriorTag, CCamLocalCoord, frame);
    }

    wxString strPovDensityBox;
    
    mCConfActiveSlot.mBody = orgActive; 
    
    double min[3], max[3];
    double minRoot[3] = {0, 0, 0}, maxRoot[3] = {1, 1, 1};
    int selfIndex = -1;
    
    std::vector < std::vector < double > > VClipMinNoUse, VClipMaxNoUse;
    bool bClipping = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bClipping |= GetPFilterCtrl(i)->GetClippingInformation(_("Volume"), minRoot, maxRoot, &VClipMinNoUse, &VClipMaxNoUse, NULL, NULL, (i==0));
    
    pCDrawingBlock->ConvertMinAndMaxToChild(min, max, minRoot, maxRoot);
    LOOP_XYZ if (min[xyz] > 1 || max[xyz] < 0) {
	pCDrawingBlock->mbVisibility = FALSE;
	return;// wxT("");
    }
    
    // Ŝ^[QbgɂFogXg
    std::vector < FilterBase_cls* > VCPFog;
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	VCPFog.push_back(GetPFilterCtrl(i)->GetPFogTargettingMe(_("Volume")));
	while (VCPFog[VCPFog.size() - 1] != NULL)
	    VCPFog.push_back(GetPFilterCtrl(i)->GetPFogTargettingMe(_("Volume"), FALSE));
    }
    
    wxString strDifference;
    wxString strCoordinateInv = pCDrawingBlock->GetPovStrInverseCoordinate();
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	strDifference << GetPFilterCtrl(i)->GetPovStrDiffObj(strCoordinateInv);
    }
    
    strDifference << pCDrawingBlock->GetPovStrDifference();
    
    strPovDensityBox << PovVolume_cls::GetPovStrBox
	(wxT("BOX_") + strTag, 
	 min, max, bClipping, VCPFog, &VstrInteriorTag, 
	 strDifference, pCDrawingBlock);
    
    ofstream << wxConvFile.cWX2WC(strPovDensityBox);
}

wxString DF3SeqCtrl_cls::
GetPovStrPolygonObject(std::vector < DrawingBlock_cls > &VCDrawingCls){
    wxString strObj;
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	wxString strDF3Index; strDF3Index << i;
	int isoIndex = 0;
	for (std::list<FilterBase_cls*>::iterator itr = GetPFilterCtrl(i)->mVpCFilter.begin();
	     itr != GetPFilterCtrl(i)->mVpCFilter.end(); itr++){
	    if ((*itr)->mFilterType == _FILTER_ISOSURFACE){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		    (*itr)->IsClippingEmpty() != TRUE)
		    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
			strObj << VCDrawingCls[i].GetPovStrDataBlock
			    (wxT(""), _DBOX_ISOSURF, strIndex);
		isoIndex++;
	    }
	    else if ((*itr)->mFilterType == _FILTER_CROSS_SECTION){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		    (*itr)->IsClippingEmpty() != TRUE)
		    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
			strObj << VCDrawingCls[i].GetPovStrDataBlock
			    (wxT(""), _DBOX_CROSS_SECTION, strIndex);
		isoIndex++;
	    }
	    else if ((*itr)->mFilterType == _FILTER_POV_OBJECT){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		    (*itr)->IsClippingEmpty() != TRUE)
		    strObj << VCDrawingCls[0].GetPovStrDataBlock
			(wxT(""), _DBOX_POVOBJ, strIndex);
	    }
	    else if ((*itr)->mFilterType == _FILTER_THREEVIEW){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		    (*itr)->IsClippingEmpty() != TRUE)
		    
		    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
			strObj << VCDrawingCls[i].GetPovStrDataBlock
			    (wxT(""), _DBOX_THREE_VIEW, strIndex,
			     ((PovThreeView_cls*)(*itr))->GetPCData()->mCbShow0.mpbBody, ((PovThreeView_cls*)(*itr))->GetPCData()->mCbShow1.mpbBody);
		isoIndex++;
	    }
	    else if ((*itr)->mFilterType == _FILTER_STREAMLINE){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE)
		    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
			strObj << VCDrawingCls[i].GetPovStrDataBlock
			    (wxT(""), _DBOX_STREAMLINE, strIndex);
		isoIndex++;
	    }
	    else if ((*itr)->mFilterType == _FILTER_VELOCITY_FIELD){
		wxString strIndex; strIndex << (*itr)->mstrId;
		if ((*itr)->GetPData()->mCbValid.mbBody == TRUE)
		    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
			strObj << VCDrawingCls[i].GetPovStrDataBlock
			    (wxT(""), _DBOX_VELOCITYFIELD, strIndex);
		isoIndex++;
	    }
	    else if ((*itr)->mFilterType == _FILTER_VOLUME)
		isoIndex++;
	}
    }
    return strObj;
}

// ------

void DF3SeqCtrl_cls::
ReadStrLineAllDF3Sequence(std::wistream &istream, std::wstring strLine){
    mCConfActiveSlot.ReadStrLine(strLine);
    GetPDF3SeqActive()->ReadStrLineSequence(istream, strLine);
}

wxString DF3SeqCtrl_cls::
SerializeAllDF3Sequence(){
    wxString strInfo;
    int activeSlotOrg = mCConfActiveSlot.mBody;
    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	mCConfActiveSlot.mBody = i;
	strInfo << mCConfActiveSlot.Serialize();
	strInfo << mpCDF3Seq[i]->SerializeFile();
	strInfo << mpCDF3Seq[i]->mpCFilterCtrl->SerializeConf();
    }
    mCConfActiveSlot.mBody = activeSlotOrg;
    strInfo << mCConfActiveSlot.Serialize();
    return strInfo;
}

void DF3SeqCtrl_cls::
ReadStrLineRenderParam(std::wstring strLine){
    mCConfListRender.ReadStrLineAll(strLine);
    //MsgBox(NULL, wxConvFile.cWC2WX(strLine.c_str()), wxT("!!"));
}

wxString DF3SeqCtrl_cls::
SerializeRenderParam(){
    return mCConfListRender.SerializeAll();
}

void DF3SeqCtrl_cls::
ReadStrLineSceneConf(std::wstring strLine){
    mCConfListScene.ReadStrLineAll(strLine);
}

wxString DF3SeqCtrl_cls::
SerializeSceneConf(){
    return mCConfListScene.SerializeAll();
}

DF3SeqCtrl_cls* DF3SeqCtrl_cls::mspCThis = NULL;
TimeCtrl_cls* DF3SeqCtrl_cls::mspCTimeLink = NULL;
TreeBoxRule_cls* DF3SeqCtrl_cls::mspCTreeBoxRuleLink = NULL;
