
#ifndef _MORPH_OPEN_CLOSE_H_
#define _MORPH_OPEN_CLOSE_H_

#define _SE_MAX 25

#define _MORPH_DILATION 0
#define _MORPH_EROSION 1
#define _MORPH_OPENING 2
#define _MORPH_CLOSING 3
#define _MORPH_OPENING_EEDD 4
#define _MORPH_CLOSING_DDEE 5

#include "../../COMMON/clsMultiThreadCtrl.h"
#include "../../CONFITEM/clsConfItem.h"
#include "../../FILTER/clsFilterBase.h"
#include "../../WIDGETCTRL/clsMultiNumericalInputs.h"

class MorphOpenClose_cls;

class MorphOpenCloseGUI_cls : public FilterGUI_cls {
 public:
    MorphOpenCloseGUI_cls(MorphOpenClose_cls *pCParent);
    
    MorphOpenClose_cls *mpCParent;

    void SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow,
		      unsigned int index);
    MACRO_DefVirtualCustmFunctions;

    int mIndexKernelSize, mIndexAsymSize;
    
    static void  FuncHookHideItem(void *pArg, void *pEvent);
};

class MorphOpenCloseData_cls : public FilterData_cls {
 public:
    MorphOpenCloseData_cls();

    void ReDefineList();
    ConfItemUINT_cls mCType;
    ConfItemUINT_cls mCKernelType;
    ConfItemINT_cls mCKernelSize;
    ConfItemINTARRAY_cls mCKernelAssimSize;

    void MakeFilterInstanceAndAddToList
	(int index, FilterCtrl_cls *pCParentCtrl);
};

class ThreadMorphDilation : public MultiThreadThread_cls {
 public:
    ThreadMorphDilation(wxThreadKind kind);
    void SetParameters
	(MorphOpenClose_cls *pCParent, 
	 unsigned char ***pFrom, unsigned char ***pTo, int size[3],
	 int seSize_x, int seSize_y, int seSize_z,
	 char se[_SE_MAX][_SE_MAX][_SE_MAX],
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum,
	 int iniZ, int finZ);

    MorphOpenClose_cls *mpCParent;
    int mSize[3];
    unsigned char ***mpFrom;
    unsigned char ***mpTo;
    char (*mse)[_SE_MAX][_SE_MAX];
    int mSeSize_x, mSeSize_y, mSeSize_z;
    int mxOffset, myOffset, mzOffset;
    int mIniZ, mFinZ;
    
    int mseListNum, mseDiffNum;
    char *mpseList, *mpseDiff;
    
    inline unsigned char ThreadSeLoopDilationEdge
	(unsigned char *pFrom, 
	 unsigned char max, char *pList, int listNum,
	 const int indexListMax, char pMaxPos[3]);
    inline unsigned char ThreadSeLoopDilationCenter
	(unsigned char *pFrom, 
	 unsigned char max, char *pList, int listNum,
	 const int indexListMax, char pMaxPos[3]);
    void KernelLoopDilationEdge
	(int y, int z, int xc, int yc, int zc, int iniX, int finX, 
	 unsigned char *pTo, unsigned char *pFrom, char *pse, 
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum );
    void KernelLoopDilationCenter
	(int y, int z, int xc, int yc, int zc, 
	 unsigned char *pTo, unsigned char *pFrom, char *pse, 
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum );

    virtual ExitCode Entry();
};

class ThreadMorphErosion : public MultiThreadThread_cls {
 public:
    ThreadMorphErosion(wxThreadKind kind);
    void SetParameters
	(MorphOpenClose_cls *pCParent, 
	 unsigned char ***pFrom, unsigned char ***pTo, int size[3],
	 int seSize_x, int seSize_y, int seSize_z,
	 char se[_SE_MAX][_SE_MAX][_SE_MAX],
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum,
	 int iniZ, int finZ);

    MorphOpenClose_cls *mpCParent;
    int mSize[3];
    unsigned char ***mpFrom;
    unsigned char ***mpTo;
    char (*mse)[_SE_MAX][_SE_MAX];
    int mSeSize_x, mSeSize_y, mSeSize_z;
    int mxOffset, myOffset, mzOffset;
    int mIniZ, mFinZ;

    int mseListNum, mseDiffNum;
    char *mpseList, *mpseDiff;
    
    inline unsigned char ThreadSeLoopErosionEdge
	(unsigned char *pFrom, 
	 unsigned char min, char *pList, int listNum,
	 const int indexListMax, char pMaxPos[3]);
    inline unsigned char ThreadSeLoopErosionCenter
	(unsigned char *pFrom, 
	 unsigned char min, char *pList, int listNum,
	 const int indexListMax, char pMaxPos[3]);
    void KernelLoopErosionEdge
	(int y, int z, int xc, int yc, int zc, int iniX, int finX, 
	 unsigned char *pTo, unsigned char *pFrom, char *pse, 
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum);
    void KernelLoopErosionCenter
	(int y, int z, int xc, int yc, int zc,
	 unsigned char *pTo, unsigned char *pFrom, char *pse, 
	 char *pseList, int seListNum, char *pseDiff, int seDiffNum);

    virtual ExitCode Entry();
};

class MorphOpenClose_cls : public FilterBase_cls, public MultiThreadCtrl_cls {
 public:
    MorphOpenClose_cls(FilterCtrl_cls *pCParentCtrl,
		       MorphOpenCloseData_cls *pCInitData = NULL);
    ~MorphOpenClose_cls();
    
    MorphOpenCloseData_cls mCData;
    FilterData_cls*  GetPData();

    int mSize[3];
    int mSeSize_x, mSeSize_y, mSeSize_z;
    unsigned char ***mpObj;
    unsigned char ***mpObjB;
    
    virtual void AddFilterItemPolymorph(wxBoxSizer *pBoxSizer, FilterPanels_cls *pCFilterPanels, unsigned int index);
    MorphOpenCloseData_cls* GetPCData();

    //--
    
    void Dilation3D(unsigned char ***pFrom, unsigned char ***pTo,
		    unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]);
    void Erosion3D(unsigned char ***pFrom, unsigned char ***pTo,
		   unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]);
    bool ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]);
    
    void MakeSphericalSE(unsigned char se_org[_SE_MAX][_SE_MAX][_SE_MAX]);

    //--

    wxString GetFilterStatusStr();
    wxString SerializeConf();
    
    static MorphOpenCloseData_cls ReadDataBlock(std::wistream &istream);
};


#endif
