// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsClipping.h"

#include "clsClipping.h"

#include "../DF3/clsDF3SeqGUI.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../../wxWidgetEnum.h"

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#ifdef WIN32
#define WXENDL wxT("\r\n")
#else
#define WXENDL wxT("\n")
#endif

MACRO_FilterConstructor(FilterClipping, WithTarget,);

void FilterClippingGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 80;

    wxString listTarget[1] = { _("Volume") };
    pCurr->mComboWidth = 108;
    mIndexTarget = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mIndexOnList,
	 _("Clip target"), TRUE, 1, listTarget, 0);
    
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    //mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);

    wxString targetStr = mpCParent->GetLocaleStrLabel(mpCParent->GetPCData()->mCConfTargetStr.msBody);
    UpdateListMain();
    SetTargetLabelStr(targetStr);
}

#include "../MYAPI/MsgBox.h"

MACRO_SetCustomFunctions(FilterClippingGUI_cls,,,);
MACRO_HideAndShowEmpty(FilterClippingGUI_cls);

// -----------------

FilterClippingData_cls::
FilterClippingData_cls(FilterCtrl_cls* pCCtrl,
		       bool bConnectToStaticParam) : 
FilterWithParameterData_cls(pCCtrl)
{
    // ʂ̃Xgɒǉ
    mCConfList.Add(&mCConfTargetStr, wxT("ClippingTargetStr="), _("Volume"));

    Parameters_cls CParam(6, L"ClippingBox");
    double boxsize[6] = {0, 0, 0, 1, 1, 1};
    CParam.ExecCommandInitSet(boxsize);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void FilterClippingData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCConfTargetStr, wxT("ClippingTargetStr="));
}

MACRO_MakeInstanceAndAddToList(FilterClipping)

wxString FilterClippingData_cls::
SerializeConfPolymorph(){
    return msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

void FilterClippingData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void FilterClippingData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* FilterClippingData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > FilterClippingData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > FilterClippingData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// ----------

#include "../MYAPI/MsgBox.h"

FilterClipping_cls::
FilterClipping_cls(FilterCtrl_cls *pCParentCtrl, FilterClippingData_cls *pCInitData) :
    FilterWithParameter_cls(pCParentCtrl)
{
    mpCData = new FilterClippingData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *(FilterClippingData_cls*)mpCData = *pCInitData;
    ConstructorCommon(pCParentCtrl, _("Clipping"), _FILTER_CLIPPING);
    mpCGUIPanel = new FilterClippingGUI_cls(this);
    // TargettFileterTRUEŏĂ̂ŖI
    mbWithTexture = FALSE;
}

FilterClipping_cls::
~FilterClipping_cls(){
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
    if (MainGLCanvas_cls::mspCSelectedItem == mpCData->GetPParamActive()) 
	MainGLCanvas_cls::SelectItem(NULL);
}

MACRO_TargetUpdateFunctions(FilterClipping_cls);

void FilterClipping_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls *pCDLG = 
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));
    
    pCDLG->mTextWidth = 180;
    pCDLG->CombinePointer(_GP_BOOL, &msCConfClippingFollowTracking.mbBody, wxT("Follow Tracking"));
    pCDLG->FinishSettingDLG(FALSE);
}

MACRO_GetPCData(FilterClipping);

MACRO_AddFilterItemPolymorphP(FilterClipping,);
MACRO_InitializeStaticMemberP(FilterClipping);

// ------

wxString FilterClipping_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("Clipping::") << mpCData->mCbValid.mbBody;
    
    if (mpCData->mCbValid.mbBody == FALSE)
	strStatus<< wxT("::");
    else
	strStatus << 
	    mpCData->GetPInterpolated()[0] << wxT(":") <<
	    mpCData->GetPInterpolated()[1] << wxT(":") <<
	    mpCData->GetPInterpolated()[2] << wxT(":") <<
	    mpCData->GetPInterpolated()[3] << wxT(":") <<
	    mpCData->GetPInterpolated()[4] << wxT(":") <<
	    mpCData->GetPInterpolated()[5] << wxT("::");
    
    return strStatus;
}

wxString FilterClipping_cls::
SerializeConf(){
    UpdateCommonData();
    UpdateTargetInfo();
    /*
    GetPCData()->mCConfTargetStr.msBody
	= GetLocaleIndependentStrLabel(GetTargetLabelStr());
    */
    return mpCData->SerializeConf(mFilterType);
}

MACRO_GetNewPReadDataBlock(FilterClipping)

ConfItemBOOL_cls FilterClipping_cls::msCConfClippingFollowTracking(wxT("ClippingFollowTracking="), wxT("1"));
