// -*- c -*- 
// (setq c-basic-offset 4)
// "clsBoxCopy.h"

#include "clsDataRotate.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"

#include "../MYAPI/MsgBox.h"

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

MACRO_FilterConstructor(DataRotate,,)

void DataRotateGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);
    
#ifdef WIN32
    pCurr->mTextWidth = 60;
#else
    pCurr->mTextWidth = 80;
#endif
    
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Axis/Degree"));

    // lݒ肵Ă
    mpCParent->GetPCData()->GetPInterpolated()[0] = 0;

    pCurr->mComboWidth = 48;
    wxString listAxis[3] = { wxT("X"), wxT("Y"), wxT("Z")};
    pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->GetPCData()->mCAxis.mBody, wxT(""), TRUE, 3, listAxis, 0);
    pCurr->CombinePointer(_GP_DOUBLE, &mpCParent->GetPCData()->GetPInterpolated()[0]);
    
    pCurr->mProportion = 1;
    //pCurr->AlignRight(indexCustomA);
    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
}

MACRO_SetCustomFunctions(DataRotateGUI_cls,,,)
MACRO_HideAndShowEmpty(DataRotateGUI_cls);

// -----

ThreadRotate::
ThreadRotate(wxThreadKind kind) : MultiThreadThread_cls(kind) {}

void ThreadRotate::
SetParameters(DataRotate_cls *pCParent,
	      unsigned char *pTo, unsigned char *pFrom, 
	      unsigned int size[3],
	      int S, int T, int U, int layerIni, int layerFin,
	      float *pRotTableFloatX, float *pRotTableFloatY,
	      float *pRotTableFloatZ){
    mpCParent = pCParent;
    mpTo = pTo;
    mpFrom = pFrom;
    LOOP_XYZ mSize[xyz] = size[xyz];
    mS = S; mT = T; mU = U;
    mLayerIni = layerIni;
    mLayerFin = layerFin;

    mpRotTableFloatX = pRotTableFloatX;
    mpRotTableFloatY = pRotTableFloatY;
    mpRotTableFloatZ = pRotTableFloatZ;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

unsigned char ThreadRotate::
GetSmoothedValue(float x, float y, float z){
   
    unsigned char *pFrom = mpFrom;
    
    int coordx = x, coordy = y, coordz = z;
    float fx = x-coordx;
    float fy = y-coordy;
    float fz = z-coordz;

    int lOffsetZ = mSize[X] * mSize[Y];
    int lOffsetY = mSize[X];
    int lBaseIndex = coordz * lOffsetZ + coordy * lOffsetY + coordx;

#ifdef _DEBUG
    if (lBaseIndex + lOffsetZ + lOffsetY + 1 >= mSize[X] * mSize[Y] * mSize[Z])
	printf("!! %f %f %f\n", x, y, z);
#endif
	
    const float P000(pFrom[lBaseIndex]);
    const float P001(pFrom[lBaseIndex+1]);
    const float P010(pFrom[lBaseIndex+lOffsetY]);
    const float P011(pFrom[lBaseIndex+lOffsetY+1]);
    const float P100(pFrom[lBaseIndex+lOffsetZ]); 
    const float P101(pFrom[lBaseIndex+lOffsetZ+1]);
    const float P110(pFrom[lBaseIndex+lOffsetZ+lOffsetY]);
    const float P111(pFrom[lBaseIndex+lOffsetZ+lOffsetY+1]);

#ifdef _DEBUG
    if (lBaseIndex + lOffsetZ + lOffsetY + 1 >= mSize[X] * mSize[Y] * mSize[Z])
	printf("!!end\n");
#endif
    
    return  
	(1.0f-fz) * ( (1.0f-fy) * ((1.0f-fx)*P000 + fx*P001)  +
		      fy        * ((1.0f-fx)*P010 + fx*P011)) +
	fz        * ( (1.0f-fy) * ((1.0f-fx)*P100 + fx*P101)  +
		      fy        * ((1.0f-fx)*P110 + fx*P111));
}

wxThread::ExitCode ThreadRotate::
Entry(){
    float degree = mpCParent->GetPCData()->GetPInterpolated()[0];
    
    bool matSX, matSY, matSZ;
    bool matTX, matTY, matTZ;
    bool matUX, matUY, matUZ;

    switch (mS){
    case X:
	matSX = 1; matSY = 0; matSZ = 0; break;
    case Y:
	matSX = 0; matSY = 1; matSZ = 0; break;
    case Z:
	matSX = 0; matSY = 0; matSZ = 1; break;
    }

    switch (mT){
    case X:
	matTX = 1; matTY = 0; matTZ = 0; break;
    case Y:
	matTX = 0; matTY = 1; matTZ = 0; break;
    case Z:
	matTX = 0; matTY = 0; matTZ = 1; break;
    }

    switch (mU){
    case X:
	matUX = 1; matUY = 0; matUZ = 0; break;
    case Y:
	matUX = 0; matUY = 1; matUZ = 0; break;
    case Z:
	matUX = 0; matUY = 0; matUZ = 1; break;
    }

    int sizeX = mSize[X];
    int sizeXY = mSize[X] * mSize[Y];
    int sizeS = mSize[mS], sizeT = mSize[mT], sizeU = mSize[mU];
    
    float factorX = (mSize[X]-1)/2;
    float factorY = (mSize[Y]-1)/2;
    float factorZ = (mSize[Z]-1)/2;

    unsigned char *pFrom = mpFrom;
    unsigned char *pTo = mpTo;

    // STł̃e[uQƂȂ
    // [v̐[łif炷ߍ璷
    if (mpCParent->GetPCData()->mCbBilinear.mbBody == TRUE &&
	mpCParent->GetPCData()->mCbTrimOutside.mbBody == TRUE)
	for (int cu = mLayerIni; cu < mLayerFin; cu++){
	    float cnu = (2.0f * cu - sizeU) / sizeU;
	    
	    int dcxfromu = cu * matUX;
	    int dcyfromu = cu * matUY;
	    int dczfromu = cu * matUZ;
	    float dnxfromu = cnu * matUX * factorX;
	    float dnyfromu = cnu * matUY * factorY;
	    float dnzfromu = cnu * matUZ * factorZ;
	    
	    for (int ct = 0; ct < sizeT; ct++){
		int cx = ct * matTX + dcxfromu;
		int cy = ct * matTY + dcyfromu;
		int cz = ct * matTZ + dczfromu;
		int indexST = ct * sizeS;
		for (int cs = 0; cs < sizeS; cs++, indexST++){
		    int cindex = cz * sizeXY + cy * sizeX + cx;
		    
		    // trimȂǂĂif͕Kvc
		    if (mpRotTableFloatX[indexST] == -1){
			mpTo[cindex] = 0;
			cx += matSX; cy += matSY; cz += matSZ;
			continue;
		    }
		    
		    // ]̍WST̃e[uĂU̕ǉ];
		    float nx = mpRotTableFloatX[indexST] + dnxfromu;
		    float ny = mpRotTableFloatY[indexST] + dnyfromu;
		    float nz = mpRotTableFloatZ[indexST] + dnzfromu;
		    
		    pTo[cindex] = GetSmoothedValue(nx, ny, nz);
		    cx += matSX; cy += matSY; cz += matSZ;
		}
	    }
	}
    else if (mpCParent->GetPCData()->mCbBilinear.mbBody == FALSE &&
	     mpCParent->GetPCData()->mCbTrimOutside.mbBody == TRUE)
	for (int cu = mLayerIni; cu < mLayerFin; cu++){
	    float cnu = (2.0f * cu - sizeU) / sizeU;
	    
	    int dcxfromu = cu * matUX;
	    int dcyfromu = cu * matUY;
	    int dczfromu = cu * matUZ;
	    float dnxfromu = cnu * matUX * factorX;
	    float dnyfromu = cnu * matUY * factorY;
	    float dnzfromu = cnu * matUZ * factorZ;
	    
	    for (int ct = 0; ct < sizeT; ct++){
		int cx = ct * matTX + dcxfromu;
		int cy = ct * matTY + dcyfromu;
		int cz = ct * matTZ + dczfromu;
		int indexST = ct * sizeS;
		for (int cs = 0; cs < sizeS; cs++, indexST++){
		    int cindex = cz * sizeXY + cy * sizeX + cx;
		    
		    // trimȂǂĂif͕Kvc
		    if (mpRotTableFloatX[indexST] == -1){
			pTo[cindex] = 0;
			cx += matSX; cy += matSY; cz += matSZ;
			continue;
		    }
		    
		    // ]̍WST̃e[uĂU̕ǉ];
		    float nx = mpRotTableFloatX[indexST] + dnxfromu;
		    float ny = mpRotTableFloatY[indexST] + dnyfromu;
		    float nz = mpRotTableFloatZ[indexST] + dnzfromu;
		    
		    int nindex = 
			((int)nz) * sizeXY + ((int)ny) * sizeX + ((int)nx);
		    pTo[cindex] = pFrom[nindex];
		    cx += matSX; cy += matSY; cz += matSZ;
		}
	    }
	}
    else if (mpCParent->GetPCData()->mCbBilinear.mbBody == TRUE &&
	     mpCParent->GetPCData()->mCbTrimOutside.mbBody == FALSE)
	for (int cu = mLayerIni; cu < mLayerFin; cu++){
	    float cnu = (2.0f * cu - sizeU) / sizeU;
	    
	    int dcxfromu = cu * matUX;
	    int dcyfromu = cu * matUY;
	    int dczfromu = cu * matUZ;
	    float dnxfromu = cnu * matUX * factorX;
	    float dnyfromu = cnu * matUY * factorY;
	    float dnzfromu = cnu * matUZ * factorZ;
	    
	    for (int ct = 0; ct < sizeT; ct++){
		int cx = ct * matTX + dcxfromu;
		int cy = ct * matTY + dcyfromu;
		int cz = ct * matTZ + dczfromu;
		int indexST = ct * sizeS;
		for (int cs = 0; cs < sizeS; cs++, indexST++){
		    int cindex = cz * sizeXY + cy * sizeX + cx;
		    
		    // ]̍WST̃e[uĂU̕ǉ];
		    float nx = mpRotTableFloatX[indexST] + dnxfromu;
		    float ny = mpRotTableFloatY[indexST] + dnyfromu;
		    float nz = mpRotTableFloatZ[indexST] + dnzfromu;
		    
		    pTo[cindex] = GetSmoothedValue(nx, ny, nz);
		    cx += matSX; cy += matSY; cz += matSZ;
		}
	    }
	}
    else
	for (int cu = mLayerIni; cu < mLayerFin; cu++){
	    float cnu = (2.0f * cu - sizeU) / sizeU;
	    
	    int dcxfromu = cu * matUX;
	    int dcyfromu = cu * matUY;
	    int dczfromu = cu * matUZ;
	    float dnxfromu = cnu * matUX * factorX;
	    float dnyfromu = cnu * matUY * factorY;
	    float dnzfromu = cnu * matUZ * factorZ;
	    
	    for (int ct = 0; ct < sizeT; ct++){
		int cx = ct * matTX + dcxfromu;
		int cy = ct * matTY + dcyfromu;
		int cz = ct * matTZ + dczfromu;
		int indexST = ct * sizeS;
		for (int cs = 0; cs < sizeS; cs++, indexST++){
		    int cindex = cz * sizeXY + cy * sizeX + cx;
		    
		    // ]̍WST̃e[uĂU̕ǉ];
		    float nx = mpRotTableFloatX[indexST] + dnxfromu;
		    float ny = mpRotTableFloatY[indexST] + dnyfromu;
		    float nz = mpRotTableFloatZ[indexST] + dnzfromu;
		    
		    int nindex = 
			((int)nz) * sizeXY + ((int)ny) * sizeX + ((int)nx);
		    pTo[cindex] = pFrom[nindex];
		    cx += matSX; cy += matSY; cz += matSZ;
		}
	    }
	}

    *mpbReadyFlag = TRUE;
    return 0;
}

// -----

DataRotateData_cls::
DataRotateData_cls(FilterCtrl_cls* pCCtrl,
		   bool bConnectToStaticParam) :
FilterWithParameterData_cls(pCCtrl)
{
    // ʂ̃Xgɒǉ
    mCConfList.Add(&mCAxis, wxT("DataRotateAxis="), L"0");
    //mCConfList.Add(&mCDegree, wxT("DataRotateDegree="), L"0");
    mCConfList.Add(&mCbTrimOutside, wxT("DataRotateTrim="), L"1");
    mCConfList.Add(&mCbBilinear, wxT("DataRotateBilinear="), L"1");

    Parameters_cls CParam(1, L"DataRotate");
    double initParam[1] = {0};
    CParam.ExecCommandInitSet(initParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void DataRotateData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCAxis, wxT("DataRotateAxis="));
    //mCConfList.Add(&mCDegree , wxT("DataRotateDegree="));
    mCConfList.Add(&mCbTrimOutside, wxT("DataRotateTrim="));
    mCConfList.Add(&mCbBilinear, wxT("DataRotateBilinear="));
}

MACRO_MakeInstanceAndAddToList(DataRotate)

wxString DataRotateData_cls::
SerializeConfPolymorph(){
    return msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

void DataRotateData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void DataRotateData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* DataRotateData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > DataRotateData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > DataRotateData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// -----

#include "../MYAPI/MsgBox.h"

DataRotate_cls::
DataRotate_cls(FilterCtrl_cls *pCParentCtrl, DataRotateData_cls *pCInitData) 
: FilterWithParameter_cls(pCParentCtrl){

    mpCData = new DataRotateData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *(DataRotateData_cls*)mpCData = *pCInitData;
    mbProcessData = TRUE;
    ConstructorCommon(pCParentCtrl, _("Data Rotate"), _FILTER_DATA_ROTATE);
    mpCGUIPanel = new DataRotateGUI_cls(this);

    mTableS = mTableT = 0;
    mpRotTableFloatX = mpRotTableFloatY = mpRotTableFloatZ = NULL;
}

DataRotate_cls::
~DataRotate_cls(){
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
    
    if (mpRotTableFloatX != NULL){
	delete mpRotTableFloatX;
	delete mpRotTableFloatY; 
	delete mpRotTableFloatZ; 
    }
}

MACRO_AddFilterItemPolymorphP(DataRotate,);
MACRO_InitializeStaticMemberP(DataRotate);
MACRO_GetPCData(DataRotate);

// ----

void DataRotate_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls* pCDLG =
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));

    pCDLG->mTextWidth = 180;
    pCDLG->CombinePointer(_GP_BOOL, &GetPCData()->mCbTrimOutside.mbBody,
			  wxT("trim outside"));
    pCDLG->CombinePointer(_GP_BOOL, &GetPCData()->mCbBilinear.mbBody,
			  wxT("Bilinear / NearestNeibour"));

    pCDLG->FinishSettingDLG(FALSE);
}

// ----

void DataRotate_cls::
MakeTable(int S, int T, int U, unsigned int sizeXYZ[3], int sizeS, int sizeT){
    
    if (sizeS != mTableS || sizeT != mTableT){
	delete mpRotTableFloatX; mpRotTableFloatX = NULL;
	delete mpRotTableFloatY; mpRotTableFloatY = NULL;
	delete mpRotTableFloatZ; mpRotTableFloatZ = NULL;
    }

    if (mpRotTableFloatX == NULL){
	mpRotTableFloatX = new float [sizeS * sizeT];
	mpRotTableFloatY = new float [sizeS * sizeT];
	mpRotTableFloatZ = new float [sizeS * sizeT];
    }

    mTableS = sizeS; mTableT = sizeT;

    bool matSX, matSY, matSZ;
    bool matTX, matTY, matTZ;
   
    switch (S){
    case X:
	matSX = 1; matSY = 0; matSZ = 0; break;
    case Y:
	matSX = 0; matSY = 1; matSZ = 0; break;
    case Z:
	matSX = 0; matSY = 0; matSZ = 1; break;
    }

    switch (T){
    case X:
	matTX = 1; matTY = 0; matTZ = 0; break;
    case Y:
	matTX = 0; matTY = 1; matTZ = 0; break;
    case Z:
	matTX = 0; matTY = 0; matTZ = 1; break;
    }

    float aspect = msABoxScale[T] / msABoxScale[S]; 
    float aspectinv = 1.0f/aspect;
    float cosRad = cos(GetPCData()->GetPInterpolated()[0] * M_PI / 180);
    float sinRad = sin(GetPCData()->GetPInterpolated()[0] * M_PI / 180);

    int sizeX = sizeXYZ[X], sizeY = sizeXYZ[Y], sizeZ = sizeXYZ[Z];
    float sizeSinv = 1.0f/sizeS;
    float factorX = (sizeX - 1) / 2;
    float factorY = (sizeY - 1) / 2;
    float factorZ = (sizeZ - 1) / 2;

    if (GetPCData()->mCbTrimOutside.mbBody == TRUE)
	for (int ct = 0; ct < sizeT; ct++){
	    int indexT = ct * sizeS;
	    float cnt = (2.0f * ct - sizeT) / sizeT;
	    float cntaspect = cnt*aspect;
	    for (int cs = 0; cs < sizeS; cs++){
		int index = indexT + cs;
		float cns = (2.0f * cs - sizeS) * sizeSinv;
		float nns = cns * cosRad - cntaspect * sinRad;
		float nnt = cns * sinRad * aspectinv + cnt * cosRad;
		
		float nnx = nns*matSX + nnt*matTX;
		float nny = nns*matSY + nnt*matTY;
		float nnz = nns*matSZ + nnt*matTZ;
		
		float nx = (nnx + 1) * factorX;
		float ny = (nny + 1) * factorY;
		float nz = (nnz + 1) * factorZ;
		
		// ͂ݏof[^-1ƂB
		if (nx < 0 || nx > sizeX - 2 ||
		    ny < 0 || ny > sizeY - 2 ||
		    nz < 0 || nz > sizeZ - 2) {
		    nx = ny = nz = -1;
		}
		
		mpRotTableFloatX[index] = nx;
		mpRotTableFloatY[index] = ny;
		mpRotTableFloatZ[index] = nz;
	    }
	}
    else
	for (int ct = 0; ct < sizeT; ct++){
	    int indexT = ct * sizeS;
	    float cnt = (2.0f * ct - sizeT) / sizeT;
	    float cntaspect = cnt*aspect;
	    for (int cs = 0; cs < sizeS; cs++){
		int index = indexT + cs;
		float cns = (2.0f * cs - sizeS) * sizeSinv;
		float nns = cns * cosRad - cntaspect * sinRad;
		float nnt = cns * sinRad * aspectinv + cnt * cosRad;
		
		float nnx = nns*matSX + nnt*matTX;
		float nny = nns*matSY + nnt*matTY;
		float nnz = nns*matSZ + nnt*matTZ;
		
		float nx = (nnx + 1) * factorX;
		float ny = (nny + 1) * factorY;
		float nz = (nnz + 1) * factorZ;
		
		if (nx < 0) nx = 0;
		if (nx > sizeX - 2.01) nx = sizeX - 2.01;
		if (ny < 0) ny = 0;
		if (ny > sizeY - 2.01) ny = sizeY - 2.01;
		if (nz < 0) nz = 0;
		if (nz > sizeZ - 2.01) nz = sizeZ - 2.01;
		
		mpRotTableFloatX[index] = nx;
		mpRotTableFloatY[index] = ny;
		mpRotTableFloatZ[index] = nz;
	    }
	}
}

bool DataRotate_cls::
ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3]){
    int totalSize = size[X] * size[Y] * size[Z];
    
    if (mpCData->mCbValid.mbBody == FALSE){
	memcpy(pTo, pFrom, totalSize);
	return FALSE;
    }

    /*
      // ̏ƂB
      // FIX_ME!! W̕ϊȂ̂1{NZĂ
    float degree = ((DataRotateData_cls*)mpCData)->GetPInterpolated()[0];
    if (degree == 0){
	mTargetValue = 0;
	memcpy(pTo, pFrom, totalSize);
	return TRUE;
    }
    */
    
    int S, T, U;
    switch (GetPCData()->mCAxis.mBody){
    case X:
	S = Y; T = Z; U = X; break;
    case Y:
	S = Z; T = X; U = Y; break;
    case Z:
	S = X; T = Y; U = Z; break;
    }

    GbGLLockForMultithread = TRUE;
    MakeTable(S, T, U, size, size[S], size[T]);
    int layerStep = 16;

    InitializeThreadCtrl();

    if (mThreadNum == 1) layerStep = size[U];
    
    for (int layerIni = 0; layerIni < size[U]; layerIni += layerStep){
	int layerFin = layerIni + layerStep;
	if (layerFin > (int)size[U]) layerFin = size[U];

	ThreadRotate *pCThread = 
	    new ThreadRotate(wxTHREAD_DETACHED);
	pCThread->SetParameters(this, pTo, pFrom, size,
				S, T, U, layerIni, layerFin,
				mpRotTableFloatX, mpRotTableFloatY,
				mpRotTableFloatZ);
	WaitForThreadsAndRun(pCThread);
    }
    
    WaitAllThreadsEnd();

    GbGLLockForMultithread = FALSE;

    return TRUE;
}

// ----

wxString DataRotate_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("DataRotate::") <<
	mpCData->mCbValid.mbBody;
    
    if (GetPCData()->mCbValid.mbBody == FALSE)
	strStatus<< wxT("::");
    else
	strStatus << wxT(":") <<
	    GetPCData()->mCAxis.mBody << wxT(":") << 
	    mpCData->GetPInterpolated()[0] << wxT("::");
    
    return strStatus;
}

MACRO_SerializeConfP(DataRotate)
MACRO_GetNewPReadDataBlock(DataRotate)
