// (setq c-basic-offset 4)

#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/clsDF3Sequence.h"
#include "../MAINCANVAS/clsFilterPanels.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../POVRAY/clsPovFog.h"
#include "../POVRAY/clsPovIsosurface.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsButtonEX.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "clsFilterBase.h"
#include "clsFilterCtrl.h"
#include "clsClipping.h"

#include "../MYAPI/MsgBox.h"

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#ifdef WIN32
#define WXENDL wxT("\r\n")
#else
#define WXENDL wxT("\n")
#endif

FilterGUI_cls::
FilterGUI_cls(FilterBase_cls *pCParent){
    mpCParentBase = pCParent;
    mpCMNumPanel = NULL;
}

FilterGUI_cls::
~FilterGUI_cls(){
    if (mpCMNumPanel != NULL){
	delete mpCheckBoxValid;
	delete mpBTNShow;
	delete mpCMNumPanel;
    }
}

MultiNumericalInputsPanel_cls* FilterGUI_cls::
InitializeMNumPanel(wxWindow *pCParentWindow){
    mpCMNumPanel = new MultiNumericalInputsPanel_cls
	(&mpCMNumPanel, pCParentWindow, wxID_ANY);
    mpCMNumPanel->SetDefaultButtonType(_MN_NO_BUTTONS);
    wxString strForMinimize = wxT("[+]");
    SetFilterBaseCommonSettings(strForMinimize + mpCParentBase->mstrLabel,
				&mpCParentBase->GetPData()->mCbValid.mbBody);
    return mpCMNumPanel;
}

void FilterGUI_cls::
SetFilterBaseCommonSettings(wxString strTitle, bool *pbValid){
    int boxWidth = 2;
    mpCMNumPanel->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mpCMNumPanel->AssignValidCheckBoxBoth(mpCheckBoxValid, wxID_ANY,
					  pbValid);
    mpCMNumPanel->GetPLastVCPointer()->mItemWidth = 16;

    wxString strForMinimize = wxT("[+]");
    strForMinimize << mpCParentBase->GetPData()->mCstrCustomName.msBody;
    mpCMNumPanel->AssignMinimizeButtonBoth(mpBTNShow, wxID_ANY,
					   strForMinimize/*strTitle*/);
    mpCMNumPanel->mpBTNMinimize->Connect
	(mpCMNumPanel->mpBTNMinimize->GetId(), wxEVT_RIGHT_DOWN,
	 wxMouseEventHandler(MultiNumericalInputsPanel_cls::OnMouse),
	 NULL, mpCMNumPanel->GetEventHandler());
    
    mpCMNumPanel->GetParent()->
	Connect(mpCMNumPanel->mpCheckBoxValid->GetId(),
		wxEVT_COMMAND_CHECKBOX_CLICKED,
		wxCommandEventHandler(FilterPanels_cls::OnCheckBoxValid));

    mpCMNumPanel->GetParent()->
	Connect(mpCMNumPanel->mpBTNMinimize->GetId(),
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(FilterPanels_cls::OnButtonShowItem));

    mpCMNumPanel->SetFuncHook(&FilterCtrl_cls::FuncHookValidityCheckbox,
			      _MN_HOOK_CHECKBOX);
    mpCMNumPanel->SetFuncHook(&FilterCtrl_cls::FuncHookMouse, _MN_HOOK_MOUSE);
    mpCMNumPanel->SetFuncHook(&FilterCtrl_cls::FuncHookShow, _MN_HOOK_SHOW);
}

void FilterGUI_cls::
SetValueFLOAT(int index, float value){
    mpCMNumPanel->mVCPointers[index].SetValueFloat(value);
}

void FilterGUI_cls::
SetValueINT(int index, int value){
    mpCMNumPanel->mVCPointers[index].SetValueInt(value);
}

void FilterGUI_cls::
SetValueBOOL(int index, bool value){
    mpCMNumPanel->mVCPointers[index].SetValueBool(value);
}

DF3SeqCtrl_cls*  FilterGUI_cls::mspCDF3SeqCtrlLink = NULL;

// -----------------------

FilterData_cls::
FilterData_cls(){
    mCConfList.Clear();
    mCConfList.Add(&mCbShow,     wxT("FilterShow="), L"1");
    mCConfList.Add(&mCbValid,    wxT("FilterValid="), L"1");
    mCConfList.Add(&mCstrCustomName, wxT("FilterName=")); // lȂ
}

FilterData_cls::
~FilterData_cls(){
}

void FilterData_cls::
ReDefineListBase(){
    mCConfList.Clear();
    mCConfList.Add(&mCbShow,  wxT("FilterShow="));
    mCConfList.Add(&mCbValid,  wxT("FilterValid="));
    mCConfList.Add(&mCstrCustomName, wxT("FilterName="));
}

wxString FilterData_cls::
SerializeConfPolymorph(){
    return wxT("");
}

wxString FilterData_cls::
SerializeConf(int mFilterType){
    mCConfList.mbSetWithInitValue = FALSE;
    ReDefineList();

    // FIX_ME!! LINUX ł wistream œ{͓ǂ߂ȂB
    // [Aʂ̎iɑȂƍsȂcH
#ifndef WIN32
    wxString strName = mCstrCustomName.msBody;
    mCstrCustomName.msBody = wxT("!!SKIP!!");
#endif
    wxString strConf = 	msDataBlockTagStr[mFilterType] + WXENDL +
	mCConfList.SerializeAll() +
	SerializeConfPolymorph() + 
	wxT("--END") + WXENDL;
#ifndef WIN32
    mCstrCustomName.msBody = strName;
#endif
    return strConf;

}

void FilterData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
}

void FilterData_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    mCConfList.ReadStrLineAll(strLine);
    ReadStrLinePolymorph(istream, strLine);
}

wxString FilterData_cls::msDataBlockTagStr[_MAX_FILTER] = {
    wxT("--Isosurface"),
    wxT("--Fog"),
    wxT("--BoxCopy"),
    wxT("--ClipBox"),
    wxT("--Volume"),
    wxT("--BlurGauss"),
    wxT("--LevelAndGamma"),
    wxT("--InvDiffuse"),
    wxT("--FixPosition"),
    wxT("--DataRotate"),
    wxT("--LayerColorCurve"),
    wxT("--CrossSection"),
    wxT("--PovObject"),
    wxT("--CenterOfMass"),
    wxT("--PovDiffObject"),
    wxT("--MorphOpenClose"),
    wxT("--PovThreeView"),
    wxT("--PovStreamLine"),
    wxT("--PovVelocityField")
};

// --------

FilterBase_cls::
FilterBase_cls(FilterCtrl_cls *pCParentCtrl){
    mpCParentCtrl = pCParentCtrl;
    mbProcessData = FALSE;
    mbWithParameter = FALSE;
    mbWithSelect = FALSE;
    mbWithTexture = FALSE;
}

FilterBase_cls::
~FilterBase_cls(){
    if (mpCGUIPanel != NULL) delete mpCGUIPanel;
}

void FilterBase_cls::
ConstructorCommon(FilterCtrl_cls *pCParentCtrl, wxString strDefaultLabel,
		  int filterType){
    mFilterType = filterType;
    mstrId = pCParentCtrl->GetIdString(mFilterType);
    mstrLabel << strDefaultLabel << mstrId;
    if (GetPData()->mCstrCustomName.msBody == wxT("NONE"))
	GetPData()->mCstrCustomName.msBody = mstrLabel;
}

int FilterBase_cls::
GetDF3Index(){
    return mpCParentCtrl->mpCDF3SeqParent->mSelfIndex;
}

int FilterBase_cls::
GetRemappedIndex(){
    return mpCParentCtrl->mspCDF3SeqCtrlLink->GetRemappedIndex(GetDF3Index());
}

double FilterBase_cls::
GetRemappedIndexDouble(){
    return mpCParentCtrl->mspCDF3SeqCtrlLink->GetRemappedIndexDouble(GetDF3Index());
}

bool FilterBase_cls::
IsClippingEmpty(){
    double min[3] = {0, 0 ,0};
    double max[3] = {1, 1, 1};
    GetClippingInformation(min, max);
    LOOP_XYZ if (min[xyz] >= max[xyz]) return TRUE;
    return FALSE;
}

void FilterBase_cls::
GetClippingInformation(double min[3], double max[3], wxString queryStr){
    int selfIndex = mpCParentCtrl->GetSelfIndex(this);
    std::vector < std::vector < double > > VClipMinNouse, VClipMaxNouse;
    if (queryStr == wxT("")) queryStr = GetPData()->mCstrCustomName.msBody;
    mpCParentCtrl->GetClippingInformation(queryStr, min, max, &VClipMinNouse, &VClipMaxNouse);
}

wxString FilterBase_cls::
GetLocaleIndependentStrLabel(wxString strLabel){
    strLabel.Replace(_("Solid"), wxT("Solid"));
    strLabel.Replace(_("Empty"), wxT("Empty"));
    strLabel.Replace(_("Volume"), wxT("Volume"));
    strLabel.Replace(_("Isosurface"), wxT("Isosurface"));
    strLabel.Replace(_("FixPosition"), wxT("FixPosition"));
    strLabel.Replace(_("Global"), wxT("Global"));
    strLabel.Replace(_("Object"), wxT("Object"));
    strLabel.Replace(_("Streamline"), wxT("Streamline"));
    strLabel.Replace(_("VelocityField"), wxT("VelocityField"));
    return strLabel;
}

wxString FilterBase_cls::
GetLocaleStrLabel(wxString strLabel){
    strLabel.Replace(wxT("Solid"), _("Solid"));
    strLabel.Replace(wxT("Empty"), _("Empty"));
    strLabel.Replace(wxT("Volume"), _("Volume"));
    strLabel.Replace(wxT("Isosurface"), _("Isosurface"));
    strLabel.Replace(wxT("FixPosition"), _("FixPosition"));
    strLabel.Replace(wxT("Global"), _("Global"));
    strLabel.Replace(wxT("Object"), _("Object"));
    strLabel.Replace(wxT("Streamline"), _("Streamline"));
    strLabel.Replace(wxT("VelocityField"), _("VelocityField"));
    return strLabel;
}

void FilterBase_cls::
SelectPanel(bool bSelected, bool bExclusive, bool bTerminate){
    mpCGUIPanel->mbSelected = bSelected;
    mpCGUIPanel->mpCMNumPanel->SetHilight(bSelected);
    if (bSelected == TRUE && bTerminate == FALSE) {// TLPanelƂ̃[vif
	mspCDF3SeqGUI->DeactivateAllPanel();
	mspCDF3SeqGUI->ActivatePanel(mstrLabel, bExclusive, TRUE);
    }
}

void FilterBase_cls::
RefreshTop(){
    mpCParentCtrl->FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

wxString FilterBase_cls::
GetTargetLabelStr(){
    if (mbWithSelect == FALSE) return wxT("");
    else 
	return ((FilterWithTargetGUI_cls*)mpCGUIPanel)->GetTargetLabelStr();
}

void FilterBase_cls::
AdvancedSettingCommon(bool bRefresh){
    SetMNumAdvanced(mpCGUIPanel->mpCMNumPanel);
    mpCMNumAdvanced->UpdateValues();
    mpCMNumAdvanced->ShowModal();
    delete mpCMNumAdvanced;
    if (bRefresh == TRUE)
	FilterCtrl_cls::FuncHookRefreshTop(mpCGUIPanel->mpCMNumPanel, NULL);
}

void FilterBase_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
}

void FilterBase_cls::
AddFilterItemPolymorph(wxBoxSizer *pBoxSizer, FilterPanels_cls *pCFilterPanels,
		       unsigned int index){
}

void FilterBase_cls::
AddFilterItem(FilterPanels_cls *pCFilterPanels, unsigned int index){

    if (pCFilterPanels->IsItemExist(mpCGUIPanel->mpCMNumPanel) == TRUE){
	MsgBox(pCFilterPanels, "Exist", "!!");
	return;
    }

    mpBoxSizer = (wxBoxSizer*)pCFilterPanels->GetSizer();
	//pCFilterPanels->mpBoxSizer;

    pCFilterPanels->Freeze();
    AddFilterItemPolymorph(mpBoxSizer, pCFilterPanels, index);
    pCFilterPanels->AddItemList(mpCGUIPanel->mpBTNShow, 
				/*mpCGUIPanel->mpObjectForDelete,*/
				mpCGUIPanel->mpCParentBase,
				mpCGUIPanel->mpCMNumPanel);
    mpCGUIPanel->mpCMNumPanel->ShowPanel(GetPData()->mCbShow.mbBody);
    mpBoxSizer->Layout();
    pCFilterPanels->Thaw();
    if (mspCDF3SeqGUI != NULL) mspCDF3SeqGUI->UpdatePanelLayout();
    
}

void FilterBase_cls::
RemoveFilterItem(){
    //mpBoxSizer->Detach(mpCGUIPanel->mpCMNumPanel);
    //mpBoxSizer->Detach(mpCGUIPanel->mpBTNShow);
    //mpBoxSizer->Detach(mpCGUIPanel->mpCheckBoxValid);

    //tH[JX܂܏Ɩ肪鎞B
    mpCGUIPanel->mpCMNumPanel->SetFocusIgnoringChildren(); 
    delete mpCGUIPanel;
    mpCGUIPanel = NULL;
    if (mspCDF3SeqGUI != NULL) mspCDF3SeqGUI->UpdatePanelLayout();
}

/*
void FilterBase_cls::
UpdateList(){
    //((FilterWithTargetGUI_cls*)mpCGUIPanel)->UpdateList();
}
*/

void FilterBase_cls::
UpdateCommonData(){
    GetPData()->mCbShow.mbBody = mpCGUIPanel->mpCMNumPanel->IsShownOnScreen();
}

DF3SeqGUI_cls*   FilterBase_cls::mspCDF3SeqGUI = NULL;
DF3SeqCtrl_cls*  FilterBase_cls::mspCDF3SeqCtrlLink = NULL;
Preference_cls*  FilterBase_cls::mspCPreferenceLink = NULL;
PovRay_cls*      FilterBase_cls::mspCPovRayLink = NULL;
TreeBoxRule_cls* FilterBase_cls::mspCTreeBoxRuleLink = NULL;
double           FilterBase_cls::msABoxScale[3] = {1.0, 1.0, 1.0};
float            FilterBase_cls::msBestLineWidth = 1.0f;
bool             FilterBase_cls::msbConstructedFromMenu = FALSE;
PositionAndDirection_cls *FilterBase_cls::mspCCamPos = NULL;

bool FilterObject_cls::msbRetex = FALSE;
