// (setq c-basic-offset 4)

#include "clsFilterCtrl.h"
#include "clsBlurGauss.h"
#include "clsInvDiffuse.h"
#include "clsLevelAndGamma.h"
#include "clsClipping.h"
#include "clsFixPosition.h"
#include "clsDataRotate.h"
#include "clsLayerColorCurve.h"
#include "ANALYZER/clsCenterOfMass.h"
#include "ANALYZER/clsMorphOpenClose.h"
#include "FILTERUNDO/FilterUndo.h"

#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../DF3/GL/clsPreviewProperty.h"
#include "../DRAWBOX/clsBoxCopy.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../MAINCANVAS/clsFilterPanels.h"
#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../DRAWBOX/clsBoxCopy.h"
#include "../PARAMETER/clsParametersPanel.h"
#include "../PARAMETER/clsTimeLinePanel.h"
#include "../POVRAY/clsPovDiffObj.h"
#include "../POVRAY/clsPovFog.h"
#include "../POVRAY/clsPovCrossSection.h"
#include "../POVRAY/clsPovIsosurface.h"
#include "../POVRAY/clsPovObject.h"
#include "../POVRAY/clsPovVolume.h"
#include "../POVRAY/clsPovThreeView.h"
#include "../POVRAY/clsTreeBoxRule.h"
#include "../POVRAY/VELOCITY/clsPovStreamLine.h"
#include "../POVRAY/VELOCITY/clsPovVelocityField.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include "../MYAPI/MsgBox.h"

#include "../../clsTimeCtrl.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

extern long GTestVariableLong;
extern int GTestVariableInt;

#define X 0
#define Y 1
#define Z 2

#define COMPMAX(a,b) ((a)>(b)?(a):(b))
#define COMPMIN(a,b) ((a)<(b)?(a):(b))


#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

FilterCtrl_cls::
FilterCtrl_cls(DF3Sequence_cls *pCDF3SeqParent){
    mpCDF3SeqParent = pCDF3SeqParent;
    mSelfDF3Index = mpCDF3SeqParent->mSelfIndex;
    mbBufferAllocated = FALSE;
    mBufferAllocatedSize = 0;
    mVstrFilterStatusSum.resize(_DF3_SEQ_MAX);
}

FilterCtrl_cls::
~FilterCtrl_cls(){
    for (std::list< FilterBase_cls* >::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	delete (*itr);

    if (mbBufferAllocated == TRUE){
	delete mpBuffer[0];
	delete mpBuffer[1];
    }
    
    //delete FilterGUI_cls::mspPopupMenu;
}

void FilterCtrl_cls::
SetFilterPanelsLinkMutual(FilterPanels_cls *pCFilterPanels){
    mpCFilterPanelsLink = pCFilterPanels;
    mpCFilterPanelsLink->SetFilterCtrlLink(this);
}

int FilterCtrl_cls::
GetIndexFromTag(wxString strTag){
    int index = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mstrLabel == strTag) return index;
	index++;
    }
    return -1;
}

int FilterCtrl_cls::
GetSelfIndex(FilterBase_cls *pC){
    int index = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr) == pC) return index;
	index++;
    }
    return -1;
}

wxString FilterCtrl_cls::
GetIdString(unsigned int type){
    int id = 0;
    wxString strId; strId << mSelfDF3Index << id;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end();){
	if ((*itr)->mFilterType == type && (*itr)->mstrId  == strId){
	    id++;
	    strId = wxT(""); strId << mSelfDF3Index << id;
	    itr = mVpCFilter.begin();
	    continue; 
	}
	itr++;
    }
    return strId;
}

wxString FilterCtrl_cls::
GetIdStringFromTag(wxString strTag){
    std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mstrLabel == strTag)
	    return (*itr)->mstrId;
    }
    return wxT("");
}

std::vector < Parameters_cls* > FilterCtrl_cls::
GetVpCPrimParam(){
    std::vector < Parameters_cls* > VpCPrimParam;
    std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    VpCPrimParam.push_back
		( ((PovIsosurface_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION)
	    VpCPrimParam.push_back
		( ((PovCrossSection_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_DATA_ROTATE)
	    VpCPrimParam.push_back
		( ((DataRotate_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    VpCPrimParam.push_back
		( ((PovObject_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_POV_DIFFOBJECT)
	    VpCPrimParam.push_back
		( ((PovDiffObject_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_THREEVIEW)
	    VpCPrimParam.push_back
		( ((PovThreeView_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_STREAMLINE)
	    VpCPrimParam.push_back
		( ((PovStreamLine_cls*)(*itr))->mpCData->GetPParamActive());
	if ((*itr)->mFilterType == _FILTER_VELOCITY_FIELD)
	    VpCPrimParam.push_back
		( ((PovVelField_cls*)(*itr))->mpCData->GetPParamActive());
    }
    return VpCPrimParam;
}

void FilterCtrl_cls::
InitializeStaticMembers(Preference_cls  *pCPreference,
			DF3SeqCtrl_cls  *pCDF3SeqCtrl, 
			PovRay_cls* pCPov, TreeBoxRule_cls *pCTreeRule,
			PreviewProperty_cls* pCPrevProp){
    mspCPreferenceLink = pCPreference;
    mspCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mspCPreviewPropertyLink = pCPrevProp;
    //PovIsosurface_cls::SetMNumStaticMembers(wxGetTopLevelParent((wxPanel*)pCDF3SeqCtrl->mpCGUI));
    FilterBase_cls::mspCDF3SeqGUI = pCDF3SeqCtrl->mpCGUI;
    FilterBase_cls::mspCDF3SeqCtrlLink = pCDF3SeqCtrl;
    FilterBase_cls::mspCPovRayLink = pCPov;
    FilterBase_cls::mspCTreeBoxRuleLink = pCTreeRule;
    FilterGUI_cls::mspCDF3SeqCtrlLink = pCDF3SeqCtrl;

    for (int i = 0; i < _DF3_MAX_SLOT; i++){
	PovIsosurface_cls::InitializeStaticMember(i);
	FilterClipping_cls::InitializeStaticMember(i);
	LevelAndGamma_cls::InitializeStaticMember(i);
	FilterFixPosition_cls::InitializeStaticMember(i);
	DataRotate_cls::InitializeStaticMember(i);
	LayerColorCurve_cls::InitializeStaticMember(i);
	PovCrossSection_cls::InitializeStaticMember(i);
	PovObject_cls::InitializeStaticMember(i);
	PovDiffObject_cls::InitializeStaticMember(i);
	PovThreeView_cls::InitializeStaticMember(i);
	PovStreamLine_cls::InitializeStaticMember(i);
	PovVelField_cls::InitializeStaticMember(i);
    }

    for (int i = 0; i < _DF3_SEQ_MAX; i++) 
	PovStreamLine_cls::msAIndexList[i] = i;

    FilterBase_cls::mspCPreferenceLink = pCPreference;
}

bool FilterCtrl_cls::
DoesFilterExist(int type, int condition, FilterBase_cls **ppFilter, int skip){
    int count = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mpCGUIPanel->mFilterType == type)
	    if (condition == _FILTER_INCLUDE_INVALID ||
		(*itr)->GetPData()->mCbValid.mbBody == TRUE){
		count++;
		if (count > skip){
		    if (ppFilter != NULL) *ppFilter = *itr;
		    return TRUE;
		}
	    }
    return FALSE;
}

bool FilterCtrl_cls::
DoesSameFilterNameExist(wxString strTag, FilterBase_cls *pCException){
    bool bExist = FALSE;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr) == pCException) continue;
	if ((*itr)->mstrLabel == strTag) bExist = TRUE;
	if ((*itr)->GetPData()->mCstrCustomName.msBody == strTag) bExist = TRUE;
    }
    return bExist;
}

bool FilterCtrl_cls::
DoesVolumeBoxRequired(){
    bool bRet = FALSE;
    bRet |= GetVolumeValidity();

    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mpCGUIPanel->mFilterType == _FILTER_FOG){
	    if (((PovFog_cls*)(*itr))->mCData.mCConfTargetStr.msBody.Matches(_("Volume")))
		bRet = TRUE;
	}
    }	
    
    return bRet;
}

FilterBase_cls* FilterCtrl_cls::
GetPFilterBase(int type){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mpCGUIPanel->mFilterType == type)  return (*itr);
    return NULL;
}

FilterData_cls* FilterCtrl_cls::
GetPFilterData(int type){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mpCGUIPanel->mFilterType == type)
	    return (*itr)->GetPData();
    return NULL;
}

std::list<FilterBase_cls*>::iterator FilterCtrl_cls::
GetIterator(int index){
    int i = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if (i == index) return itr;
	i++;
    }
    return mVpCFilter.end();
}

void FilterCtrl_cls::
ClearData(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    ((PovIsosurface_cls*)(*itr))->ClearIsosurface();
	
    }
}

void FilterCtrl_cls::
ClearFilter(){
    unsigned int filterNum = (unsigned int)mVpCFilter.size();
    for (unsigned int index = 0; index < filterNum; index++)
	DeleteFilterItem(0);
}

#define MACRO_DELETE(Filter,TAG)\
    case _FILTER_##TAG:							\
    GCUndoList.SetNewCommand						\
    (new ComFilterParamRemove_cls(this, index),				\
     new Com##Filter##ParamAdd_cls(this, index, (*itr)->GetPData()));	\
    break;

void FilterCtrl_cls::
DeleteFilter(void *pObjectForDelete){
    int index = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if (pObjectForDelete == (*itr)/*->mpCGUIPanel->mpCParent*/){
	    // ADHOC::SpinCtrlFocusďƂ܂̂ŐɒDĂ
	    (*itr)->mpCGUIPanel->mpCMNumPanel->SetFocusIgnoringChildren();
	    switch((*itr)->mFilterType){
		//MACRO_DELETE(Isosurface,ISOSURFACE);
	    case _FILTER_ISOSURFACE:
		((PovIsosurface_cls*)(*itr))->UpdateTargetInfo();
		GCUndoList.SetNewCommand
		    (new ComFilterParamRemove_cls(this, index),
		     new ComIsosurfaceParamAdd_cls(this, index, (*itr)->GetPData()));
		break;
		MACRO_DELETE(CrossSection,CROSS_SECTION);
		//MACRO_DELETE(Fog,FOG);
	    case _FILTER_FOG:
		((PovFog_cls*)(*itr))->UpdateTargetInfo();
		GCUndoList.SetNewCommand
		    (new ComFilterParamRemove_cls(this, index),
		     new ComFogParamAdd_cls(this, index, (*itr)->GetPData()));
		break;
		MACRO_DELETE(PovObject,POV_OBJECT);
		MACRO_DELETE(BoxCopy,BOXCOPY);
		//MACRO_DELETE(Clipping,CLIPPING);
	    case _FILTER_CLIPPING:
		((FilterClipping_cls*)(*itr))->UpdateTargetInfo();
		GCUndoList.SetNewCommand
		    (new ComFilterParamRemove_cls(this, index),
		     new ComClippingParamAdd_cls(this, index, (*itr)->GetPData()));
		break;
		MACRO_DELETE(Volume,VOLUME);
		MACRO_DELETE(BlurGauss,BLURING);
		MACRO_DELETE(InvDiffuse,INV_DIFFUSE);
		MACRO_DELETE(LevelAndGamma,LEVEL_GAMMA);
		MACRO_DELETE(FixPosition,FIX_POSITION);
		MACRO_DELETE(DataRotate,DATA_ROTATE);
		MACRO_DELETE(LayerColorCurve,LAYER_COLOR_CURVE);
		MACRO_DELETE(CenterOfMass,CENTEROFMASS);
		MACRO_DELETE(MorphOpenClose, MORPH_OPENCLOSE)
		MACRO_DELETE(PovDiffObject,POV_DIFFOBJECT);
		MACRO_DELETE(ThreeView,THREEVIEW);
		MACRO_DELETE(PovStreamLine, STREAMLINE);
		MACRO_DELETE(PovVelField, VELOCITY_FIELD);
	    }
	    
	    GCUndoList.ExecCurrent();
	    break;
	}
	index++;
    }
}

void FilterCtrl_cls::
Rename(){
    FilterBase_cls *pSelectedItem = GetPSelectedItem();
    if (pSelectedItem == NULL) return;
    
    wxString strName = pSelectedItem->GetPData()->mCstrCustomName.msBody;
    wxString strMsg = _("Change from 1");
    strMsg << strName << _("Change from 2");
    wxPoint mousePoint = wxGetMousePosition();
    wxString strNewLabel = wxGetTextFromUser
	(strMsg, _("Rename"), 
	 strName, pSelectedItem->mpCGUIPanel->mpCMNumPanel,
	 mousePoint.x-120, mousePoint.y-24, FALSE);
    if (strNewLabel == wxT("")) return;
    while (mspCDF3SeqCtrlLink->DoesSameFilterNameExistInAllDF3(strNewLabel, pSelectedItem) == TRUE)
	strNewLabel += wxT("@");
    
    pSelectedItem->GetPData()->mCstrCustomName.msBody = strNewLabel;
    UpdateList();
    
    wxString strForMinimize = wxT("[+]");
    strForMinimize << strNewLabel;
    pSelectedItem->mpCGUIPanel->mpCMNumPanel->
	SetMinimizeButtonLabel(strForMinimize);
    if (pSelectedItem->mbWithParameter == TRUE){
	((FilterWithParameter_cls*)pSelectedItem)->mpCParamPanel->mpCTimeLine->mstrCustomName = strNewLabel;
	((FilterWithParameter_cls*)pSelectedItem)->mpCParamPanel->mpCTimeLine->Refresh(TRUE);
    }

}

void FilterCtrl_cls::
AdvancedSetting(){
    FilterBase_cls *pSelectedItem = GetPSelectedItem();
    if (pSelectedItem == NULL) return;
    switch(pSelectedItem->mFilterType){
	//RefreshȂ
    case _FILTER_CLIPPING: 
    	pSelectedItem->AdvancedSettingCommon();	break;
    case _FILTER_FIX_POSITION:
	pSelectedItem->AdvancedSettingCommon(); break;
	//Refresh
    case _FILTER_LEVEL_GAMMA:
    case _FILTER_POV_OBJECT:
    case _FILTER_DATA_ROTATE: 
    case _FILTER_STREAMLINE:
    case _FILTER_VELOCITY_FIELD:
	pSelectedItem->AdvancedSettingCommon(TRUE); break;
	//̂ق̌ʂ
    case _FILTER_FOG:
	PovFog_cls::mspCConnectedToDLG = 
	    ((PovFog_cls*)pSelectedItem);
	pSelectedItem->AdvancedSettingCommon(TRUE); break;
	break;
    case _FILTER_VOLUME:	
	PovVolume_cls::mspCConnectedToDLG = 
	    ((PovVolume_cls*)pSelectedItem);
	pSelectedItem->AdvancedSettingCommon();
	break;
    case _FILTER_ISOSURFACE: 
	PovIsosurface_cls::mspCConnectedToDLG = 
	    ((PovIsosurface_cls*)pSelectedItem);
	((PovIsosurface_cls*)pSelectedItem)->AdvancedSetting();
	break;
    case _FILTER_CROSS_SECTION:
	((PovCrossSection_cls*)pSelectedItem)->AdvancedSetting();
	break;
    case _FILTER_THREEVIEW:
	((PovThreeView_cls*)pSelectedItem)->AdvancedSetting();
	break;
    }
}

void FilterCtrl_cls::
SwapFilterList(int index, int direction){
    std::list<FilterBase_cls*>::iterator itr = GetIterator(index);
    if (direction == -1 && itr == mVpCFilter.begin()) return;
    std::list<FilterBase_cls*>::iterator itr2 = itr;
    if (direction == -1) itr2--;
    if (direction == +1) itr2++;
    if (itr2 == mVpCFilter.end()) return;
    std::swap(*itr, *itr2);
}

void FilterCtrl_cls::
SwapFilterExecute(int index, int direction){
    SwapFilterList(index, direction);
    mpCFilterPanelsLink->SwapItemList(index, direction);
    mpCFilterPanelsLink->UpdateSizer();
    mpCFilterPanelsLink->UpdatePanelLayout();
    mpCFilterPanelsLink->Refresh(TRUE);
}

void FilterCtrl_cls::
SwapFilter(int direction){
    int index = GetSelectedItemIndex();
    if (index < 0) return; 
    if (index + direction < 0) return;
    if (index + direction >= (int)mVpCFilter.size()) return;
    GCUndoList.SetNewCommand
	(new ComFilterParamSwap_cls(this, index, direction),
	 new ComFilterParamSwap_cls(this, index+direction, -direction));
    GCUndoList.ExecCurrent();
}

void FilterCtrl_cls::
ChangeConfFLOATExecute(wxString mstrTag, int mIndex, float mValue){
    FilterBase_cls *pCFilterBase = GetPFromTag(mstrTag);

    pCFilterBase->mpCGUIPanel->SetValueFLOAT(mIndex, mValue);
    pCFilterBase->mpCGUIPanel->mpCMNumPanel->UpdateValuesAll();
}

void FilterCtrl_cls::
ChangeConfINTExecute(wxString mstrTag, int mIndex, int mValue){
    FilterBase_cls *pCFilterBase = GetPFromTag(mstrTag);

    pCFilterBase->mpCGUIPanel->SetValueINT(mIndex, mValue);
    pCFilterBase->mpCGUIPanel->mpCMNumPanel->UpdateValuesAll();
    pCFilterBase->mpCGUIPanel->HideAndShowItems();
}

void FilterCtrl_cls::
ChangeConfBOOLExecute(wxString mstrTag, int mIndex, bool mValue){
    FilterBase_cls *pCFilterBase = GetPFromTag(mstrTag);

    pCFilterBase->mpCGUIPanel->SetValueBOOL(mIndex, mValue);
    pCFilterBase->mpCGUIPanel->mpCMNumPanel->UpdateValuesAll();
    pCFilterBase->mpCGUIPanel->HideAndShowItems();
}


int FilterCtrl_cls::
GetSelectedItemIndex(){
    int index = 0;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mpCGUIPanel->mbSelected == TRUE)
	    return index;
	index++;
    }
    return -1;
}

FilterBase_cls* FilterCtrl_cls::
GetPSelectedItem(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mpCGUIPanel->mbSelected == TRUE)
	    return (*itr);
    }
    return NULL;
}

void* FilterCtrl_cls::
GetPSelectedItemDeleteObject(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mpCGUIPanel->mbSelected == TRUE)
	    return (*itr)/*->mpCGUIPanel->mpObjectForDelete*/;

    return NULL;
}

void FilterCtrl_cls::
UpdateTarget(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_CLIPPING)
	    ((FilterClipping_cls*)(*itr))->UpdateTarget();
	if ((*itr)->mFilterType == _FILTER_FOG)
	    ((PovFog_cls*)(*itr))->UpdateTarget();
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE){
	    ((PovIsosurface_cls*)(*itr))->UpdateTarget();
	}
    }
}

void FilterCtrl_cls::
UpdateColors(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    ((PovVolume_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION)
	    ((PovCrossSection_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    ((PovIsosurface_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    ((PovObject_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_THREEVIEW)
	    ((PovThreeView_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_STREAMLINE)
	    ((PovStreamLine_cls*)(*itr))->UpdateColors();
	if ((*itr)->mFilterType == _FILTER_VELOCITY_FIELD)
	    ((PovVelField_cls*)(*itr))->UpdateColors();
    }
}

void FilterCtrl_cls::
UpdateColorBar(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    ((PovVolumeGUI_cls*)(*itr)->mpCGUIPanel)->RefreshColorBar();
	if ((*itr)->mbWithTexture == TRUE)
	    ((FilterWithTextureGUI_cls*)(*itr)->mpCGUIPanel)->RefreshColorBar();
    }
}

void FilterCtrl_cls::
UpdateList(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mbWithSelect == TRUE){
	    ((FilterWithTargetGUI_cls*)(*itr)->mpCGUIPanel)->UpdateList();
	}
}

void FilterCtrl_cls::
UpdateParameters(double doubleIndex){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mbWithParameter == TRUE)
	    ((FilterWithParameter_cls*)(*itr))->UpdateParameters(doubleIndex);
}

/*
void FilterCtrl_cls::
UpdateHideAndShows(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_FOG)
	    ((PovFogGUI_cls*)(*itr)->mpCGUIPanel)->HideAndShowItems();
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    ((PovObjectGUI_cls*)(*itr)->mpCGUIPanel)->HideAndShowItems();
    }
}
*/

void FilterCtrl_cls::
DeselectFilter(wxString strLabel){
    std::list<FilterBase_cls*>::iterator itr;
    for (itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if (strLabel == (*itr)->mstrLabel || strLabel == wxT("all"))
	    (*itr)->SelectPanel(FALSE, FALSE);
}

void FilterCtrl_cls::
SelectFilter(wxString strLabel, bool bExclusive, bool bTerminate){
    std::list<FilterBase_cls*>::iterator itr, itr2;
    for (itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	 if (strLabel == (*itr)->mstrLabel)
	     for (itr2 = mVpCFilter.begin(); itr2 != mVpCFilter.end(); itr2++)
		 if (itr2 == itr) 
		     (*itr2)->SelectPanel(TRUE, bExclusive, bTerminate);
		 else             
		     (*itr2)->SelectPanel(FALSE, bExclusive, bTerminate);
}

void FilterCtrl_cls::
FilterCustomA(void *pObjectForCustom){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if (pObjectForCustom == (*itr)->mpCGUIPanel->mpObjectForCustomA){
	    (*itr)->mpCGUIPanel->CustomA();
	    break;
	}
}

void FilterCtrl_cls::
FilterCustomB(void *pObjectForCustom){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if (pObjectForCustom == (*itr)->mpCGUIPanel->mpObjectForCustomB){
	    (*itr)->mpCGUIPanel->CustomB();
	    break;
	}
}

void FilterCtrl_cls::
FilterCustomC(void *pObjectForCustom){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if (pObjectForCustom == (*itr)->mpCGUIPanel->mpObjectForCustomC){
	    (*itr)->mpCGUIPanel->CustomC();
	    break;
	}
}

// ------

wxString FilterCtrl_cls::
GetPovStrFog(){
    wxString strFog;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FOG)
	    strFog << ((PovFog_cls*)(*itr))->GetPovStrFog(PovVolume_cls::GetPovStrCommonDensityModifier(L"\t", 0));

    return strFog;
}

wxString FilterCtrl_cls::
GetPovStrGlobalObject(wxString strTag,
		      PositionAndDirection_cls CCamPos,
		      std::vector < wxString > VstrFilePath){
    wxString strGlobalObject;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    strGlobalObject << ((PovObject_cls*)(*itr))->
		GetPovStrObject(_POV_OBJECT_GLOBAL_CALL, strTag, CCamPos,
				VstrFilePath[mSelfDF3Index]);
    }
    return strGlobalObject;
}

// ^C}bvpɂQDF3`
wxString FilterCtrl_cls::
GetPovStrDensityFile(wxString strTag, std::vector < wxString > VstrFilePath,
		     bool bTimeRemap, float remapFactor){
    wxString strPovDensityFile;
    
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME) 
	    strPovDensityFile << ((PovVolume_cls*)(*itr))->
		GetPovStrDensityFile(strTag, VstrFilePath, 
				     bTimeRemap, remapFactor);

    return strPovDensityFile;
}

void FilterCtrl_cls::
GetPovStrDProfile(std::wfstream &ofstream,
		  wxString strTag, std::vector < wxString > VstrFilePath,
		  DrawingBlock_cls *pCDrawingBlock,
		  bool bTimeRemap, float remapFactor){

    wxString strPovDProfile;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME) 
	    strPovDProfile << ((PovVolume_cls*)(*itr))->
		GetPovStrDProfile(strTag, VstrFilePath,
				  pCDrawingBlock, bTimeRemap, remapFactor);
    
    ofstream << wxConvFile.cWX2WC(strPovDProfile);
}

void FilterCtrl_cls::
GetPovStrDensityBox(std::wfstream &ofstream,
		    double baseAlpha, double doubleIndex,
		    wxString strTag, wxString strFilePath,
		    wxString strGeometry, DrawingBlock_cls *pCDrawingBlock,
		    std::vector < wxString > *pVstrInteriorTag,
		    PositionAndDirection_cls CCamLocalCoord,
		    int frame){
    
    unsigned int treeDivideParam = pCDrawingBlock->mTreeDivideParam;
    unsigned int depth = pCDrawingBlock->mDepth;
    wxString strPovDensityBox;
    
    strPovDensityBox << wxT("// Data") << mSelfDF3Index << wxT("\n");

    double clipForPeriodicBoundary[3];
    LOOP_XYZ 
	if (mpCDF3SeqParent->GetDataSize(xyz) == 0)
	    clipForPeriodicBoundary[xyz] = 1;
	else
	    clipForPeriodicBoundary[xyz] = 
		1.0 - 1.0/mpCDF3SeqParent->GetDataSize(xyz);
    
    double min[3] = {0, 0, 0}, max[3] = {1, 1, 1};
    int selfIndex = -1;
    std::vector < std::vector < double > > VClipMin, VClipMax;
    bool bClipping = FALSE;
    for (int i = 0; i < _DF3_MAX_SLOT; i++)
	bClipping |= GetClippingInformation(_T("Volume"), min, max, &VClipMin, &VClipMax);
    
    wxString strInteriorTag;

    // {[͖{̂ƂIsosurfaceĂ΂邱Ƃ̂ŁA
    // ɏo͂Ă
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_VOLUME) {
	    strPovDensityBox << ((PovVolume_cls*)(*itr))->
		GetPovStrVolumeBox(strTag, strGeometry, 
				   treeDivideParam, depth,
				   baseAlpha, doubleIndex, pCDrawingBlock);
	    
	    if (((PovVolume_cls*)(*itr))->mCData.mCbValid.mbBody == TRUE){
		strInteriorTag << L"\tINTERIOR_" << (*itr)->mstrId + strTag << L"()" << wxT("\n");
		(*pVstrInteriorTag).push_back(strInteriorTag); 
	    }
	}
    }
    
    ofstream << wxConvFile.cWX2WC(strPovDensityBox);

    // Object
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	
	wxString strPovObject;
	
	switch ((*itr)->mFilterType){
	case _FILTER_ISOSURFACE:
	    ((PovIsosurface_cls*)(*itr))->
		GetPovStrIsosurfBox(&strPovObject,
				    strTag, strFilePath,
				    strGeometry, treeDivideParam, depth,
				    baseAlpha, clipForPeriodicBoundary,
				    pCDrawingBlock);
	    break;
	case _FILTER_CROSS_SECTION:
	    strPovObject << ((PovCrossSection_cls*)(*itr))->
		GetPovStrCrossSection
		(strTag, strFilePath, strGeometry, treeDivideParam, depth,
		 baseAlpha, clipForPeriodicBoundary, CCamLocalCoord, frame,
		 pCDrawingBlock);
	    break;
	case _FILTER_POV_OBJECT:
	    strPovObject << ((PovObject_cls*)(*itr))->
		GetPovStrObject(_POV_OBJECT_LOCAL_CALL, strTag, CCamLocalCoord,
				strFilePath);
	    break;
	case _FILTER_THREEVIEW:
	    strPovObject << ((PovThreeView_cls*)(*itr))->
		GetPovStrThreeViewColorMap(strTag);
	    LOOP_XYZ
		strPovObject << ((PovThreeView_cls*)(*itr))->
		GetPovStrThreeView
		(xyz, strTag, strFilePath, strGeometry, treeDivideParam, depth,
		 baseAlpha, clipForPeriodicBoundary, CCamLocalCoord, frame,
		 pCDrawingBlock);
	    
	    break;
	case _FILTER_STREAMLINE:
	    ((PovStreamLine_cls*)(*itr))->
		GetPovStrStreamLines
		(ofstream, strTag, strGeometry, frame, pCDrawingBlock);
	    break; 
	case _FILTER_VELOCITY_FIELD:
	    ((PovVelField_cls*)(*itr))->
		GetPovStrVelocityField
		(ofstream, strTag, strGeometry, frame, pCDrawingBlock);
	    break;
	}

	//ofstream << wxConvFile.cWX2WC(strPovObject);
	ofstream << (const char*)strPovObject.ToAscii();
    }
    
}

wxString FilterCtrl_cls::
GetPovStrDiffObj(wxString strCoordinateInv){
    wxString strPovDiffObj;
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_POV_DIFFOBJECT) 
	    strPovDiffObj <<
		((PovDiffObject_cls*)(*itr))->
		GetPovStrDiffObj(strCoordinateInv);
    }
    return strPovDiffObj;
}

// ------

// p[^t̏ꍇ
#define MACRO_CallReadDBlock(Tag,Filter)\
    loc = strLine.find(Tag);		   \
    if (loc != std::wstring::npos){\
	Filter##Data_cls *pCData = Filter##_cls::GetNewPReadDataBlock(this, istream); \
	pCData->MakeFilterInstanceAndAddToList(-1, this);		\
	delete pCData;\
    }

// Œl݂̂̏ꍇ
#define MACRO_CallReadDBlock2(Tag,Filter)\
    loc = strLine.find(Tag);\
    if (loc != std::wstring::npos){\
    Filter##Data_cls CData = Filter##_cls::ReadDataBlock(istream); \
    CData.MakeFilterInstanceAndAddToList(-1, this);		\
    }

void FilterCtrl_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    wxString *strTagList = FilterData_cls::msDataBlockTagStr;
    std::wstring::size_type loc;
    MACRO_CallReadDBlock2(strTagList[_FILTER_FOG],PovFog);
    MACRO_CallReadDBlock2(strTagList[_FILTER_BOXCOPY],BoxCopy);
    MACRO_CallReadDBlock2(strTagList[_FILTER_VOLUME],PovVolume);
    MACRO_CallReadDBlock2(strTagList[_FILTER_BLURING],BlurGauss);
    MACRO_CallReadDBlock2(strTagList[_FILTER_INV_DIFFUSE],InvDiffuse);
    MACRO_CallReadDBlock2(strTagList[_FILTER_CENTEROFMASS], FilterCenterOfMass);
    MACRO_CallReadDBlock2(strTagList[_FILTER_MORPH_OPENCLOSE],MorphOpenClose);
    MACRO_CallReadDBlock(strTagList[_FILTER_ISOSURFACE],PovIsosurface);
    MACRO_CallReadDBlock(strTagList[_FILTER_CLIPPING],FilterClipping);
    MACRO_CallReadDBlock(strTagList[_FILTER_LEVEL_GAMMA],LevelAndGamma);
    MACRO_CallReadDBlock(strTagList[_FILTER_FIX_POSITION],FilterFixPosition);
    MACRO_CallReadDBlock(strTagList[_FILTER_DATA_ROTATE],DataRotate);
    MACRO_CallReadDBlock(strTagList[_FILTER_LAYER_COLOR_CURVE],LayerColorCurve);
    MACRO_CallReadDBlock(strTagList[_FILTER_CROSS_SECTION],PovCrossSection);
    MACRO_CallReadDBlock(strTagList[_FILTER_POV_OBJECT],PovObject);
    MACRO_CallReadDBlock(strTagList[_FILTER_POV_DIFFOBJECT],PovDiffObject);
    MACRO_CallReadDBlock(strTagList[_FILTER_THREEVIEW],PovThreeView);
    MACRO_CallReadDBlock(strTagList[_FILTER_STREAMLINE],PovStreamLine);
    MACRO_CallReadDBlock(strTagList[_FILTER_VELOCITY_FIELD],PovVelField);
}

wxString FilterCtrl_cls::
SerializeConf(){
    wxString strRet;
    if (mVpCFilter.size() != 0){
	for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	     itr != mVpCFilter.end(); itr++){
	    strRet += (*itr)->SerializeConf();
	}
    }
    strRet += PovIsosurface_cls::SerializeStaticConf();
    strRet += PovVolume_cls::SerializeStaticConf();
    
    return strRet;
}

// --

void FilterCtrl_cls::
DeleteSelectedItem(){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end();){
	if ((*itr)->mpCGUIPanel->mbSelected == TRUE){
	    DeleteFilter(*itr);
	    itr = mVpCFilter.end();
	    continue;
	}
	itr++;
    }
    
    if (msOmitUpdateList == 0) UpdateList();

    mpCFilterPanelsLink->UpdatePanelLayout();
}

void FilterCtrl_cls::
DeleteFilterItem(int index){
    std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
    for (int i = 0; i < index; i++) itr++;
    mpCFilterPanelsLink->RemoveItemList((*itr)->mpCGUIPanel->mpCMNumPanel);
    (*itr)->RemoveFilterItem();
    delete (*itr);
    mVpCFilter.erase(itr);
    
    if (msOmitUpdateList == 0) UpdateList();

    mpCFilterPanelsLink->UpdatePanelLayout();
}

void FilterCtrl_cls::
AddToList(int index, FilterBase_cls *pCFilter){
    if (index == -1)
	pCFilter->AddFilterItem(mpCFilterPanelsLink, -1);
    else
	pCFilter->AddFilterItem(mpCFilterPanelsLink,
				_FILTER_PANEL_OFFSET + index * 2);
    
    if (index == -1)
	mVpCFilter.push_back(pCFilter);
    else{
	std::list< FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	for (int i = 0; i < index; i++) itr++;
	mVpCFilter.insert(itr, pCFilter);
    }
    
    // CHECK_ME!! ŐݒĂ
    // UNDOŎsƃftHg̒lɂȂĂ܂BmCConfTargetStȑ͉ŎH
    /*
    wxString strLabel;
    if (pCFilter->mFilterType == _FILTER_CLIPPING){
	strLabel = ((FilterClipping_cls*)pCFilter)->GetLocaleStrLabel(((FilterClipping_cls*)pCFilter)->GetPCData()->mCConfTargetStr.msBody);
	MsgBox(NULL, strLabel, strLabel);
    }
    */
    
    if (msOmitUpdateList == 0) UpdateList();

    /*
    if (pCFilter->mFilterType == _FILTER_CLIPPING)
	((FilterClippingGUI_cls*)pCFilter->mpCGUIPanel)->SetTargetLabelStr(strLabel);
    */

    mpCFilterPanelsLink->UpdatePanelLayout();
}

// ꂾNULLł̎ďôŎc
void FilterCtrl_cls::
AddVolume(int index, PovVolumeData_cls *pInitData){
    PovVolume_cls *pC = new PovVolume_cls(this, pInitData);
    AddToList(index, pC);
}

// -------------

void FilterCtrl_cls::
DrawItems(double center[3], double scale[3]){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    ((PovIsosurface_cls*)(*itr))->DrawBox(center, scale);
	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION)
	    ((PovCrossSection_cls*)(*itr))->DrawBox(center, scale);
	if ((*itr)->mFilterType == _FILTER_CENTEROFMASS)
	    ((FilterCenterOfMass_cls*)(*itr))->DrawBox
		(mspCDF3SeqCtrlLink->GetCurrentFrame(), center, scale);
	if ((*itr)->mFilterType == _FILTER_THREEVIEW)
	    ((PovThreeView_cls*)(*itr))->DrawBox(center, scale);
	if ((*itr)->mFilterType == _FILTER_STREAMLINE)
	    ((PovStreamLine_cls*)(*itr))->DrawBox
		(mspCDF3SeqCtrlLink->GetCurrentFrame());
	if ((*itr)->mFilterType == _FILTER_VELOCITY_FIELD)
	    ((PovVelField_cls*)(*itr))->DrawBox();
    }
}

void FilterCtrl_cls::
DrawObject(double scale, bool bPositiveObject, bool bNegativeObject){
    for (std::list<FilterBase_cls*>::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT && bPositiveObject)
	    ((PovObject_cls*)(*itr))->DrawGlobal(scale);
	if ((*itr)->mFilterType == _FILTER_POV_DIFFOBJECT && bNegativeObject)
	    ((PovDiffObject_cls*)(*itr))->DrawGlobal(scale);
	/*
	if ((*itr)->mFilterType == _FILTER_THREEVIEW){
	    if (bNegativeObject == TRUE) glDepthMask(GL_FALSE);
	    double center[3] = {0, 0, 0};
	    double scaleVector[3] = {1, 1, 1};
	    ((PovThreeView_cls*)(*itr))->DrawBox(center, scaleVector);
	}
	*/
    }
}

double FilterCtrl_cls::
GetRemappedDoubleIndex(){
    double doubleIndex = mspCDF3SeqCtrlLink->mspCTimeLink->GetRemappedIndexDouble(_TIME_CTRL_CURRENT);
    
    if (doubleIndex > mpCDF3SeqParent->mMaxIndex) 
	doubleIndex = mpCDF3SeqParent->mMaxIndex;
    return doubleIndex;
}

int  FilterCtrl_cls::
GetRemappedTimeIndex(){
    unsigned int timeIndex = mspCDF3SeqCtrlLink->GetRemappedIndex(0/*FIX_ME!!*/);
    if (timeIndex > mpCDF3SeqParent->mMaxIndex) 
	timeIndex = mpCDF3SeqParent->mMaxIndex;
    return timeIndex;
}

bool FilterCtrl_cls::
GetVolumeValidity(){
    bool bRetval = FALSE;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    if ((*itr)->GetPData()->mCbValid.mbBody == TRUE)
		bRetval = TRUE;
    return bRetval;
}

int FilterCtrl_cls::
GetVolumeColorType(){
    int colorType = -1;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    colorType = ((PovVolumeData_cls*)(*itr)->GetPData())->mCColoring.mBody;
    return colorType;
}

int FilterCtrl_cls::
GetVolumeDataType(){
    int dataType = 0;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    dataType = ((PovVolumeData_cls*)(*itr)->GetPData())->mCDataForColoring.mBody;
    return dataType;
}

double FilterCtrl_cls::
GetVolumeParameter(int type){
    if (GetVolumeValidity() == FALSE) return 0;

    double param = 0;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME){
	    if (type == _VOLUME_EMISSION)
		param = ((PovVolumeData_cls*)(*itr)->GetPData())->mCEmission.mBody;
	    if (type == _VOLUME_ABSORPTION)
		param = ((PovVolumeData_cls*)(*itr)->GetPData())->mCAbsorption.mBody;
	    if (type == _VOLUME_SCATTER)
		param = ((PovVolumeData_cls*)(*itr)->GetPData())->mCScatter.mBody;
	}
    return param;
}

float FilterCtrl_cls::
GetVolumeColor(int index){
    float color = 0;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    color = ((PovVolumeData_cls*)(*itr)->GetPData())->mCMonoColor.mpBody[index];
    return color;
}

void FilterCtrl_cls::
GetTranslate(double pTrans[3], int frame, int frameMax, unsigned int size[3]){
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FIX_POSITION &&
	    (*itr)->GetPData()->mCbValid.mbBody == TRUE){
	    ((FilterFixPosition_cls*)(*itr))->GetTranslate(pTrans, frame, frameMax, size);
	    return;
	}
}

bool FilterCtrl_cls::
GetClippingInformation(wxString queryStr, double min[3], double max[3],
		       std::vector < std::vector < double > > *pVVMin,
		       std::vector < std::vector < double > > *pVVMax,
		       std::vector < bool > *pVActive,
		       std::vector < bool > *pVVisible,
		       bool bFlagForTransrate,
		       wxString strStopLabel, bool *pbFlagSkip){
    
    bool bRetval = FALSE;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++){
	
	std::vector < double > lVMin, lVMax;
	bool bActive, bVisible;
	
	if ((*itr)->mFilterType == _FILTER_CLIPPING){
	    FilterClipping_cls* pCClip = (FilterClipping_cls*)(*itr);
	    Parameters_cls* pParam = pCClip->mpCData->GetPParamActive();
	    
	    if (pCClip->mstrLabel == strStopLabel) {
		if (pbFlagSkip != NULL) *pbFlagSkip = TRUE;
		return FALSE; }
	    if (pCClip->mpCData->mCbValid.mbBody == FALSE) continue;

	    wxString targetStr = ((FilterWithTargetGUI_cls*)pCClip->mpCGUIPanel)->GetTargetLabelStr();
	    if (queryStr == targetStr || queryStr == wxT("_FILTER_ALLINFO") ||
		targetStr == wxT("All")){
		bRetval = TRUE;
		double minB[3], maxB[3], tmpmin[3], tmpmax[3];
		LOOP_XYZ{
		    minB[xyz] = pParam->mpInterpolatedValue[xyz];
		    maxB[xyz] = pParam->mpInterpolatedValue[3+xyz]; 
		}
		LOOP_XYZ{
		    tmpmin[xyz] = min[xyz]; tmpmax[xyz] = max[xyz];
		    min[xyz] += (tmpmax[xyz] - tmpmin[xyz]) * minB[xyz];
		    max[xyz] -= (tmpmax[xyz] - tmpmin[xyz]) * (1.0 - maxB[xyz]);
		}
		LOOP_XYZ lVMin.push_back(min[xyz]);
		LOOP_XYZ lVMax.push_back(max[xyz]);
		bActive = pCClip->mpCGUIPanel->mbSelected;
		bVisible = pParam->mCbVisible.mbBody;
	    }
	}

	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION) {
	    PovCrossSection_cls* pCHeight = (PovCrossSection_cls*)(*itr);
	    PovCrossSectionData_cls* pCData = (PovCrossSectionData_cls*)(pCHeight->mpCData);
	    Parameters_cls* pParam = pCHeight->mpCData->GetPParamActive();

	    if (pCHeight->mstrLabel == strStopLabel) {
		if (pbFlagSkip != NULL) *pbFlagSkip = TRUE;
		return FALSE; }
	    if (pCData->mCbValid.mbBody == FALSE) continue;

	    if (pCData->mCCliping.mBody != 0){
		bRetval = TRUE;
		switch(pCData->mCCliping.mBody){
		case 1:
		    LOOP_XYZ min[xyz] = COMPMAX(min[xyz], 0.0001); 
		    LOOP_XYZ max[xyz] = COMPMIN(max[xyz], 
						pCHeight->GetPos(xyz, 0.9999));
		    break;
		case 2:
		    LOOP_XYZ min[xyz] = COMPMAX(min[xyz], 
						pCHeight->GetPos(xyz, 0.0001));
		    LOOP_XYZ max[xyz] = COMPMIN(max[xyz], 0.9999);
		    break;
		}
		LOOP_XYZ lVMin.push_back(min[xyz]);
		LOOP_XYZ lVMax.push_back(max[xyz]);
		bActive = pCHeight->mpCGUIPanel->mbSelected;
		bVisible = 
		    pParam->mCbVisible.mbBody & 
		    pCData->mCbDrawFrame.mbBody;
	    }
	}

	if (lVMin.size() != 0){
	    pVVMin->push_back(lVMin);
	    pVVMax->push_back(lVMax);
	    if (pVActive != NULL) pVActive->push_back(bActive);
	    if (pVVisible != NULL) pVVisible->push_back(bVisible);
	}
    }
    
    if (bFlagForTransrate == TRUE &&
	FilterClipping_cls::msCConfClippingFollowTracking.mbBody == FALSE){
	
	LOOP_XYZ {
	    if (min[xyz] != 0)
		min[xyz] -= DrawingBlock_cls::msTranslate[xyz]; }
	LOOP_XYZ {
	    if (max[xyz] != 1)
		max[xyz] -= DrawingBlock_cls::msTranslate[xyz]; }
	
	LOOP_XYZ {
	    if (min[xyz] < 0) min[xyz] = 0; 
	    if (min[xyz] > 1) min[xyz] = 1;
	    if (max[xyz] < 0) max[xyz] = 0;
	    if (max[xyz] > 1) max[xyz] = 1; }

	if (pVVMin->size() > 0){
	    LOOP_XYZ (*pVVMin)[0][xyz] = min[xyz];
	    LOOP_XYZ (*pVVMax)[0][xyz] = max[xyz];
	}
    }
    
    return bRetval;
}

double FilterCtrl_cls::
GetFogAbsorption(){
    double value = 0;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FOG)
	    if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		(*itr)->GetTargetLabelStr() == _("Volume"))
		value = ((PovFogData_cls*)(*itr)->GetPData())->mCAbsorption.mBody;
    return value;
}

double FilterCtrl_cls::
GetFogEmission(){
    double value = 0;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FOG)
	    if ((*itr)->GetPData()->mCbValid.mbBody == TRUE &&
		(*itr)->GetTargetLabelStr() == _("Volume") )
		value = ((PovFogData_cls*)(*itr)->GetPData())->mCEmission.mBody;
    return value;
}

double FilterCtrl_cls::
GetFogDistAndClr(float pColor[3], wxString strTag){
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FOG){
	    PovFog_cls *pFog = (PovFog_cls*)(*itr);
	    if (pFog->GetPData()->mCbValid.mbBody == TRUE &&
		pFog->GetTargetLabelStr() == strTag &&
		pFog->GetPCData()->mCFogGlobalType.mBody == 0){
		if (strTag == _("Global"))
		    pFog->GetColorAtInf(pColor);
		else
		    LOOP_RGB pColor[rgb] = 
			pFog->GetPCData()->mCColor.mpBody[rgb];
		return 
		    pFog->GetFogDistance();//mCDistance.mBody;
	    }
	    /*
	      LOOP_XYZ pColor[xyz] = 
	      pFog->GetPCData()->mCColor.mpBody[xyz];
	      pFog->GetFogDistance();//mCDistance.mBody;
	    */
	}
    return -1;
}

double FilterCtrl_cls::
GetTotalLevelMapping(double value, FilterBase_cls *pCStop){
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++){
	if ((*itr) == pCStop) break;
	if ((*itr)->mFilterType == _FILTER_LEVEL_GAMMA)
	    value = ((LevelAndGamma_cls*)(*itr))->GetLevelMapping(value);
    }
    return value;
}

#include <wx/stopwatch.h>

extern wxString GstrDebug;

// FIX_ME ͂܂Ƃ߂Ĕ
bool FilterCtrl_cls::
IsReprocessRequired(int remappedTimeIndex,
		    double remappedTimeIndexDouble,
		    FilterBase_cls *pAtThisFilter){
    if (mSelfDF3Index == 0) GstrDebug = wxT("");
    bool bRequired = FALSE;
    wxString strFilterStatusSum;

    strFilterStatusSum << remappedTimeIndexDouble << wxT("::");

    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++){
	strFilterStatusSum << (*itr)->GetFilterStatusStr();

	if ((*itr) == pAtThisFilter) break;
	
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    bRequired |=
		((PovIsosurface_cls*)(*itr))->IsReprocessRequired();
	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION)
	    bRequired |=
		((PovCrossSection_cls*)(*itr))->IsReprocessRequired();
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    bRequired |=
		((PovObject_cls*)(*itr))->IsReprocessRequired();
	if ((*itr)->mFilterType == _FILTER_THREEVIEW)
	    bRequired |=
		((PovThreeView_cls*)(*itr))->IsReprocessRequired();
    }
    
    
    if (pAtThisFilter != NULL){
	if (pAtThisFilter->mstrFilterStatus != strFilterStatusSum)
	    bRequired = TRUE;
	pAtThisFilter->mstrFilterStatus = strFilterStatusSum;
    }
    else {
	if (mVstrFilterStatusSum[remappedTimeIndex] != strFilterStatusSum)
	    bRequired = TRUE;
	mVstrFilterStatusSum[remappedTimeIndex] = strFilterStatusSum;
    }
   
    if (mSelfDF3Index == 0)
	GstrDebug << mVstrFilterStatusSum[remappedTimeIndex] << wxT("\n") << 
	    strFilterStatusSum << wxT("!!") << bRequired;

    return bRequired;
}

bool FilterCtrl_cls::
IsNextFilterIsVolume(std::list<FilterBase_cls*>::iterator itrCur){
    for (std::list<FilterBase_cls*>::iterator
	     itr = itrCur; itr != mVpCFilter.end(); itr++){
	if (itr == itrCur) continue;
	if ((*itr)->mFilterType == _FILTER_VOLUME)
	    return TRUE;
	if ((*itr)->mbProcessData == TRUE)
	    return FALSE;
    }
    return FALSE;
}

void FilterCtrl_cls::
ProcessXYZData(unsigned char *pTo, unsigned char *pFrom, unsigned int size[3],
	       bool bTimeRemap, float remapFactor, unsigned char *pFromNext){
    
    long int totalSize = size[X] * size[Y] * size[Z];
    
    // f[^up̃͒fЉȂ悤g܂킷B
    if (mbBufferAllocated == TRUE && totalSize != mBufferAllocatedSize){
	delete mpBuffer[0];
	delete mpBuffer[1];
	mbBufferAllocated = FALSE;
    }

    if (mbBufferAllocated == FALSE){
	mpBuffer[0] = new unsigned char [totalSize];
	mpBuffer[1] = new unsigned char [totalSize];
	mbBufferAllocated = TRUE;
	mBufferAllocatedSize = totalSize;
    }
    
    mBufferIndex = 0;

    int current = 0; // r̃f[^u̎wpindex
    unsigned char *pSource = NULL, *pTarget = NULL;
    bool bCopyToPTOFinished = FALSE;

    wxStopWatch CWatch;
    CWatch.Start();

    if (bTimeRemap == TRUE && remapFactor != 0 && pFromNext != NULL){
	pSource = mpBuffer[0];
	mBufferIndex = 1;
	float remapFactorInv = 1.0f-remapFactor;
	for (int i = 0; i < totalSize; i++)
	    mpBuffer[0][i] = pFrom[i] * remapFactorInv 
		+ pFromNext[i] * remapFactor;
    }
    else
	pSource = pFrom;
    
    //wxString strFilterStat;
    // FIX_ME
    // strFilterStat << remappedTimeIndexDouble << wxT("::");
    
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++){
	//strFilterStat << (*itr)->GetFilterStatusStr();

	if (current > 0) pSource = pTarget;
	
	if ((*itr)->mFilterType == _FILTER_VOLUME){
	    if (pTo != pSource){
		memcpy(pTo, pSource, totalSize);
		bCopyToPTOFinished = TRUE;
	    }
	}
	    
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE)
	    ((PovIsosurface_cls*)(*itr))->RefreshFilter(pSource/*,strFilterStat*/); 
	
	if ((*itr)->mFilterType == _FILTER_POV_OBJECT)
	    ((PovObject_cls*)(*itr))->RefreshFilter(pSource, size);

	if ((*itr)->mFilterType == _FILTER_CROSS_SECTION)
	    ((PovCrossSection_cls*)(*itr))->RefreshFilter(pSource);
	
	if ((*itr)->mFilterType == _FILTER_FIX_POSITION)
	    ((FilterFixPosition_cls*)(*itr))->DoCalc
		(mspCDF3SeqCtrlLink->GetCurrentFrame(), pSource, size);

	if ((*itr)->mFilterType == _FILTER_CENTEROFMASS)
	    ((FilterCenterOfMass_cls*)(*itr))->DoCalc
		(mspCDF3SeqCtrlLink->GetCurrentFrame(), pSource, size);

	if ((*itr)->mFilterType == _FILTER_THREEVIEW)
	    ((PovThreeView_cls*)(*itr))->RefreshFilter(pSource);
	
	// tB^i܂Ȃꍇ
	// pTarget == pSource Ȃ̂ŁÃ[vłv
    
	// tB^Pi
	if ((*itr)->mbProcessData == TRUE){
	    
	    if ((*itr)->GetPData()->mCbValid.mbBody == FALSE) continue;

	    // 𖞂ƂpTargetpToɂ΃Rs[Ȃ
	    if (IsNextFilterIsVolume(itr) == TRUE)
		pTarget = pTo; 
	    else{
		pTarget = mpBuffer[mBufferIndex];
		mBufferIndex = 1 - mBufferIndex;
	    }
	    	
	    bool bProcessed = FALSE;
	    if ((*itr)->mFilterType == _FILTER_BLURING){
		((BlurGauss_cls*)(*itr))->ProcessXYZData
		    (pTarget, pSource, size, totalSize,
		     mspCPreferenceLink->mCbSSEFilter.mbBody);
		bProcessed = TRUE; }
	    if ((*itr)->mFilterType == _FILTER_LEVEL_GAMMA){
		((LevelAndGamma_cls*)(*itr))->
		    ProcessXYZData(pTarget, pSource, size);
		bProcessed = TRUE; }
	    if ((*itr)->mFilterType == _FILTER_INV_DIFFUSE){
		((InvDiffuse_cls*)(*itr))->
		    ProcessXYZData(pTarget, pSource, size);
		bProcessed = TRUE; }
	    if ((*itr)->mFilterType == _FILTER_DATA_ROTATE){
		((DataRotate_cls*)(*itr))->
		    ProcessXYZData(pTarget, pSource, size);
		bProcessed = TRUE; }
	    if ((*itr)->mFilterType == _FILTER_LAYER_COLOR_CURVE){
		((LayerColorCurve_cls*)(*itr))->
		    ProcessXYZData(pTarget, pSource, size);
		bProcessed = TRUE; }
	    if ((*itr)->mFilterType == _FILTER_MORPH_OPENCLOSE){
		((MorphOpenClose_cls*)(*itr))->
		    ProcessXYZData(pTarget, pSource, size);
		bProcessed = TRUE; }
	    
	    if (pTarget == pTo && bProcessed == TRUE){
		bCopyToPTOFinished = TRUE;
	    }

	    current++;
	}
    }

    GTestVariableLong = CWatch.Time();
    GTestVariableInt++;
}

int FilterCtrl_cls::
CountDataFilterNumber(){
    int number = 0;
    
    if (mspCDF3SeqCtrlLink->mspCTimeLink->
	mCbUseTimeRemap.mbBody == TRUE) number++;

    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mbProcessData == TRUE)
	    if ((*itr)->GetPData()->mCbValid.mbBody == TRUE) number++;
    
    return number;
}

bool FilterCtrl_cls::
IsIsosurfTargetingMeExist(wxString queryStr){
    bool bRetval = FALSE;
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_ISOSURFACE){
	    bool bValid = ((PovIsosurface_cls*)(*itr))->mpCData->mCbValid.mbBody;
	    wxString targetStr = (*itr)->GetTargetLabelStr();
	    if (targetStr == queryStr && bValid == TRUE) bRetval = TRUE;
	}
    return bRetval;
}

FilterBase_cls* FilterCtrl_cls::
GetPFogTargettingMe(wxString queryStr, bool bStart){
    static std::list<FilterBase_cls*>::iterator itr;
    if (bStart == TRUE) 
	itr = mVpCFilter.begin();
    else
	itr++;
    for (; itr != mVpCFilter.end(); itr++)
	if ((*itr)->mFilterType == _FILTER_FOG){
	    bool bValid = ((PovFog_cls*)(*itr))->mCData.mCbValid.mbBody;
	    wxString targetStr = (*itr)->GetTargetLabelStr();
	    if (queryStr == targetStr && bValid == TRUE)
		return (*itr);
	}
    return NULL;
}

FilterBase_cls* FilterCtrl_cls::
GetPFromTag(wxString strLabel){
    int indexTarget = GetIndexFromTag(strLabel);
    if (indexTarget == -1) return NULL;
    int index = 0;
    for (std::list< FilterBase_cls* >::iterator itr = mVpCFilter.begin();
	 itr != mVpCFilter.end(); itr++){
	if (index == indexTarget) return (*itr);
	index++;
    }
    return NULL;
}

wxString FilterCtrl_cls::
GetColorMaps(){
    wxString strClrMap;
    std::vector <int> mVAlready;

    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++){
	if ((*itr)->mFilterType == _FILTER_VOLUME){
	    PovVolume_cls *pPov = (PovVolume_cls*)(*itr);
	    int index = pPov->mCData.mCColoring.mBody;
	    
	    int dataForColor = pPov->mCData.mCDataForColoring.mBody;
	    float absorption = pPov->mCData.mCCompAbsorption.mBody;
	    bool bAlready = FALSE;
	    for (size_t i = 0; i < mVAlready.size(); i++)
		if (mVAlready[i] == index) bAlready = TRUE;
	    if (bAlready == TRUE) continue;

	    int df3index = mSelfDF3Index;
	    bool bMultiplyLevel = TRUE;
	    unsigned int shadePattern = mspCPreviewPropertyLink->mCShadingPattern.mBody;
	    if (dataForColor != 0){
		float densityColor[3] = {1, 1, 1};
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (mSelfDF3Index, index, _CLR_NOSHADE, 0, densityColor);
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (mSelfDF3Index, index, _CLR_SHADE, absorption, densityColor, TRUE, shadePattern);
		bMultiplyLevel = FALSE;
		mSelfDF3Index = dataForColor - 1;
	    }

	    // bMultiplyLevelFALSE͎̎ȊO^[Qbg
	    if (index == 0){
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (df3index, 0, _CLR_NOSHADE, 0, 
		     pPov->mCData.mCMonoColor.mpBody, bMultiplyLevel);
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (df3index, 0, _CLR_SHADE, absorption, 
		     pPov->mCData.mCMonoColor.mpBody, bMultiplyLevel, shadePattern);
	    }
	    else{
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (df3index, index, _CLR_NOSHADE, 0, 
		     NULL, bMultiplyLevel);
		strClrMap << GCColoring.GetPovStrColorMapMacro
		    (df3index, index, _CLR_SHADE, absorption, 
		     NULL, bMultiplyLevel, shadePattern);
	    }
	    
	    mSelfDF3Index = df3index; // ɖ߂
	}
    }

    return strClrMap;
}

void FilterCtrl_cls::
FuncHookRetexture(void *pArg, void *WXUNUSED(pEvent)){
    wxCommandEvent command = wxCommandEvent(wxEVT_COMMAND_MENU_SELECTED, menuItemReTexture);
    wxGetTopLevelParent((MultiNumericalInputsPanel_cls*)pArg)->AddPendingEvent(command);
}

void FilterCtrl_cls::
FuncHookRefreshTop(void *pArg, void *WXUNUSED(pEvent)){
    wxCommandEvent command = wxCommandEvent(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    wxGetTopLevelParent((MultiNumericalInputsPanel_cls*)pArg)->AddPendingEvent(command);
}

FilterGUI_cls* FilterCtrl_cls::
GetPFilterGUI(void *pArg){
    for (std::list<FilterBase_cls*>::iterator
	     itr = mVpCFilter.begin(); itr != mVpCFilter.end(); itr++)
	if (pArg == (*itr)->mpCGUIPanel->mpCMNumPanel)
	    return (*itr)->mpCGUIPanel;
    return NULL;
}

void FilterCtrl_cls::
SelectLastItem(){
    std::list<FilterBase_cls*>::iterator itr = mVpCFilter.end();
    itr--;
    SelectItem((*itr)->mpCGUIPanel, FALSE);
}

void FilterCtrl_cls::
SelectItem(FilterGUI_cls *pTgt, bool bAdditive){
    DF3SeqCtrl_cls::DeselectAllFilter();
    
    pTgt->mpCParentBase->SelectPanel(TRUE, !bAdditive);
    mspCDF3SeqCtrlLink->mCConfActiveSlot.mBody = mSelfDF3Index;
    
    if (pTgt->mpCParentBase->mbWithTexture == TRUE){
	((FilterWithTextureGUI_cls*)pTgt)->OpenTextureDLG(FALSE);
    }
    
    wxCommandEvent com = wxCommandEvent
	(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshPanels);
    wxGetTopLevelParent(pTgt->mpCMNumPanel)->AddPendingEvent(com);
}

void FilterCtrl_cls:: 
FuncHookMouse(void *pArg, void *pEvent){
    FilterGUI_cls *pTgt = DF3SeqCtrl_cls::GetPFilterGUI(pArg);
    FilterBase_cls *pBase = pTgt->mpCParentBase;
    wxMouseEvent event = *(wxMouseEvent*)pEvent;
    FilterCtrl_cls *pThis = pTgt->mpCParentBase->mpCParentCtrl;

    // _uNbN
    if (event.ButtonDClick() && pBase->mbWithParameter == TRUE)
	((FilterWithParameter_cls*)pBase)->mpCParamPanel->mpCTimeLine->
	    CreateKeyDLG(8, 24, pTgt->mpCMNumPanel->GetScreenPosition());
    
    if (event.LeftDown() == TRUE || event.RightDown() == TRUE){
	pThis->SelectItem(pTgt, event.m_shiftDown);
	if (event.RightDown() == TRUE && event.RightIsDown() == TRUE)
	    pThis->mpCFilterPanelsLink->PopUpFilterMenu();
    }
}

void FilterCtrl_cls::
FuncHookValidityCheckbox(void *pArg, void *pEvent){
    FilterGUI_cls *pTgt = DF3SeqCtrl_cls::GetPFilterGUI(pArg);
    FilterBase_cls *pBase = pTgt->mpCParentBase;
    if (pBase->mbProcessData == TRUE)
	//FuncHookRetexture(pArg, pEvent);
	FuncHookRefreshTop(pArg, pEvent);
    else
	FuncHookRefreshTop(pArg, pEvent);
}

void FilterCtrl_cls::
FuncHookShow(void *pArg, void *pbShow){
    FilterGUI_cls *pTgt = DF3SeqCtrl_cls::GetPFilterGUI(pArg);
    DF3SeqCtrl_cls::DeselectAllFilter();

    wxCommandEvent com1 = wxCommandEvent(wxEVT_COMMAND_MENU_SELECTED, 
					 menuItemRefreshPanels);
    if (pTgt != NULL){
	pTgt->mpCParentBase->SelectPanel(TRUE);
	pTgt->mpCParentBase->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
	wxWindow* pCTop = wxGetTopLevelParent(pTgt->mpCMNumPanel);
	pCTop->AddPendingEvent(com1);
    }
    
}

void FilterCtrl_cls::
FuncHookAssign(void *pArg, void *pEvent){
    FilterBase_cls *pTgtBase = DF3SeqCtrl_cls::GetPFilterGUI(pArg)->mpCParentBase;
    if (pTgtBase->mbWithParameter == TRUE)
	((FilterWithParameter_cls*)pTgtBase)->SetKeyIfChanged(_PARAM_CURRENT_DINDEX);

    if (pTgtBase->mFilterType == _FILTER_LEVEL_GAMMA)
	mspCPreviewPropertyLink->mpCGUI->UpdateGraph();
}

void FilterCtrl_cls::
FuncHookChanged(void *pArg, void *pEvent){
    FilterGUI_cls *pFilterGUI = DF3SeqCtrl_cls::GetPFilterGUI(pArg);
    FilterBase_cls *pFilter = pFilterGUI->mpCParentBase;
    FilterCtrl_cls *pCtrl = pFilter->mpCParentCtrl;
    int index = *(unsigned int *)pEvent;

    int type = pFilterGUI->mpCMNumPanel->mVCPointers[index].mType;
    
    // XC_[𓮂Ɗ{IFLOAT̂łł悢H
    if (type == _GP_FLOAT || type == _GP_FLOAT_SLIDER){
	float prevValue, newValue;
	prevValue = pFilterGUI->mpCMNumPanel->mCachedFloat;
	newValue = pFilterGUI->mpCMNumPanel->mVCPointers[index].GetValueFloat();

	GCUndoList.SetNewCommand
	    (new ComFilterConfChangeFLOAT(pCtrl, pFilter->mstrLabel, index, newValue),
	     new ComFilterConfChangeFLOAT(pCtrl, pFilter->mstrLabel, index, prevValue));
	GCUndoList.SkipCurrent();
    }

    if (type == _GP_INTEGER || type == _GP_INTEGER_SPIN || 
	type == _GP_INTEGER_SLIDER ||
	type == _GP_UNSIGNED_INT_LIST || type == _GP_UNSIGNED_INT || 
	type == _GP_UINT_SPIN){
	int prevValue, newValue;
	prevValue = pFilterGUI->mpCMNumPanel->mCachedInt;
	newValue = pFilterGUI->mpCMNumPanel->mVCPointers[index].GetValueInt();
	
	GCUndoList.SetNewCommand
	    (new ComFilterConfChangeINT(pCtrl, pFilter->mstrLabel, index, newValue),
	     new ComFilterConfChangeINT(pCtrl, pFilter->mstrLabel, index, prevValue));
	GCUndoList.SkipCurrent();
    }

    if (type == _GP_BOOL){
	int prevValue, newValue;
	prevValue = pFilterGUI->mpCMNumPanel->mCachedBool;
	newValue = pFilterGUI->mpCMNumPanel->mVCPointers[index].GetValueBool();

	GCUndoList.SetNewCommand
	    (new ComFilterConfChangeBOOL(pCtrl, pFilter->mstrLabel, index, newValue),
	     new ComFilterConfChangeBOOL(pCtrl, pFilter->mstrLabel, index, prevValue));
	GCUndoList.SkipCurrent();
    }

    //MsgBoxf(NULL, "Change!!", "%d-th slot; %s, %d th item from %f to %f", pCtrl->mSelfDF3Index, pFilter->mstrId.char_str(), index, prevValue, newValue);
    
}

void FilterCtrl_cls::
PushListOmit(){
    msOmitUpdateList++;
}

void FilterCtrl_cls::
PopListOmit(){
    msOmitUpdateList--;
}

// -----

DF3SeqCtrl_cls   *FilterCtrl_cls::mspCDF3SeqCtrlLink;
Preference_cls   *FilterCtrl_cls::mspCPreferenceLink;
PreviewProperty_cls *FilterCtrl_cls::mspCPreviewPropertyLink;
int               FilterCtrl_cls::msOmitUpdateList = 0;
int               FilterCtrl_cls::msStereoTmpFileIOMode = 3;

// -------------

// p[^t̃tB^
#define MACRO_ComFilterParamAdd(Short,Long)\
    Com##Short##ParamAdd_cls::\
    Com##Short##ParamAdd_cls(FilterCtrl_cls *pCTgt, int index, void *pCData):\
    ComFilterParamBase_cls(pCTgt, index){\
    if (pCData == NULL)\
	mpCData = new Long##Data_cls(pCTgt, TRUE);\
    else{\
	mpCData = new Long##Data_cls(pCTgt, FALSE);\
	*(Long##Data_cls*)mpCData = *(Long##Data_cls*)pCData;\
    }\
}

// Œl݂̂̃tB^
#define MACRO_ComFilterParamAdd2(Short,Long)\
    Com##Short##ParamAdd_cls::\
    Com##Short##ParamAdd_cls(FilterCtrl_cls *pCTgt, int index, void *pCData):\
    ComFilterParamBase_cls(pCTgt, index){\
    mpCData = new Long##Data_cls();\
    if (pCData != NULL)\
	*(Long##Data_cls*)mpCData = *(Long##Data_cls*)pCData;\
    }

MACRO_ComFilterParamAdd(Isosurface,PovIsosurface)
MACRO_ComFilterParamAdd(PovObject,PovObject)
MACRO_ComFilterParamAdd(CrossSection,PovCrossSection)
MACRO_ComFilterParamAdd2(Fog,PovFog)
MACRO_ComFilterParamAdd2(BoxCopy,BoxCopy)
MACRO_ComFilterParamAdd2(BlurGauss,BlurGauss)
MACRO_ComFilterParamAdd2(InvDiffuse,InvDiffuse)
MACRO_ComFilterParamAdd(LevelAndGamma,LevelAndGamma)
MACRO_ComFilterParamAdd(Clipping,FilterClipping)
MACRO_ComFilterParamAdd2(Volume,PovVolume)
MACRO_ComFilterParamAdd(FixPosition,FilterFixPosition)
MACRO_ComFilterParamAdd(DataRotate,DataRotate)
MACRO_ComFilterParamAdd(LayerColorCurve,LayerColorCurve)
MACRO_ComFilterParamAdd2(CenterOfMass,FilterCenterOfMass)
MACRO_ComFilterParamAdd2(MorphOpenClose,MorphOpenClose)
MACRO_ComFilterParamAdd(PovDiffObject,PovDiffObject)
MACRO_ComFilterParamAdd(ThreeView,PovThreeView)
MACRO_ComFilterParamAdd(PovStreamLine, PovStreamLine)
MACRO_ComFilterParamAdd(PovVelField, PovVelField)
