// -*- c++ -*- 
// (setq c-basic-offset 4)
// "clsFixPosition.h"

#include "clsFixPosition.h"

#include "../DF3/clsDF3SeqGUI.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../MAINCANVAS/clsPreference.h"
#include "../MYAPI/MsgBox.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../../clsTimeCtrl.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define ENDLWX wxT("\n")

MACRO_FilterConstructor(FilterFixPosition,,)

void FilterFixPositionGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){
    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);
    
    pCurr->mTextWidth = -1;
    int boxWidth = 2;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
    int indexDoCalc = pCurr->CombinePointer
	(_GP_BOOL, &mpCParent->mbDoCalc, _("Do Calc"));
    int indexCustomA = pCurr->CombinePointerButton
	(buttonCtrlPanelItemsCustomA, _("Copy Result"));

    boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Pos"));
    pCurr->CombinePointer(_GP_DOUBLE, 
			  &mpCParent->mpCData->GetPInterpolated()[0]);
    pCurr->CombinePointer(_GP_DOUBLE, 
			  &mpCParent->mpCData->GetPInterpolated()[1]);
    pCurr->CombinePointer(_GP_DOUBLE, 
			  &mpCParent->mpCData->GetPInterpolated()[2]);

    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    
    mpObjectForCustomA = pCurr->mVCPointers[indexCustomA].mpPointerCtrl;

    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookRefreshTop, _MN_HOOK_ASSIGN);

    ((wxCheckBox*)(pCurr->mVCPointers[indexDoCalc].mpPointerCtrl))
	->SetToolTip(_("calc on retexture"));
    ((wxButton*)(pCurr->mVCPointers[indexCustomA].mpPointerCtrl))
	->SetToolTip(_("copy to TimeLine"));
    
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
}

MACRO_SetCustomFunctions(FilterFixPositionGUI_cls,CopyResult(),,)
MACRO_HideAndShowEmpty(FilterFixPositionGUI_cls);

void FilterFixPositionGUI_cls::
CopyResult(){
    mpCParent->CopyResult();
}

// -----------------

FilterFixPositionData_cls::
FilterFixPositionData_cls(FilterCtrl_cls* pCCtrl,
			  bool bConnectToStaticParam) :
FilterWithParameterData_cls(pCCtrl)
{
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
    
    Parameters_cls CParam(3, L"FixPosition");
    double iniParam[3] = {0, 0, 0};
    CParam.ExecCommandInitSet(iniParam);

    // !! todo !! _MAX_PARAMETER_NUMBER 𒴂̏
    if (bConnectToStaticParam == TRUE){
	mParamNumber = -1;
	for (size_t i = 0; i < msVbParamUsed[mpCCtrl->mSelfDF3Index].size(); i++)
	    if (msVbParamUsed[mpCCtrl->mSelfDF3Index][i] == FALSE){
		mParamNumber = (int)i;
		msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].CopyFromP(&CParam);
		msVbParamUsed[mpCCtrl->mSelfDF3Index][mParamNumber] = TRUE;
		i = _MAX_PARAMETER_NUMBER;
	    }
	if (mParamNumber == -1){
	    mParamNumber = (int)msVbParamUsed[mpCCtrl->mSelfDF3Index].size();
	    msVCParam[mpCCtrl->mSelfDF3Index].push_back(CParam);
	    msVbParamUsed[mpCCtrl->mSelfDF3Index].push_back(TRUE);
	}
    }
}

void FilterFixPositionData_cls::
ReDefineList(){
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCAxisSwitch, wxT("TrackingAxisOnOff="), L"1 1 1");
}

MACRO_MakeInstanceAndAddToList(FilterFixPosition)

wxString FilterFixPositionData_cls::
SerializeConfPolymorph(){
    return msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].Serialize();
}

void FilterFixPositionData_cls::
ReadStrLinePolymorph(std::wistream &istream, std::wstring strLine){
    msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber].ReadStrLine(istream, strLine);
}

void FilterFixPositionData_cls::
InitializeStaticMember(int df3index){
    msVCParam[df3index].reserve(_MAX_PARAMETER_NUMBER);
    msVbParamUsed[df3index].reserve(_MAX_PARAMETER_NUMBER);
}

Parameters_cls* FilterFixPositionData_cls::
GetPParamActive(){
    return &msVCParam[mpCCtrl->mSelfDF3Index][mParamNumber];
}

std::vector < Parameters_cls > FilterFixPositionData_cls::msVCParam[_DF3_MAX_SLOT];
std::vector < bool > FilterFixPositionData_cls::msVbParamUsed[_DF3_MAX_SLOT];

// ----------

FilterFixPosition_cls::
FilterFixPosition_cls(FilterCtrl_cls *pCParentCtrl, FilterFixPositionData_cls *pCInitData) :
    FilterWithParameter_cls(pCParentCtrl) 
{
    mpCData = new FilterFixPositionData_cls(pCParentCtrl, (pCInitData == NULL));
    if (pCInitData != NULL) *(FilterFixPositionData_cls*)mpCData = *pCInitData;
    
    ConstructorCommon(pCParentCtrl, _("FixPosition"), _FILTER_FIX_POSITION);
    mpCGUIPanel = new FilterFixPositionGUI_cls(this);
    mbAllocatedDataPrev = FALSE;
    mbDoCalc = FALSE;
    for (int i = 0; i < 9999; i++) 
	mADifference[i][0] = mADifference[i][1] = mADifference[i][2] = 0;
    mFrameIndexPrev = -2;
}

FilterFixPosition_cls::
~FilterFixPosition_cls(){
    if (mspCDF3SeqGUI != NULL)
	mspCDF3SeqGUI->mpCPanelList->Remove(mstrLabel);
    if (MainGLCanvas_cls::mspCSelectedItem == mpCData->GetPParamActive()) 
	MainGLCanvas_cls::SelectItem(NULL);
    if (mbAllocatedDataPrev == TRUE) delete mpDataPrev;
}

MACRO_AddFilterItemPolymorphP(FilterFixPosition,);
MACRO_InitializeStaticMemberP(FilterFixPosition);
MACRO_GetPCData(FilterFixPosition);

void FilterFixPosition_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls *pCDLG = 
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));

    pCDLG->mTextWidth = 100;

    int boxWidth = 3;
    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Track On/Off"));
    pCDLG->CombinePointer(_GP_BOOL, 
			  &GetPCData()->mCAxisSwitch.mpbBody[X], _("X axis"));
    pCDLG->CombinePointer(_GP_BOOL, 
			  &GetPCData()->mCAxisSwitch.mpbBody[Y], _("Y axis"));
    pCDLG->CombinePointer(_GP_BOOL, 
			  &GetPCData()->mCAxisSwitch.mpbBody[Z], _("Z axis"));
    
    pCDLG->mMinimumWidth = 300;
    pCDLG->FinishSettingDLG(FALSE);
}

// ------

ThreadCalcDifference::
ThreadCalcDifference() 
  : wxThread(wxTHREAD_JOINABLE) {
    mbValid = FALSE;
    mbRunning = FALSE;
}

void ThreadCalcDifference::
SetParameters(FilterFixPosition_cls *pCParent,
	      unsigned int size[3],
	      double xIni, double xFin,
	      double yIni, double yFin,
	      double zIni, double zFin,
	      int offsetX, int offsetY, int offsetZ) {
    mbValid = TRUE;
    mpCParent = pCParent;    
    LOOP_XYZ mSize[xyz] = size[xyz];
    int trackRangeX = _TRACK_RANGE_X;
    int trackRangeY = _TRACK_RANGE_Y;
    int trackRangeZ = _TRACK_RANGE_Z;
    // zɔf[^̂adhocɏC
    if (size[Z] < 100) trackRangeZ = _TRACK_RANGE_Z / 2; 
    if (size[Z] < 30) trackRangeZ = _TRACK_RANGE_Z / 2; 

    mOffsetX = offsetX; mOffsetY = offsetY; mOffsetZ = offsetZ;
    mIniX = mSize[X] * xIni; mFinX = mSize[X] * xFin;
    mIniY = mSize[Y] * yIni; mFinY = mSize[Y] * yFin;
    mIniZ = mSize[Z] * zIni; mFinZ = mSize[Z] * zFin;
    if (mIniX - trackRangeX < 0) mIniX = trackRangeX;
    if (mFinX + trackRangeX > (int)mSize[X]) mFinX = (int)mSize[X]-trackRangeX;
    if (mIniY - trackRangeY < 0) mIniY = trackRangeY;
    if (mFinY + trackRangeY > (int)mSize[Y]) mFinY = (int)mSize[Y]-trackRangeY;
    if (mIniZ - trackRangeZ < 0) mIniZ = trackRangeZ;
    if (mFinZ + trackRangeZ > (int)mSize[Z]) mFinZ = (int)mSize[Z]-trackRangeZ;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadCalcDifference::
Entry(){

    int difference = 0;

    printf("thread start %d %d %d\n", mOffsetX, mOffsetY, mOffsetZ);
    
    unsigned int *pDataCurr = mpCParent->mpDataCurr;
    unsigned int *pDataPrev = mpCParent->mpDataPrev;

    int coordOffsetY = mSize[X];
    int coordOffsetZ = mSize[X] * mSize[Y];
    int indexOffset = mOffsetZ * coordOffsetZ + mOffsetY * coordOffsetY + mOffsetX;
    
    int finX = mFinX, finY = mFinY, finZ = mFinZ;

    for (int z = mIniZ; z < finZ; z++){
	int indexZ = z * coordOffsetZ;
	for (int y = mIniY; y < finY; y++){
	    int indexZY = indexZ + y * coordOffsetY;
	    for (int x = mIniX; x < finX; x++){
		int indexZYX = indexZY + x;
		difference += abs((int)pDataCurr[indexZYX] -
				  (int)pDataPrev[indexZYX + indexOffset]);
	    }
	}
    }
    
    mpCParent->mMatrixDifference[mOffsetZ+_TRACK_RANGE_Z][mOffsetY+_TRACK_RANGE_Y][mOffsetX+_TRACK_RANGE_X] = difference;
    
    return 0;
}

// ------

void FilterFixPosition_cls::
DoCalc(int frame, unsigned char *pSource, unsigned int size[3]){
    if (mpCData->mCbValid.mbBody == FALSE) return;

    GbGLLockForMultithread = TRUE;

    long int totalSize = size[X] * size[Y] * size[Z];
    int matrixSize = (_TRACK_RANGE_Z*2+1)*(_TRACK_RANGE_Y*2+1)*(_TRACK_RANGE_X*2+1);
    int *pFin = mMatrixDifference[0][0] + matrixSize;
    for (int* p = mMatrixDifference[0][0]; p < pFin; p++) *p = -1; // -1 ͏O

    mpDataCurr = new unsigned int [totalSize]; // prevƂĎĝŎĂȂƂȂ
    for (int i = 0; i < totalSize; i++)
	mpDataCurr[i] = pSource[i];
    
    int threadNum = mspCPreferenceLink->mCThreadNumFilter.mBody;
    int offsetX = 0, offsetY = 0, offsetZ = 0;

    double min[3] = {0, 0 ,0}, max[3] = {1, 1, 1};
    GetClippingInformation(min, max);

    if (mbDoCalc == TRUE && 
	mbAllocatedDataPrev == TRUE &&
	mFrameIndexPrev == frame - 1){

	bool bFoundMinimum = FALSE;

	while (bFoundMinimum == FALSE){

	    ThreadCalcDifference ACThread[7];
	    for (int i = 0; i < 7; i++){
		int matrixIndexX = offsetX+msAdX[i]+_TRACK_RANGE_X;
		int matrixIndexY = offsetY+msAdY[i]+_TRACK_RANGE_Y;
		int matrixIndexZ = offsetZ+msAdZ[i]+_TRACK_RANGE_Z;

		if (mMatrixDifference[matrixIndexZ][matrixIndexY][matrixIndexX] == -1)
		    ACThread[i].SetParameters
			(this, size, min[X], max[X], min[Y], max[Y], min[Z], max[Z],
			 offsetX+msAdX[i], offsetY+msAdY[i], offsetZ+msAdZ[i]);	
	    }
	    
	    int index = 0;
	    while (index < 7){
		int thread = 0;
		while (thread < threadNum){
		    if (ACThread[index].mbValid == TRUE){
			ACThread[index].mbRunning = TRUE;
			ACThread[index].Run(); thread++; 
		    }
		    index++;
		    if (index >= 7) break;
		}
		    		
		for (int i = 0; i < 7; i++)
		    if (ACThread[i].mbRunning == TRUE){
			ACThread[i].Wait(); 
			ACThread[i].mbRunning = FALSE;
		    }
	    }

	    int minimumIndex = 0;
	    int minimum = mMatrixDifference
		[offsetZ+_TRACK_RANGE_Z][offsetY+_TRACK_RANGE_Y][offsetX+_TRACK_RANGE_X];
	    for (int i = 1; i < 7; i++){
		int matrixIndexX = offsetX+msAdX[i]+_TRACK_RANGE_X;
		int matrixIndexY = offsetY+msAdY[i]+_TRACK_RANGE_Y;
		int matrixIndexZ = offsetZ+msAdZ[i]+_TRACK_RANGE_Z;
		if (mMatrixDifference[matrixIndexZ][matrixIndexY][matrixIndexX] < minimum){
		    minimumIndex = i;
		    minimum = mMatrixDifference[matrixIndexZ][matrixIndexY][matrixIndexX];
		}
	    }
	    
	    if (minimumIndex == 0) bFoundMinimum = TRUE;
	    offsetX += msAdX[minimumIndex];
	    offsetY += msAdY[minimumIndex];
	    offsetZ += msAdZ[minimumIndex];

	    if (offsetX <= -10 || offsetX >= 10) bFoundMinimum = TRUE;
	    if (offsetY <= -10 || offsetY >= 10) bFoundMinimum = TRUE;
	    if (offsetZ <= -10 || offsetY >= 10) bFoundMinimum = TRUE;

	}
	
	if (GetPCData()->mCAxisSwitch.mpbBody[X] == TRUE)
	    mADifference[frame][X] = offsetX;
	else
	    mADifference[frame][X] = 0;

	if (GetPCData()->mCAxisSwitch.mpbBody[Y] == TRUE)
	    mADifference[frame][Y] = offsetY;
	else
	    mADifference[frame][Y] = 0;

	if (GetPCData()->mCAxisSwitch.mpbBody[Z] == TRUE)
	    mADifference[frame][Z] = offsetY;
	else
	    mADifference[frame][Z] = 0;
	
	printf("offset= %d %d %d\n", offsetX, offsetY, offsetZ);
    }
    
    if (mbAllocatedDataPrev == TRUE) delete mpDataPrev;
    mpDataPrev = mpDataCurr;
    mbAllocatedDataPrev = TRUE;
    mFrameIndexPrev = frame;

    GbGLLockForMultithread = FALSE;
}

void FilterFixPosition_cls::
CopyResult(){
    int maxFrame = mpCParentCtrl->mspCDF3SeqCtrlLink->GetMaxFrame();
    Parameters_cls *pCTMP = new Parameters_cls(3, L"FixPosition");
    pCTMP->mCbTimeDependent.mbBody = TRUE;
    double pZero[3] = {0, 0, 0};
    pCTMP->ExecCommandSetKey(0, pZero, _PARAM_STEP, pZero, pZero);
    for (int i = 0; i <= maxFrame; i++){
	double doubleIndex = (double)i / maxFrame;
	double pValue[3] = {mADifference[i][0], 
			    mADifference[i][1],
			    mADifference[i][2]};
	if (i > 0 && (
	    mADifference[i][0] != mADifference[i-1][0] || 
	    mADifference[i][1] != mADifference[i-1][1] || 
	    mADifference[i][2] != mADifference[i-1][2]))
	    pCTMP->ExecCommandSetKey(doubleIndex, pValue, _PARAM_STEP,
				     pZero, pZero);
    }
    Parameters_cls *pParam = mpCData->GetPParamActive();
    pParam->CopyParameterFromP(pCTMP);
    delete pCTMP;

    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    wxGetTopLevelParent(mpCGUIPanel->mpCMNumPanel)->AddPendingEvent(com);
}

void FilterFixPosition_cls::
GetTranslate(double pTrans[3], int frame, int frameMax, unsigned int size[3]){
    LOOP_XYZ pTrans[xyz] = 0;
    LOOP_XYZ if (size[xyz] == 0) return; 

    for (int i = 0; i <= frame; i++){
	double doubleIndex = (double)i / frameMax;
	double remappedIndexDouble = 
	    mpCParentCtrl->mspCDF3SeqCtrlLink->mspCTimeLink->GetRemappedIndexDouble(i);
	if (fmod(remappedIndexDouble+1e-5, 1) < 2e-5){
	    pTrans[0] += mpCData->GetPParamActive()->GetPInterpolatedValue(doubleIndex)[0];
	    pTrans[1] += mpCData->GetPInterpolated()[1];
	    pTrans[2] += mpCData->GetPInterpolated()[2];
	}
    }
    LOOP_XYZ pTrans[xyz] /= size[xyz];
}

// ------

wxString FilterFixPosition_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("FixPosition::") << GetPCData()->mCbValid.mbBody;
    return strStatus;
}

wxString FilterFixPosition_cls::
SerializeConf(){
    UpdateCommonData();
    return mpCData->SerializeConf(mFilterType);
}

MACRO_GetNewPReadDataBlock(FilterFixPosition)

int FilterFixPosition_cls::msAdX[7] = {0, -1, +1,  0,  0,  0,  0};
int FilterFixPosition_cls::msAdY[7] = {0,  0,  0, -1, +1,  0,  0};
int FilterFixPosition_cls::msAdZ[7] = {0,  0,  0,  0,  0, -1, +1};

