// "clsDrawPath.h"

#include "clsDrawPath.h"
#include "../GLOBJ/GLOBJECTS.h"
#include "../MYGL/clsGLGuides.h"
#include "../PARAMETER/clsParameters.h"

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

bool DrawPath_cls::
DrawSelectedKey(GLGuides_cls* pCGLGuide, Parameters_cls *pParam, 
		Parameters_cls *pParamTgt, Parameters_cls *pParamTop,
		int displayList, double dpos[3], double scale, int flag,
		int type, double dRot[4]){
    bool bRet = FALSE;
    PositionAndDirection_cls CTmp, CParentObjPos;
    for (unsigned int i = 0; i < pParam->GetKeyNumber(); i++){
	if (pParam->mVCKeys[i].mbSelected == FALSE) continue;
	
	double tmpPos[3], tmpTop[3] = {0, 0, 1}, tmpAnc[3] = {0, 0, 0};
	double doubleIndex = pParam->GetKeyFrameDoubleIndex(i);
	LOOP_XYZ tmpPos[xyz] = pParam->GetPKeyFrameValue(i)[xyz] + dpos[xyz];
	if (pParamTop != NULL)
	    LOOP_XYZ tmpTop[xyz] = 
		pParamTop->GetPInterpolatedValue(doubleIndex)[xyz];
	if (pParamTgt != NULL)
	    LOOP_XYZ tmpAnc[xyz] = 
		pParamTgt->GetPInterpolatedValue(doubleIndex)[xyz];
	
	LOOP_XYZ CTmp.mPos[xyz] = tmpPos[xyz];
	LOOP_XYZ CTmp.mTop[xyz] = tmpTop[xyz];
	CTmp.PointAt(tmpAnc);

	pCGLGuide->SetArrowDir(1, 1, 1);
	if (type == _GUIDES_TRANSLATE){
	    LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	    pCGLGuide->DrawGuide(tmpPos, 2, 1, NULL, CParentObjPos, type);
	}
	
	if (type == _GUIDES_ROTATE &&
	    pCGLGuide->mRotationDim == 1){
	    LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	    LOOP_XYZ CParentObjPos.mTgt[xyz] = CTmp.mTgt[xyz];
	    LOOP_XYZ CParentObjPos.mTop[xyz] = CTmp.mTop[xyz];
	    LOOP_XYZ CParentObjPos.mRgt[xyz] = CTmp.mRgt[xyz];
	    pCGLGuide->DrawGuide(tmpPos, 2, 1, NULL, CParentObjPos, type, 
				 dRot);
	}
	
	if (type == _GUIDES_ROTATE &&
	    pCGLGuide->mRotationDim == 3){
	    LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	    LOOP_XYZ CParentObjPos.mTgt[xyz] = CTmp.mTop[xyz];
	    LOOP_XYZ CParentObjPos.mTop[xyz] = CTmp.mTgt[xyz];
	    CParentObjPos.NormalizeAndOrthogonalize();
	    pCGLGuide->DrawGuide(tmpPos, 2, 1, NULL, CParentObjPos, type, 
				 dRot);
	}

	//pCGLGuide->DrawGuide(tmpPos, 2, 1, NULL, CTmp);
	bRet = TRUE;
    }
    return bRet;
}

void DrawPath_cls::
DrawCurrent(GLGuides_cls* pCGLGuide, PositionAndDirection_cls CPosDir,
	    std::vector < int > VDisplayList,
	    double dpos[3], double scale, double *pAnc,
	    PositionAndDirection_cls CParentObjPos,
	    int type, double dScale[3], double dRot[4]){
    
    LOOP_XYZ CPosDir.mPos[xyz] += dpos[xyz];
    if (pAnc != NULL)
	CPosDir.PointAt(pAnc);
    
    glPushMatrix();
    GLOBJ::SetMatrix3dv3(CPosDir.mPos, CPosDir.mTgt, CPosDir.mTop);
    glScalef(scale, scale, scale);
    if (type == _GUIDES_ROTATE &&
	pCGLGuide->mRotationDim == 1)
	glRotated(dRot[0], 0, 0, 1);
    
    for (int i = 0; i < VDisplayList.size(); i++)
	GLOBJ::Object::Draw(VDisplayList[i]);

    glPopMatrix();

    pCGLGuide->SetArrowDir(1, 1, 1);
    if (type == _GUIDES_TRANSLATE){
	LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	pCGLGuide->DrawGuide(CPosDir.mPos, 2, 1, NULL, CParentObjPos, type);
    }

    if (type == _GUIDES_ROTATE &&
	pCGLGuide->mRotationDim == 1){
	LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	LOOP_XYZ CParentObjPos.mTgt[xyz] = CPosDir.mTgt[xyz];
	LOOP_XYZ CParentObjPos.mTop[xyz] = CPosDir.mTop[xyz];
	LOOP_XYZ CParentObjPos.mRgt[xyz] = CPosDir.mRgt[xyz];
	pCGLGuide->DrawGuide(CPosDir.mPos, 2, 1, NULL, CParentObjPos, type, 
			     dRot);
    }

    if (type == _GUIDES_ROTATE &&
	pCGLGuide->mRotationDim == 3){
	LOOP_XYZ CParentObjPos.mPos[xyz] = 0;
	LOOP_XYZ CParentObjPos.mTgt[xyz] = CPosDir.mTop[xyz];
	LOOP_XYZ CParentObjPos.mTop[xyz] = CPosDir.mTgt[xyz];
	CParentObjPos.NormalizeAndOrthogonalize();
	pCGLGuide->DrawGuide(CPosDir.mPos, 2, 1, NULL, CParentObjPos, type, 
			     dRot);
    }
}
