// "clsGLGuides.h"
// (setq c-basic-offset 4)

#include "clsGLGuides.h"

#include "../GLOBJ/GLOBJTransCube.h"
#include "../GLCANVAS/clsMainGLCanvas.h"

#include <limits.h>

#define X 0
#define Y 1
#define Z 2

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

GLGuides_cls::
GLGuides_cls(int rotationDim){
    mArrowDir[0] = mArrowDir[1] = mArrowDir[2] = 1;
    mRotationDim = rotationDim;
}

void GLGuides_cls::
SetArrowDir(double xlength, double ylength, double zlength){
    mArrowDir[X] = xlength;
    mArrowDir[Y] = ylength;
    mArrowDir[Z] = zlength;
}

void GLGuides_cls::
DrawArrow(double length, int type){
    if (length == 0) return;
    double arrow;
    if (length > 0) arrow = length - 0.6;
    if (length < 0) arrow = length + 0.6;
    switch (type){
    case _GUIDES_TRANSLATE:
	glBegin(GL_TRIANGLE_FAN);
	glVertex3d(length, 0, 0);
	glVertex3d(arrow,  0,      -0.1);
	glVertex3d(arrow, -0.0707, -0.0707);
	glVertex3d(arrow, -0.1,     0);
	glVertex3d(arrow, -0.0707,  0.0707);
	glVertex3d(arrow,  0,       0.1);
	glVertex3d(arrow,  0.0707,  0.0707);
	glVertex3d(arrow,  0.1,     0);
	glVertex3d(arrow,  0.0707,  -0.0707);
	glVertex3d(arrow,  0,      -0.1);
	glEnd();
	break;
    case _GUIDES_SCALE:
	glBegin(GL_QUAD_STRIP);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glVertex3d(length + 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, -0.2, -0.2);
	glVertex3d(length + 0.2, +0.2, +0.2);
	glVertex3d(length + 0.2, +0.2, -0.2);
	glVertex3d(length - 0.2, +0.2, +0.2);
	glVertex3d(length - 0.2, +0.2, -0.2);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glEnd();
	glBegin(GL_QUADS);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, +0.2, +0.2);
	glVertex3d(length - 0.2, +0.2, +0.2);
	
	glVertex3d(length - 0.2, +0.2, -0.2);
	glVertex3d(length + 0.2, +0.2, -0.2);
	glVertex3d(length + 0.2, -0.2, -0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glEnd();
	break;
    }
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(length, 0, 0);
    glEnd();
}

void GLGuides_cls::
DrawGuide(double pos[3], double scale, int index, double *pScaleXYZ,
	  PositionAndDirection_cls CParentObjPos, int type, double *YPR){
    if (type == _GUIDES_ROTATE){
	DrawGuideRotation(pos, scale, index, pScaleXYZ, CParentObjPos, YPR);
	return; 
    }
    
    glPushMatrix();
    glTranslated(pos[X], pos[Y], pos[Z]);
    if (pScaleXYZ == NULL)
	glScaled(scale * 0.05, scale * 0.05, scale * 0.05);
    else
	glScaled(scale * 0.05 * pScaleXYZ[X],
		 scale * 0.05 * pScaleXYZ[Y],
		 scale * 0.05 * pScaleXYZ[Z]);

    GLOBJ::SetMatrix3dv3Inv(CParentObjPos.mPos, 
			    CParentObjPos.mTgt,
			    CParentObjPos.mTop);

    glPushName(index); // <--index
    glPushName(1);
    glColor4d(1, 0, 0, 1);
    DrawArrow(mArrowDir[X]*2, type);
    glPopName();
    glPushName(2);
    glColor4d(0, 1, 0, 1);
    glRotated(90, 0, 0, 1); DrawArrow(mArrowDir[Y]*2, type); glRotated(-90, 0, 0, 1);
    glPopName();
    glPushName(3);
    glColor4d(0.2, 0.2, 1, 1);
    glRotated(-90, 0, 1, 0); DrawArrow(mArrowDir[Z]*2, type); glRotated(90, 0, 1, 0);
    glPopName();
    glPopName(); // -->index
    glPopMatrix();
}

void GLGuides_cls::
DrawAxisAndCalcDirection(double X0, double Y0, double Z0,
			 double X1, double Y1, double Z1){
    double matrixModelView[16];
    double matrixProjection[16];
    int    matrixViewport[4];
    double posU0, posV0, posW0;
    double posU1, posV1, posW1;

    glBegin(GL_LINES);
    glVertex3d(X0, Y0, Z0); glVertex3d(X1, Y1, Z1);
    glEnd();
    
    glGetDoublev(GL_MODELVIEW_MATRIX, matrixModelView);
    glGetDoublev(GL_PROJECTION_MATRIX, matrixProjection);
    glGetIntegerv(GL_VIEWPORT, matrixViewport);
    gluProject(X0, Y0, Z0,
	       matrixModelView, matrixProjection, matrixViewport,
	       &posU0, &posV0, &posW0);
    gluProject(X1, Y1, Z1,
	       matrixModelView, matrixProjection, matrixViewport,
	       &posU1, &posV1, &posW1);
    mbDirectionInv = (posW0 > posW1) ? TRUE : FALSE;
}

void GLGuides_cls::
DrawRing(double radius, int axis){
    glBegin(GL_LINE_LOOP);
    for (int i = 0; i < 360; i+=15){
	float rcosphi = radius * cos(3.14159 * i / 180);
	float rsinphi = radius * sin(3.14159 * i / 180);
	glVertex3d(rcosphi * msTableU[axis][X] + rsinphi * msTableV[axis][X],
		   rcosphi * msTableU[axis][Y] + rsinphi * msTableV[axis][Y], 
		   rcosphi * msTableU[axis][Z] + rsinphi * msTableV[axis][Z]);
    }
    glEnd();
}

void GLGuides_cls::
DrawGuideRotation(double pos[3], double scale, int index,
		  double *pScaleXYZ,
		  PositionAndDirection_cls CParentObjPos,
		  double YPR[4]){
    glPushMatrix();
    glTranslated(pos[X], pos[Y], pos[Z]);
    if (pScaleXYZ == NULL)
	glScaled(scale * 0.05, scale * 0.05, scale * 0.05);
    else
	glScaled(scale * 0.05 * pScaleXYZ[X],
		 scale * 0.05 * pScaleXYZ[Y],
		 scale * 0.05 * pScaleXYZ[Z]);

    GLOBJ::SetMatrix3dv3/*Inv*/(CParentObjPos.mPos, 
				CParentObjPos.mTgt,
				CParentObjPos.mTop);

    glPushName(index); // <--index
    glLineWidth(2);

    int dimension = 1;

    if (dimension <= mRotationDim){
	glRotated(YPR[0], 0, 0 ,1);
	if (MainGLCanvas_cls::msMoveMode == _ROTATE_MOVE_Y){
	    glColor4d(1, 0.2, 0.2, 1);
	    DrawAxisAndCalcDirection(0, 0, -6, 0, 0, 6);
	    glLineWidth(3);
	}
	else{
	    glColor4d(0.7, 0.1, 0.1, 1);
	    glLineWidth(2);
	}
	
	glPushName(1);
	DrawRing(1.8, Z);
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0);
	glVertex3d(0, 1.8, 0);
	glEnd();
	glPopName();
    }
    dimension++;

    if (dimension <= mRotationDim){
	glRotated(YPR[1], 1, 0, 0);
	if (MainGLCanvas_cls::msMoveMode == _ROTATE_MOVE_P){
	    glColor4d(0.2, 1, 0.2, 1);
	    DrawAxisAndCalcDirection(-6, 0, 0, 6, 0, 0);
	    glLineWidth(3);
	}
	else{
	    glColor4d(0.1, 0.7, 0.1, 1);
	    glLineWidth(2);
	}
	
	glPushName(2);
	DrawRing(1.5, X);
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0);
	glVertex3d(0, 0, 1.5);
	glEnd();
	glPopName();
    }
    dimension++;

    if (dimension <= mRotationDim){
	glRotated(YPR[2], 0, 1, 0);
	if (MainGLCanvas_cls::msMoveMode == _ROTATE_MOVE_R){
	    glColor4d(0.3, 0.3, 1, 1);
	    DrawAxisAndCalcDirection(0, -6, 0, 0, 6, 0);
	    glLineWidth(3);
	}
	else {
	    glColor4d(0.1, 0.1, 0.8, 1);    
	    glLineWidth(2);
	}
	
	glPushName(3);
	DrawRing(Y, 1.2);
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0);
	glVertex3d(1.2, 0, 0);
	glEnd();
	glPopName();
    }
    dimension++;

    if (dimension <= mRotationDim){
	if (MainGLCanvas_cls::msMoveMode == _ROTATE_MOVE_Y2){
	    glColor4d(1, 0.2, 0.2, 1);
	    DrawAxisAndCalcDirection(0, 0, -6, 0, 0, 6);
	    glLineWidth(3);
	}
	else {
	    glColor4d(0.7, 0.1, 0.1, 1);
	    glLineWidth(2);
	}
	
	glPushName(4);
	DrawRing(0.9, Z);
	glEnd();
	glPopName();
    }
    
    glPopName(); // -->index

    glLineWidth(1);
    
    glPopMatrix();
};

void GLGuides_cls::
DrawGuideArrowBox(double posBox[3], double scale,
		  double boxScale[3], bool bLimitBox){
    
    double scaleInv[3], center[3], tmpPos[3];
    LOOP_XYZ{
	if (boxScale[xyz] == 0) boxScale[xyz] = 0.1;
	scaleInv[xyz] = 1.0/boxScale[xyz];
    }

    if (bLimitBox == TRUE){
	LOOP_XYZ if (posBox[xyz] < 0) posBox[xyz] = 0;
	LOOP_XYZ if (posBox[xyz+3] < 0) posBox[xyz+3] = 0;
	LOOP_XYZ if (posBox[xyz] > 1) posBox[xyz] = 1;
	LOOP_XYZ if (posBox[xyz+3] > 1) posBox[xyz+3] = 1;
    }
    
    GLOBJ::DrawWireBoxMinMax(posBox, posBox+3);
    LOOP_XYZ center[xyz] = (posBox[xyz] + posBox[xyz+3]) / 2.0;

    PositionAndDirection_cls CTmp;

    SetArrowDir(1, 1, 1);
    DrawGuide(center, scale*0.4, 1, scaleInv, CTmp);
    
    tmpPos[X] = posBox[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(-1, 0, 0);
    DrawGuide(tmpPos, scale*0.25, 2, scaleInv, CTmp);
    tmpPos[X] = posBox[X+3]; tmpPos[Y] = center[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(1, 0, 0);
    DrawGuide(tmpPos, scale*0.25, 3, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = posBox[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(0, -1, 0);
    DrawGuide(tmpPos, scale*0.25, 4, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = posBox[Y+3]; tmpPos[Z] = center[Z];
    SetArrowDir(0, 1, 0);
    DrawGuide(tmpPos, scale*0.25, 5, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = posBox[Z];
    SetArrowDir(0, 0, -1);
    DrawGuide(tmpPos, scale*0.25, 6, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = posBox[Z+3];
    SetArrowDir(0, 0, 1);
    DrawGuide(tmpPos, scale*0.25, 7, scaleInv, CTmp);
}

std::vector <int> GLGuides_cls::
GetObjectFromList(){
    int hitObj = 0;
    int hitGuide = 0;
    mHitNum = glRenderMode(GL_RENDER); // ZNg[hIGqbg𓾂
    
    GLuint *pBuf = mBuffer4Selection;
    GLuint depthMinMax = UINT_MAX;
    for (int i = 0; i < mHitNum; i++){
	GLuint objNum = pBuf[0]; //pBuf[3]ȍ~int񂪕ł邩
	GLuint depthMin = pBuf[1];
	GLuint depthMax = pBuf[2];
	GLuint CurObj   = pBuf[3]; 
	GLuint CurGuide = 0;
	if (objNum == 2) CurGuide = pBuf[4];
	if (CurObj > 0 && CurObj < 8){
	    hitObj      = CurObj;
	    hitGuide    = CurGuide;
	    depthMinMax = depthMin;
	}
	pBuf += (3 + objNum);
    }
    std::vector <int> retVec;
    retVec.push_back(hitObj); retVec.push_back(hitGuide);
    return retVec;
}

void GLGuides_cls::
InitializeProjection(){
    glSelectBuffer(100, mBuffer4Selection);
    glRenderMode(GL_SELECT);
    glInitNames();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
}

std::vector < int > GLGuides_cls::
GetSelection(){

    std::vector < int > retVec = GetObjectFromList();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    return retVec;
}

int GLGuides_cls::msTableU[3][3] = {{0,1,0}, {0,0,1}, {1,0,0}};
int GLGuides_cls::msTableV[3][3] = {{0,0,1}, {1,0,0}, {0,1,0}};
