// -*- c -*-
//(setq c-basic-offset 4)

#include "clsGLPerspective.h"

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#define X 0
#define Y 1
#define Z 2

const double GLPerspective_cls::sDefaultCameraPos[7][3] = 
{{-1.12,-0.8,+0.56},
 {-1,  0,  0}, { +1,  0,  0},
 { 0, -1,  0}, {  0, +1,  0},
 { 0,  0, -1}, {  0,  0, +1}
};

const double GLPerspective_cls::sDefaultCameraTop[7][3] = 
{{ 0,  0, +1},
 { 0,  0, +1}, {  0,  0, +1},
 { 0,  0, +1}, {  0,  0, +1},
 { 0, +1,  0}, {  0, +1,  0}
};

GLPerspective_cls::
GLPerspective_cls(int width, int height, double deg, double depthMIN, double depthMAX){
    mbPerspective = TRUE;
    mDeg = deg;
    mDepthMIN = depthMIN;
    mDepthMAX = depthMAX;
    mWidth = width; mHeight = height;
}

double GLPerspective_cls::
GetAspectRatio(){
    return (double)mWidth / mHeight;
}

void GLPerspective_cls::
SetViewportOnGL(){
    glViewport(0, 0, (GLint) mWidth, (GLint) mHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
}

void GLPerspective_cls::
SetPerspectiveOnGL(PositionAndDirection_cls* pCConfigure){
    if (mbPerspective)
	gluPerspective(mDeg, GetAspectRatio(), mDepthMIN, mDepthMAX);
    else{
	float vsize = (float)(pCConfigure->GetDistanceToAnc() * 0.5);
	glOrtho(-vsize * GetAspectRatio(), vsize * GetAspectRatio(),
		-vsize, vsize, -100, 100);
    }
}

void GLPerspective_cls::
SetModelOrProjectionViewOnGL(PositionAndDirection_cls* pCConfigure, GLenum mode){
    double mat[16];
    glMatrixMode(mode);
    glLoadIdentity();
   
    mat[0]  =  pCConfigure->mRgt[X];
    mat[1]  =  pCConfigure->mTop[X];
    mat[2]  = -pCConfigure->mTgt[X];
    mat[3]  = 0;
    mat[4]  =  pCConfigure->mRgt[Y];
    mat[5]  =  pCConfigure->mTop[Y];
    mat[6]  = -pCConfigure->mTgt[Y];
    mat[7]  = 0;
    mat[8]  =  pCConfigure->mRgt[Z];
    mat[9]  =  pCConfigure->mTop[Z];
    mat[10] = -pCConfigure->mTgt[Z];
    mat[11] = mat[12] = mat[13] = mat[14] = 0;
    mat[15] = 1;
    glMultMatrixd(mat);
    
    glTranslated(-pCConfigure->mPos[X],
		 -pCConfigure->mPos[Y],
		 -pCConfigure->mPos[Z]);

}

void GLPerspective_cls::
SetWindowSize(int width, int height){
    mWidth  = width;
    mHeight = height;
}

void GLPerspective_cls::
Push2DMatrix(int width, int height){
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glViewport( 0, 0, width, height);
    glLoadIdentity();
    gluOrtho2D(0, width, 0, height);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
}

void GLPerspective_cls::
Pop2DMatrix(){
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW); 
}
