// -*- c -*-
// "clsPositionAndDirection.cpp"

#ifndef _POSITION_AND_DIRECTION_H_
#define _POSITION_AND_DIRECTION_H_

#define _DIRECTION_FREE 0
#define _DIRECTION_X_PLUS  1
#define _DIRECTION_X_MINUS 2
#define _DIRECTION_Y_PLUS  3
#define _DIRECTION_Y_MINUS 4
#define _DIRECTION_Z_PLUS  5
#define _DIRECTION_Z_MINUS 6

#define _ROT_INVERSE 1

#define _POS_AND_DIR_MOVE_X 0
#define _POS_AND_DIR_MOVE_Y 1
#define _POS_AND_DIR_MOVE_Z 2
#define _POS_AND_DIR_MOVE_FREE 4

#define _ROTATE_MOVE_Y 0
#define _ROTATE_MOVE_P 1
#define _ROTATE_MOVE_R 2
#define _ROTATE_MOVE_Y2 3


#include <string>
#include <wx/wx.h>

class Direction_cls{ 
 public:
    double mTgt [3], mTop [3], mRgt [3];
    double mTgtV[3], mTopV[3], mRgtV[3];

    Direction_cls(double tgtX =  0, double tgtY =  0, double tgtZ = -1,
		  double topX =  0, double topY =  1, double topZ =  0,
		  double rgtX =  1, double rgtY =  0, double rgtZ =  0);
    Direction_cls(Direction_cls &CDir);

    void Setv(double tgt[3], double top[3]);
    void SetvUnitZ(int direction = 1);
    void RejectInvalidDirection();
    void NormalizeAndOrthogonalize();
    
    void SetDirection(unsigned int type);

    // -----
    void   RotateYPR(double Yaw, double Pitch, double Roll);
    void   RotateYPR3d(double* pYPR);
    void   RotateYPR4d(double* pYPRY, bool bInverse = FALSE);
    void   RotateYaw(double Yaw);
    void   RotatePitch(double Pitch);
    void   RotateRoll(double Roll);

    void   RotTransformToThisCoordinate(double vec[3]);

    bool ReadStrLine(std::wstring strLine);
    wxString Serialize();
};

bool operator==(const Direction_cls& CLeft, const Direction_cls& CRight);

class Position_cls{ 
 public:
    double mPos[3];
    double mVel[3];
    
    Position_cls(double posX =  0, double posY =  0, double posZ =  0,
		 double velX =  0, double velY =  0, double VelY =  0);
    Position_cls(const Position_cls &CPos);
    void Set(Position_cls &CPos);
    void SetZero();

    bool ReadStrLine(std::wstring strLine);
    wxString Serialize();
};

bool operator==(const Position_cls& CLeft, const Position_cls& CRight);

class PositionAndDirection_cls : public Position_cls, public Direction_cls { 
 public:
    
    PositionAndDirection_cls
	(double posX =  0, double posY =  0, double posZ =  0,
	 double velX =  0, double velY =  0, double VelY =  0,
	 double tgtX =  0, double tgtY =  0, double tgtZ = -1,
	 double topX =  0, double topY =  1, double topZ =  0,
	 double rgtX =  1, double rgtY =  0, double rgtZ =  0);
    PositionAndDirection_cls
	(Position_cls &CPos, Direction_cls &CDir);

    static bool msbMatrixMove[5][3];
    
    void SetSensitivity(double sensitivity);
    void SetAspectRatio(double aspectRatio);

    double GetDistanceToAnc();
    double GetProjectedDistanceToAnc();

 private:
    double mAncPos[3];

 public:
    void   SetAncPos(double pos[3]);
    void   TrackBall(float prevX, float prevY, float currX, float currY);
    void   Zoom     (float prevX, float prevY, float currX, float currY);
    void   Translate(float prevX, float prevY, float currX, float currY, int mode = _POS_AND_DIR_MOVE_FREE, PositionAndDirection_cls *pCParentCoord = NULL);
    void   PointAt(double point[3]);
    void   PointAtAnc();

    void RotateAroundOriginYPR(double Yaw, double Pitch, double Roll);
    void RotateVectorZ(double pVec[3], double cosYaw, double sinYaw);
    void RotateAroundOriginYaw(double Yaw);
    void RotateVectorY(double pVec[3], double cosPitch, double sinPitch);
    void RotateAroundOriginPitch(double Pitch);
    void RotateVectorX(double pVec[3], double cosRoll, double sinRoll);
    void RotateAroundOriginRoll(double Roll);

    void GLSetMatrixToThisCoordinate();
    PositionAndDirection_cls GetPosAndDirectionInThisCoordinate
	(PositionAndDirection_cls CPosAndDir);
    //void LogPosdata();

    bool ReadStrLine(std::wistream &istream);
    wxString Serialize();
		   
 public:
    double mSensitivity;
    double mATranslateScale[3];
    double mAspectRatio;
};

bool operator==(const PositionAndDirection_cls& CLeft,
		const PositionAndDirection_cls& CRight);
#endif
