
#ifndef _TEXTURE_PROPERTY_H_
#define _TEXTURE_PROPERTY_H_

#include "../CONFITEM/clsConfItem.h"
#include "../POVRAY/clsPovSerialize.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/clsDialogEXKey.h"
#include "../WIDGETCTRL/clsDialogController.h"
#include "../WIDGETCTRL/clsMinimizeButton.h"
#include "../MYGL/clsPositionAndDirection.h"

#include <vector>
#include <wx/wx.h>

#define _MAPPING_GLOBAL 0
#define _MAPPING_LOCAL 1

#define _IMAGEMAP_PLANER 0
#define _IMAGEMAP_SPHERICAL 1
#define _IMAGEMAP_CYLINDER 2

#define _TEX_NORMAL_MAX 2

#define _TEX_MAPPING_VALUE 0
#define _TEX_MAPPING_PATTERN 1
#define _TEX_MAPPING_IMAGE 2
#define _TEX_MAPPING_PLUGINS 3

#define _TRANSMIT_TRANSMIT 1
#define _TRANSMIT_FILTER 2
#define _TRANSMIT_BOTHEFFECT 3

#define _TEX_RENDER_ALPHA 0
#define _TEX_RENDER_ADD 1

#define _TEX_NO_MIRROR 0
#define _TEX_MIRROR_MIRROR 1
#define _TEX_MIRROR_METAL 2

#define _TEX_PATTERN_AGATE 0
#define _TEX_PATTERN_AOI 1
#define _TEX_PATTERN_BOZO 2
#define _TEX_PATTERN_BRICK 3
#define _TEX_PATTERN_CELLS 4
#define _TEX_PATTERN_CHECKER 5
#define _TEX_PATTERN_CRACKLE 6
#define _TEX_PATTERN_CYLINDRICAL 7
#define _TEX_PATTERN_DENTS 8
#define _TEX_PATTERN_GRADIENT 9
#define _TEX_PATTERN_GRANITE 10
#define _TEX_PATTERN_HEXAGON 11
#define _TEX_PATTERN_RADIAL 12

#define _NORMAL_PATTERN_FLAT 0
#define _NORMAL_PATTERN_AGATE 1
#define _NORMAL_PATTERN_BOZO 2
#define _NORMAL_PATTERN_BUMPS 3
#define _NORMAL_PATTERN_BRICK 4
#define _NORMAL_PATTERN_CELLS 5
#define _NORMAL_PATTERN_CHECKER 6
#define _NORMAL_PATTERN_CRACKLE 7
#define _NORMAL_PATTERN_DENTS 8
#define _NORMAL_PATTERN_FACETS 9
#define _NORMAL_PATTERN_MARBLE 10
#define _NORMAL_PATTERN_RIDGED_MF 11

#define _TEX_PLUGINGS_GRASSTEX 0

class FilterWithParameter_cls;
class TexProperty_cls;
class TexPropertyCtrl_cls;
class PanelColorBar_cls; 
class PovRay_cls;

class TexPropertyGUI_cls : public wxDialogEXKey, public MinimizeButton_cls {
 public:
    TexPropertyGUI_cls(TexPropertyCtrl_cls* pCParent,
		       wxWindow* pCWindow,
		       wxPoint pos, wxSize size);
    ~TexPropertyGUI_cls();

    static TexPropertyCtrl_cls *mspCParent;
    static TexPropertyGUI_cls *mspCThis;

    void OnMouse(wxMouseEvent &event);
    void OnMenuMinimize(wxCommandEvent &event);
    void OnButtonDefault     (wxCommandEvent &event);
    void OnButtonFadeColor   (wxCommandEvent &event);
    void OnButtonShowInterior(wxCommandEvent &event);
    void OnButtonShowBump    (wxCommandEvent &event);
    void OnButtonShowMirror  (wxCommandEvent &event);
    void OnButtonShowOptions (wxCommandEvent &event);
    void OnButtonHelp        (wxCommandEvent &event);

    int  mIndexNormalDegree[_TEX_NORMAL_MAX];
    int  mIndexNormalScale[_TEX_NORMAL_MAX];
    int  mIndexNormalTurb[_TEX_NORMAL_MAX];
    int  mIndexColorType, mIndexColorBar, mIndexFadeClrBtn, mIndexPluginHelp;
    int  mIndexMappingType, mIndexPatternType, mIndexPluginType;
    int  mIndexPluginMappingOptional;
    int  mIndexMappingPos, mIndexMappingScale, mIndexMappingTurb, mIndexMappingScaleAndTurb;
    int  mIndexMappingAxis, mIndexMappingRepeat, mIndexMappingAxisAndRepeat;
    int  mIndexReflexMin, mIndexReflexMax, mIndexReflexType;
    int  mIndexAlpha, mIndexRenderType;
    int  mIndexFadeDistance, mIndexIOR, mIndexDispersion;
    int  mIndexImagePath, mIndexImageType;
    void UpdateColorList();
    void HideAndShow();

    static void HookAssign(void *pArg, void *pEvent);
    static void HookCombo(void *pArg, void *pEvent);
    static void HookPanelClick(void *pArg, void *pEvent);

    static double GetColorR(double value, void *pClue);
    static double GetColorG(double value, void *pClue);
    static double GetColorB(double value, void *pClue);

    PanelColorBar_cls *mpCPanelColorBar;
    MultiNumericalInputsPanel_cls* mpCMNumMain;
    MultiNumericalInputsPanel_cls* mpCMNumInterior;
    MultiNumericalInputsPanel_cls* mpCMNumBump;
    MultiNumericalInputsPanel_cls* mpCMNumMirror;
    MultiNumericalInputsPanel_cls* mpCMNumOptions;
    void SetMNumMain(wxBoxSizer *pBoxSizer);
    void SetMNumBump(wxBoxSizer *pBoxSizer);
    void SetMNumMirror(wxBoxSizer *pBoxSizer);
    void SetMNumInterior(wxBoxSizer *pBoxSizer);
    void SetMNumOptions(wxBoxSizer *pBoxSizer);

    wxButtonEX *mpCBTNInterior, *mpCBTNBump, *mpCBTNMirror, *mpCBTNOptions;
    //bool mbMinimizeInterior, mbMinimizeBump;
    //bool mbMinimizeMirror, mbMinimizeOptions;
    void UpdateButtonMinimize();

 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class TexPropertyCtrl_cls :
public DialogControl_cls
{
 public:
    TexPropertyCtrl_cls(wxWindow *pC, wxBitmap *pBitmap);
    ~TexPropertyCtrl_cls();
    TexPropertyGUI_cls* mpCGUI;
    wxBitmap *mpBitmapLink;
    MultiNumericalInputsPanel_cls* mpCParentMNumPanel;
    wxRect mCScreenRect;

    void UpdateColorType();
    void UpdateColorBar();
    void ConnectProperty(TexProperty_cls *pCConnected, wxString strItemName,
			 MultiNumericalInputsPanel_cls *pCParentMNumPanel = NULL);
    void UpdateLinkedProperty();
    TexProperty_cls *mpCConnected;
    wxString mstrItemName;
    unsigned int   mDummyColorType;
    float mDummySpecular, mDummyDiffuse, mDummyDiffuseBack;
    float mDummyAmbient,  mDummyRoughness;
    unsigned int   mDummyTransmitType;
    float mDummyColor[3];
    unsigned int   mDummyMappingType;
    unsigned int   mDummyMappingCoord;
    unsigned int   mDummyPatternType;
    unsigned int   mDummyPluginType;
    float mDummyMappingPos[3];
    float mDummyMappingScale, mDummyMappingTurb;
    unsigned int   mDummyMappingAxis;
    bool  mbDummyMappingRepeat;
    bool  mbDummyMappingOptional;
    float mGrayColor[3];
    float mDummyAlpha;
    float mDummySubsurfaceFactor;
    unsigned int   mDummyRenderType;
    
    float mDummyFadeDistance;
    float mDummyFadeColor[3], mDummySubsurfaceColor[3];
    float mDummyIOR;
    float mDummyDispersion;

    int   mDummyReflectionType;
    float mDummyReflectionMin, mDummyReflectionMax;
    
    int   mDummyNormalPattern[_TEX_NORMAL_MAX];
    float mDummyNormalDegree[_TEX_NORMAL_MAX];
    float mDummyNormalScale[_TEX_NORMAL_MAX];
    float mDummyNormalTurb[_TEX_NORMAL_MAX];
    
    wxString mstrDummyImagePath;
    int   mDummyImageMappingType;
    int   mDummyImageColoringType;

    bool  mDummyImageOn, mDummyShadowOn;
    bool  mDummyRadiosityOn, mDummyReflectionOn;

    void  Apply();

    void UpdateItems(bool bForce = FALSE);
    void UpdateColorList();
    void SetToDefault();
    void InitGUI(wxWindow *pC);

    void OpenFadeColorDLG();
    void OpenColorDLG(int x, int y);

    //bool  UpdateValues();
    bool  IsGUIShownOnScreen();
};

class TexProperty_cls : public PovSerialize_cls {
 public:
    TexProperty_cls();
    void LinkFilterParameter(FilterWithParameter_cls *pCFilterParamLinked,
			     int linkedParamIndex);
    void LinkAlphaParameter(double *pAlpha);
    
    static PovRay_cls *mspCPovRayLink;
    static double msBoxMinWorld[3], msBoxMaxWorld[3];
    static double msBoxRotParam[4]; 
    static double g2boxA[3], g2boxB[3], gcenter[3];
    static void SetBoxGeom(double min[3], double max[3], double rotParam[4]);
    static bool IsRotationRequired();
    static void ConvertGlobal2LocalNoRot(double posGlobal[3], double posLocal[3]);
    static void ConvertGlobal2Local(double posGlobal[3], double posLocal[3]);
    static double msRotZ0, msRotX0, msRotY0, msRotZ1;
    static double msCosZ1, msSinZ1;
    static double msCosY0, msSinY0;
    static double msCosX0, msSinX0;
    static double msCosZ0, msSinZ0;

    double *mpAlphaLinked;    
    FilterWithParameter_cls *mpCFilterParamLinked;
    int mLinkedParamIndex;
    double GetLinkedParam();
    
    bool IsHalfMapping();
    bool IsUVPatternReady();

    unsigned int *mpColorIndexLinked; // N悪!=0 ȂJX^J[
    
    ConfItemList_cls       mCConfList;
    ConfItemUINT_cls       mCColorType;
    ConfItemFLOATARRAY_cls mCColor;
    ConfItemUINT_cls       mCMappingType;
    ConfItemUINT_cls       mCMappingCoord;
    ConfItemUINT_cls       mCPatternType;
    ConfItemUINT_cls       mCPluginType;
    ConfItemFLOATARRAY_cls mCMappingPos;
    ConfItemFLOAT_cls      mCMappingScale, mCMappingTurb;
    ConfItemUINT_cls       mCMappingAxis;
    ConfItemBOOL_cls       mCbMappingRepeat;
    ConfItemBOOL_cls       mCbMappingOptional;
    ConfItemFLOAT_cls      mCSpecular, mCDiffuse, mCDiffuseBack;
    ConfItemFLOAT_cls      mCAmbient, mCRoughness, mCAlpha;
    ConfItemFLOAT_cls      mCSubsurfaceFactor;
    ConfItemUINT_cls       mCTransmitType;
    ConfItemUINT_cls       mCRenderType;
    ConfItemFLOATARRAY_cls mCFadeColor, mCSubsurfaceColor;
    ConfItemFLOAT_cls      mCFadeDistance;
    ConfItemFLOAT_cls      mCIOR;
    ConfItemFLOAT_cls      mCDispersion;
    ConfItemUINT_cls       mCReflectionType;
    ConfItemFLOAT_cls      mCReflectionMin, mCReflectionMax;
    ConfItemUINT_cls       mCNormalPattern[_TEX_NORMAL_MAX];
    ConfItemFLOAT_cls      mCNormalDegree[_TEX_NORMAL_MAX];
    ConfItemFLOAT_cls      mCNormalScale[_TEX_NORMAL_MAX];
    ConfItemFLOAT_cls      mCNormalTurb[_TEX_NORMAL_MAX];
    ConfItemSTR_cls        mCStrImagePath;
    ConfItemUINT_cls       mCImageMappingType;
    ConfItemUINT_cls       mCImageColoringType;
    ConfItemBOOL_cls       mCbShadowOn, mCbRadiosityOn;
    ConfItemBOOL_cls       mCbImageOn, mCbReflectionOn;

    static ConfItemBOOL_cls msCConfBTNInterior;
    static ConfItemBOOL_cls msCConfBTNBump;
    static ConfItemBOOL_cls msCConfBTNMirror;
    static ConfItemBOOL_cls msCConfBTNOptions;
    
    void SetToDefault();
    void ReDefineList();

    std::vector < float > GetV4AmbientColor(float value = 0);
    std::vector < float > GetV4DiffuseColor(float value = 0);
    std::vector < float > GetV4SpecularColor(float value = 0);
    
    void CalcScale(float scale[3], double normScale,
		   bool bGlobalCoord, int rotAxis);
    wxString GetPovStrFinish(bool bShading);
    
    wxString mstrDensityFile;
    wxString mstrTranslationForDensityFile;
    std::vector < wxString > GetVPovStrColorMap(int df3index, float trans);
    std::vector < wxString > GetVPovStrColors(int df3index, float trans);
    void SetDensityFile(wxString strDensityFile);
    void SetTranslationForDensityFile(wxString strTranslate);
    wxString GetPovStrNormalAll
	(float scaleFactorX = 1, float scaleFactorY = 1, float scaleFactorZ = 1,
	 bool bShading = TRUE, bool bGlobalCoord = FALSE, 
	 int rotAxis = -1);
    wxString GetPovStrTex
	(float alpha, wxString strFinishId, wxString strColorMap = wxT(""),
	 float scaleFactorX = 1, float scaleFactorY = 1, float scaleFactorZ = 1,
	 bool bShading = TRUE, bool bGlobalCoord = FALSE, 
	 int rotAxis = -1, 
	 wxString strClr1 = wxT(""), wxString strClr2 = wxT(""),
	 wxString strClr3 = wxT(""));
    wxString GetPovStrObjectOption();
    wxString GetPovStrPluginsParameters(PositionAndDirection_cls CCamPos,
					wxString strClr1, wxString strClr2,
					wxString strClr3);
    wxString GetPovStrPluginsInObject(wxString strFinishId);
    wxString GetPovStrInterior();
    wxString GetPovStrNormal(int index, float scale[3]);
    wxString GetPovStrPattern(wxString strClrMap, 
			      wxString strClr1, wxString strClr2,
			      wxString strClr3);

    // ----

    wxString Serialize();
    static wxString SerializeStaticConf();
    void ReadStrLine(std::wstring strLine);
    static void ReadStrLineStaticConf(std::wstring strLine);
};

#endif
