// -*- c -*-
// "clsKeyFrame.h"

#include "clsKeyFrame.h"

#include <fstream>
#include <sstream>

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#ifdef WIN32
#define ENDLL L"\r\n"
#else
#define ENDLL L"\n"
#endif

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

KeyFrame_cls::
KeyFrame_cls() : 
    mDimention(0), mType(_PARAM_HERMITE_AUTO) 
{
    mpValue  = NULL;
    mpDValue1 = NULL;
    mpDValue2 = NULL;
    mbSelected = FALSE;
}

KeyFrame_cls::
KeyFrame_cls(unsigned int dimention):
    mDimention(dimention), mType(_PARAM_HERMITE_AUTO)
{
    if (mDimention > 0){
	mpValue  = new double [mDimention];
	mpDValue1 = new double [mDimention];
	mpDValue2 = new double [mDimention];
    }
    mbSelected = FALSE;
}

KeyFrame_cls::
~KeyFrame_cls(){
    if (mpValue == NULL) return;
    if (mDimention > 0){
	delete [] mpValue;
	delete [] mpDValue1;
	delete [] mpDValue2;
    }
}

KeyFrame_cls:: // Rs[RXgN^
KeyFrame_cls(const KeyFrame_cls &obj){
    mDimention = 0;
    mpValue = NULL;
    FullCopy(obj);
}

void KeyFrame_cls::
CopyWithoutConstruct(const KeyFrame_cls &obj){
    if (mDimention != obj.mDimention) throw ("dimension discrepancy");
    mbSelected   = obj.mbSelected;
    mType        = obj.mType;
    mDoubleIndex = obj.mDoubleIndex;
    for (unsigned int i = 0; i < mDimention; i++){
	mpValue[i]  = obj.mpValue[i];
	mpDValue1[i] = obj.mpDValue1[i];
	mpDValue2[i] = obj.mpDValue2[i];
    }
}

void KeyFrame_cls::
FullCopy(const KeyFrame_cls& obj){
    if (mpValue != NULL && mDimention > 0){
	delete [] mpValue;
	delete [] mpDValue1;
	delete [] mpDValue2; }

    mDimention = obj.mDimention;
    mbSelected = obj.mbSelected;
    mType      = obj.mType;

    if (mDimention > 0){
	mpValue    = new double [mDimention];
	mpDValue1  = new double [mDimention];
	mpDValue2  = new double [mDimention];
    }
    mDoubleIndex = obj.mDoubleIndex;
    for (unsigned int i = 0; i < mDimention; i++){
	mpValue[i]  = obj.mpValue[i];
	mpDValue1[i] = obj.mpDValue1[i];
	mpDValue2[i] = obj.mpDValue2[i];
    }
}

int KeyFrame_cls::operator > (const KeyFrame_cls obj){
    return mDoubleIndex > obj.mDoubleIndex; 
}

unsigned int KeyFrame_cls::
GetDimension(){
    return mDimention;
}

void KeyFrame_cls::
Set(double doubleIndex, unsigned int type,
    double *pFrom, double *pFromD1, double *pFromD2, 
    /*unsigned int type, */bool bSelected){
    mDoubleIndex = doubleIndex;
    mType = type;
    mbSelected = bSelected;
    for (unsigned int i = 0; i < mDimention; i++){
	mpValue[i]  = pFrom[i];
	if (pFromD1 != NULL)
	    mpDValue1[i] = pFromD1[i];
	else
	    mpDValue1[i] = 0;
	if (pFromD2 != NULL)
	    mpDValue2[i] = pFromD2[i];
	else
	    mpDValue2[i] = 0;
    }
}

void KeyFrame_cls::
SetValue(double *pFrom){
    for (unsigned int i = 0; i < mDimention; i++)
	mpValue[i] = pFrom[i];
}

bool KeyFrame_cls::
Select(double doubleIndex, double resolution, bool bTest){
    if (doubleIndex < mDoubleIndex + resolution &&
	doubleIndex > mDoubleIndex - resolution){
	if (bTest == FALSE) mbSelected = TRUE;
	return TRUE; }
    return FALSE;
}

bool KeyFrame_cls::
Toggle(double doubleIndex, double resolution){
    if (doubleIndex < mDoubleIndex + resolution &&
	doubleIndex > mDoubleIndex - resolution){
	mbSelected = !mbSelected;
	return TRUE;
    }
    return FALSE;
}

void KeyFrame_cls::
DrawKey(wxAutoBufferedPaintDC &dc, int x, int mode, bool bOnPanel){
    switch (mode){
    case _KEY_NORMAL:
	if (mbSelected == TRUE)
	    dc.SetBrush(wxBrush(wxT("YELLOW")));
	else
	    dc.SetBrush(*wxGREY_BRUSH);
	break;
    case _KEY_DRAG:
	if (bOnPanel == TRUE){
	    dc.SetPen(*wxGREY_PEN);
	    dc.SetBrush(*wxGREY_BRUSH);
	}
	else{
	    dc.SetPen(*wxLIGHT_GREY_PEN);
	    dc.SetBrush(*wxTRANSPARENT_BRUSH);
	}
    }

    static wxPoint ADiaPoints[4] = {
	wxPoint(-4, 0), wxPoint(0, -4), wxPoint(4, 0), wxPoint(0, 4)};

    switch(mType){
    case _PARAM_LINEAR:
	dc.DrawPolygon(4, ADiaPoints, x, 8); 
	break;
    case _PARAM_HERMITE_AUTO:
	dc.DrawRoundedRectangle(x-3, 5, 6, 6, 2); 
	break;
    case _PARAM_HERMITE_ADJ:
	dc.DrawCircle(x, 8, 4); 
	break;
    case _PARAM_HERMITE:
	dc.DrawCircle(x, 8, 4); dc.DrawCircle(x, 8, 2); 
	break;
    case _PARAM_STEP:
	dc.DrawRectangle(x-4, 5, 8, 8); 
	break;
    }
};

wxString KeyFrame_cls::
Serialize(){
    std::wstringstream ssOut;
    ssOut.precision(9);
    ssOut << mDoubleIndex << L" " << mDimention << L" " << mType;
    for (unsigned int i = 0; i < mDimention; i++)
	ssOut << L" " << mpValue[i];
    for (unsigned int i = 0; i < mDimention; i++)
	ssOut << L" " << mpDValue1[i];
    for (unsigned int i = 0; i < mDimention; i++)
	ssOut << L" " << mpDValue2[i];
    ssOut << ENDLL;
    return wxConvFile.cWC2WX(ssOut.str().c_str());
}

void KeyFrame_cls::
ReadStrLine(std::wstring strLine){
    std::wistringstream istream(strLine);
    istream >> mDoubleIndex >> mDimention >> mType; 
    for (unsigned int i = 0; i < mDimention; i++)
	istream >> mpValue[i];
    for (unsigned int i = 0; i < mDimention; i++)
	istream >> mpDValue1[i];
    for (unsigned int i = 0; i < mDimention; i++)
	istream >> mpDValue2[i];
    return;
}

void KeyFrame_cls::
InitializeStaticMembers(){
    msAstrInterp[0] = _("LINEAR");
    msAstrInterp[1] = _("HERMITE (AUTO)");
    msAstrInterp[2] = _("HERMITE (SEMIAUTO)");
    msAstrInterp[3] = _("HERMITE");
    msAstrInterp[4] = _("STEP");
}

wxString KeyFrame_cls::msAstrInterp[5];
