// -*- c -*-
// "clsMultiParameters.cpp"

#ifndef _MULTI_PARAMETERS_H_
#define _MULTI_PARAMETERS_H_

#include "../CONFITEM/clsConfItem.h"
#include "../WIDGETCTRL/clsDialogEXKey.h"
#include "../MAINCANVAS/clsDrawItem.h"
#include "clsParameters.h"

#include <vector>
#include <string>

#include "../MYGL/clsDrawPath.h"

class DF3SeqGUI_cls;
class GLGuides_cls;
class MultiParameters_cls;
class MultiNumericalInputsPanel_cls;

#define _MAX_ITEM_NUMBER 64

class MultiParametersGUI_cls : 
public wxDialogEXKey, DrawPath_cls
{ 
 public:
    MultiParametersGUI_cls(MultiParameters_cls *pCParent,
			   wxWindow *pC,  wxString strTitle,
			   wxPoint pos, wxSize size, 
			   wxBitmap *pBitmap = NULL, long style = wxDEFAULT_DIALOG_STYLE);

    std::vector < int > mVCDisplayList;
    std::vector < wxIcon > mVIcon;
    
    MultiParameters_cls* mpCParent;

    MultiNumericalInputsPanel_cls *mpCMNumMain;
    MultiNumericalInputsPanel_cls *mpCMNumParam;
    
    int mIndexIcon, mIndexAdd, mIndexDelete;
    
    void SetMatrix(Parameters_cls *pParam, int keyIndex,
		   double dpos[3], int flag);
    void DrawSelectedPath(GLGuides_cls *pCGLGuide, double dpos[3], unsigned int listGL, double scale, unsigned int pathIndex, PositionAndDirection_cls CParentObjPos);
    
    void DrawPath(unsigned int i, bool bSelectedPanel,
		  unsigned int listGL, double scale);

    void SetMNumMain(wxBoxSizer *pBoxSizer, 
		     unsigned int idAdd, unsigned int idRemove);
    virtual void SetMNumParam(wxBoxSizer *pBoxSizer) = 0;

    virtual void UpdateComboBoxPropagate() = 0;
    void OnSelector(wxCommandEvent &event);
    void UpdateComboBox();
    
    void OnButtonAdd(wxCommandEvent &event);
    void OnButtonRemove(wxCommandEvent &event);
    bool RemoveItem();
    virtual void UpdateActiveOnTL() = 0;
    void OnButtonClose     (wxCommandEvent &event);
};

class MultiParameters_cls : public DrawItem_cls {
 public:
    MultiParameters_cls(std::wstring strItemNum, std::wstring strTag,
			unsigned int dimension, bool bConnectToTLPanel = TRUE);
    ~MultiParameters_cls();

    bool mbConnectToTLPanel;

    std::wstring mstrTag;
    wxString mstrLabelForPanel;
    unsigned int mDimension;
    double *mpInit;

    static void InitializeStaticMembers(DF3SeqGUI_cls* pCDF3SeqGUI);
    static DF3SeqGUI_cls* mspCDF3SeqGUI;
    
    ConfItemUINT_cls mCItemNum;
    unsigned int mActiveItem;
    std::vector < Parameters_cls > mVCParam;
    
    void SetInitParam(double *pInit);

    unsigned int GetActiveItemKeyNum();
    Parameters_cls *GetPActiveItem();

    //wxString* GetPItemListStringVolatile();

    void AddItem();
    void RemoveItem();
    void AdjustAutoControl();

    void AddItemProc(Parameters_cls *pCNewParam, unsigned int index);
    void RemoveItemProc(unsigned int index);
    
    //--
    void HideTLPanel(int index);
    virtual void ReadStrLinePropagate();
    void ReadStrLine(std::wistream &istream, std::wstring strLine);
    wxString Serialize();
    
    static void FuncHookItemNumberChanged(void* pArg, void* pEventOvwt);
    
    virtual void AddItemPropagate() = 0;
    virtual void RemoveItemPropagate() = 0;
};

// --------------------------------

#include "../UNDO/clsUndoManager.h"

class ComMultiParamBase_cls{ 
 public:
    ComMultiParamBase_cls(MultiParameters_cls *pCTgt);

    MultiParameters_cls *mpCTgt;
};

class ComMultiParamAdd_cls : public CommandItem_cls, ComMultiParamBase_cls {
 public:
    ComMultiParamAdd_cls(MultiParameters_cls *pCTgt, Parameters_cls *pCNewParam,
		      unsigned int index);
    ~ComMultiParamAdd_cls();

    unsigned int mIndex;
    Parameters_cls *mpCParamMemento;
    bool Execute(bool bFlag = _COMMAND_REDO);
};

class ComMultiParamRemove_cls : public CommandItem_cls, ComMultiParamBase_cls {
 public:
    ComMultiParamRemove_cls(MultiParameters_cls *pCTgt, unsigned int index);
    
    unsigned int mIndex;
    bool Execute(bool bFlag = _COMMAND_REDO);
    
};

#endif
