// -*- c -*-
// "clsTimeLinePanel.h"
// (setq c-basic-offset 4)

#include "clsTimeLinePanel.h"
#include "clsParameters.h"
#include "clsParametersPanel.h"
#include "../POVRAY/clsLightSource.h"
#include "../DRAWBOX/clsHotSpot.h"
#include "../DF3/clsDF3SeqGUI.h"
#include "../DF3/TL/clsParametersPanelList.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include <wx/dcbuffer.h>

#ifdef WIN32
#define _MEM_PANEL_OFFSET 14
#else
#define _MEM_PANEL_OFFSET 10
#endif

#ifdef WIN32
#define WCHARCONV(x) (wxString)wxConvLibc.cMB2WC(x)
#else
#define WCHARCONV(x) (wxString)wxConvUTF8.cMB2WC(x)
#endif

extern wxFrame* GpCMainFrame;
extern FrameMain_cls* GpCMainFrameCLS;

BEGIN_EVENT_TABLE(TimeLinePanel_cls, wxPanel)
     EVT_PAINT       (TimeLinePanel_cls::OnPaint)
     EVT_MOUSE_EVENTS(TimeLinePanel_cls::OnMouse)
     EVT_KEY_DOWN    (TimeLinePanel_cls::OnKeyDown)
     EVT_MENU        (TLMenuInsert, TimeLinePanel_cls::OnMenuInsert)
     EVT_MENU        (TLMenuEditKey, TimeLinePanel_cls::OnMenuEditKey)
     EVT_MENU        (TLMenuHide,   TimeLinePanel_cls::OnMenuHide)
     EVT_MENU        (TLMenuImport, TimeLinePanel_cls::OnMenuImport)
     EVT_MENU        (TLMenuCustom01, TimeLinePanel_cls::OnMenuCustom)
END_EVENT_TABLE()

TimeLinePanel_cls::
     TimeLinePanel_cls(Parameters_cls *pCParam, ParametersPanel_cls *parent, /*wxString strTag,*/ wxString strCustomName, wxPoint pos, wxSize size, wxWindowID id) :
    wxPanel(parent, id, pos, size, wxWANTS_CHARS),
    mpCParam(pCParam), mpCParentPanel(parent),
	/*mstrTag(strTag),*/ mstrCustomName(strCustomName),
    mbActive(FALSE), mbSelected(FALSE),
    mpCCustomTarget(NULL)
{
    this->SetBackgroundColour(wxColour(216, 216, 216));
    mOffset = _MEM_PANEL_OFFSET;
    mbCaptureMouse = FALSE;
    mbDragging = FALSE;
    mbOnPanel = FALSE;
    mSelected = -1;

    mpCMenu = new wxMenu;
    mpCMenu->Append(TLMenuInsert, _("Insert/Remove Key"));
    mpCMenu->Append(TLMenuEditKey, _("Edit Key"));
    mpCMenu->Append(TLMenuHide  , _("Hide this panel"));
    //mpCMenu->Append(TLMenuImport, _("Import from file"));

    if (msbIsLoadIcon == FALSE){
	msbIsLoadIcon = TRUE;
#ifdef WIN32
	msCIcon.LoadFile(wxT("ICONS/Numerical.ico"), wxBITMAP_TYPE_ICO);
#else
	msCIcon.LoadFile(wxT("ICONS/Numerical.png"), wxBITMAP_TYPE_PNG);
#endif
    }

    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
}

TimeLinePanel_cls::
~TimeLinePanel_cls(){
    delete mpCMenu;
}

void TimeLinePanel_cls::
InitializeStaticMembers(LightSRC_cls* pCLightSRC,
			HotSpots_cls* pCHotSpot,
			DF3SeqGUI_cls* pCDF3SeqGUI){
    mspCLightSRCLink = pCLightSRC;
    mspCHotSpotLink = pCHotSpot;
    mspCDF3SeqGUILink = pCDF3SeqGUI;
}

void TimeLinePanel_cls::
LinkParameter(Parameters_cls *pCParam){
    mpCParam = pCParam;
}

void TimeLinePanel_cls::
AppendMenu(wxWindow *pCCustomTarget, int menuID,
	   wxString strTitle, int type){
    mpCCustomTarget = pCCustomTarget;
    mCustumMenuID = menuID;
    switch(type){
    case _TIMELINE_MENU_APPEND:
	mpCMenu->Append(TLMenuCustom01, strTitle);
	break;
    case _TIMELINE_MENU_INSERT:
	mpCMenu->Append(TLMenuCustom01, strTitle);
	break;
    }
}

void TimeLinePanel_cls::
GrayOutUnusedMenu(){
    mpCMenu->FindItemByPosition(1)->SetItemLabel(_("Edit Key"));
    if (mpCParam == NULL){
	mpCMenu->FindItemByPosition(0)->Enable(FALSE);
	mpCMenu->FindItemByPosition(1)->Enable(FALSE);
	return; }
    if (mpCParam->mCbTimeDependent.mbBody == FALSE){
	mpCMenu->FindItemByPosition(0)->Enable(FALSE);
	mpCMenu->FindItemByPosition(1)->Enable(TRUE);
	mpCMenu->FindItemByPosition(1)->SetItemLabel(_("Edit Value"));
	return; }
    if (mSelected == -1){
	mpCMenu->FindItemByPosition(0)->Enable(TRUE);
	mpCMenu->FindItemByPosition(1)->Enable(FALSE);
	return; }
    if (mpCParam->GetSelectedKeyNum() == 1){
	mpCMenu->FindItemByPosition(0)->Enable(TRUE);
	mpCMenu->FindItemByPosition(1)->Enable(TRUE);
	return; }
    mpCMenu->FindItemByPosition(0)->Enable(TRUE);
    mpCMenu->FindItemByPosition(1)->Enable(FALSE);
    
}

void TimeLinePanel_cls::
SetDoubleIndex(int maxFrame, int AAMin, int AAMax, double doubleIndex,
	       double doubleIndexAAMin, double doubleIndexAAMax,
	       double doubleIndexWAMin, double doubleIndexWAMax){
    msMaxFrame = maxFrame;
    msAAMin = AAMin;
    msAAMax = AAMax;
    msCurrentDoubleIndex = doubleIndex;
    msDoubleIndexAAMin   = doubleIndexAAMin;
    msDoubleIndexAAMax   = doubleIndexAAMax;
    msDoubleIndexWAMin   = doubleIndexWAMin;
    msDoubleIndexWAMax   = doubleIndexWAMax;
}

void TimeLinePanel_cls::
Activate(bool bActive, bool bExclusive){
    bool prev = mbActive;
    mbActive = bActive;
    if (bExclusive == TRUE) mbSelected = FALSE;
    if (mbSelected == FALSE) mpCParam->SelectKeyFrameRegion(0, 1, FALSE);
    if (mbActive == TRUE) mbSelected = TRUE;
    if (mbActive != prev){
	
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	wxGetTopLevelParent(this)->AddPendingEvent(com);
	
	mspCLightSRCLink->UpdateActiveFromTag(/*mstrTag*/mstrCustomName);
	mspCHotSpotLink->UpdateActiveFromTag(/*mstrTag*/mstrCustomName);
    }

}

void TimeLinePanel_cls::
OnPaint(wxPaintEvent& event){
    // must always be here
    //wxPaintDC dc(this);
    
    wxAutoBufferedPaintDC dc(this);
    
    dc.SetBackground(wxBrush(wxColor(216, 216, 216), wxSOLID));
    dc.Clear();
    
    SetMemberSize(GetClientSize());

    int posWAmin = GetCoordinateFromDouble(msDoubleIndexWAMin);
    int posWAmax = GetCoordinateFromDouble(msDoubleIndexWAMax);

    if (msDoubleIndexWAMax == 0)
	posWAmax = GetDeviceX(1);

    dc.SetPen(*wxTRANSPARENT_PEN);
    if (mbActive == TRUE)
	dc.SetBrush(wxBrush(wxColour(191,216,216)));
    else
	if (mbSelected == TRUE)
	    dc.SetBrush(wxBrush(wxColour(216, 232, 232)));
	else
	    dc.SetBrush(*wxWHITE_BRUSH);
    
    dc.DrawRectangle(posWAmin, 0, posWAmax - posWAmin, mHeight);

    if (mbDragging == TRUE && mbActive == TRUE && mSelected == -1){
	int posDrag0 = GetCoordinateFromDouble(mRegionOrigin);
	int posDrag1 = GetCoordinateFromDouble(mDragDoubleIndex);
	if (posDrag0 != posDrag1){
	    dc.SetPen(wxPen(wxColour(128, 128, 128)));
	    dc.DrawRectangle(posDrag0, 2, posDrag1-posDrag0, mHeight-4);
	}
    }

    if (mbActive == TRUE)
	dc.SetTextForeground(wxColour(64, 64, 64));
    else
	if (mbSelected == TRUE)
	    dc.SetTextForeground(wxColour(96, 96, 96));
	else
	    dc.SetTextForeground(wxColour(128, 128, 128));
    
    dc.SetFont(wxFont(11, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL));
    dc.DrawText(mstrCustomName, _MEM_PANEL_OFFSET + 5, -2);
    
    if (mbActive == TRUE)
	dc.SetTextForeground(wxColour(64, 64, 64));
    else
	dc.SetTextForeground(*wxLIGHT_GREY);
    wxString Mes;
    
    dc.SetPen(msCPenVeryLightGrey);
    dc.DrawRectangle(0, 0, mWidth, 1);

    if (mpCParam == NULL) {
	Mes << wxT("Not Assigned");
	dc.DrawText(Mes, _MEM_PANEL_OFFSET + 100, -2);
	event.Skip();
	return;
    }

    if (msbShowParameter == TRUE){
	if (mpCParam->mCDimension.mBody > 1) Mes << wxT("[ ");
	for (unsigned int i = 0; i < mpCParam->mCDimension.mBody; i++){
	    Mes << mpCParam->GetPInterpolatedValue(msCurrentDoubleIndex)[i];
	    if (mpCParam->mCDimension.mBody > 1 && (i != mpCParam->mCDimension.mBody - 1)) Mes << wxT(", ");
	}
	if (mpCParam->mCDimension.mBody > 1) Mes << wxT("] ");
	dc.DrawText(Mes, _MEM_PANEL_OFFSET + 140, -2);
    }
    
    if (mpCParam->mCbTimeDependent.mbBody == TRUE){
	dc.SetPen(*wxBLACK_PEN);
	for (unsigned int i = 0; i < mpCParam->GetKeyNumber(); i++){
	    CalcCoordinateFromDouble(mpCParam->mVCKeys[i].mDoubleIndex);
	    mpCParam->mVCKeys[i].DrawKey(dc, mX0, _KEY_NORMAL);
	}
    }

    if (mbDragging == TRUE /*&& mbCaptureMouse == TRUE && mSelected != -1*/)
	for (unsigned int i = 0; i < mpCParam->GetKeyNumber(); i++)
	    if (mpCParam->mVCKeys[i].mbSelected == TRUE){
		CalcCoordinateFromDouble
		    (mDragDoubleIndex 
		     - mDragDoubleIndexOrigin
		     //- mpCParam->mVCKeys[mSelected].mDoubleIndex
		     + mpCParam->mVCKeys[i].mDoubleIndex);
		mpCParam->mVCKeys[i].DrawKey(dc, mX0, _KEY_DRAG, mbOnPanel);
	    }
    
    dc.SetPen(*wxRED_PEN);
    CalcCoordinateFromDouble(msCurrentDoubleIndex);
    dc.DrawRectangle(mX0, 0, 1, mHeight);

    /*
    Mes = wxT("");
    Mes << mDragDoubleIndex << wxT(":") << mDraggingFrame ;
    dc.SetPen(*wxGREY_PEN);
    dc.DrawText(Mes, _MEM_PANEL_OFFSET + 500, 0);
    */

    event.Skip();
}

void TimeLinePanel_cls::
CreateKeyDLGAtMousePos(){
    if (mSelected != -1 && mpCParam->GetSelectedKeyNum() == 1){
	CalcCoordinateFromDouble(mpCParam->mVCKeys[mSelected].mDoubleIndex);
	mMenuPosX = mX0 + 60;
    }
    CreateKeyDLG(mMenuPosX-40, mMenuPosY-20);
}

void TimeLinePanel_cls::
CreateKeyDLG(int xpos, int ypos, wxPoint posPanel){
    if (mpCParam == NULL) return;
    bool bMainFrameRefresh = FALSE;

    if (posPanel == wxDefaultPosition) 
	posPanel = GetScreenPosition();
    else // Filter$BB&$+$i8F$P$l$?$iA*Br$5$l$?(BKey$B$H$O4X78$J$/$9$k(B
	mSelected = -1; 

    MultiNumericalInputs_cls *pCDialog;
    pCDialog = new MultiNumericalInputs_cls
	(&pCDialog, this, wxID_ANY, mpCParam->mLabelStr, 
	 posPanel + wxPoint(xpos, ypos));
    
    pCDialog->SetIcon(msCIcon);

    mpParamCache = new double [mpCParam->mCDimension.mBody];
    mpDValue1Cache = new double [mpCParam->mCDimension.mBody];
    mpDValue2Cache = new double [mpCParam->mCDimension.mBody];
    mpTypeCache = new unsigned int;

    if (mSelected == -1)
	for (unsigned int i = 0; i < mpCParam->mCDimension.mBody; i++)
	    mpParamCache[i] = mpCParam->GetPInterpolatedValue(msCurrentDoubleIndex)[i];
    else
	for (unsigned int i = 0; i < mpCParam->mCDimension.mBody; i++)
	    mpParamCache[i] = mpCParam->mVCKeys[mSelected].mpValue[i];

    int candid = mpCParam->GetKeyFrameIndexFromDouble(msCurrentDoubleIndex);
    *mpTypeCache = mpCParam->GetInterpTypeFromDouble(msCurrentDoubleIndex);

    mspCurrentThis = this;
    double doubleFrame;
    if (mpCParam->mCbTimeDependent.mbBody == FALSE)
	doubleFrame = 0;
    else if (mSelected == -1)
	doubleFrame = msCurrentDoubleIndex;
    else
	doubleFrame = mpCParam->mVCKeys[mSelected].mDoubleIndex;
    int frame = (msMaxFrame * doubleFrame) + 1e-5;
    ConnectParameters(pCDialog, &frame);
    pCDialog->MakeGeometry();
    pCDialog->UpdateValuesAll(TRUE);
    ShowAndHideParameters(pCDialog, NULL);
    if (mpCParam->mCbTimeDependent.mbBody && mpCParam->IsKeyFrame(doubleFrame))
	pCDialog->ColoringItem(-1, wxColour(255, 255, 192));
    pCDialog->SetFuncHook(&ShowAndHideParameters, _MN_HOOK_COMBOBOX);
    if (pCDialog->FinishSettingDLG(TRUE, _MN_DLG_MODAL, 
				   _MN_DLG_NOMAKEGEOM) == wxID_OK){
	KeyFrame_cls CKey(mpCParam->mCDimension.mBody);//, CKeySlave(0);
	double frameDouble;
	if (msMaxFrame == 0) frameDouble = 0;
	else frameDouble = (double)frame / msMaxFrame;
	CKey.Set(frameDouble, *mpTypeCache,
		 mpParamCache, mpDValue1Cache, mpDValue2Cache);
	// $B$b$7%+%a%i$J$i%N!<%^%i%$%:=hM}$,I,MW(B
	if (mpCParam->mLabelStr == L"CamPos"){
	    GCUndoList.OpenBlock();
	    mpCParam->SetKey(CKey);
	    GpCMainFrameCLS->NormalizeActiveCameraPath();
	    GCUndoList.CloseBlock();
	}
	else
	    mpCParam->SetKey(CKey);
	    
	
	mpCParam->GetPInterpolatedValue(msCurrentDoubleIndex);
	bMainFrameRefresh = TRUE;
	if (mSelected != -1) mpCParam->mVCKeys[mSelected].mbSelected = TRUE;
    }
    
    delete mpTypeCache;
    delete [] mpParamCache;
    delete [] mpDValue1Cache;
    delete [] mpDValue2Cache;

    delete pCDialog;

    if (bMainFrameRefresh == TRUE){
	if (mpCParam->mbUpdateCamera == TRUE){
	    wxCommandEvent com = wxCommandEvent
		(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshCamera);
	    wxGetTopLevelParent(this)->AddPendingEvent(com);
	}
	wxCommandEvent com = wxCommandEvent
	    (wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	wxGetTopLevelParent(this)->AddPendingEvent(com);
    }
}

void TimeLinePanel_cls::
ConnectParameters(MultiNumericalInputs_cls* pCDLG, int *pFrame){
    int boxWidth = 2;

    pCDLG->mVCPointers.clear();

    pCDLG->SetDefaultButtonType(_MN_OK_CANCEL_DEFAULT);
    pCDLG->mMinimumWidth = 170;
    pCDLG->mTextWidth = 60;

    pCDLG->CombinePointer(_GP_INTEGER, pFrame, wxT("Frame"), FALSE);

    for (unsigned int i = 0; i < mpCParam->mCDimension.mBody; i++)
	pCDLG->CombinePointer(_GP_DOUBLE, &mpParamCache[i],
			      mpCParam->mVstrItemLabel[i]);
    
    pCDLG->CombinePointer(_GP_UNSIGNED_INT_LIST, mpTypeCache, 
			  wxT(""), TRUE, 5, KeyFrame_cls::msAstrInterp, *mpTypeCache);
    
    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth); //dim+2

    pCDLG->CombinePointer(_GP_DOUBLE_SLIDER, &mpDValue2Cache[0]); //dim+3
    pCDLG->SetRange(0, 30, _MN_LASTITEM, 10);
    pCDLG->CombinePointer(_GP_DOUBLE, &mpDValue2Cache[0]);
    
    for (unsigned int i = 0; i < mpCParam->mCDimension.mBody; i++)
	pCDLG->CombinePointer(_GP_DOUBLE, &mpDValue1Cache[i],
			      mpCParam->mVstrItemLabelD1[i]);
}

#include "../MYAPI/MsgBox.h"

void TimeLinePanel_cls::
ShowAndHideParameters(void* pArg, void *pEvent){
    // $B<!85(Bproof$B$K$9$k(B!!
    MultiNumericalInputs_cls* pCDLG = (MultiNumericalInputs_cls*)pArg;
    TimeLinePanel_cls* pC = mspCurrentThis;
    int dim = pC->mpCParam->mCDimension.mBody;
    
    pCDLG->HideItem(dim + 3); //slider
    pCDLG->HideItem(dim + 4); //value
    for (int i = 0; i < dim; i++) pCDLG->HideItem(dim + 5 + i); // dvalue
				      
    int candid = pC->mpCParam->GetKeyFrameIndexFromDouble(msCurrentDoubleIndex);
    // $B4{B8$N%U%l!<%`30$G$OB.EY$N8uJd$r<+F0Jd4V$NCM$K$9$k!#(B
    if (candid != -1)
	for (int i = 0; i < dim; i++)
	    pC->mpDValue1Cache[i] = pC->mpCParam->mVCKeys[candid].mpDValue1[i];
    else{
	int prev = pC->mpCParam->GetAdjacentKey(msCurrentDoubleIndex, -1);
	int next = pC->mpCParam->GetAdjacentKey(msCurrentDoubleIndex, +1);
	pC->mpCParam->CalcVelocity(prev, next, pC->mpDValue1Cache);
    }

    if (*(pC->mpTypeCache) == _PARAM_HERMITE_ADJ){
	pC->mpDValue2Cache[0] = 0;
	int candid = pC->mpCParam->GetKeyFrameIndexFromDouble(msCurrentDoubleIndex);
	if (candid != -1)
	    if (pC->mpCParam->mVCKeys[candid].mType == _PARAM_HERMITE_ADJ){
		pC->mpDValue2Cache[0] = pC->mpCParam->mVCKeys[candid].mpDValue2[0];
	    }
	pCDLG->ShowItem(dim + 3);
	pCDLG->ShowItem(dim + 4);
    }

    if (*(pC->mpTypeCache) == _PARAM_HERMITE)
	for (int i = 0; i < dim; i++) pCDLG->ShowItem(dim + 5 + i);

    if (pC->mpCParam->mCbTimeDependent.mbBody == FALSE){
	pCDLG->HideItem(0);
	pCDLG->HideItem(dim+1);
    }

    pCDLG->Layout();
    pCDLG->SetSizerAndFit(pCDLG->mpBoxSizerSrc);
    pCDLG->UpdateValues();
}

void TimeLinePanel_cls::
OnMouse(wxMouseEvent& event){

    bool bMainFrameRefresh = FALSE;
    double doubleIndex = CalcDoubleIndexFromPosition(event.GetX());
    mMenuPosX = event.GetX(); mMenuPosY = event.GetY();
    
    if (event.ButtonUp(wxMOUSE_BTN_RIGHT)){
	GetParent()->AddPendingEvent(event);
	GrayOutUnusedMenu();
	PopupMenu(mpCMenu, mMenuPosX-2, mMenuPosY-2);
    }

    if (event.LeftDClick() == TRUE) CreateKeyDLGAtMousePos();

    if (mpCParam == NULL) {
	GetParent()->AddPendingEvent(event);
	event.Skip();
	return;
    }

    if (event.ButtonDown(wxMOUSE_BTN_RIGHT)/* && mSelected == -1*/){
	if (mpCParam->OnWhichKeyFrame(doubleIndex, CalcResolutionFor1Pixel() * 3) != -1)
	    mSelected = mpCParam->SelectKeyFrame(doubleIndex, CalcResolutionFor1Pixel() * 3, FALSE, !event.m_shiftDown);
	bMainFrameRefresh = TRUE;
	if (mbSelected == FALSE)
	    GetParent()->AddPendingEvent(event);
    }

    // ---Dragging---
    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	mRegionOrigin = doubleIndex;
	mSelected = mpCParam->SelectKeyFrame(doubleIndex, CalcResolutionFor1Pixel() * 3, (mbActive == FALSE), !event.m_shiftDown);
	if (mSelected != -1 && mpCParam->mCbTimeDependent.mbBody == TRUE){
	    if (event.m_controlDown && !event.m_altDown){
		if (mpCParam->mVCKeys[mSelected].mType == _PARAM_HERMITE_AUTO)
		    //mpCParam->SetKeyType(mSelected, _PARAM_LINEAR);
		    mpCParam->SetKeyTypeMulti(_PARAM_LINEAR);
		else
		    //mpCParam->SetKeyType(mSelected, _PARAM_HERMITE_AUTO);
		    mpCParam->SetKeyTypeMulti(_PARAM_HERMITE_AUTO);
		bMainFrameRefresh = TRUE;
	    }
	    
	    /*
	      // $B%$!<%:78?t$r#0$K$9$k=hM}$rIU$1$J$$$H!"JQ$JJd4V$K$J$k!#(B
	    if (event.m_controlDown && event.m_shiftDown){
		mpCParam->SetKeyTypeMulti(_PARAM_HERMITE_ADJ);
		bMainFrameRefresh = TRUE;
	    }
	    */
	    
	    if (event.m_controlDown && event.m_altDown){
		//mpCParam->SetKeyType(mSelected, _PARAM_STEP);
		mpCParam->SetKeyTypeMulti(_PARAM_STEP);
		bMainFrameRefresh = TRUE;
	    }
	    
	    mDragDoubleIndexOrigin = mpCParam->mVCKeys[mSelected].mDoubleIndex;
	    mDragDoubleIndex = mDragDoubleIndexOrigin;
	    if (mbActive == FALSE)
		GetParent()->AddPendingEvent(event);
	}
	else{
	    GetParent()->AddPendingEvent(event);
	}
	CaptureMouse();
	mbCaptureMouse = TRUE;
	mbOnPanel = TRUE;
	Refresh(FALSE);
    }

    if (event.Dragging() == TRUE){
	mbDragging = event.m_leftDown;
	mDragDoubleIndex = 
	    Quantize(CalcDoubleIndexFromPosition(event.GetX()));
	if (event.m_shiftDown){
	    int frame = mDragDoubleIndex * msMaxFrame;
	    mDraggingFrame = frame;
	    int snipWidth = (msAAMax - msAAMin) / 100;
	    for (int candid = frame - snipWidth; candid <= frame + snipWidth; candid++){
		if (candid < 0) continue;
		if (candid >= (int)msVBSnippingTag.size()) continue;
		if (msVBSnippingTag[candid] == TRUE) frame = candid;
	    }
	    mDragDoubleIndex = Quantize(CalcDoubleIndexFromFrame(frame));
	}
	if (mSelected != -1)
	    mspCDF3SeqGUILink->mpCPanelList->DragAllPanel
		(mbDragging, mDragDoubleIndexOrigin, mDragDoubleIndex);
	else
	    Refresh(TRUE);
	mbOnPanel = (event.GetY() > -2 * mHeight && event.GetY() < mHeight * 3);
    }
    else
	mspCDF3SeqGUILink->mpCPanelList->DragAllPanel
	    (FALSE, mDragDoubleIndexOrigin, mDragDoubleIndex);
	//mbDragging = FALSE;

    if (event.ButtonUp(wxMOUSE_BTN_LEFT)){
	if (mbCaptureMouse == TRUE){
	    ReleaseMouse();
	    mbCaptureMouse = FALSE;
	    if (mSelected == -1)
		if (mpCParam->
		    SelectKeyFrameRegion(mRegionOrigin, doubleIndex) == TRUE)
		    bMainFrameRefresh = TRUE;
	    
	    if (mbOnPanel == TRUE && mSelected != -1){
		GCUndoList.OpenBlock();
		if (mpCParam->MoveTimeSelectedKey
		    (mDragDoubleIndex - mDragDoubleIndexOrigin)
		    == TRUE){
		    mspCDF3SeqGUILink->mpCPanelList->MoveTimeSelectedKeyAllPanel
			(mDragDoubleIndex - mDragDoubleIndexOrigin, this);
		}
		else // $B0\F0$,L5$1$l$P!"B>$N(Bkey$B$N(Bdeselect
		    mSelected = mpCParam->SelectKeyFrame(doubleIndex, CalcResolutionFor1Pixel() * 3, event.m_shiftDown, TRUE);		
		GCUndoList.CloseBlock();
		bMainFrameRefresh = TRUE;
	    }
	}
	mbDragging = FALSE;
	Refresh(TRUE);
    }

    if (bMainFrameRefresh == TRUE){
	wxCommandEvent dummy(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
	GpCMainFrame->AddPendingEvent(dummy);
    }
    
    event.Skip();
}

void TimeLinePanel_cls::
OnKeyDown(wxKeyEvent& event){
    int code = event.GetKeyCode();
    
    if (code == WXK_DELETE){
	mpCParam->DeleteSelectedKey();
	GpCMainFrame->Refresh(FALSE);
	return;
    }

    if (event.GetModifiers() == wxMOD_CONTROL){
	switch(code){
	case 'X':
	    mpCParam->CopyKeyFrameToClipBoard(); 
	    if (mpCParam->DeleteSelectedKey() == TRUE)
		GpCMainFrame->Refresh(FALSE);
	    return; // $B=hM}$7$?$iH4$1$k(B
	case 'C':
	    mpCParam->CopyKeyFrameToClipBoard(); 
	    return; // $B=hM}$7$?$iH4$1$k(B
	case 'V':
	    mpCParam->PasteKeyFrameFromClipBoard();
	    Refresh(TRUE);
	    return; // $B=hM}$7$?$iH4$1$k(B
	}
    }
    
    GetGrandParent()->AddPendingEvent(event);
}

void TimeLinePanel_cls::
OnMenuInsert(wxCommandEvent& WXUNUSED(event)){
    mpCParentPanel->EmuCheckBox();
    mSelected = mpCParam->GetSelectedKeyIndex();
}

void TimeLinePanel_cls::
OnMenuEditKey(wxCommandEvent& WXUNUSED(event)){
    CreateKeyDLGAtMousePos();
}

void TimeLinePanel_cls::
OnMenuHide(wxCommandEvent& WXUNUSED(event)){
    mpCParentPanel->HidePanel();
}

void TimeLinePanel_cls::
OnMenuImport(wxCommandEvent& WXUNUSED(event)){
    ;
}

void TimeLinePanel_cls::
OnMenuCustom(wxCommandEvent& event){
    if (mpCCustomTarget != NULL){
	event.SetId(mCustumMenuID);
	mpCCustomTarget->AddPendingEvent(event);
    }
}

// -----

int TimeLinePanel_cls::
GetCoordinateFromDouble(double doubleIndex){
    if (msDoubleIndexAAMax == msDoubleIndexAAMin)
	return GetDeviceX(0);
    double indexWA = 
	(doubleIndex - msDoubleIndexAAMin) / 
	(msDoubleIndexAAMax - msDoubleIndexAAMin);
    return GetDeviceX((float)indexWA);
}

void TimeLinePanel_cls::
CalcCoordinateFromDouble(double doubleIndex){
    mX0 = GetCoordinateFromDouble(doubleIndex);
}

double TimeLinePanel_cls::
CalcResolutionFor1Pixel(){
    return (msDoubleIndexAAMax - msDoubleIndexAAMin) / (mWidth - 2 * mOffset);
}

double TimeLinePanel_cls::
Quantize(double doubleIndex){
    return mpCParam->Quantize(doubleIndex); 
}

double TimeLinePanel_cls::
CalcDoubleIndexFromPosition(int x){
    double coordinate = ((double)x - mOffset) / (mWidth - 2 * mOffset);
    double doubleIndex = coordinate * (msDoubleIndexAAMax - msDoubleIndexAAMin) + 
	msDoubleIndexAAMin;
    
    return doubleIndex;
}

double TimeLinePanel_cls::
CalcDoubleIndexFromFrame(int frame){
    return (double)frame / msMaxFrame;
}

bool   TimeLinePanel_cls::msbIsLoadIcon = FALSE;
wxIcon TimeLinePanel_cls::msCIcon;
bool   TimeLinePanel_cls::msbShowParameter = TRUE;
wxPen  TimeLinePanel_cls::msCPenVeryLightGrey(wxColour(236, 236, 236));
std::vector < bool > TimeLinePanel_cls::msVBSnippingTag;
double TimeLinePanel_cls::msCurrentDoubleIndex;
double TimeLinePanel_cls::msDoubleIndexAAMin;
double TimeLinePanel_cls::msDoubleIndexAAMax;
double TimeLinePanel_cls::msDoubleIndexWAMin;
double TimeLinePanel_cls::msDoubleIndexWAMax;
int    TimeLinePanel_cls::msMaxFrame;
int    TimeLinePanel_cls::msAAMin;
int    TimeLinePanel_cls::msAAMax;
TimeLinePanel_cls* TimeLinePanel_cls::mspCurrentThis;
LightSRC_cls* TimeLinePanel_cls::mspCLightSRCLink = NULL;
HotSpots_cls* TimeLinePanel_cls::mspCHotSpotLink = NULL;
DF3SeqGUI_cls* TimeLinePanel_cls::mspCDF3SeqGUILink = NULL;
