// "clsMarchingCube.h"
// (setq c-basic-offset 4)

// The main routine of this class used
// Marching Cubes Example Program 
// by Cory Bloyd (corysama@yahoo.com)
// which is publishd as public domain 

#include "clsMarchingCube.h"
#include <math.h>

#include "../../MYAPI/MsgBox.h"

//SSE2
#include <malloc.h>
#include <emmintrin.h>
//SSE3
#ifdef WIN32
#include <intrin.h>     
#else
#include <pmmintrin.h>
#endif

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

extern int GTestVariableIntPoly[1024];

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2

ThreadMarching::
ThreadMarching(unsigned char *pData, 
	       float *pTmpVertices, float *pTmpNormal,
	       int *pArrayIndex) 
  : wxThread(wxTHREAD_JOINABLE),
    mpData(pData), mpTmpVertices(pTmpVertices), mpTmpNormal(pTmpNormal),
    mpArrayIndex(pArrayIndex)
{
}

void ThreadMarching::
SetParameters(unsigned int size[3], 
	      int initX, int finX,
	      int initZ, int finZ,
	      int *pSkip, 
	      float targetValue, bool bUseSSE){

    LOOP_XYZ mSize[xyz] = size[xyz];
    mInitX = initX; mFinX = finX;
    mInitZ = initZ; mFinZ = finZ;
    mpSkip = pSkip;
    mTargetValue = targetValue;
    mbUseSSE = bUseSSE;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

// finds the point of intersection between two points
GLfloat ThreadMarching::
GetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired){
    GLdouble fDelta = fValue2 - fValue1;
    if (fDelta == 0.0) return 0.5;
    return (fValueDesired - fValue1)/fDelta;
}

float ThreadMarching::
GetInterpolatedValue(const unsigned char *pData,
		     const int lBaseIndex, 
		     const float fx, const float fy, const float fz){

    // ŌNomalizêŁANormalizesv
    // Ǐ֐ǂނقo֐ǂ݂ɍs葬íHj
    
    register const unsigned int lOffsetZ(mOffsetZ);
    register const unsigned int lOffsetY(mOffsetY);
    
    const float P000(pData[lBaseIndex]);
    const float P001(pData[lBaseIndex+1]);
    const float P010(pData[lBaseIndex+lOffsetY]);
    const float P011(pData[lBaseIndex+lOffsetY+1]);
    const float P100(pData[lBaseIndex+lOffsetZ]); 
    const float P101(pData[lBaseIndex+lOffsetZ+1]);
    const float P110(pData[lBaseIndex+lOffsetZ+lOffsetY]);
    const float P111(pData[lBaseIndex+lOffsetZ+lOffsetY+1]);
    
    return  
	(1.0f-fz) * ( (1.0f-fy) * ((1.0f-fx)*P000 + fx*P001)  +
		      fy       * ((1.0f-fx)*P010 + fx*P011)) +
	fz        * ( (1.0f-fy) * ((1.0f-fx)*P100 + fx*P101)  +
		      fy       * ((1.0f-fx)*P110 + fx*P111));
}

void ThreadMarching::
GetNormalPrecise(GLfloat ANorm[4], const unsigned char *pData, 
		 const int x, const GLfloat fX, 
		 const int y, const GLfloat fY,
		 const int z, const GLfloat fZ){
    
    const float baseValue
	(GetInterpolatedValue(pData, mBaseIndex, fX, fY, fZ));

    ANorm[X] = ANorm[Y] = baseValue;
    ANorm[X] -= GetInterpolatedValue(pData, mBaseIndex+1, fX, fY, fZ);
    ANorm[Y] -= GetInterpolatedValue(pData, mBaseIndex+mOffsetY, fX, fY, fZ);

    if (z == mSize[Z] - 2) // /* ̏ꂵ̂if */
	ANorm[Z] = 
	    GetInterpolatedValue(pData, mBaseIndex-mOffsetZ, fX, fY, fZ) -
	    baseValue;
    else
	ANorm[Z] = 
	    baseValue -
	    GetInterpolatedValue(pData, mBaseIndex+mOffsetZ, fX, fY, fZ);

    float norm = sqrt(ANorm[X]*ANorm[X] + ANorm[Y]*ANorm[Y] + ANorm[Z]*ANorm[Z]);

    if (norm != 0){ // ̏ꂵ̂if 
	ANorm[X] /= norm;
	ANorm[Y] /= norm;
	ANorm[Z] /= norm;
    }
    else
	ANorm[Z] = 1;
}

void ThreadMarching::
GetNormalPreciseSSE(GLfloat ANorm[4], const unsigned char *pData, 
		    const int x, const GLfloat fX, 
		    const int y, const GLfloat fY,
		    const int z, const GLfloat fZ){
    
    const float baseValue
	(GetInterpolatedValue(pData, mBaseIndex  , fX, fY, fZ));
    
    ANorm[Y] = ANorm[X] = baseValue;
    ANorm[X] -= GetInterpolatedValue(pData, mBaseIndex+1, fX, fY, fZ);
    ANorm[Y] -= GetInterpolatedValue(pData, mBaseIndex+mOffsetY, fX, fY, fZ);

    if (z == mSize[Z] - 2) // /* ̏ꂵ̂if */
	ANorm[Z] = 
	    GetInterpolatedValue(pData, mBaseIndex-mOffsetZ, fX, fY, fZ) -
	    baseValue;
    else
	ANorm[Z] = 
	    baseValue -
	    GetInterpolatedValue(pData, mBaseIndex+mOffsetZ, fX, fY, fZ);

    ANorm[3] = 0; // SSÊ߂Ƀ_~[

    __m128 simd_tmp; 
    __m128 simd_value;

    simd_tmp = _mm_loadu_ps( ANorm ); // 4߂0
    simd_value = _mm_mul_ps(simd_tmp, simd_tmp);
    simd_value = _mm_hadd_ps(simd_value, simd_value); // x+y, z+0, x+y, z+0;
    simd_value = _mm_hadd_ps(simd_value, simd_value); // x+y+z+0 * 4

    float pDist2[4];
    _mm_storeu_ps(pDist2, simd_value);
    if (pDist2[0] == 0){ ANorm[2] = 1; return; }

    simd_value = _mm_rsqrt_ps(simd_value); //(norminv, ...);
    simd_tmp = _mm_mul_ps(simd_tmp, simd_value);
    _mm_storeu_ps(ANorm, simd_tmp);
}

int ThreadMarching::
MarchCubeMainThread(const unsigned char *pData, 
		    const int x, const int y, const int z){

    GLint iFlagIndex(0), iEdgeFlags;
    GLfloat fOffset;
    GLfloat afCubeValue[8];
    GLfloat AAEdgeVertex[12][3];
    GLfloat AAEdgeNorm[12][4]; // SSÊ߂4߂̓_~[
    
    const int lOffsetZ(mOffsetZ);
    const int lOffsetY(mOffsetY);
    mBaseIndex = z*lOffsetZ + y*lOffsetY + x;
	
    //Make a local copy of the values at the cube's corners
    const unsigned char *pDataIndex = pData + mBaseIndex;
    const GLint *pOffsetIndex = &VOffset[0][0];
    const GLint *pOffsetIndexEnd = pOffsetIndex + 24;
    GLfloat *pfCubeValueIndex = afCubeValue;
    for (; pOffsetIndex < pOffsetIndexEnd; 
	 pOffsetIndex+=3, pfCubeValueIndex++){
	*pfCubeValueIndex = (GLfloat) *(pDataIndex + 
					pOffsetIndex[Z]*lOffsetZ +
					pOffsetIndex[Y]*lOffsetY +
					pOffsetIndex[X]);
    }
    
    //Find which vertices are inside of the surface and which are outside
    for(int iVertexTest = 0; iVertexTest < 8; iVertexTest++) 
	if(afCubeValue[iVertexTest] <= mTargetValue255) 
	    iFlagIndex |= 1<<iVertexTest;
    
    iEdgeFlags = aiCubeEdgeFlags[iFlagIndex];
    if(iEdgeFlags == 0) {
	if (afCubeValue[0] > mTargetValue255)
	    return 1;
	else
	    return 2;
    }

    //Find the point of intersection of the surface with each edge
    //Then find the normal to the surface at those points
    if (mbUseSSE == TRUE) // if͂ł邾[v̊OɋNc
	for(int iEdge = 0; iEdge < 12; iEdge++) {
	    const GLint *pEdgeConnection = a2iEdgeConnection[iEdge];
	    if(iEdgeFlags & (1<<iEdge)) { //if intersection on this edge
		fOffset = 
		    GetOffset(afCubeValue[ pEdgeConnection[0] ], 
			      afCubeValue[ pEdgeConnection[1] ], 
			      mTargetValue255);
		
		// 0-1͈̔͂ł̍W
		float fX = 
		    VOffset[pEdgeConnection[0]][X] +
		    fOffset * a2fEdgeDirection[iEdge][X];
		float fY = 
		    VOffset[pEdgeConnection[0]][Y] +
		    fOffset * a2fEdgeDirection[iEdge][Y];
		float fZ = 
		    VOffset[pEdgeConnection[0]][Z] +
		    fOffset * a2fEdgeDirection[iEdge][2];
		
		AAEdgeVertex[iEdge][X] = (float)(x + fX) / mSize[X];
		AAEdgeVertex[iEdge][Y] = (float)(y + fY) / mSize[Y];
		AAEdgeVertex[iEdge][Z] = (float)(z + fZ) / mSize[Z];
		
		GetNormalPreciseSSE(AAEdgeNorm[iEdge], pData,
				    x, fX, y, fY, z, fZ);
	    }
	}
    else
	for(int iEdge = 0; iEdge < 12; iEdge++) {
	    const GLint *pEdgeConnection = a2iEdgeConnection[iEdge];
	    if(iEdgeFlags & (1<<iEdge)) { //if intersection on this edge
		fOffset = 
		    GetOffset(afCubeValue[ pEdgeConnection[0] ], 
			      afCubeValue[ pEdgeConnection[1] ], 
			      mTargetValue255);
		
		// 0-1͈̔͂ł̍W
		float fX = 
		    VOffset[pEdgeConnection[0]][X] +
		    fOffset * a2fEdgeDirection[iEdge][X];
		float fY = 
		    VOffset[pEdgeConnection[0]][Y] +
		    fOffset * a2fEdgeDirection[iEdge][Y];
		float fZ = 
		    VOffset[pEdgeConnection[0]][Z] +
		    fOffset * a2fEdgeDirection[iEdge][2];
		
		AAEdgeVertex[iEdge][X] = (float)(x + fX) / mSize[X];
		AAEdgeVertex[iEdge][Y] = (float)(y + fY) / mSize[Y];
		AAEdgeVertex[iEdge][Z] = (float)(z + fZ) / mSize[Z];
		
		GetNormalPrecise(AAEdgeNorm[iEdge], pData,
				 x, fX, y, fY, z, fZ);
	    }
	}

    //Store triangles that were found.  There can be up to five per cube

    //int ArrayIndex = *mpArrayIndex;

    for(int iTriangleIndex = 0; iTriangleIndex < 15; iTriangleIndex+=3) {
	if(a2iTriangleConnectionTable[iFlagIndex][iTriangleIndex] < 0) break;

	// [v̍Ő[Ȃ̂œWJ
	GLint iVertexX = a2iTriangleConnectionTable[iFlagIndex][iTriangleIndex];
	memcpy(mpTmpVertices + *mpArrayIndex, AAEdgeVertex[iVertexX], 12);
	memcpy(mpTmpNormal + *mpArrayIndex, AAEdgeNorm[iVertexX], 12);
	*mpArrayIndex+=3;

	GLint iVertexY = a2iTriangleConnectionTable[iFlagIndex][iTriangleIndex+1];
	memcpy(mpTmpVertices + *mpArrayIndex, AAEdgeVertex[iVertexY], 12);
	memcpy(mpTmpNormal + *mpArrayIndex, AAEdgeNorm[iVertexY], 12);
	*mpArrayIndex+=3;

	GLint iVertexZ = a2iTriangleConnectionTable[iFlagIndex][iTriangleIndex+2];
	memcpy(mpTmpVertices + *mpArrayIndex, AAEdgeVertex[iVertexZ], 12);
	memcpy(mpTmpNormal + *mpArrayIndex, AAEdgeNorm[iVertexZ], 12);
	*mpArrayIndex+=3;
    }

    //*mpArrayIndex = ArrayIndex;
    
    return 0; // EʗL
}

int ThreadMarching::
MarchCubeLOOPX1(const unsigned char *pData, int y, int z){
    // 0EL 1,2  E & targetValue ̏㉺ bitFiledI
    int prevStateX = 0, currStateX = 0;
    int boxnum = 0;
    for (int x = mInitX + 1; x < mFinX; x+=2){
	prevStateX = currStateX;
	currStateX = MarchCubeMainThread(mpData, x, y, z);
	boxnum += (currStateX == 0);
	if (currStateX != 0 &&
	    prevStateX == currStateX) continue;
	boxnum += (MarchCubeMainThread(mpData, x-1, y, z) == 0);
    }

    /*
    if ((mFinX - mInitX) % 2 == 1)
	MarchCubeMainThread(mpData, mFinX-1, y, z);
    */

    if (boxnum == 0) return currStateX;
    else return 0;
}

/*
void ThreadMarching::
MarchCubeLOOPX2(const unsigned char *pData,
		int y, int z){
    // 0EL 1,2  E & targetValue ̏㉺ bitFiledI
    int prevStateX = 0, currStateX = 0;
    for (int x = mInitX + 1; x < mFinX; x+=2){
	prevStateX = currStateX;
	currStateX = MarchCubeMainThread(mpData, x, y, z);
	if (currStateX != 0 &&
	    prevStateX == currStateX) continue;
	MarchCubeMainThread(mpData, x-1, y, z);
    }

    if ((mFinX - mInitX) % 2 == 1)
	MarchCubeMainThread(mpData, mFinX-1, y, z);
}
*/

int ThreadMarching::
MarchCubeLOOPY(const unsigned char *pData, int z){
    // 0,1,2  & Ă̂ŁA12̒lɂȂ̂ݏȗ
    int prevStateY = 3, currStateY = 3;
    int linenum = 0;
    for (int y = 1; y < mSize[Y]-1; y+=2){
	prevStateY = currStateY;
	currStateY = MarchCubeLOOPX1(mpData, y, z);
	//currStateZ &= currStateY;
	linenum += (currStateY == 0);
	if ((prevStateY == 1 && currStateY == 1) ||
	    (prevStateY == 2 && currStateY == 2)) {
	    mpSkip[Y]++; continue; }
	
	linenum += (MarchCubeLOOPX1(mpData, y-1, z) == 0);
    }

    if ((mSize[Y] - 1) % 2 == 1)
	MarchCubeLOOPX1(mpData, mSize[Y]-2, z);

    if (linenum == 0) return currStateY;
    else return 0;
}


wxThread::ExitCode ThreadMarching::
Entry(){
    mOffsetZ = mSize[X] * mSize[Y];
    mOffsetY = mSize[X];
    mTargetValue255 = mTargetValue * 255;
	
    // 0,1,2  & Ă̂ŁA12̒lɂȂ̂ݏȗ
    int prevStateZ = 3, currStateZ = 3;

    if (mFinZ - mInitZ < 4){
	for (int z = mInitZ; z < mFinZ; z++)
	    MarchCubeLOOPY(mpData, z);
	return 0; //ȗłȂ
    }
    
    // z̃C[2̔{łȂƑʖ
    for (int z = mInitZ+1; z < mFinZ; z+=2){
	prevStateZ = currStateZ;
	currStateZ = MarchCubeLOOPY(mpData, z);

	//printf ("prev %d curr %d\n", prevStateZ, currStateZ);
	if ((prevStateZ == 1 && currStateZ == 1) ||
	    (prevStateZ == 2 && currStateZ == 2)) {
	    //printf ("skip!!\n"); 
	    mpSkip[Z]++; continue; }

	MarchCubeLOOPY(mpData, z-1);
    }

    // ]
    if ((mFinZ - mInitZ) % 2 == 1)
	MarchCubeLOOPY(mpData, mFinZ-1);
    
    return 0;
}

//lists the positions, relative to vertex0, of each of the 8 vertices
const GLint ThreadMarching::VOffset[8][3] = {
    {0, 0, 0},{1, 0, 0},{1, 1, 0},{0, 1, 0},
    {0, 0, 1},{1, 0, 1},{1, 1, 1},{0, 1, 1}
};

//lists the index of the endpoint vertices for each of the 12 edges
const GLint ThreadMarching::a2iEdgeConnection[12][2] = {
    {0,1}, {1,2}, {2,3}, {3,0},
    {4,5}, {5,6}, {6,7}, {7,4},
    {0,4}, {1,5}, {2,6}, {3,7}
};

//lists the direction vector (vertex1-vertex0) for each edge
const GLfloat ThreadMarching::a2fEdgeDirection[12][3] = {
    {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
    {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
    {0.0, 0.0, 1.0},{0.0, 0.0, 1.0},{ 0.0, 0.0, 1.0},{0.0,  0.0, 1.0}
};

// _[8]  256 ʂ̏ԂA_ׂ̂[12]bitmaskœ
const GLint ThreadMarching::aiCubeEdgeFlags[256]= {
    0x000, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00, 
    0x190, 0x099, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90, 
    0x230, 0x339, 0x033, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30, 
    0x3a0, 0x2a9, 0x1a3, 0x0aa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0, 
    0x460, 0x569, 0x663, 0x76a, 0x066, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60, 
    0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0x0ff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0, 
    0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x055, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950, 
    0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0x0cc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0, 
    0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0x0cc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0, 
    0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x055, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650, 
    0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0x0ff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0, 
    0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x066, 0x76a, 0x663, 0x569, 0x460, 
    0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0x0aa, 0x1a3, 0x2a9, 0x3a0, 
    0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x033, 0x339, 0x230, 
    0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x099, 0x190, 
    0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x000
};

//  For example: a2iTriangleConnectionTable[3] list the 2 triangles formed when corner[0] 
//  and corner[1] are inside of the surface, but the rest of the cube is not.
const GLint ThreadMarching::a2iTriangleConnectionTable[256][16] =  {
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
    {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
    {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
    {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
    {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
    {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
    {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
    {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
    {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
    {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
    {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
    {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
    {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
    {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
    {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
    {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
    {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
    {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
    {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
    {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
    {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
    {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
    {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
    {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
    {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
    {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
    {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
    {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
    {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
    {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
    {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
    {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
    {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
    {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
    {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
    {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
    {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
    {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
    {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
    {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
    {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
    {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
    {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
    {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
    {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
    {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
    {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
    {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
    {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
    {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
    {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
    {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
    {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
    {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
    {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
    {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
    {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
    {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
    {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
    {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
    {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
    {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
    {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
    {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
    {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
    {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
    {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
    {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
    {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
    {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
    {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
    {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
    {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
    {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
    {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
    {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
    {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
    {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
    {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
    {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
    {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
    {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
    {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
    {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
    {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
    {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
    {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
    {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
    {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
    {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
    {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
    {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
    {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
    {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
    {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
    {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
    {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
    {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
    {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
    {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
    {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
    {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
    {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
    {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
    {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
    {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
    {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
    {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
    {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
    {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
    {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
    {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
    {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
    {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
    {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
    {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
    {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
    {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
    {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
    {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
    {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
    {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
    {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
    {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
    {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
    {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
    {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
    {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};

// ----

void MarchingCube_cls::
Clear(int sizeX, int sizeZ){
    for (unsigned int i = 0; i < mVV_Vertices.size(); i++){
	mVV_Vertices[i].clear();
	mVV_Normals[i].clear();
    }

    mVV_Vertices.clear();
    mVV_Normals.clear();

    mVV_Vertices.resize(2 * sizeZ);
    mVV_Normals.resize(2 * sizeZ);

    for (unsigned int i = 0; i < mVV_Vertices.size(); i++){
	mVV_Vertices[i].clear();
	mVV_Normals[i].clear();
    }
}

bool MarchingCube_cls::
ThreadControlMain(unsigned char *pData, int *pSkip,
		  int xLayerBase, int zLayerBase,
		  bool bUseSSE, int threadNum, const int layerFactor){

    
    float *pTmpVertices[8];
    float *pTmpNormal[8];
    int arrayIndex[8];
    ThreadMarching *pCThread[8];
    bool bAllocation = TRUE;
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){

	int zLayerIndex = zLayerBase + threadIndex * layerFactor;
	int zIni = zLayerIndex;
	int zFin = zLayerIndex + layerFactor;
	int xIni = xLayerBase;
	int xFin = xLayerBase + mSize[X]/1;
	if (xFin == mSize[X]) xFin = mSize[X] - 1;
	int xSize = xFin - xIni;
	//printf("x %d %d z %d %d\n", xIni, xFin, zIni, zFin);

	if (zIni >= mSize[Z]-1) continue;
	if (zFin >= mSize[Z]-1) zFin = mSize[Z] - 1;
	
	// Kvobt@vZ
	// }bNX1box5*3ǁAɂ܂ł͂ȂƉ
	try{ pTmpVertices[threadIndex] = 
		new float [xSize * mSize[Y] * layerFactor*6]; }
	catch(std::bad_alloc e){
	    bAllocation = FALSE; threadIndex = threadNum; continue; 
	}

	try{ pTmpNormal  [threadIndex] = 
		new float [xSize * mSize[Y] * layerFactor*6]; }
	catch(std::bad_alloc e){
	    delete pTmpVertices[threadIndex];
	    bAllocation = FALSE; threadIndex = threadNum; continue; 
	}

	pCThread    [threadIndex] = 
	    new ThreadMarching(pData,
			       pTmpVertices[threadIndex], 
			       pTmpNormal[threadIndex],
			       &arrayIndex[threadIndex]);
	arrayIndex  [threadIndex] = 0;
	
	pCThread[threadIndex]->
	    SetParameters(mSize, xIni, xFin, zIni, zFin, pSkip,
			  mTargetValue, bUseSSE);
    }

    if (bAllocation == FALSE) return FALSE;
        
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	if (zLayerBase + threadIndex * layerFactor >= mSize[Z]-1) continue;
	pCThread[threadIndex]->Run();
    }
    
    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	if (zLayerBase + threadIndex * layerFactor >= mSize[Z]-1) continue;
	pCThread[threadIndex]->Wait(); 
    }

    // ŏIIɂ̓xNg̗pāAreallocɕύX

    for (int threadIndex = 0; threadIndex < threadNum; threadIndex++){
	int zLayerIndex = zLayerBase + threadIndex * layerFactor;
	if (zLayerIndex >= mSize[Z]-1) continue;

	mVV_Vertices[mVerticesIndex].resize(arrayIndex[threadIndex]);
	mVV_Normals[mVerticesIndex].resize(arrayIndex[threadIndex]);
		
	//[v̍œKǂ̂ȂWin(Pen4)memcpy̕xH
	if (arrayIndex[threadIndex] != 0){
	    memcpy
		(&mVV_Vertices[mVerticesIndex][0],
		 pTmpVertices[threadIndex], sizeof(float)*arrayIndex[threadIndex]);
	    memcpy
		(&mVV_Normals[mVerticesIndex][0],
		 pTmpNormal[threadIndex], sizeof(float)*arrayIndex[threadIndex]);
	}

	delete pCThread[threadIndex];
	delete pTmpVertices[threadIndex];
	delete pTmpNormal[threadIndex];
	
	GTestVariableIntPoly[mVerticesIndex] = arrayIndex[threadIndex];
	mVerticesIndex++;
    }

    return TRUE;
}

void MarchingCube_cls::
ConstructVertices(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
		  bool bUseSSE, int threadNum, const int layerFactor){
    GbGLLockForMultithread = TRUE;

    int ASkip[3] = {0, 0 ,0};

    Clear(sizeX, sizeZ);
    mSize[X] = sizeX; mSize[Y] = sizeY; mSize[Z] = sizeZ;
    mTargetValue255 = mTargetValue * 255;

    mVerticesIndex = 0;
    for (int xLayerBase = 0; xLayerBase < mSize[X]; xLayerBase += mSize[X]/1)
	for (int zLayerBase = 0; zLayerBase < mSize[Z]-1;
	     zLayerBase += threadNum * layerFactor){
	    
	    if (ThreadControlMain(pData, ASkip,
				  xLayerBase, zLayerBase,
				  bUseSSE, threadNum, layerFactor) == FALSE){
		
		/*
		  MsgBox(NULL, "Isosurface", 
		  "Memory Allocation for temporal buffer fails.");
		*/
	       GbGLLockForMultithread = FALSE;
	       return;
	     }
	    
	}
	
	
    LOOP_XYZ { mScalePrev[xyz] = 1;}

    GbGLLockForMultithread = FALSE;

    printf ("Skip=%d %d %d\n", ASkip[X], ASkip[Y], ASkip[Z]);
}

void MarchingCube_cls::
GLDraw(double scale[3], bool bCullFace, GLuint showType){

    //float scalefactor = 0.5 * sqrt(scale[X] * scale[X] + scale[Y] * scale[Y] + scale[Z] * scale[Z]);

    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindTexture(GL_TEXTURE_3D, 0);

    if (bCullFace == TRUE)
	glEnable(GL_CULL_FACE);

    if (!(scale[X] == mScalePrev[X] && scale[Y] == mScalePrev[Y] && scale[Z] == mScalePrev[Z])){
	for (unsigned int index = 0; index < mVV_Normals.size(); index++){
	    std::vector < float > *pNormal = &mVV_Normals[index];
	    for (unsigned int i = 0; i < mVV_Normals[index].size() ; i += 3)
		LOOP_XYZ (*pNormal)[i+xyz] *= (scale[xyz] / mScalePrev[xyz]);
	}
	LOOP_XYZ mScalePrev[xyz] = scale[xyz];
    }
    
    for (unsigned int index = 0; index < mVV_Normals.size(); index++){
	std::vector < float > *pNormal = &mVV_Normals[index];
	std::vector < float > *pVertices = &mVV_Vertices[index];
    
	if (mVV_Normals[index].size() != 0){
	    glEnableClientState(GL_VERTEX_ARRAY);
	    glEnableClientState(GL_NORMAL_ARRAY);
	    
	    glVertexPointer(3, GL_FLOAT, 0, &(*pVertices)[0]);
	    glNormalPointer(GL_FLOAT, 0, &(*pNormal)[0]);
	
	    glDrawArrays(showType, 0, mVV_Normals[index].size()/3);
	    
	    glDisableClientState(GL_VERTEX_ARRAY);
	    glDisableClientState(GL_NORMAL_ARRAY);
	}

	/*
	glBegin(GL_TRIANGLES);
    	for (unsigned int i = 0; i < mVV_Normals[index].size() ; i += 3){
	    glNormal3fv(&(*pNormal)[i]);
	    glVertex3fv(&(*pVertices)[i]);
	}
	glEnd();
	*/
    }

    glDisable(GL_CULL_FACE);
}

//lists the positions, relative to vertex0, of each of the 8 vertices
