// "clsMarchingCube.cpp"
// (setq c-basic-offset 4)

#ifndef _ISOSURFICE_MARCHINGCUBE_H_
#define _ISOSURFICE_MARCHINGCUBE_H_

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif


#include <vector>
#include <wx/thread.h>

// The main routine of this class used
// Marching Cubes Example Program 
// by Cory Bloyd (corysama@yahoo.com)
// which is publicshd as public domain 

class MarchingCube_cls;

class ThreadMarching : public wxThread {
 public:
    ThreadMarching(unsigned char *pData, 
		   float *pTmpVertices, float *pTmpNormal,
		   int *pArrayIndex);
    void SetParameters(unsigned int size[3], 
		       int initX, int finX,
		       int initZ, int finZ,
		       int *pSkip,
		       float targetValue, bool bUseSSE);

    GLfloat GetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired);
    inline float GetInterpolatedValue
	(const unsigned char *pData,
	 const int lBaseIndex, const float fx, const float fy, const float fz);
    void GetNormalPrecise
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);
    void GetNormalPreciseSSE
	(GLfloat ANorm[3], const unsigned char *pData,
	 const int x, const GLfloat fX, 
	 const int y, const GLfloat fY,
	 const int z, const GLfloat fZ);

    int MarchCubeMainThread(const unsigned char *pData,
			    const int x, const int y, const int z);
    
    
    int MarchCubeLOOPX1(const unsigned char *pData, int y, int z);
    //void MarchCubeLOOPX2(const unsigned char *pData, int y, int z);
    int MarchCubeLOOPY(const unsigned char *pData, int z);
   
    virtual ExitCode Entry();

    unsigned char * const mpData;
    float * const mpTmpVertices;
    float * const mpTmpNormal;
    int   * const mpArrayIndex;

    int   * mpSkip;
    
    int mSize[3];
    int mInitX, mFinX;
    int mInitZ, mFinZ;

    int mOffsetZ, mOffsetY;
    int mBaseIndex;
    float mTargetValue, mTargetValue255;
    bool mbUseSSE;

    static const GLint   VOffset[8][3];
    static const GLint   a2iEdgeConnection[12][2];
    static const GLfloat a2fEdgeDirection[12][3];

    static const GLint aiCubeEdgeFlags[256];
    static const GLint a2iTriangleConnectionTable[256][16];
};

class MarchingCube_cls{
 public:

    std::vector < std::vector < float > > mVV_Vertices;
    std::vector < std::vector < float > > mVV_Normals;
      
    //std::vector < float* > mVpVertices;
    //std::vector < float* > mVpNormals;
    //std::vector < int > mVAllocSize;

    //int mSizeX, mSizeY, mSizeZ;
    unsigned int mSize[3];
    int mVerticesIndex;
   
    double mScalePrev[3];
    
    
    float mTargetValue, mTargetValue255;
    
    void Clear(int sizeX, int sizeZ);
    bool ThreadControlMain(unsigned char *pData, int *pSkip,
			   int xLayerBase, int zLayerBase,
			   bool bUseSSE, int threadNum, const int layerFactor);
    void ConstructVertices
	(unsigned char *pData, int sizeX, int sizeY, int sizeZ,
	 bool bUseSSE, int threadNum, const int layerFactor);
    void GLDraw(double scale[3], bool bCullFace, GLuint showType);
}; 

#endif

