// "clsPovObjectPlugin.h"
// (setq c-basic-offset 4)

#include "clsPovObjectPlugin.h"

#ifdef WIN32
#include <GL/glu.h>
#include <GL/glext.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#define ENDLWX wxT("\n")

extern int displayListSphere;

void PovObjectPlugin_cls::
DrawGLPlugin(int type, int tree){
    switch(type){
    case _PLUGIN_TOMTREE:
	DrawGLPluginTomTree(tree);
	break;
    }
}

void PovObjectPlugin_cls::
DrawGLPluginTomTree(int tree){

    glPushMatrix();
    
    float birkColor[4] = {0.4, 0.4, 0.4, 1};
    float leafColorDif[4] = {0.1, 0.6, 0.05, 1};
    float leafColorAmb[4] = {0.1, 0.2, 0.05, 1};
    float zeroColor[4] = {0, 0, 0, 0};
    float shininess = 90;

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, birkColor);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, birkColor);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, birkColor);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shininess);

    glTranslated(0, 0, 0.4);    
    glScaled(0.06, 0.06, 0.8);
    DrawBox();

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, leafColorAmb);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, leafColorDif);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, zeroColor);

    glEnable( GL_NORMALIZE );
    switch(tree){
    case _TOMTREE_BIRCH:
	glTranslated(0, 0, 0.4);
	glScaled(5, 5, 0.5); break;
    case _TOMTREE_OLIVE:
	glTranslated(0, 0, 0.4);
	glScaled(7, 7, 0.5); break;
    case _TOMTREE_PALM:
	glTranslated(0, 0, 0.6);
	glScaled(6, 6, 0.4); break;
    }
    
    glCallList(displayListSphere);
   
    /*
    glTranslated(0.95, -0.78, 0);
    for (int i = 0; i < 50; i++){
	glTranslated(0, 1.5, 0.3);
	glRotated(70, 0, 0, 1);
	glScaled(0.9, 0.9, 0.9);
	DrawBox();
    }
    */

    glPopMatrix();
}

wxString PovObjectPlugin_cls::
GetPovStrPluginParameters(int type, int tree, int quality){
    wxString strParam;

    switch(type){
    case _PLUGIN_TOMTREE:
	strParam << GetPovStrPluginParametersTomTree(tree, quality);
	break;
    }

    return strParam;
}

wxString PovObjectPlugin_cls::
GetPovStrPluginObject(int type, int tree){
    wxString strObject;

    switch(type){
    case _PLUGIN_TOMTREE:
	strObject << GetPovStrPluginObjectTomTree(tree);
	break;
    }

    return strObject;
}

wxString PovObjectPlugin_cls::
GetPovStrPluginParametersTomTree(int tree, int quality){
    wxString strParam;

    wxString strTree;
    switch(tree){
    case _TOMTREE_BIRCH: strTree << wxT("BIRCH"); break;
    case _TOMTREE_OLIVE: strTree << wxT("OLIVE"); break;
    case _TOMTREE_PALM : strTree << wxT("PALM"); break;
    }
    if (quality == 0) strTree << wxT("_LQ");

    strParam << wxT("#include \"TomTree/") << strTree << wxT(".inc\"") << ENDLWX;
    strParam << wxT("#include \"TomTree/TOMTREE.inc\"") << ENDLWX;
    
    return strParam;
}

wxString PovObjectPlugin_cls::
GetPovStrPluginObjectTomTree(int tree){
    wxString strObject;

    strObject << ENDLWX;
    strObject << wxT("light_group {") << ENDLWX
	      << wxT("light_source { <0, 0, 10> color rgb <1.5, 1.5, 1.5> shadowless }") << ENDLWX;
    strObject << wxT("\tobject { TREE rotate <90, 0, 0> scale 1.5 double_illuminate }") << ENDLWX;
    
    return strObject;
}
