
// "clsPovCrossSection.h"
// (setq c-basic-offset 4)

#include "clsPovCrossSectionCommon.h"

#include "../MYAPI/MsgBox.h"

#include "../MAINCANVAS/clsPreference.h"

#include <GL/glu.h>

extern bool GbGLLockForMultithread;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

ThreadCalcNormal::
ThreadCalcNormal() : wxThread(wxTHREAD_JOINABLE) {}

void ThreadCalcNormal::
SetParameters(std::vector < std::vector < unsigned char > > *pVVField,
	      std::vector < std::vector < double > > *pVVNormalU,
	      std::vector < std::vector < double > > *pVVNormalV,
	      std::vector < std::vector < double > > *pVVNormalW,
	      int sizeU, int iniV, int finV,
	      double embossU, double embossV){
    mpVVField = pVVField;
    mpVVNormalU = pVVNormalU;
    mpVVNormalV = pVVNormalV;
    mpVVNormalW = pVVNormalW;
    mSizeU = sizeU; mIniV = iniV; mFinV = finV;
    mEmbossU = embossU; mEmbossV = embossV;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadCalcNormal::
Entry(){
    std::vector < std::vector < unsigned char > > *pVVField = mpVVField;
    std::vector < std::vector < double > > *pVVNormalU = mpVVNormalU;
    std::vector < std::vector < double > > *pVVNormalV = mpVVNormalV;
    std::vector < std::vector < double > > *pVVNormalW = mpVVNormalW;
    
    const int sizeUminus = mSizeU - 1;
    const int finV = mFinV;
    for (int v = mIniV; v < finV; v++){
	unsigned char *pVFieldLine = &((*pVVField)[v][0]);
	unsigned char *pVFieldLineP = &((*pVVField)[v+1][0]);
	unsigned char *pVFieldLineM = &((*pVVField)[v-1][0]);
	double *pVNormalLineU = &((*pVVNormalU)[v][0]);
	double *pVNormalLineV = &((*pVVNormalV)[v][0]);
	for (int u = 1; u < sizeUminus; u++){
	    pVNormalLineU[u] = (double)
		(pVFieldLine[u+1] - pVFieldLine[u-1])/2;
	    pVNormalLineV[u] = (double)
		(pVFieldLineP[u] - pVFieldLineM[u])/2;
	}
    }

    int lu = 0;
    for (int v = mIniV; v < finV; v++){
	(*pVVNormalU)[v][lu] = (double)
	    ((*pVVField)[v][lu+1] - (*pVVField)[v][lu]);
	(*pVVNormalV)[v][lu] = (double)
	    ((*pVVField)[v+1][lu] - (*pVVField)[v-1][lu]) / 2;
    }
    
    lu = mSizeU-1;
    for (int v = mIniV; v < finV; v++){
	(*pVVNormalU)[v][lu] = (double)
	    ((*pVVField)[v][lu] - (*pVVField)[v][lu-1]);
	(*pVVNormalV)[v][lu] = (double)
	    ((*pVVField)[v+1][lu] - (*pVVField)[v-1][lu]) / 2;
    }

    const int sizeU = mSizeU;
    const double embossU = -mEmbossU, embossV = -mEmbossV;
    for (int v = mIniV; v < finV; v++){
	double *pVNormalLineU = &((*pVVNormalU)[v][0]);
	double *pVNormalLineV = &((*pVVNormalV)[v][0]);
	double *pVNormalLineW = &((*pVVNormalW)[v][0]);
	for (int u = 0; u < sizeU; u++){
	    pVNormalLineU[u] *= embossU;
	    pVNormalLineV[u] *= embossV;
	    double scale = 
		sqrt(pVNormalLineU[u] * pVNormalLineU[u] +
		     pVNormalLineV[u] * pVNormalLineV[u] + 1);
	    pVNormalLineW[u] = 1.0 / scale;
	    pVNormalLineU[u] *= pVNormalLineW[u];
	    pVNormalLineV[u] *= pVNormalLineW[u];
	}
    }

    return 0;
}

// ----------------------

PovCrossSectionCommon_cls::
PovCrossSectionCommon_cls(FilterCtrl_cls *pCParentCtrl) : 
    FilterWithParameter_cls(pCParentCtrl)
{
    mSizeX = mSizeY = mSizeZ = 0;
    
}

void PovCrossSectionCommon_cls::
DrawVertexArray(int colorType, int sizeUm, int sizeVm, 
		float *pTexture, float *pNormals, float *pVertices,
		unsigned int *pPolyIndices, bool bFlip){
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, pVertices);
    glNormalPointer(GL_FLOAT, 0, pNormals);
    if (colorType != 0) {
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(1, GL_FLOAT, 0, pTexture);
    }

    if (bFlip == TRUE)
	for (int v = 0; v < sizeVm; v++){
	    int offsetV0 = v * sizeUm;
	    int offsetV1 = (v+1) * sizeUm;
	    for (int u = 0; u < sizeUm ; u++){
		pPolyIndices[u*2 + 0] = offsetV1 + u;
		pPolyIndices[u*2 + 1] = offsetV0 + u;
	    }
	    glDrawElements(GL_QUAD_STRIP, sizeUm*2,
			   GL_UNSIGNED_INT, pPolyIndices);
	}
    else
	for (int v = 0; v < sizeVm; v++){
	    int offsetV0 = v * sizeUm;
	    int offsetV1 = (v+1) * sizeUm;
	    for (int u = 0; u < sizeUm ; u++){
		pPolyIndices[u*2 + 0] = offsetV0 + u;
	    pPolyIndices[u*2 + 1] = offsetV1 + u;
	    }
	    glDrawElements(GL_QUAD_STRIP, sizeUm*2,
		       GL_UNSIGNED_INT, pPolyIndices);
	}
    
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
}

void PovCrossSectionCommon_cls::
ConstructVerticesCoreX(unsigned char *pData, double layer, 
		       std::vector < std::vector < unsigned char > > *pVVField){
    
    int x = (int)(layer + 1e-6);
    if (x >= mSizeX) x = mSizeX - 1;
    if (x < 0) x = 0;
    int x2 = x+1;
    if (x2 >= mSizeX) x2 = mSizeX - 1;

    float fracx2 = layer - x;
    float fracx1 = 1.0f - fracx2;
    
    if (fracx2 != 0)
	for (int z = 0; z < mSizeZ; z++){
	    int lBaseIndex = mSizeX * mSizeY * z;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[z]);
	    for (int y = 0; y < mSizeY; y++){
		int lOffsetY = mSizeX * y;
		(*pVFieldLine)[y] = 
		    fracx1 * pData[lBaseIndex + lOffsetY + x]+
		    fracx2 * pData[lBaseIndex + lOffsetY + x2];
	    }
	}
    else
	for (int z = 0; z < mSizeZ; z++){
	    int lBaseIndex = mSizeX * mSizeY * z;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[z]);
	    for (int y = 0; y < mSizeY; y++){
		int lOffsetY = mSizeX * y;
		(*pVFieldLine)[y] = pData[lBaseIndex + lOffsetY + x];
	    }
	}
}

void PovCrossSectionCommon_cls::
ConstructVerticesCoreY(unsigned char *pData, double layer, 
		       std::vector < std::vector < unsigned char > > *pVVField){
    int y = (int)(layer + 1e-6);
    if (y >= mSizeY) y = mSizeY - 1;
    if (y < 0) y = 0;
    int y2 = y+1;
    if (y2 >= mSizeY) y2 = mSizeY - 1;

    float fracy2 = layer - y;
    float fracy1 = 1.0f - fracy2;

    int lOffsetY = mSizeX * y;
    int lOffsetY2 = mSizeX * y2;
    int sizeXY = mSizeX * mSizeY;
    if (fracy2 != 0)
	for (int z = 0; z < mSizeZ; z++){
	    int lBaseIndex = sizeXY * z;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[z]);
	    for (int x = 0; x < mSizeX; x++)
		(*pVFieldLine)[x] = 
		    fracy1 * pData[lBaseIndex + lOffsetY + x] +
		    fracy2 * pData[lBaseIndex + lOffsetY2 + x];
	}
    else
	for (int z = 0; z < mSizeZ; z++){
	    int lBaseIndex = sizeXY * z;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[z]);
	    for (int x = 0; x < mSizeX; x++)
		(*pVFieldLine)[x] = pData[lBaseIndex + lOffsetY + x];
	}
}

void PovCrossSectionCommon_cls::
ConstructVerticesCoreZ(unsigned char *pData, double layer, 
		       std::vector < std::vector < unsigned char > > *pVVField){
    int z = (int)(layer + 1e-6);
    if (z >= mSizeZ) z = mSizeZ - 1;
    if (z < 0) z = 0;
    int z2 = z+1;
    if (z2 >= mSizeZ) z2 = mSizeZ - 1;

    float fracz2 = layer - z;
    float fracz1 = 1.0f - fracz2;

    int lBaseIndex = mSizeX * mSizeY * z;
    int lBaseIndex2 = mSizeX * mSizeY * z2;
    
    int sizeX = mSizeX;
    if (fracz2 != 0)
	for (int y = 0; y < mSizeY; y++){
	    int lOffsetY = sizeX * y;
	    int base1 = lBaseIndex + lOffsetY;
	    int base2 = lBaseIndex2 + lOffsetY;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[y]);
	    for (int x = 0; x < sizeX; x++){
		(*pVFieldLine)[x] = 
		    fracz1 * pData[base1 + x] +
		    fracz2 * pData[base2 + x];
	    }
	}
    else
	for (int y = 0; y < mSizeY; y++){
	    int lOffsetY = sizeX * y;
	    int base1 = lBaseIndex + lOffsetY;
	    int base2 = lBaseIndex2 + lOffsetY;
	    std::vector < unsigned char > *pVFieldLine = &((*pVVField)[y]);
	    for (int x = 0; x < sizeX; x++)
		(*pVFieldLine)[x] = pData[base1 + x];
	}
}

void PovCrossSectionCommon_cls::
CalcNormal(std::vector < std::vector < unsigned char > > *pVVField,
	   std::vector < std::vector < double > > *pVVNormalU,
	   std::vector < std::vector < double > > *pVVNormalV,
	   std::vector < std::vector < double > > *pVVNormalW,
	   int sizeU, int sizeV, double emboss, bool bInvert){
    GbGLLockForMultithread = TRUE;
    
    // ܂SȊO
    int lv = 0;
    unsigned char *pVFieldLine = &((*pVVField)[lv][0]);
    unsigned char *pVFieldLineP = &((*pVVField)[lv+1][0]);
    double *pVNormalLineU = &((*pVVNormalU)[lv][0]);
    double *pVNormalLineV = &((*pVVNormalV)[lv][0]);
    for (int u = 1; u < sizeU-1; u++){
	pVNormalLineU[u] = (double)(pVFieldLine[u+1] - pVFieldLine[u-1]) / 2;
	pVNormalLineV[u] = (double)(pVFieldLineP[u] - pVFieldLine[u]);
    }
    
    lv = sizeV - 1;
    pVFieldLine = &((*pVVField)[lv][0]);
    unsigned char *pVFieldLineM = &((*pVVField)[lv-1][0]);
    pVNormalLineU = &((*pVVNormalU)[lv][0]);
    pVNormalLineV = &((*pVVNormalV)[lv][0]);
    for (int u = 1; u < sizeU-1; u++){
	pVNormalLineU[u] = (double)(pVFieldLine[u+1] - pVFieldLine[u-1]) / 2;
	pVNormalLineU[u] = (double)(pVFieldLine[u] - pVFieldLineM[u]);
    }

    int lu = 0; lv = 0;
    (*pVVNormalU)[lv][lu] = (double)((*pVVField)[lv][lu+1]-(*pVVField)[lv][lu]);
    (*pVVNormalV)[lv][lu] = (double)((*pVVField)[lv+1][lu]-(*pVVField)[lv][lu]);
    
    lu = sizeU-1; lv = 0;
    (*pVVNormalU)[lv][lu] = (double)((*pVVField)[lv][lu]-(*pVVField)[lv][lu-1]);
    (*pVVNormalV)[lv][lu] = (double)((*pVVField)[lv+1][lu]-(*pVVField)[lv][lu]);

    lu = 0; lv = sizeV-1;
    (*pVVNormalU)[lv][lu] = (double)((*pVVField)[lv][lu+1]-(*pVVField)[lv][lu]);
    (*pVVNormalV)[lv][lu] = (double)((*pVVField)[lv][lu]-(*pVVField)[lv-1][lu]);

    lu = sizeU-1; lv = sizeV-1;
    (*pVVNormalU)[lv][lu] = (double)((*pVVField)[lv][lu]-(*pVVField)[lv][lu-1]);
    (*pVVNormalV)[lv][lu] = (double)((*pVVField)[lv][lu]-(*pVVField)[lv-1][lu]);

    double embossU = emboss * sizeU / 4096;
    double embossV = emboss * sizeV / 4096;

    // S}`Xbh
    ThreadCalcNormal *pThread[_MAX_THREAD_NUM];
    int threadNum = mspCPreferenceLink->GetThreadNumFilter();

    int lineStep = 64;

    for (int lineBase = 1; lineBase < sizeV-1; lineBase += threadNum * lineStep){
	for (int thread = 0; thread < threadNum; thread++){
	    int lineIni = lineBase + thread * lineStep;
	    int lineFin = lineBase + thread * lineStep + lineStep;
	    if (lineIni > (int)sizeV-2) continue;
	    if (lineFin > (int)sizeV-1) lineFin = sizeV-1;

	    pThread[thread] = new ThreadCalcNormal;
	    pThread[thread]->SetParameters
		(pVVField, pVVNormalU, pVVNormalV, pVVNormalW,
		 sizeU, lineIni, lineFin, embossU, embossV);
	}

	for (int thread = 0; thread < threadNum; thread++){
	    int lineIni = lineBase + thread * lineStep;
	    if (lineIni > (int)sizeV-2) continue;
	    pThread[thread]->Run();
	}
	    
	for (int thread = 0; thread < threadNum; thread++){
	    int lineIni = lineBase + thread * lineStep;
	    if (lineIni > (int)sizeV-2) continue;
	    pThread[thread]->Wait();
	}

	for (int thread = 0; thread < threadNum; thread++){
	    int lineIni = lineBase + thread * lineStep;
	    if (lineIni > (int)sizeV-2) continue;
	    delete pThread[thread];
	}

    }
	    
    GbGLLockForMultithread = FALSE;

    // ŌɎc
    lv = 0;
    pVNormalLineU = &((*pVVNormalU)[lv][0]);
    pVNormalLineV = &((*pVVNormalV)[lv][0]);
    double *pVNormalLineW = &((*pVVNormalW)[lv][0]);
    for (int u = 0; u < sizeU; u++){
	pVNormalLineU[u]*= - embossU;
	pVNormalLineV[u]*= - embossU;
	double scale = sqrt( pVNormalLineU[u] * pVNormalLineU[u] +
			     pVNormalLineV[u] * pVNormalLineV[u] + 1);
	pVNormalLineW[u] = 1.0 / scale;
	pVNormalLineU[u] *= pVNormalLineW[u];
	pVNormalLineV[u] *= pVNormalLineW[u];
    }

    lv = sizeV - 1;
    pVNormalLineU = &((*pVVNormalU)[lv][0]);
    pVNormalLineV = &((*pVVNormalV)[lv][0]);
    pVNormalLineW = &((*pVVNormalW)[lv][0]);
    for (int u = 0; u < sizeU; u++){
	pVNormalLineU[u] *= - embossU;
	pVNormalLineV[u] *= - embossV;
	double scale = sqrt( pVNormalLineU[u] * pVNormalLineU[u] +
			     pVNormalLineV[u] * pVNormalLineV[u] + 1);
	pVNormalLineW[u] = 1.0 / scale;
	pVNormalLineU[u] *= pVNormalLineW[u];
	pVNormalLineV[u] *= pVNormalLineW[u];
    }

    // FIX_ME!!
    // x,y,zɂčlȂɍ̂ŁAx,y ̎ɔ]ȂƂȂB
    // Best::{璼A
    // Better::SvZtɂŷƂ]ɂƏ}V
    if (bInvert == TRUE)
	for (int lv = 0; lv < sizeV; lv++){
	    pVNormalLineU = &((*pVVNormalU)[lv][0]);
	    pVNormalLineV = &((*pVVNormalV)[lv][0]);
	    pVNormalLineW = &((*pVVNormalW)[lv][0]);
	    for (int u = 0; u < sizeU; u++){
		pVNormalLineU[u] *= -1;
		pVNormalLineV[u] *= -1;
		pVNormalLineW[u] *= -1;
	    }
	}
}

void PovCrossSectionCommon_cls::
RecalcNorm(std::vector < std::vector < double > > *pVVNormalU,
	   std::vector < std::vector < double > > *pVVNormalV,
	   std::vector < std::vector < double > > *pVVNormalW,
	   double scale[3], int sizeU, int sizeV,
	   int baseAxis, int axisU, int axisV, int axisW){
    if (scale[X] == mScalePrev[baseAxis][X] && 
	scale[Y] == mScalePrev[baseAxis][Y] && 
	scale[Z] == mScalePrev[baseAxis][Z]) return;
    
    double lScale[3], lScalePrev[3];
    LOOP_XYZ lScale[xyz] = scale[xyz];
    LOOP_XYZ lScalePrev[xyz] = mScalePrev[baseAxis][xyz];
    LOOP_XYZ if (lScale[xyz] < 0) lScale[xyz] *= -1;
    LOOP_XYZ if (lScalePrev[xyz]< 0) lScalePrev[xyz] *= -1;
    
    double scaleU = mScalePrev[baseAxis][axisU];
    double scaleV = mScalePrev[baseAxis][axisV];
    double scaleW = mScalePrev[baseAxis][axisW];
    for (int v = 0; v < sizeV; v++)
	for (int u = 0; u < sizeU; u++){
	    (*pVVNormalU)[v][u] *= lScale[axisU] / scaleU;
	    (*pVVNormalV)[v][u] *= lScale[axisV] / scaleV;
	    (*pVVNormalW)[v][u] *= lScale[axisW] / scaleW;
	}
    
    LOOP_XYZ mScalePrev[baseAxis][xyz] = scale[xyz];
}
