// -*- c -*- 
// (setq c-basic-offset 4)
// "clsPovRay.h"

#include "clsPovRay.h"
#include "clsPovVolume.h"
#include "clsPovIsosurface.h"
#include "clsPovCrossSection.h"
#include "clsPovThread.h"
#include "clsStereoGram.h"
#include "OPTIONAL/clsPovBG.h"
#include "OPTIONAL/clsPovFocalBlur.h"
#include "OPTIONAL/clsPovRadiosity.h"
#include "OPTIONAL/clsPovPhotons.h"

#include "../COMMON/clsFilename.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsClipping.h"
#include "../GLCANVAS/clsGLCanvasCtrl.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
#include "../MYAPI/MsgBox.h"
#include "../MAINCANVAS/clsPreviewPanel.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"
#include "../WIDGETCTRL/EX/clsStaticTextEX.h"
#include "../WIDGETCTRL/EX/clsFileDialogEX.h"
extern wxFileDialogEX*   GpCFileDialog;

#include "../../clsCameraPath.h"
#include "../../clsFrameMain.h"
#include "../../clsMain.h"
#include "../../clsTimeCtrl.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#ifndef WIN32
#include <unistd.h>
#endif

#include <sstream>

#include <wx/colordlg.h>
#include <wx/file.h>
#include <wx/filename.h>
#include <wx/statline.h>
#include <wx/valgen.h>
extern wxColourDialogEX* GpCClrDialog;

#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include "../WIDGETCTRL/EX/clsButtonEX.h"
#include "../WIDGETCTRL/EX/clsCheckBoxEX.h"
#include "../WIDGETCTRL/clsGraphPanel.h"
#include "../WIDGETCTRL/clsAtoiTextCtrl.h"
#include "../WIDGETCTRL/clsAtoiSpinCtrl.h"
#include "../WIDGETCTRL/clsAtoWXSTextCtrl.h"

#ifdef WIN32
#define POVFILE wxT("pvengine*.exe files (pvengine.exe)|pvengine*.exe|all files|*")
#define POVDEFWX wxT("c:\\Program Files\\POV-Ray for Windows v3.6\\bin\\pvengine.exe")
#define SEPARATOR L"\\"
#define SEPARATORWX wxT("\\")
//#define POVOPTIONWX wxT(" /RENDER ")
#define POVOPTIONWX wxT("")
#define SNWPRINTF _snwprintf
#else
#define POVFILE wxT("povray files (povray)|*povray*")
#define POVDEFWX wxT("/usr/bin/povray")
#define SEPARATOR L"/"
#define SEPARATORWX wxT("/")
//#define POVOPTION " &"
#define POVOPTIONWX wxT("")
#define SNWPRINTF snprintf
#endif

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

#define R 0
#define G 1
#define B 2
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)

#define ENDL std::endl
#define ENDLWX wxT("\n")

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

// ----

// ÓIɃCxgт
BEGIN_EVENT_TABLE(PovRayGUI_cls, wxDialog)
EVT_SIZE        (wxDialogEXKey::OnSize)
EVT_MOUSE_EVENTS(PovRayGUI_cls::OnMouse)
EVT_BUTTON(btnPov01, PovRayGUI_cls::OnButtonPathPovray)
EVT_BUTTON(btnPov02, PovRayGUI_cls::OnButtonWorkingDir)
EVT_BUTTON(btnPovExportSnap, PovRayGUI_cls::OnButtonExportSnap)
EVT_BUTTON(btnPovExportAnim, PovRayGUI_cls::OnButtonExportAnim)
EVT_BUTTON(btnPovRendSnap, PovRayGUI_cls::OnButtonRenderSnap)
EVT_BUTTON(btnPovRendAnim, PovRayGUI_cls::OnButtonRenderAnim)
EVT_BUTTON(btnPovApply   , PovRayGUI_cls::OnButtonApply)
EVT_BUTTON(btnPovDefault , PovRayGUI_cls::OnButtonDefault)
EVT_BUTTON(btnPovAdvanced, PovRayGUI_cls::OnButtonAdvanced)
EVT_BUTTON(wxID_CANCEL   , PovRayGUI_cls::OnButtonClose)
EVT_BUTTON(btnPovVolumePref, PovRayGUI_cls::OnButtonShowItemVolume)
EVT_BUTTON(btnPovIsosurfPref, PovRayGUI_cls::OnButtonShowItemIsosurf)
EVT_BUTTON(btnPovRadiosity, PovRayGUI_cls::OnButtonShowItemRadios)
EVT_BUTTON(btnPovPhotons, PovRayGUI_cls::OnButtonShowItemPhotons)
EVT_BUTTON(btnPovFocalBlur, PovRayGUI_cls::OnButtonShowItemFocalBlur)
EVT_BUTTON(btnPovBG,        PovRayGUI_cls::OnButtonShowItemBG)
EVT_BUTTON(btnPovBG0,       PovRayGUI_cls::OnButtonBGColor0)
EVT_BUTTON(btnPovBG1,       PovRayGUI_cls::OnButtonBGColor1)
EVT_BUTTON(btnPovPluginHelp, PovRayGUI_cls::OnButtonPluginHelp)
EVT_BUTTON(btnPovBGN1,       PovRayGUI_cls::OnButtonBGNColor1)
EVT_BUTTON(btnPovBGN2,       PovRayGUI_cls::OnButtonBGNColor2)
EVT_BUTTON(btnPovBGN3,       PovRayGUI_cls::OnButtonBGNColor3)
EVT_BUTTON(btnPovFormatPref, PovRayGUI_cls::OnButtonShowItemFormat)
EVT_MENU(menuItemMinmizeButton, PovRayGUI_cls::OnMenuMinimize)
END_EVENT_TABLE()

PovRayGUI_cls::
PovRayGUI_cls(PovRay_cls* pCParent, wxWindow *pC,
	      wxPoint pos, wxSize size) :
wxDialogEXKey(pC, wxID_ANY, _("Pov-Ray Rendering Configure"), 
	      pos, size, pCParent->mpBitmapLink) {
	 
    mspCThis = this;
    mpCParent = pCParent;
    SetScreenRectPos(&pCParent->mCScreenRect);
    
    // TCU[
    wxBoxSizer *pBoxSizerSrc    = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerConf1  = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerConf2 = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerConf3 = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerConf4 = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerParam  = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamVolume  = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamIsosurf = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamRadios = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamPhotons = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamFocalBlur = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamBG = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamFormat = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerParamType = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerBTMA    = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer *pBoxSizerBTMB    = new wxBoxSizer(wxHORIZONTAL);
    
    // CONFwithWX
#ifdef WIN32
    int textWidth = 85;
#else
    int textWidth = 110;
#endif
    mpCParent->mCConfPathPovray.
	SetItem(this, _("Pov-Ray Path="), wxT("POVRAY_PATH="),
		mpCParent->mCConfPathPovray.msBody, wxSize(textWidth,-1));
    mpCParent->mCConfWorkingDir.
	SetItem(this, _("Work Directory="), wxT("WORKING_DIR="), 
		mpCParent->mCConfWorkingDir.msBody, wxSize(textWidth,-1));
    mpCParent->mCConfSnapFile.
	SetItem(this, _("Snap File="), wxT("POV_SNAP_FILE="), 
		wxT("povsnap%04d"), wxSize(textWidth + 15,-1));
    mpCParent->mCConfAnimFile.
	SetItem(this, _("Anim File="), wxT("POV_ANIM_FILE="), 
		wxT("povanim%04d"), wxSize(textWidth + 15,-1));
    
    mpCParent->mCConfPathPovray.mpCBody->
	SetEventOnTab(this, wxEVT_COMMAND_BUTTON_CLICKED, btnPovApply);
    mpCParent->mCConfWorkingDir.mpCBody->
	SetEventOnTab(this, wxEVT_COMMAND_BUTTON_CLICKED, btnPovApply);
    
    //PATH
    pBoxSizerConf1->Add( 5, -1);
    mpCCheckPovray = new wxPanel(this, wxID_ANY, wxDefaultPosition, wxSize(10,10), wxBORDER_SUNKEN);
    pBoxSizerConf1->Add(mpCCheckPovray, 0, wxALIGN_CENTER_VERTICAL);
    pBoxSizerConf1->Add( 5, -1);
    mpCParent->mCConfPathPovray.AddToSizer(pBoxSizerConf1);
    wxButton *pCButtonPathPovRay = new wxButton
	(this, btnPov01, wxT("..."), wxDefaultPosition, wxSize(25, -1));
    pBoxSizerConf1->Add( pCButtonPathPovRay, 0, wxALIGN_CENTER_VERTICAL);

    pBoxSizerConf2->Add( 5, -1);
    mpCCheckWorkDir = new wxPanel(this, wxID_ANY, wxDefaultPosition, wxSize(10,10), wxBORDER_SUNKEN);
    mpCCheckWorkDir->SetToolTip(_("red if not valid."));
    pBoxSizerConf2->Add(mpCCheckWorkDir, 0, wxALIGN_CENTER_VERTICAL);
    pBoxSizerConf2->Add( 5, -1);
    mpCParent->mCConfWorkingDir.AddToSizer(pBoxSizerConf2);
    wxButton *pCButtonWorkingDir = new wxButton
	(this, btnPov02, wxT("..."), wxDefaultPosition, wxSize(25, -1));
    pBoxSizerConf2->Add( pCButtonWorkingDir, 0, wxALIGN_CENTER_VERTICAL);

    pBoxSizerConf3->Add( 8, -1);
    mpCParent->mCConfSnapFile.AddToSizer(pBoxSizerConf3);
    mpCParent->mpCSnapNumber = new AtoiSpinCtrl_cls (this, wxID_ANY, 0);
    pBoxSizerConf3->Add( mpCParent->mpCSnapNumber, 0, wxALIGN_CENTER_VERTICAL);
    pBoxSizerConf4->Add( 8, -1);
    mpCParent->mCConfAnimFile.AddToSizer(pBoxSizerConf4);
    wxCheckBoxEX *pCAbsolutePath = new wxCheckBoxEX(this, wxID_ANY, _("Absolute Path"), wxGenericValidator(&mpCParent->mCbConfAbsolutePath.mbBody));
    pBoxSizerConf4->Add(pCAbsolutePath, 0, wxALIGN_CENTER_VERTICAL);

    //
    wxStaticLine *pCStaticLine1 = new wxStaticLine(this);

    SetMNumCameraParamAdvanced();

    SetMNumVolume(pBoxSizerParamVolume); // VOLUME_PARAM
    pBoxSizerParam->Add(pBoxSizerParamVolume, 0, wxEXPAND);

    SetMNumIsosurf(pBoxSizerParamIsosurf); // ISOSURF_PARAM
    pBoxSizerParam->Add(pBoxSizerParamIsosurf, 0, wxEXPAND);



    mpCParent->mpCRadiosity->SetMNumRadiosity
	(pBoxSizerParamRadios, this, &mpCParent->mCConfWorkingDir.msBody); 
    pBoxSizerParam->Add(pBoxSizerParamRadios, 0, wxEXPAND);

    mpCParent->mpCPhotons->SetMNumPhotons
	(pBoxSizerParamPhotons, this,  &mpCParent->mCConfWorkingDir.msBody); 
    pBoxSizerParam->Add(pBoxSizerParamPhotons, 0, wxEXPAND);

    mpCParent->mpCFocalBlur->SetMNumFocalBlur
	(pBoxSizerParamFocalBlur, this); 
    pBoxSizerParam->Add(pBoxSizerParamFocalBlur, 0, wxEXPAND);

    mpCParent->mpCBG->SetMNumBG
	(pBoxSizerParamBG, this);
    pBoxSizerParam->Add(pBoxSizerParamBG, 0, wxEXPAND);

    SetMNumFormat(pBoxSizerParamFormat); // FORMAT_PARAM
    pBoxSizerParam->Add(pBoxSizerParamFormat, 0, wxEXPAND);

    SetMNumTypeParam(pBoxSizerParamType);
    pBoxSizerParam->Add(pBoxSizerParamType, 0, wxEXPAND);
    
    // BTM
    wxButtonEX *pCButtonExportSnap = 
	new wxButtonEX(this, btnPovExportSnap, wxT("ExportSnap"));
    /*
    wxButtonEX *pCButtonExportAnim = 
	new wxButtonEX(this, btnPovExportAnim, wxT("ExportAnim"));
    */
    mpCButtonExportAnim = 
	new wxButtonEX(this, btnPovExportAnim, wxT("ExportAnim"));
    wxButtonEX *pCButtonRenderSnap = 
	new wxButtonEX(this, btnPovRendSnap,   wxT("RenderSnap"));
    mpCButtonRenderAnim = 
	new wxButtonEX(this, btnPovRendAnim,   wxT("RenderAnim"));
    
    wxButtonEX *pCButtonAdv   = new wxButtonEX (this, btnPovAdvanced, 
						_("Advanced"));
    wxButtonEX *pCButtonDEF   = new wxButtonEX (this, btnPovDefault, 
						_("Set to Defalt"));
    wxButtonEX *pCButtonClose = new wxButtonEX (this, wxID_CANCEL,
						wxT("Close"));
    pBoxSizerBTMA->Add( pCButtonExportSnap, 0, wxEXPAND);
    pBoxSizerBTMA->Add( mpCButtonExportAnim, 0, wxEXPAND);
    pBoxSizerBTMA->Add( pCButtonRenderSnap, 0, wxEXPAND);
    pBoxSizerBTMA->Add( mpCButtonRenderAnim, 0, wxEXPAND);
    //pBoxSizerBTMB->Add( -1, -1, 1, wxEXPAND);
    pBoxSizerBTMB->Add( pCButtonAdv,    0, wxEXPAND);
    pBoxSizerBTMB->Add( -1, -1, 1, wxEXPAND);
    pBoxSizerBTMB->Add( pCButtonDEF,    0, wxEXPAND);
    pBoxSizerBTMB->Add( pCButtonClose,  0, wxEXPAND);

    // SRC
    pBoxSizerSrc->Add( pBoxSizerConf1, 0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerConf2, 0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerConf3, 0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerConf4, 0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerParam, 0, wxEXPAND);
    pBoxSizerSrc->Add( pCStaticLine1, 0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerBTMA,  0, wxEXPAND);
    pBoxSizerSrc->Add( pBoxSizerBTMB,  0, wxEXPAND);

    pBoxSizerSrc->Add(pBoxSizerSrc->GetMinSize().GetWidth(), -1, 0, 0);
    SetSizerAndFit(pBoxSizerSrc);

    // -----



    int minimumWidth = GetClientSize().GetWidth() - 8;
    minimumWidth = mpCMNumVolume->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumIsosurf->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumFormat->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumTypeParam->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCParent->mpCRadiosity->mpCMNumRadiosity->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCParent->mpCPhotons->mpCMNumPhotons->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCParent->mpCFocalBlur->mpCMNumFocalBlur->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCParent->mpCBG->mpCMNumBG->GetPaddingWidth(minimumWidth);
    
    mpCMNumVolume->SetPaddingWidth(minimumWidth);
    mpCMNumIsosurf->SetPaddingWidth(minimumWidth);
    mpCMNumFormat->SetPaddingWidth(minimumWidth);
    mpCMNumTypeParam->SetPaddingWidth(minimumWidth);
    mpCParent->mpCRadiosity->mpCMNumRadiosity->SetPaddingWidth(minimumWidth);
    mpCParent->mpCPhotons->mpCMNumPhotons->SetPaddingWidth(minimumWidth);
    mpCParent->mpCFocalBlur->mpCMNumFocalBlur->SetPaddingWidth(minimumWidth);
    mpCParent->mpCBG->mpCMNumBG->SetPaddingWidth(minimumWidth);



    //UpdateValues();
    pCButtonClose->SetFocus();
}

PovRayGUI_cls::
~PovRayGUI_cls(){
    MakeModal(FALSE);
    mpCParent->mCConfWorkingDir.ReleaseWX();
    mpCParent->mCConfPathPovray.ReleaseWX();
    mpCParent->mCConfSnapFile.ReleaseWX();
    mpCParent->mCConfAnimFile.ReleaseWX();
}

void PovRayGUI_cls::
UpdateComboboxList(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumTypeParam;

    if (mpCParent->mpCParentFrame->mpCStereo->mCType.mBody == _STEREO_NONE){
	mpCParent->mCStereoEye.mBody = 0;
	//pCurr->mVCPointers[mMNumIndexStereo].GetPCtrlWindow()->Hide();
    }
    else{
	mpCParent->mCStereoEye.mBody = 
	    mpCParent->mpCParentFrame->mpCStereo->mCEye.mBody;
	//pCurr->mVCPointers[mMNumIndexStereo].GetPCtrlWindow()->Show();
    }
    
    wxString pListCamera[_PATH_NUM_MAX + 5] = 
	{_("Follow Active Camera"), _("Same view to Preview0"), _("Same view to Preview1"), _("Same view to Preview2"), _("Same view to Preview3")};
    for (int i = 0; i < _PATH_NUM_MAX; i++)
	pListCamera[i + 5] << wxT("Cam") << i;

    pCurr->mVCPointers[mMNumIndexCamSelect].CombinePointer
	(_GP_UNSIGNED_INT_LIST, NULL,
	 _("Camera View"), TRUE, 5 + mpCParent->mpCParentFrame->mpCPathCtrl->mCCameraNum.mBody, pListCamera);
}

void PovRayGUI_cls::
SetMNumVolume(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumVolume, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNVolume, btnPovVolumePref,
				    _("[+]Volume Render Settings"));
    int boxWidth = 2;

    pCurr->mTextWidth    = -1;
    pCurr->mComboWidth = 108;
    pCurr->mItemWidth = 108;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Volume Method"));
    wxString strListInterp[3] = {
	wxT("0: None"),wxT("1: Tri-Linear"),wxT("2: Tri-Cubic")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			  &PovVolume_cls::msCInterpolate.mBody,
			  wxT(""), TRUE, 3, strListInterp, 1);
    // ftHg Adaptive Œ
    wxString strListSampling[3] = {
	wxT("1: Random"),wxT("2: Evenly"),wxT("3: Adaptive")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			  &PovVolume_cls::msCMethod.mBody,
			  wxT(""), TRUE, 3, strListSampling, 3, 1);

    mIndexIntervals = pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &PovVolume_cls::msCInterval.mBody,wxT("Intervals"));
    pCurr->SetRange(1, 63);
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Samples(min,max)"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, 
			  &PovVolume_cls::msCSampleMin.mBody);
    pCurr->SetRange(4, 2048);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, 
			  &PovVolume_cls::msCSampleMax.mBody);
    pCurr->SetRange(8, 2048);

    mIndexVolumeAA = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("AALevel/Threshold"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, 
			  &PovVolume_cls::msCAALevel.mBody);
    pCurr->SetRange(2, 16);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &PovVolume_cls::msCAAThreshould.mBody);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    //pCurr->mProportion = 1;

    pCurr->FinishSetting(pBoxSizer, FALSE);
    
    pCurr->SetFuncHook(HookHideAndShow, _MN_HOOK_COMBOBOX);
}

void PovRayGUI_cls::
SetMNumIsosurf(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumIsosurf, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNIsosurf, btnPovIsosurfPref,
				    _("[+]Isosurface Render Settings"));
    pCurr->mTextWidth     = -1;
    pCurr->mComboWidth = 100;
    
    wxString outputType[3] =
	{_("Link to .df3"), wxT("Link to *.pov (Mesh2)"),
	 wxT("Pack in pov")};
    mIndexIsosurfaceType = pCurr->
	CombinePointer(_GP_UNSIGNED_INT_LIST, &PovIsosurface_cls::msCOutputType.mBody,
		       _("isosurface type"), TRUE, 3, outputType);

    mIndexIsosurfacePath = pCurr->
	CombinePointer(_GP_WXSTRING_PATH, &PovIsosurface_cls::msCMesh2Path.msBody,
		       wxT("Mesh2 File"), TRUE);
    pCurr->CombinePointer(_GP_BOOL, &PovIsosurface_cls::msCbMesh2Sequential.mbBody,
			  wxT("Sequential (%04d)"));
    
    int indexMaxtrace = pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &PovIsosurface_cls::msCMaxTrace.mBody, wxT("max_trace"));
    pCurr->SetRange(0, 128);

    mIndexAccuracy = pCurr->CombinePointer
	(_GP_FLOAT, &PovIsosurface_cls::msCAccuracy.mBody, _("Accuracy"));

    int boxWidth = 2;
    mIndexGradient = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("max_gradient"));
    
    wxString maxgradientType[2] =
	{wxT("static"), wxT("dynamic")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &PovIsosurface_cls::msCTypeGradient.mBody,
			  wxT("max_gradient"), TRUE, 2, maxgradientType);
    mIndexStatic = pCurr->CombinePointer
	(_GP_FLOAT, &PovIsosurface_cls::msCMaxGradient.mBody);
    
    boxWidth = 3;
    mIndexDynamic = pCurr->
	CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("dynamic"));
    pCurr->CombinePointer(_GP_FLOAT, &PovIsosurface_cls::msCEvaluate.mpBody[0]);
    pCurr->CombinePointer(_GP_FLOAT, &PovIsosurface_cls::msCEvaluate.mpBody[1]);
    pCurr->CombinePointer(_GP_FLOAT, &PovIsosurface_cls::msCEvaluate.mpBody[2]);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    //pCurr->mProportion = 1;
    pCurr->FinishSetting(pBoxSizer, FALSE);

    pCurr->SetFuncHook(HookHideAndShow, _MN_HOOK_COMBOBOX);

    pCurr->mVCPointers[indexMaxtrace].
	mpCTextLabel->SetToolTip(_("0: all_intersections"));
    pCurr->mVCPointers[mIndexAccuracy].
	mpCTextLabel->SetToolTip(_("Lower value means better and slower rendering"));
    pCurr->mVCPointers[mIndexGradient+1].
	GetPCtrlWindow()->SetToolTip(_("Tip max_gradient"));
    pCurr->mVCPointers[mIndexGradient+2].
	GetPCtrlWindow()->SetToolTip(_("Higher value means better and slower rendering"));

    pCurr->mVCPointers[mIndexDynamic].
	mpCTextLabel->SetToolTip(_("dynamic Tip"));
    pCurr->mVCPointers[mIndexDynamic+1].
	GetPCtrlWindow()->SetToolTip(_("evaluate param P0"));
    pCurr->mVCPointers[mIndexDynamic+2].
	GetPCtrlWindow()->SetToolTip(_("evaluate param P1"));
    pCurr->mVCPointers[mIndexDynamic+3].
	GetPCtrlWindow()->SetToolTip(_("evaluate param P2"));
}

void PovRayGUI_cls::
SetMNumFormat(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumFormat, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNFormat, btnPovFormatPref,
				    _("[+]Format setting"));
    
    int boxWidth = 2;
    pCurr->mTextWidth     = -1;
    pCurr->mComboWidth = 108;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Output Format"));

    wxString listTemplate[7] = {_("Custom"), wxT("1024x768"), wxT("800x600"), wxT("640x480"), wxT("480x360"), wxT("320x240"), wxT("240x180")};
    mIndexGeomTemplate = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mGeometryTemplate, wxT(""),
	 TRUE, 7, listTemplate);

    wxString listFormat[8] = {wxT("PNG(3x8bit)"), wxT("PNG(3x12bit)"), wxT("PNG(3x16bit)"), wxT("SYS(BMP/PICT)"), wxT("PPM"), wxT("JPEG (ver.3.7)"), wxT("EXR (ver.3.7)"), wxT("HDR( ver.3.7)") };
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, &mpCParent->mCOutputFormat.mBody, wxT(""), TRUE, 8, listFormat);
    
    mIndexOutputGeometry = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Width,Height"));
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCWidth .mBody);
    pCurr->SetRange(40, 8192);
    pCurr->CombinePointer(_GP_INTEGER_SPIN, &mpCParent->mCHeight.mBody);
    pCurr->SetRange(30, 8192);
    
    //pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Option"));
    boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    int indexAntalias = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbAntiAlias.mbBody,
					      wxT("Antialias(+A)"));
    int indexJitter = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbJitter.mbBody,
					    wxT("Jitter(+J)"));
    int indexDisplay = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbDisplay.mbBody,
					     wxT("Display(+D)"));

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    int indexPause = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbPause.mbBody,
					     wxT("Pause(+P)"));
    int indexPreview = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbPreview.mbBody,
			  wxT("Preview(+SP8)"));
    int indexAlpha = pCurr->CombinePointer(_GP_BOOL, &mpCParent->mCbPreview.mbBody,
			  wxT("Alpha(+UA)"));
#ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    //pCurr->mProportion = 1;
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->mVCPointers[indexAntalias].
	GetPCtrlWindow()->SetToolTip(_("Antialias to decrease jaggies"));
    pCurr->mVCPointers[indexJitter].
	GetPCtrlWindow()->SetToolTip(_("Add noise to decrease jaggies"));
    pCurr->mVCPointers[indexDisplay].
	GetPCtrlWindow()->SetToolTip(_("Show rendering image on the display"));
    pCurr->mVCPointers[indexPause].
	GetPCtrlWindow()->SetToolTip(_("Pause when the rendering is finished"));
    pCurr->mVCPointers[indexPreview].
	GetPCtrlWindow()->SetToolTip(_("Show low resolution test renders"));
    pCurr->mVCPointers[indexAlpha].
	GetPCtrlWindow()->SetToolTip(_("Sets alpha output on"));

    pCurr->SetFuncHook(&HookGeometryTemplate, _MN_HOOK_COMBOBOX);
}

void PovRayGUI_cls::
SetMNumTypeParam(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumTypeParam, this);

    int boxWidth = 1;
    pCurr->mTextWidth     = -1;
    pCurr->mComboWidth = 108;

    wxString listRegion[3] = {_("Render Working Area"), _("Render All Scene"), _("Custum")};
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Region to render"));
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			  &mpCParent->mCRegionType.mBody, 
			  _("Region to render"), TRUE, 3, listRegion);

    wxString pListCamera[_PATH_NUM_MAX + 5] = 
	{_("Follow Active Camera"), _("Same view to Preview0"), _("Same view to Preview1"), _("Same view to Preview2"), _("Same view to Preview3")};
    for (int i = 0; i < _PATH_NUM_MAX; i++)
	pListCamera[i + 5] << wxT("Cam") << i;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Camera View"));
    mMNumIndexCamSelect = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCCameraType.mBody,
 	 _("Camera View"), TRUE,
	 5 + mpCParent->mpCParentFrame->mpCPathCtrl->mCCameraNum.mBody, 
	 pListCamera);

    wxString listStereo[4] = {_("No Stereo"), _("LEFT"), _("RIGHT"), _("Render/Export Both")};
    //pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Stereogram"));
    mMNumIndexStereo = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCStereoEye.mBody, 
	 _("Stereogram"), TRUE, 4, listStereo);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);
}

void PovRayGUI_cls::
SetMNumCameraParamAdvanced(){
    mpCMNumCameraParamAdvanced = new MultiNumericalInputs_cls
	(&mpCMNumCameraParamAdvanced, this, wxID_ANY,
	 _("Advanced PovRay setting"), 
	 wxDefaultPosition, wxDefaultSize);
    
    MultiNumericalInputs_cls* pCurr = mpCMNumCameraParamAdvanced;

    pCurr->SetDefaultButtonType(_MN_OK_CANCEL_DEFAULT);

#ifdef WIN32
    pCurr->mTextWidth     = 120;
#else
    pCurr->mTextWidth     = 140;
#endif

    wxString strPovVersionList[2] = {
	wxT("PovRay 3.6"), wxT("PovRay 3.7")};
    mIndexVersion = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, 
	 &mpCParent->mCPovRayVersion.mBody,
	 _("PovRay Version"), TRUE, 2, strPovVersionList);
    int boxWidth = 2;
    mIndexGamma = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Gamma(Disp,File)"));
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCDisplayGamma.mBody);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCFileGamma.mBody);
    /*
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCAssumedGamma.mBody,
			  wxT("assumed_gamma"));
    */

    mIndexThread = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Thread,blocksize"));
    pCurr->CombinePointer(_GP_UINT_SPIN, &mpCParent->mCWorkThread.mBody);
    pCurr->SetRange(0, 16);
    pCurr->CombinePointer(_GP_UNSIGNED_INT, &mpCParent->mCRenderBlockSize.mBody);

    wxString strPovBoundingMethod[2] = {
	wxT("default"), wxT("Binary Partitioning")};
    mIndexPartition = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCBoundingMethod.mBody,
	 wxT("Bounding Method"), TRUE, 2, strPovBoundingMethod);

    wxString strPovBufferSize[4] = {
	wxT("32 MB"), wxT("128 MB"), wxT("512 MB"), wxT("Never swap to HDD")};
    mIndexBufferSize = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCBufferSize.mBody,
	 wxT("Max Image Buffer"), TRUE, 4, strPovBufferSize);
    
    mIndexSSSQuality = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("SSS Quality"));
    pCurr->CombinePointer(_GP_INTEGER, &mpCParent->mCSSSQuality1.mBody);
    pCurr->CombinePointer(_GP_INTEGER, &mpCParent->mCSSSQuality2.mBody);
    
    pCurr->CombinePointer(_GP_STATIC_LINE);

    int indexMaxTrace = pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &mpCParent->mCMaxTraceLevel.mBody,
	 wxT("max_trace_level"));
    pCurr->SetRange(2, 128);

    int indexAAThreshold = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("AA Threshold"));
    
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, &mpCParent->mCAntiAliasThreshold.mBody);
    pCurr->SetRange(0, 59, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, &mpCParent->mCAntiAliasThreshold.mBody);
    
    int indexAADepth = pCurr->CombinePointer
	(_GP_INTEGER_SPIN, &mpCParent->mCAntiAliasDepth.mBody, _("AA Depth"));
    pCurr->SetRange(2, 5);

    mIndexJitter = pCurr->CombinePointer
	(_GP_FLOAT, &mpCParent->mCJitterAmount.mBody, _("Jitter Amount"));

    wxString pListRemove[2] = {
	_("not remove .pov file"), _("remove .pov file")}; 
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST,
			  &mpCParent->mCRemoveFile.mBody,
			  _("When finished"), TRUE, 2, pListRemove);

    wxString pListTweaks[2] = {
	wxT("hide"), wxT("show")}; 
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST,
			  &mpCParent->mCShowTweaks.mBody,
			  _("Tweaks"), TRUE, 2, pListTweaks);
    
    pCurr->FinishSettingDLG(FALSE);

    /*
      pCurr->mVCPointers[0].
      mpCTextLabel->SetToolTip(_("PovRay Version Tip"));
    */
    pCurr->mVCPointers[mIndexGamma].
	mpCTextLabel->SetToolTip(_("Tip for Gamma(Disp,File)"));

    pCurr->mVCPointers[mIndexThread].
	mpCTextLabel->SetToolTip(_("thread number and render block size."));
    pCurr->mVCPointers[mIndexPartition].
	mpCTextLabel->SetToolTip(_("Argorithms for partition"));
    pCurr->mVCPointers[mIndexBufferSize].
	mpCTextLabel->SetToolTip(_("Max image buffer stored in memory."));
    
    pCurr->mVCPointers[indexMaxTrace].
	mpCTextLabel->SetToolTip(_("max_trace_level"));
    pCurr->mVCPointers[indexAAThreshold].
	mpCTextLabel->SetToolTip(_("AA Threshold Tip"));
    pCurr->mVCPointers[indexAADepth].
	mpCTextLabel->SetToolTip(_("AA Depth Tip"));

    pCurr->SetFuncHook(&PovRay_cls::EnableDisablePovrayVersion, _MN_HOOK_OPEN);
    pCurr->SetFuncHook(&PovRay_cls::HookPovrayVersion, _MN_HOOK_COMBOBOX);
}

void PovRayGUI_cls::
HideAndShowVolume(){
    MultiNumericalInputsPanel_cls *pCurr = mpCMNumVolume;

    bool bAA = (PovVolume_cls::msCMethod.mBody == 3);
    pCurr->ShowItems(mIndexVolumeAA, 3, bAA);

    bool bInterval = (PovVolume_cls::msCMethod.mBody != 3);
    pCurr->ShowItems(mIndexIntervals, 1, bInterval);

    pCurr->GetSizer()->Layout();    
    mspCThis->ResizeDialog();
}

void PovRayGUI_cls::
HideAndShowIsosurf(){
    MultiNumericalInputsPanel_cls *pCurr = mpCMNumIsosurf;
    bool bShow = (PovIsosurface_cls::msCOutputType.mBody == _ISO_LINK_TO_DF3);
    pCurr->ShowItems(mIndexAccuracy, 1, bShow);
    pCurr->ShowItems(mIndexGradient, 3, bShow);
    
    bool bDynamic = (PovIsosurface_cls::msCTypeGradient.mBody == 1);
    pCurr->ShowItems(mIndexDynamic, 4, bShow & bDynamic);
    pCurr->mVCPointers[mIndexStatic].Enable(!bDynamic);

    bShow = (PovIsosurface_cls::msCOutputType.mBody == _ISO_LINK_TO_POV);
    pCurr->ShowItems(mIndexIsosurfacePath, 2, bShow);

    mspCThis->ResizeDialog();
}

void PovRayGUI_cls::
HideAndShowFormat(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumFormat;
    int indexGeom = mIndexOutputGeometry;
    int geometry = pCurr->mVCPointers[mIndexGeomTemplate].GetValueInt();
    bool bCustom = (geometry == 0);
    switch(geometry){
    case 1:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 1024; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 768; 
	break;
    case 2:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 800; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 600; 
	break;
    case 3:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 640; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 480; 
	break;
    case 4:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 480; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 360; 
	break;
    case 5:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 320; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 240; 
	break;
    case 6:
	*(int*)pCurr->mVCPointers[indexGeom+1].mpPointerVal = 240; 
	*(int*)pCurr->mVCPointers[indexGeom+2].mpPointerVal = 180; 
	break;
    }
    pCurr->ShowItems(indexGeom, 3, bCustom);
    pCurr->UpdateValues();

    ResizeDialog();
}

void PovRayGUI_cls::
HookGeometryTemplate(void *pArg, void *pEvent){
    mspCThis->HideAndShowFormat();
}

void PovRayGUI_cls::
HookHideAndShow(void *pArg, void *pEvent){
    mspCThis->HideAndShowVolume();
    mspCThis->HideAndShowIsosurf();
    mspCThis->Refresh(FALSE);
}

void PovRayGUI_cls::
UpdateDisplay(){
    mpCParent->mCConfWorkingDir.SetLabels();
    mpCParent->mCConfPathPovray.SetLabels();
}

void PovRayGUI_cls::
OnMouse(wxMouseEvent &event){
    if (event.RightDown() == TRUE){
	MACRO_PopupMinimizeButton(this,Volume);
	MACRO_PopupMinimizeButton(this,Isosurf);
	MACRO_PopupMinimizeButton(this,Format);
	MACRO_PopupMinimizeButton(mpCParent->mpCRadiosity, Radiosity);
	MACRO_PopupMinimizeButton(mpCParent->mpCPhotons, Photons);
	MACRO_PopupMinimizeButton(mpCParent->mpCFocalBlur, FocalBlur);
	MACRO_PopupMinimizeButton(mpCParent->mpCBG, BG);
    }
    event.Skip();
}

void PovRayGUI_cls::
OnMenuMinimize(wxCommandEvent &event){
    MACRO_CallMinimizeButton2(this,Volume,_("[+]Volume Render Settings"));
    MACRO_CallMinimizeButton2(this,Isosurf,_("[+]Isosurface Render Settings"));
    MACRO_CallMinimizeButton2(this,Format,_("[+]Format Settings"));
    MACRO_CallMinimizeButton2
	(mpCParent->mpCRadiosity, Radiosity, _("[+]Radiosity Settings"));
    MACRO_CallMinimizeButton2
	(mpCParent->mpCPhotons, Photons, _("[+]Photons Settings"));
    MACRO_CallMinimizeButton2
	(mpCParent->mpCFocalBlur, FocalBlur, _("[+]Focal Blur Settings"));
    MACRO_CallMinimizeButton2
	(mpCParent->mpCBG, BG, _("[+]BG Settings"));
    ResizeDialog();
    Refresh(FALSE);
}

#define MACRO_UpdateBTNMinimize(BTNName, LABEL, pCMNum)	\
    if (mpCParent->mCConfBTN##BTNName.mbBody == TRUE){\
    mpCParent->mCConfBTN##BTNName.mbBody = FALSE;\
    MinimizeButton(mpCBTN##BTNName, \
		   &mpCParent->mCConfBTN##BTNName.mbBody, LABEL,\
		   (wxButtonEX*)pCMNum->mVCPointers[0].mpPointerCtrl); }

#define MACRO_UpdateBTNMinimize2(BTNName, LABEL, pCMNum)	\
    if (mpCParent->mCConfBTN##BTNName.mbBody == TRUE){\
    mpCParent->mCConfBTN##BTNName.mbBody = FALSE;\
    MinimizeButton(mpCParent->mpC##BTNName->mpCBTN##BTNName, \
		   &mpCParent->mCConfBTN##BTNName.mbBody, LABEL,\
		   (wxButtonEX*)pCMNum->mVCPointers[0].mpPointerCtrl); }

void PovRayGUI_cls::
UpdateButtonMinimize(){
    MACRO_UpdateBTNMinimize
	(Volume, _("[+]Volume Render Settings"), mpCMNumVolume);
    MACRO_UpdateBTNMinimize
	(Isosurf, _("[+]Isosurf Render Settings"), mpCMNumIsosurf);
    MACRO_UpdateBTNMinimize
	(Format, _("[+]Format Settings"), mpCMNumFormat);
    MACRO_UpdateBTNMinimize2
	(Radiosity, _("[+]Radiosity Settings"),
	 mpCParent->mpCRadiosity->mpCMNumRadiosity);
    MACRO_UpdateBTNMinimize2
	(Photons, _("[+]Photons Settings"),
	 mpCParent->mpCPhotons->mpCMNumPhotons);
    MACRO_UpdateBTNMinimize2
	(FocalBlur, _("[+]Focal Blur Settings"),
	 mpCParent->mpCFocalBlur->mpCMNumFocalBlur);
    MACRO_UpdateBTNMinimize2
	(BG, _("[+]BG Settings"), mpCParent->mpCBG->mpCMNumBG);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonWorkingDir(wxCommandEvent &event){
    wxDirDialog *pDialog = new wxDirDialog(this, wxT("Choose Working Directory"));
    pDialog->SetPath(mpCParent->mCConfWorkingDir.msBody);
    if (pDialog->ShowModal() == wxID_OK)
	mpCParent->mCConfWorkingDir.msBody = pDialog->GetPath();
    delete pDialog;
    CheckFileExistence();
    UpdateDisplay();
}

void PovRayGUI_cls::
OnButtonPathPovray(wxCommandEvent &event){
#ifdef WIN32
    GpCFileDialog->SetMessage(_("Choose pvengine.exe or pvengine-sse2.exe file"));
#else
    GpCFileDialog->SetMessage(_("Choose povray execution file"));
#endif
    SplitPathWX_cls CPath(mpCParent->mCConfPathPovray.msBody);
    GpCFileDialog->SetInitPath(CPath.GetPathWXStr(_FN_DRIVE_DIR));
    GpCFileDialog->SetWildcard(POVFILE);
    GpCFileDialog->SetStyle(wxFD_FILE_MUST_EXIST);
    
    if (GpCFileDialog->ShowModal() == wxID_OK)
	mpCParent->mCConfPathPovray.msBody = GpCFileDialog->GetPath();
    CheckFileExistence();
    UpdateDisplay();
}

bool PovRayGUI_cls::
CheckConsistency(int type){
    wxString strWarning;
    if (PovVolume_cls::msCSampleMin.mBody > PovVolume_cls::msCSampleMax.mBody){
	PovVolume_cls::msCSampleMin.mBody = PovVolume_cls::msCSampleMax.mBody;
	strWarning << wxT("!! Detect SampleMin > SampleMax") << ENDLWX;
	strWarning << wxT("Setting SampleMin to ") << PovVolume_cls::msCSampleMax.mBody << wxT(".") << ENDLWX;
    }

    if (CheckFileExistence(type) == FALSE)
	strWarning << _("!! File path discrepancy. Please Check.") << ENDLWX;

    if (mpCParent->mpCRadiosity->IsRadiosityFormatOK() == FALSE)
	strWarning << _("!! Radiosity temporary file is not accessible.") << ENDLWX;
    
    if (strWarning.Len() > 0){
	mpCMNumVolume->UpdateValues();
	strWarning << _("Above warnings detected. Continue?") << ENDLWX;
	return (MsgBoxYesno(this, wxT("Warning"), strWarning) == TRUE);
    }

    return TRUE;
}

void PovRayGUI_cls::
OnButtonExportSnap(wxCommandEvent &event){
    Assign();
    
    if (CheckConsistency(_POV_EXPORT) == FALSE) return;

    mpCParent->CoupleToStereogram();
    UpdateComboboxList();    
    mpCMNumFormat->UpdateValues();
    mpCMNumTypeParam->UpdateValues();
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemPovSnap);
    mpCParent->mpCParentFrame->AddPendingEvent(cmd);
}

void PovRayGUI_cls::
OnButtonExportAnim(wxCommandEvent &event){
    if (mpCParent->mbExportingAnim == TRUE){
	mpCParent->mbExportingAnim = FALSE;
	return;
    }
    
    Assign();
    if (CheckConsistency(_POV_EXPORT) == FALSE) return;

    mpCParent->CoupleToStereogram();
    UpdateComboboxList();
    mpCMNumFormat->UpdateValues();
    mpCMNumTypeParam->UpdateValues();
    
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemPovAnim);
    mpCParent->mpCParentFrame->AddPendingEvent(cmd);
}

void PovRayGUI_cls::
OnButtonRenderSnap(wxCommandEvent &event){
    Assign();
    if (CheckConsistency() == FALSE) return;
    
    mpCParent->CoupleToStereogram();
    UpdateComboboxList();
    mpCMNumFormat->UpdateValues();
    mpCMNumTypeParam->UpdateValues();
    
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemRenderSnap);
    mpCParent->mpCParentFrame->AddPendingEvent(cmd);
}

void PovRayGUI_cls::
OnButtonRenderAnim(wxCommandEvent &event){
    Assign();
    if (CheckConsistency() == FALSE) return;
    
    mpCParent->CoupleToStereogram();
    UpdateComboboxList();
    mpCMNumFormat->UpdateValues();
    mpCMNumTypeParam->UpdateValues();
	
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemRenderAnim);
    mpCParent->mpCParentFrame->AddPendingEvent(cmd);
}

void PovRayGUI_cls::
OnButtonDefault(wxCommandEvent &event){
    if (MsgBoxYesno(this, _("Query"), _("Set to default ?")) == FALSE) return;
    mpCParent->SetToDefault();
    mpCParent->UpdateItems();
}

void PovRayGUI_cls::
OnButtonAdvanced(wxCommandEvent &event){
    wxCommandEvent cmd(wxEVT_COMMAND_MENU_SELECTED, menuItemConfPovRayAdvanced);
    mpCParent->mpCParentFrame->AddPendingEvent(cmd);
}

void PovRayGUI_cls::
OnButtonShowItemVolume(wxCommandEvent &event){
    mpCMNumVolume->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE); // Panelƕ`悪ꍇ邽
}

void PovRayGUI_cls::
OnButtonShowItemIsosurf(wxCommandEvent &event){
    mpCMNumIsosurf->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonShowItemRadios(wxCommandEvent &event){
    mspCThis->mpCParent->mpCRadiosity->mpCMNumRadiosity->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonShowItemPhotons(wxCommandEvent &event){
    mspCThis->mpCParent->mpCPhotons->mpCMNumPhotons->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonShowItemFocalBlur(wxCommandEvent &event){
    mspCThis->mpCParent->mpCFocalBlur->mpCMNumFocalBlur->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonShowItemBG(wxCommandEvent &event){
    mspCThis->mpCParent->mpCBG->mpCMNumBG->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonShowItemFormat(wxCommandEvent &event){
    mpCMNumFormat->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void PovRayGUI_cls::
OnButtonBGColor0(wxCommandEvent &event){
    GpCClrDialog->SetColour(mpCParent->mpCBG->GetColourBG(0));
    if (GpCClrDialog->ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGColor(GpCClrDialog->GetPColorFloat(), 0);
    /*
    wxColourDialogEX CDialog
	(this, mpCParent->mpCBG->GetColourBG(0), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGColor(CDialog.GetPColorFloat(), 0);
    */
    mpCParent->mpCBG->mpCMNumBG->UpdateValues();
}

void PovRayGUI_cls::
OnButtonBGColor1(wxCommandEvent &event){
    GpCClrDialog->SetColour(mpCParent->mpCBG->GetColourBG(1));
    if (GpCClrDialog->ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGColor(GpCClrDialog->GetPColorFloat(), 1);
    /*
    wxColourDialogEX CDialog
	(this, mpCParent->mpCBG->GetColourBG(1), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGColor(CDialog.GetPColorFloat(), 1);
    */
    mpCParent->mpCBG->mpCMNumBG->UpdateValues();
}

void PovRayGUI_cls::
OnButtonPluginHelp(wxCommandEvent &event){
    wxString strMessage;

    strMessage << _("Description of Galaxy.inc");

    MsgBox(this, strMessage, _("Plugin description"));
}

void PovRayGUI_cls::
OnButtonBGNColor1(wxCommandEvent &event){
    wxColourDialogEX CDialog
	(this, mpCParent->mpCBG->GetColourBGN(1), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGNColor(CDialog.GetPColorFloat(), 1);
    mpCParent->mpCBG->mpCMNumBG->UpdateValues();
}

void PovRayGUI_cls::
OnButtonBGNColor2(wxCommandEvent &event){
    wxColourDialogEX CDialog
	(this, mpCParent->mpCBG->GetColourBGN(2), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGNColor(CDialog.GetPColorFloat(), 2);
    mpCParent->mpCBG->mpCMNumBG->UpdateValues();
}

void PovRayGUI_cls::
OnButtonBGNColor3(wxCommandEvent &event){
    wxColourDialogEX CDialog
	(this, mpCParent->mpCBG->GetColourBGN(3), TRUE);
    if (CDialog.ShowModal() == wxID_OK)
	mpCParent->mpCBG->ChangeBGNColor(CDialog.GetPColorFloat(), 3);
    mpCParent->mpCBG->mpCMNumBG->UpdateValues();
}

void PovRayGUI_cls::
Assign(){
    wxCommandEvent dummy;
    mpCMNumVolume->OnAssign(dummy);

    mpCParent->mCConfPathPovray.OnAssign(dummy);
    mpCParent->mCConfWorkingDir.OnAssign(dummy);
    mpCParent->mCConfSnapFile.OnAssign(dummy);
    mpCParent->mCConfAnimFile.OnAssign(dummy);
    mpCParent->mpCSnapNumber->UpdateValueFromText();
    mpCParent->mSnapNumber = mpCParent->mpCSnapNumber->GetValue();
    mpCParent->CheckIllegalFileName();

    CheckFileExistence();
    UpdateDisplay();
}

// {^͂ȂƂApXύX̍ۂɌ̂ɗp
void PovRayGUI_cls::
OnButtonApply(wxCommandEvent &event){
    Assign();
}

void PovRayGUI_cls::
OnButtonClose(wxCommandEvent &event){
    mpCParent->UpdateItems();
    if (mpCParent->mbExportingAnim == TRUE) return;
    CloseEXKey();
}

bool PovRayGUI_cls::
CheckFileExistence(int type){
    bool bRet = TRUE;

    wxFileName CPovRay(mpCParent->mCConfPathPovray.msBody);
    if (CPovRay.IsFileExecutable() == TRUE){
	mpCCheckPovray->SetBackgroundColour(*wxGREEN);
	mpCCheckPovray->SetToolTip(_("green is valid."));
    }
    else{
	mpCCheckPovray->SetBackgroundColour(*wxRED);
	mpCCheckPovray->SetToolTip(_("red if not valid."));
	if (type == _POV_RENDER) // EXPORT ONLY̎FALSEłȂėǂ
	    bRet = FALSE;
    }

    wxFileName CWorkDir(mpCParent->mCConfWorkingDir.msBody + SEPARATORWX);

    if (CWorkDir.DirExists() == TRUE){
	mpCCheckWorkDir->SetBackgroundColour(*wxGREEN);
	mpCCheckWorkDir->SetToolTip(_("green is valid."));
    }
    else{
	mpCCheckWorkDir->SetBackgroundColour(*wxRED);
	mpCCheckWorkDir->SetToolTip(_("red if not valid."));
	bRet = FALSE;
    }

    mpCCheckPovray->Refresh(TRUE);
    mpCCheckWorkDir->Refresh(TRUE);

    return bRet;
}

PovRayGUI_cls* PovRayGUI_cls::mspCThis;

// -------------------------------------------

PovRay_cls::
PovRay_cls(FrameMain_cls *pC, DF3SeqCtrl_cls *pCDF3SeqCtrl,
	   wxBitmap *pBitmap) 
{
    mspCThis = this;
    mpCParentFrame = pC;
    mspCDF3SeqCtrlLink = pCDF3SeqCtrl;
    mpBitmapLink = pBitmap;
    //mspCFilterCtrlLink = pCFilterCtrl;
    mpCGUI  = NULL;

    for (int i = 0; i < _POVRAY_MAX_THREAD; i++){
	mpCThreadPovRay[i] = NULL;
	mbPovRayThreadActive[i] = FALSE;
    }

    mbRenderingAnim = FALSE;
    mbExportingAnim = FALSE;
    mbRenderingTwoSnap = FALSE;
    mCurrentAnim = 0;
    mpCThreadPovRayAnim = NULL;
    mpCRadiosity = new PovRadiosity_cls(this);
    mpCPhotons = new PovPhotons_cls;
    mpCFocalBlur = new PovFocalBlur_cls;
    mpCBG = new PovBG_cls;
    
    SetToDefault();
    InitGUI(pC); // GUI ̃CX^X

    HideAndShow();

    msAFormat[_POV_PNG_8BIT]  = L"N";
    msAFormat[_POV_PNG_12BIT] = L"N";
    msAFormat[_POV_PNG_16BIT] = L"N";
    msAFormat[_POV_SYSTEM]    = L"S";
    msAFormat[_POV_PPM]       = L"P";
    msAFormat[_POV_JPG]       = L"J";
    msAFormat[_POV_EXR]       = L"E";
    msAFormat[_POV_HDR]       = L"H";
}

PovRay_cls::
~PovRay_cls(){
    if (mpCGUI != NULL) delete mpCGUI;
    mpCGUI = NULL;

    delete mpCRadiosity;
    delete mpCPhotons;
    delete mpCFocalBlur;
    delete mpCBG;
    
    for (int i = 0; i < _POVRAY_MAX_THREAD; i++)
	if (mbPovRayThreadActive[i] == TRUE){
	    mpCThreadPovRay[i]->Wait();
	    delete mpCThreadPovRay[i]; }

    if (mpCThreadPovRayAnim != NULL){
	mpCThreadPovRayAnim->Wait();
	delete mpCThreadPovRayAnim;
    }
}

// --

void PovRay_cls::
HideAndShow(){
    mpCGUI->HideAndShowVolume();
    mpCGUI->HideAndShowIsosurf();
    mpCGUI->HideAndShowFormat();
    mpCRadiosity->HideAndShow();
    mpCPhotons->HideAndShow();
    mpCFocalBlur->HideAndShow();
    mpCBG->HideAndShow();

    if (mpCParentFrame->mpCStereo->mCType.mBody == _STEREO_NONE)
	//pCurr->mVCPointers[mMNumIndexStereo].GetPCtrlWindow()->Hide();
	mpCGUI->mpCMNumTypeParam->ShowItems(mpCGUI->mMNumIndexStereo, 1, FALSE);
    else{
	/*
	mpCParent->mCStereoEye.mBody = 
	    mpCParentFrame->mpCStereo->mCEye.mBody;
	*/
	//pCurr->mVCPointers[mMNumIndexStereo].GetPCtrlWindow()->Show();
	mpCGUI->mpCMNumTypeParam->ShowItems(mpCGUI->mMNumIndexStereo, 1, TRUE);
    }

    mspCThis->mpCGUI->ResizeDialog();
    mspCThis->mpCGUI->Refresh(FALSE);
}

void PovRay_cls::
HookRadiocity(void *pArg, void *pEvent){
    MultiNumericalInputsPanel_cls* pCPanel = (MultiNumericalInputsPanel_cls*)pArg;
    int eventId = ((wxCommandEvent*)pEvent)->GetId();
    int index0 = mspCThis->mpCRadiosity->mIndexType;
    int index1 = mspCThis->mpCRadiosity->mIndexReuse;
    if ( eventId != pCPanel->mVCPointers[index0].mID &&
	 eventId != pCPanel->mVCPointers[index1].mID)
	return;
    
    mspCThis->mpCRadiosity->HideAndShow();
    mspCThis->mpCGUI->ResizeDialog();
    mspCThis->mpCGUI->Refresh(FALSE);
}

void PovRay_cls::
HookPhotons(void *pArg, void *pEvent){
    MultiNumericalInputsPanel_cls* pCPanel = (MultiNumericalInputsPanel_cls*)pArg;
    int index = mspCThis->mpCPhotons->mIndexType;
    if (((wxCommandEvent*)pEvent)->GetId() != pCPanel->mVCPointers[index].mID)
	return;
    
    mspCThis->mpCPhotons->HideAndShow();
    mspCThis->mpCGUI->ResizeDialog();
    mspCThis->mpCGUI->Refresh(FALSE);
}

void PovRay_cls::
HookFocalBlur(void *pArg, void *pEvent){
    MultiNumericalInputsPanel_cls* pCPanel = (MultiNumericalInputsPanel_cls*)pArg;
    int index1 = mspCThis->mpCFocalBlur->mIndexType;
    int index2 = mspCThis->mpCFocalBlur->mIndexFocusLocationType;
    if (((wxCommandEvent*)pEvent)->GetId()!=pCPanel->mVCPointers[index1].mID &&
	((wxCommandEvent*)pEvent)->GetId()!=pCPanel->mVCPointers[index2].mID)
	return;
    
    mspCThis->mpCFocalBlur->HideAndShow();
    mspCThis->mpCGUI->ResizeDialog();
    mspCThis->mpCGUI->Refresh(FALSE);
}

void PovRay_cls::
HookBG(void *pArg, void *pEvent){
    MultiNumericalInputsPanel_cls* pCPanel = (MultiNumericalInputsPanel_cls*)pArg;

    int id = ((wxCommandEvent*)pEvent)->GetId();
    int indexSky = mspCThis->mpCBG->mIndexSkySphereType;
    int indexBG = mspCThis->mpCBG->mIndexBGType;
    int indexNebula = mspCThis->mpCBG->mIndexNebulaType;
    if (id != pCPanel->mVCPointers[indexSky].mID &&
	id != pCPanel->mVCPointers[indexBG].mID &&
	id != pCPanel->mVCPointers[indexNebula].mID) return;

    mspCThis->mpCBG->HideAndShow();
    if (id != pCPanel->mVCPointers[indexSky].mID){
	mspCThis->mpCGUI->ResizeDialog();
	mspCThis->mpCGUI->Refresh(FALSE); }
}

void PovRay_cls::
HookPovrayVersion(void *pArg, void *pEvent){
    MultiNumericalInputs_cls* pCDLG = (MultiNumericalInputs_cls*)pArg;
    if (((wxCommandEvent*)pEvent)->GetId() != pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexVersion].mID)
	return;
    EnableDisablePovrayVersion(pArg, NULL);
}

void PovRay_cls::
EnableDisablePovrayVersion(void* pArg, void* pDummy){
    MultiNumericalInputs_cls* pCDLG = (MultiNumericalInputs_cls*)pArg;
    /*
    int version = pCDLG->mVCPointers[pCDLG->mIndexVersion].GetValueInt();
    */
    int version = mspCThis->mCPovRayVersion.mBody;
    bool bShow = (version == 1);
    for (int i = 0; i < 3; i++){
	pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexGamma+i].Enable(bShow);
	pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexThread+i].Enable(bShow);
	pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexSSSQuality+i].Enable(bShow);
    }
    pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexPartition].Enable(bShow);
    pCDLG->mVCPointers[mspCThis->mpCGUI->mIndexBufferSize].Enable(bShow);
}

// --

void PovRay_cls::
CheckIllegalFileName(){

#ifdef WIN32
    wchar_t buf[64];
    int check = SNWPRINTF(buf, 64, mCConfSnapFile.msBody.c_str(), mSnapNumber);
#else
    char buf[64];
    wxString wxSnap = wxConvFile.cWC2WX(mCConfSnapFile.msBody.c_str());
    int check = SNWPRINTF(buf, 64, wxSnap.ToAscii(), mSnapNumber);
#endif
    if ( check >= 63 || check < 0){
	MsgBox(mpCGUI, "Warning", "Illegal File Format. Reset to default");
	mCConfSnapFile.msBody = L"povsnap%04d";
	mCConfSnapFile.SetLabels();
    }
#ifdef WIN32
    check = SNWPRINTF(buf, 64, mCConfAnimFile.msBody.c_str(), mSnapNumber);
#else
    wxSnap = wxConvFile.cWC2WX(mCConfAnimFile.msBody.c_str());
    check = SNWPRINTF(buf, 64, wxSnap.ToAscii(), mSnapNumber);
#endif
    if ( check >= 63 || check < 0){
	MsgBox(mpCGUI, "Warning", "Illegal File Format. Reset to default");
	mCConfAnimFile.msBody = L"povanim%04d";
	mCConfAnimFile.SetLabels();
    }
}

void PovRay_cls::
SetToDefault(){
    mSnapNumber = 0;
    bool bSuccess = FALSE;
#ifdef WIN32
    wchar_t buf[512];
    if (GetCurrentDirectory(512, buf) != 0) bSuccess = TRUE;
    if (bSuccess == TRUE) mCConfWorkingDir.msBody = std::wstring(buf) + SEPARATOR + L"POV";
#else
    char buf[512];
    if (getcwd (buf, 512) != NULL) bSuccess = TRUE;
    wchar_t buf2[512];
    wxConvFile.ToWChar(buf2, 512, buf, 512);
    if (bSuccess == TRUE) mCConfWorkingDir.msBody = std::wstring(buf2) + SEPARATOR + L"POV";
#endif

    mGeometryTemplate = 3;
    mCConfPathPovray.msBody = POVDEFWX;

    mCConfListDisp.Clear();
    mCConfListDisp.Add(&mCMaxTraceLevel,wxT("PovMaxTraceLevel="), L"12");
    mCConfListDisp.Add(&mCPovRayVersion, wxT("PovVersion="), L"1");
    mCConfListDisp.Add(&mCFileGamma,    wxT("PovFileGamma="), L"1");
    mCConfListDisp.Add(&mCDisplayGamma, wxT("PovDisplayGamma="), L"1");
    mCConfListDisp.Add(&mCAssumedGamma, wxT("PovAssumedGamma="), L"2.2");
    mCConfListDisp.Add(&mCWorkThread,   wxT("PovWorkThread="), L"0");
    mCConfListDisp.Add(&mCRenderBlockSize, wxT("PovRenderBlockSize="), L"32");
    mCConfListDisp.Add(&mCBoundingMethod, wxT("PovBoundingMethod="), L"0");
    mCConfListDisp.Add(&mCBufferSize, wxT("PovBufferSize="), L"1");
    mCConfListDisp.Add(&mCSSSQuality1, wxT("PovSSSQuality1="), L"50");
    mCConfListDisp.Add(&mCSSSQuality2, wxT("PovSSSQuality2"), L"50");
    
    mCConfListDisp.Add(&mCWidth,        wxT("PovCameraWidth="), L"640");
    mCConfListDisp.Add(&mCHeight,       wxT("PovCameraHeight="), L"480");
    mCConfListDisp.Add(&mCbAntiAlias,   wxT("PovAntialias="), L"0");
    mCConfListDisp.Add(&mCStereoEye,    wxT("PovStereoEye="), L"0");
    mCConfListDisp.Add(&mCAntiAliasThreshold, wxT("PovAntiAliasThreshold="), L"0.3");
    mCConfListDisp.Add(&mCAntiAliasDepth, wxT("PovAntiAliasDepth="), L"3");
    mCConfListDisp.Add(&mCRemoveFile, wxT("PovRemoveFile="), L"0");
    mCConfListDisp.Add(&mCShowTweaks, wxT("PovShowTweaks="), L"0");

    mCConfListDisp.Add(&mCbDisplay,     wxT("PovDisplay="), L"1");
    mCConfListDisp.Add(&mCbPause,       wxT("PovPauseWhenDone="), L"0");
    mCConfListDisp.Add(&mCbPreview,     wxT("PovPreview="), L"0");
    mCConfListDisp.Add(&mCbJitter,      wxT("PovJitter="), L"0");
    mCConfListDisp.Add(&mCJitterAmount, wxT("PovJitterAmount="), L"1");
    mCConfListDisp.Add(&mCbOutputAlpha,  wxT("PovOutputAlpha="), L"0");
    mCConfListDisp.Add(&mCbSubsetRow,   wxT("PovSubsetRow="), L"0");
    mCConfListDisp.Add(&mCSubsetRowStart, wxT("PovStartRow="), L"1");
    mCConfListDisp.Add(&mCSubsetRowEnd,   wxT("PovEnRow="), L"480");
    mCConfListDisp.Add(&mCRegionType,   wxT("PovRenderRegion="), L"0");
    mCConfListDisp.Add(&mCCameraType,   wxT("PovRenderCamera="), L"0");
    mCConfListDisp.Add(&mCOutputFormat, wxT("PovOutputFomat="), L"0");
        
    mCConfListLayout.Clear();
    mCConfListLayout.Add(&mCConfBTNVolume, wxT("PovMinBTNVolume="), L"0");
    mCConfListLayout.Add(&mCConfBTNIsosurf, wxT("PovMinBTNIsosurf="), L"0");
    mCConfListLayout.Add(&mCConfBTNFormat, wxT("PovMinBTNFormat="), L"0");
    mCConfListLayout.Add(&mCConfBTNRadiosity, wxT("PovMinBTNRadiosity="), L"0");
    mCConfListLayout.Add(&mCConfBTNPhotons, wxT("PovMinBTNPhotons="), L"0");
    mCConfListLayout.Add(&mCConfBTNFocalBlur, wxT("PovMinBTNFocalBlur="), L"0");
    mCConfListLayout.Add(&mCConfBTNBG, wxT("PovMinBTNBG="), L"0");

    PovVolume_cls::msCInterpolate.mBody = 1;
    PovVolume_cls::msCMethod.mBody = 3;
    PovVolume_cls::msCInterval.mBody = 1;
    PovVolume_cls::msCSampleMin.mBody = 64;
    PovVolume_cls::msCSampleMax.mBody = 128;
    PovVolume_cls::msCAALevel.mBody = 4;
    PovVolume_cls::msCAAThreshould.mBody = 0.05;
            
    mCConfListFog.Clear();
    
    mCbConfAbsolutePath.SetItem(wxT("ABSOLUTE_PATH="), L"1");
    mCbWinePov         .SetItem(wxT("PovUseWine="), L"0");

    mpCRadiosity->SetToDefault();
    mpCPhotons->SetToDefault();
    mpCFocalBlur->SetToDefault();
    mpCBG->SetToDefault();
}

void PovRay_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new PovRayGUI_cls(this, pC,
			       mCScreenRect.GetPosition(),
			       mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void PovRay_cls::
UpdateItems(bool bForce){
    if (mpCGUI == NULL) return;
    mpCGUI->mpCMNumVolume->UpdateValuesAll(bForce);
    mpCBG->mpCMNumBG->UpdateValuesAll(bForce);
    //mpCMNumCameraParam->UpdateValues();
    //mpCMNumFogParam->UpdateValues();
    mCConfPathPovray.UpdateValue();
    mCConfWorkingDir.UpdateValue();
    mpCGUI->CheckFileExistence();
    mpCGUI->UpdateDisplay();
    mpCSnapNumber->SetValueInt(mSnapNumber);
}

int PovRay_cls::
GetStereoId(int index){
    // QĂяoāA-1AXLbv
    // XeIA͕ЖڂȂQڂ-1
    // XeIo͂ȂA 1,2 ԂɕԂ
    static int stable[4][2] = {
	{0, -1}, {1, -1}, {2, -1}, {1, 2} };
    return stable[mCStereoEye.mBody][index];
}

int PovRay_cls::
GetStereoTmpFileIOMode(int index){
    // XeIŋLꎞt@C݂ƍ폜ǂ邩
    // 1:  2:폜 (rbgZ)
    static int stable[4][2] = {
	{3, 0}, {3, 0}, {3, 0}, {1, 2} };
    return stable[mCStereoEye.mBody][index];
}

wxString PovRay_cls::
GetFileStr(unsigned int type, unsigned int stereoId){
    wxString retstr = (type == _POV_ANIM) ?
	mCConfAnimFile.msBody : mCConfSnapFile.msBody;

    if (stereoId == _STEREO_LFT)
	retstr.Replace(wxT("%"), wxT("L%"));
    if (stereoId == _STEREO_RGT)
	retstr.Replace(wxT("%"), wxT("R%"));
    return retstr;
}

wxString PovRay_cls::
GetWorkingDirWithSep(bool bForceAbs){
    if (bForceAbs == TRUE || mCbConfAbsolutePath.mbBody == TRUE)
	return mCConfWorkingDir.msBody + SEPARATORWX;
    return wxT("./");
}

int PovRay_cls::
GetCameraGeometry(){
    if (mCCameraType.mBody == 0)
	return mpCParentFrame->mpCPathCtrl->GetActivePath()->mCCamGeometry.mBody;
    if (mCCameraType.mBody <= 4){
	int viewIndex = mCCameraType.mBody - 1;
	return mpCParentFrame->mpCCanvas[viewIndex]->mpCGLCanvas->mCGLGeometry.mBody;
    }
    
    return mpCParentFrame->mpCPathCtrl->mVpCPath[mCCameraType.mBody - 5]->mCCamGeometry.mBody;
}

int PovRay_cls::
GetCameraIndex(){
    if (mCCameraType.mBody == 0) return -1;
    if (mCCameraType.mBody < 5) return -2;
    return mCCameraType.mBody - 5;
}

double PovRay_cls::
GetVAngle(){
    if (mCCameraType.mBody == 0)
	return mpCParentFrame->mpCPathCtrl->GetActivePath()->mCVAngle.mBody;
    if (mCCameraType.mBody <= 4){
	int viewIndex = mCCameraType.mBody - 1;
	return mpCParentFrame->mpCCanvas[viewIndex]->mpCGLCanvas->mCGLPerspective.mDeg;
    }
    
    return mpCParentFrame->mpCPathCtrl->
	mVpCPath[mCCameraType.mBody - 5]->mCVAngle.mBody;
}

void PovRay_cls::
CoupleToStereogram(){
    mpCParentFrame->mpCStereo->mCEye.mBody = mCStereoEye.mBody;
    mpCParentFrame->mpCStereo->UpdateItems();
}

void PovRay_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    mCConfPathPovray.ReadStrLine(strLine);
    mCConfWorkingDir.ReadStrLine(strLine);
    mCConfSnapFile.ReadStrLine(strLine);
    mCConfAnimFile.ReadStrLine(strLine);
    mCConfListDisp.ReadStrLineAll(strLine);
    mCConfListLayout.ReadStrLineAll(strLine);
    mCConfListFog.ReadStrLineAll(strLine);
    PovIsosurface_cls::ReadStrLineStaticConf(strLine);
    PovVolume_cls::ReadStrLineStaticConf(strLine);
    
    mpCRadiosity->ReadStrLine(strLine);
    mpCPhotons->ReadStrLine(strLine);
    mpCFocalBlur->ReadStrLine(strLine);
    mpCBG->ReadStrLine(strLine);

    int width = mCWidth.mBody, height = mCHeight.mBody;
    if (width == 1024 && height == 768) mGeometryTemplate = 1;
    if (width == 800  && height == 600) mGeometryTemplate = 2;
    if (width == 640  && height == 480) mGeometryTemplate = 3;
    if (width == 480  && height == 360) mGeometryTemplate = 4;
    if (width == 360  && height == 240) mGeometryTemplate = 5;
    if (width == 240  && height == 180) mGeometryTemplate = 6;
}

wxString PovRay_cls::
SerializeConf(bool flag){
    if (flag == _FM_DEFAULT_CONF)
	return
	    mCConfWorkingDir.Serialize() +
	    mCConfPathPovray.Serialize() +
	    mCConfListLayout.SerializeAll();

    wxString strRet;

    strRet += mCConfSnapFile.Serialize() + mCConfAnimFile.Serialize();
    
    strRet +=
	mCConfWorkingDir.Serialize() + mCConfPathPovray.Serialize();
    
    strRet +=
	mCConfListDisp.SerializeAll() + 
	mCConfListFog.SerializeAll();

    strRet += mpCRadiosity->SerializeConf();
    strRet += mpCPhotons->SerializeConf();
    strRet += mpCFocalBlur->SerializeConf();
    strRet += mpCBG->SerializeConf();

    return strRet;
}

wxString PovRay_cls::
InsertTab(wxString strLines){
    int currPos = 0;
    strLines = wxString(wxT("\t")) << strLines;
    while (currPos != wxNOT_FOUND){
	int findPos = strLines.Mid(currPos).Find(L"\n");
	if (findPos != wxNOT_FOUND){
	    currPos += findPos;
	    if (findPos < (int)strLines.Len() - 3)
		strLines = strLines.Mid(0, currPos) + wxT("\n\t") + strLines.Mid(currPos+1);
	    currPos += 1;
	}
    }
    return strLines;
}

/*
wxString PovRay_cls::
SerializeClipping(double *pPos, wxString strLines){
    if (mCbClipSphere.mbBody == TRUE){
	wxString strBody;
	strBody << L"union {" << ENDLWX;
	strBody << PovOut3d(L"\tsphere {", pPos) << L", " 
	       << mCClipSphereRadius.mBody << ENDLWX;
	strBody << L"\t\tpigment{ rgbt <0, 0, 0, 0> }" << ENDLWX;
	strBody << L"\t\thollow" << ENDLWX;
	strBody << L"\t}" << ENDLWX;
	strBody << InsertTab(strLines);
	strBody << L"}" << ENDLWX;
	return strBody;
    }
    return strLines;
}
*/

wxString PovRay_cls::
GetPovStrCommon(){
    wxString strCommon;
    strCommon << ENDLWX;
    strCommon << wxT("#include \"functions.inc\"") << ENDLWX
	      << wxT("#include \"shapes.inc\"") << ENDLWX;

    wxString strWorkingDir = mCConfWorkingDir.msBody + SEPARATORWX;
#ifdef WIN32
    strWorkingDir.Replace(wxT("\\"), wxT("\\\\"));
#endif
    strCommon << wxT("#declare WORKINGDIR=\"") << strWorkingDir << wxT("\"") << ENDLWX;
    
    strCommon << wxT("global_settings {") << ENDLWX
	      << wxT("\tmax_trace_level ") << mCMaxTraceLevel.mBody << ENDLWX;
    if (mCPovRayVersion.mBody == _POVRAY_37)
	strCommon << wxT("\tsubsurface { samples ") << mCSSSQuality1.mBody << wxT(",") << mCSSSQuality2.mBody << wxT("}") << ENDLWX;
    //	      << wxT("\tassumed_gamma ") << mCAssumedGamma.mBody << ENDLWX;
    
    wxString strRadiosity = mpCRadiosity->GetPovStrGloabal();
    if (strRadiosity != wxT("")) strCommon << strRadiosity;
    wxString strPhotons = mpCPhotons->GetPovStrGlobal();
    if (strPhotons != wxT("")) strCommon << strPhotons;
    
    strCommon << wxT("}") << ENDLWX;

    strCommon << mpCBG->GetPovStr(mpCRadiosity->mCType.mBody != 0) << ENDLWX;

    mSplit = pow(10, PovVolume_cls::msCSplitBox.mBody);

    return strCommon;
}

wxString PovRay_cls::
SerializeCameraGeometryType(){
    switch (GetCameraGeometry()){
    case _POV_PERSPECTIVE:
	return wxT("\tperspective");
    case _POV_FISHEYE:
	return wxT("\tfisheye");
    case _POV_ORTHOGRAPHIC:
	return wxT("\torthographic");
    case _POV_PANORAMIC:
	return wxT("\tpanoramic");
    }
    return wxT("");
}

wxString PovRay_cls::
SerializeRight(){
    wxString strRight;
    strRight << L"\tright < " << -(double)mCWidth.mBody/mCHeight.mBody
	   << L", 0, 0>" << ENDLWX;
    return strRight;
}

wxString PovRay_cls::
GetPovStrCamInfo(PositionAndDirection_cls CCamPos, wxString strCamPos,
		 double doubleIndex){
    wxString strCamInfo;

    strCamInfo << wxT("camera {") << ENDLWX
	       << SerializeCameraGeometryType() << ENDLWX
	       << strCamPos
	       << SerializeRight()
	       << SerializeAngle() << ENDLWX;

    double anchorPos[3];
    double apatureFactor = 1;
    switch (mpCFocalBlur->mCFocusLocationType.mBody){
    case _POV_FOCUSPOS_ANCHORPOS:
	if (mCCameraType.mBody != 0) 
	    LOOP_XYZ anchorPos[xyz] = MainGLCanvas_cls::msCAnchorPos.mPos[xyz];
    else
	LOOP_XYZ anchorPos[xyz] = mpCParentFrame->mpCPathCtrl->GetActivePath()->mCAncPos.mpInterpolatedValue[xyz];
	break;
    case _POV_FOCUSPOS_STEREOFOCUS:
	apatureFactor = mpCParentFrame->mpCStereo->GetDistanceToFocus(doubleIndex);
	LOOP_XYZ anchorPos[xyz] = CCamPos.mPos[xyz] + CCamPos.mTgt[xyz] * apatureFactor;
	
	break;
    case _POV_FOCUSPOS_DISTANCE:
	LOOP_XYZ anchorPos[xyz] = CCamPos.mPos[xyz] + CCamPos.mTgt[xyz] * mpCFocalBlur->mCFocusDistance.mBody;
	break;
    }
    
    strCamInfo << mpCFocalBlur->GetPovStrFocalBlur(anchorPos, apatureFactor);
    strCamInfo << wxT("}") << ENDLWX;
    
    return strCamInfo;
}

wxString PovRay_cls::
GetPovStrIniFile(std::wstring strPovFile, int jpegCompress){
    wxString sInfo;
    
    sInfo << wxT("width=") << mCWidth.mBody << ENDLWX;
    sInfo << wxT("height=") << mCHeight.mBody << ENDLWX;

    /*
    if (mCbSubsetRow.mbBody == TRUE){
	sInfo << wxT("Start_Row=") << mCSubsetRowStart.mBody << ENDLWX;
	sInfo << wxT("End_Row=") << mCSubsetRowEnd.mBody << ENDLWX;
    }
    */

    sInfo << wxT("DISPLAY=") << ((mCbDisplay.mbBody == TRUE)? wxT("on") : wxT("off")) << ENDLWX;
    sInfo << wxT("Pause_When_Done=") << ((mCbPause.mbBody == TRUE)? wxT("on") : wxT("off")) << ENDLWX;
    if (mCbPreview.mbBody == TRUE)
	sInfo << wxT("Preview_Start_Size=8") << ENDLWX 
	      << wxT("Preview_End_Size=8") << ENDLWX;
    if (mCPovRayVersion.mBody == _POVRAY_37){
	sInfo << wxT("Version=3.7") << ENDLWX;
	sInfo << wxT("Display_Gamma=") << mCDisplayGamma.mBody << ENDLWX
	      << wxT("File_Gamma=") << mCFileGamma.mBody << ENDLWX;
	if (mCWorkThread.mBody != 0 || mbRenderingTwoSnap == TRUE){
	    int thread = mCWorkThread.mBody;
	    if (mbRenderingTwoSnap == TRUE){
		if (thread == 0) thread = wxThread::GetCPUCount() / 2;
		else thread /= 2;
	    }
	    sInfo << wxT("Work_Threads=") << thread << ENDLWX;
	}
	sInfo << wxT("Render_Block_Size=") << mCRenderBlockSize.mBody << ENDLWX;
	if (mCBoundingMethod.mBody == 1){
	    sInfo << wxT("Bounding_Method=2") << ENDLWX;
	}
	int buffSize[4] = {32, 128, 512, 0}; 
	sInfo << wxT("Max_Image_Buffer_Memory=") << buffSize[mCBufferSize.mBody] << ENDLWX;
    }
    sInfo << wxT("Antialias=") << mCbAntiAlias.mbBody << ENDLWX;
    if (mCbAntiAlias.mbBody == TRUE){
	sInfo << wxT("Antialias_Threshold=") << mCAntiAliasThreshold.mBody << ENDLWX;
	sInfo << wxT("Antialias_Depth=") << mCAntiAliasDepth.mBody << ENDLWX;
    }

    sInfo << wxT("Jitter=") << mCbJitter.mbBody << ENDLWX;
    if (mCbJitter.mbBody == TRUE)
	sInfo << wxT("Jitter_Amount=") << GetShortString(mCJitterAmount.mBody) << ENDLWX;
    
    if (mCOutputFormat.mBody == _POV_PNG_12BIT)
	sInfo << wxT("Bits_Per_Color=12") << ENDLWX;
    if (mCOutputFormat.mBody == _POV_PNG_16BIT)
	sInfo << wxT("Bits_Per_Color=16") << ENDLWX;
    sInfo << wxT("Output_File_Type=") << msAFormat[mCOutputFormat.mBody] << ENDLWX;
    if (mCOutputFormat.mBody == _POV_JPG)
	sInfo << wxT("Compression=") << jpegCompress << ENDLWX;
    
    sInfo << wxT("Output_Alpha=") << mCbOutputAlpha.mbBody << ENDLWX;
    
    wxString strHeader = (mCbWinePov.mbBody == TRUE) ? wxT("z:") : wxT("") ;
    wxString strInputFile;
    if (mbPovOutForRender == TRUE){
	strInputFile << wxT("\"") << strHeader << mCConfWorkingDir.msBody << SEPARATOR << strPovFile;
	sInfo << wxT("Input_File_Name=") << strInputFile << wxT(".pov\"") << ENDLWX;
    }
    else{
	strInputFile = strPovFile;
        sInfo << wxT("Input_File_Name=") << strInputFile << wxT(".pov") << ENDLWX;
    }
    
    sInfo << wxT("Output_File_Name=\"") << strHeader << GetWorkingDirWithSep() << strPovFile << wxT("\"") << ENDLWX;

    wxString strRadiosity = mpCRadiosity->GetPovStrIni();
    if (strRadiosity != wxT("")) sInfo << strRadiosity;

    if (mCRemoveFile.mBody == 1 && mbPovOutForRender == TRUE){
#ifdef WIN32
	mstrPostSceneCommand << L"del " << strInputFile << wxT(".pov\";");
	mstrPostSceneCommand << L"del " << strInputFile << wxT(".ini\";");
#else
	mstrPostSceneCommand << L"rm " << strInputFile << wxT(".pov\";");
	mstrPostSceneCommand << L"rm " << strInputFile << wxT(".ini\";");
#endif
    }
    
    return sInfo;
}

wxString PovRay_cls::
GetPovStrPhotonSetting(double alpha, double reflectMin, double reflectMax,
		       bool bKillTarget){
    return mspCThis->mpCPhotons->GetPovStrPhotonSetting
	(alpha, reflectMin, reflectMax, bKillTarget);
}

wxString PovRay_cls::
GetPovRayCommand(wxString filename){
    wxString strOption = POVOPTIONWX;
#ifdef WIN32
    if (mCbPause.mbBody == FALSE) strOption += wxT(" /NR /EXIT ");
    wxString strCommand = 
	wxT("\"") + mCConfPathPovray.msBody + wxT("\" '") + mCConfWorkingDir.msBody + SEPARATOR + filename + wxT(".ini'") + strOption;
#else
    wxString strCommand = wxT("nice ");
    strCommand <<
	mCConfPathPovray.msBody + wxT(" ") + mCConfWorkingDir.msBody + SEPARATORWX + filename + wxT(".ini") + strOption;
#endif
    return strCommand;
}

bool PovRay_cls::
RenderByPovSnap(unsigned int stereo){
    wchar_t filename[64];
    swprintf(filename, 64, GetFileStr(_POV_SNAP, stereo).c_str(), mSnapNumber);
    wxString strCommand = GetPovRayCommand(filename);
    
    mpCParentFrame->PushStatusTextMain(wxT("start render by Pov-Ray"));

    int threadIndex = -1;
    for (int testThread = 0; testThread < _POVRAY_MAX_THREAD; testThread++){
	if (mpCThreadPovRay[testThread]== NULL ||
	    mpCThreadPovRay[testThread]->mbRunning == FALSE){
	    threadIndex = testThread;
	    break; } }

    if (threadIndex == -1){
	MsgBox(mpCGUI, wxT("thread max"), wxT("thread max"));
	return FALSE; }

    if (mbPovRayThreadActive[threadIndex] == TRUE){
	mpCThreadPovRay[threadIndex]->Wait();
	delete mpCThreadPovRay[threadIndex];  
	mbPovRayThreadActive[threadIndex] = FALSE; }

    mpCThreadPovRay[threadIndex] = new ThreadPovRay;
    mpCThreadPovRay[threadIndex]->SetParameters
	(this, strCommand, mstrPostSceneCommand);
    mbPovRayThreadActive[threadIndex] = TRUE;
    mpCThreadPovRay[threadIndex]->Run();

    return TRUE;
}

void PovRay_cls::
PushBackCommands(unsigned int stereo, unsigned int frame){
    wchar_t filename[64];
    swprintf(filename, 64, GetFileStr(_POV_ANIM, stereo).c_str(), frame);
    wxString strCommand = GetPovRayCommand(filename);
    mVstrAnimCommand.push_back(strCommand);
    mVstrAnimPostSceneCommand.push_back(mstrPostSceneCommand);
}

void PovRay_cls::
ExportPovAnim(){
    mbExportingAnim = TRUE;
    mpCGUI->mpCButtonExportAnim->SetLabel(wxT("StopExport"));
    mpCGUI->mpCButtonExportAnim->SetBackgroundColour(wxColour(wxT("RED")));
    mpCGUI->mpCButtonExportAnim->SetForegroundColour(wxColour(wxT("WHITE")));
}

void PovRay_cls::
WaitingForExport(bool bFlag){
    if (bFlag == TRUE)
	mpCGUI->mpCButtonRenderAnim->SetLabel(wxT("Waiting..."));
    else
	mpCGUI->mpCButtonRenderAnim->SetLabel(wxT("RenderAnim"));
}

void PovRay_cls::
RenderByPovAnim(unsigned int init, unsigned int fin){
    mCurrentAnim = 0;
    mbRenderingAnim = TRUE;
    mpCGUI->mpCButtonRenderAnim->SetLabel(wxT("StopAnim"));
    mpCGUI->mpCButtonRenderAnim->SetBackgroundColour(wxColour(wxT("RED")));
    mpCGUI->mpCButtonRenderAnim->SetForegroundColour(wxColour(wxT("WHITE")));
}

void PovRay_cls::
ProcessRenderAnim(){
    if (mbRenderingAnim == FALSE) {
	if (mpCThreadPovRayAnim != NULL && mpCThreadPovRayAnim->mbRunning == FALSE)
	    mpCParentFrame->PushStatusTextMain(wxT("not rendering"), 2);
	return;
    }
    
    if (mpCThreadPovRayAnim != NULL && mpCThreadPovRayAnim->mbRunning == TRUE)
	return;

    if (mpCThreadPovRayAnim != NULL) {
	mpCThreadPovRayAnim->Wait();
	delete mpCThreadPovRayAnim; }

    mpCThreadPovRayAnim = new ThreadPovRay;
    mpCThreadPovRayAnim->SetParameters
	(this, mVstrAnimCommand[mCurrentAnim],
	 mVstrAnimPostSceneCommand[mCurrentAnim]);

    mpCThreadPovRayAnim->Run();
    wxString strProgress = wxT("PovRendering ");
    strProgress << mCurrentAnim+1 << wxT("/") << (int)mVstrAnimCommand.size();
    mpCParentFrame->PushStatusTextMain(strProgress, 2);
    mCurrentAnim++;
    
    if (mCurrentAnim >= (int)mVstrAnimCommand.size()) StopRenderAnim();
    
    return;
}

void PovRay_cls::
StopRenderAnim(){
    mbRenderingAnim = FALSE;
    mpCGUI->mpCButtonRenderAnim->SetLabel(wxT("RenderAnim"));
    mpCGUI->mpCButtonRenderAnim->SetBackgroundColour(wxNullColour);
    mpCGUI->mpCButtonRenderAnim->SetForegroundColour(wxNullColour);
}

void PovRay_cls::
StopExportAnim(){
    mbExportingAnim = FALSE;
    mpCGUI->mpCButtonExportAnim->SetLabel(wxT("ExportAnim"));
    mpCGUI->mpCButtonExportAnim->SetBackgroundColour(wxNullColour);
    mpCGUI->mpCButtonExportAnim->SetForegroundColour(wxNullColour);
}

//---

std::wstring PovRay_cls::
SerializeAngle(){
    wchar_t buf[255];
    double vrad = GetVAngle() * 3.14159 / 180;
    double hrad =  2 * atan ((double)mCWidth.mBody / mCHeight.mBody * tan (vrad/2));
    swprintf (buf, 255, L"\tangle %f", (hrad * 180 / 3.14159));
    return (std::wstring)buf;
}

float PovRay_cls::
GetRadianForVisCheck(){
    double y = tan (GetVAngle() / 2.0 * M_PI / 180.0);
    double x = y * (double)mCWidth.mBody / mCHeight.mBody;
    double norm = sqrt(x*x + y*y);
    return atan(norm);
}

PovRay_cls* PovRay_cls::mspCThis;
DF3SeqCtrl_cls* PovRay_cls::mspCDF3SeqCtrlLink;
