// "clsPovVolume.h"
// (setq c-basic-offset 4)

#include "clsPovVolume.h"

#include "clsPovFog.h"
#include "clsPovRay.h"
#include "clsTreeBoxRule.h"
#include "clsLightSource.h"
#include "../DF3/clsDF3Sequence.h"
#include "../DF3/clsDF3SeqCtrl.h"
#include "../DF3/GL/clsPreviewProperty.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../MAINCANVAS/clsFilterPanels.h"
#include "../WIDGETCTRL/clsMultiNumericalInputs.h"
#include "../WIDGETCTRL/clsAtoiComboCtrl.h"
#include "../WIDGETCTRL/clsGraphPanel.h"
#include "../WIDGETCTRL/EX/clsColourDialogEX.h"
#include "../../clsDrawingBox.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include <sstream>
#include <wx/file.h>
extern wxColourDialogEX* GpCClrDialog;

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGB for (int rgb = 0; rgb < 3; rgb++)
#define ENDL std::endl
#define ENDLWX wxT("\n")

// ---------------------------------

MACRO_FilterConstructor(PovVolume,,)

void PovVolumeGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, 
	     unsigned int index){

    MultiNumericalInputsPanel_cls* pCurr = InitializeMNumPanel(pCParentWindow);

    pCurr->mTextWidth = 65;

    pCurr->mItemWidth = -1;
#ifdef WIN32
    pCurr->mComboWidth = 108;
#else
    pCurr->mComboWidth = 98;
#endif
    //pCurr->mComboWidth = 40;
    //pCurr->mButtonWidth = 60;

    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT(""));
           
    mListType[0] = _("Monochrome");
    mListType[1] = _("Red-Yellow");
    mListType[2] = _("Pink");
    mListType[3] = wxString::Format(_T("%s(%s)"), _("Thin Blue"), _("Red-White"));
    mListType[4] = _("Green");
    mListType[5] = _("GreenYellow");
    mListType[6] = wxString::Format(_T("%s(%s)"), _("Wine red"), _("Green-White"));
    mListType[7] = _("purple");
    mListType[8] = _("blue");
    mListType[9] = wxString::Format(_T("%s(%s)"), _("Blue-Yellow"), _("Blue-White"));
	
    mIndexColorType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCColoring.mBody,
	 _("Coloring"), TRUE, _CLR_CUSTOM, mListType/*, 1*/);

    mpCPanelColorBar = new PanelColorBar_cls(pCurr, wxID_ANY, wxDefaultPosition, wxSize(-1, 20));
    mpCPanelColorBar->mbDisplayCurrent = FALSE;
    mpCPanelColorBar->mBorderSizeOffset[0] = 1;
    mpCPanelColorBar->SetFunction(&GetColorR);
    mpCPanelColorBar->SetFunction(&GetColorG);
    mpCPanelColorBar->SetFunction(&GetColorB);
    mpCPanelColorBar->mOffset[0] = 4;
    mpCPanelColorBar->mpClue = this;

    mIndexColorBar = pCurr->CombinePointer(_GP_WXPANEL, mpCPanelColorBar);

    wxString dataList[5] = { _("color by this"), _("color by Slot0"), _("color by Slot1"), _("color by Slot2"), _("color by Slot3")};
    mIndexDataType = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mpCParent->mCData.mCDataForColoring.mBody,
	 _("Coloring Method"), TRUE, 5, dataList);
    
    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Emission Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCEmission.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCEmission.mBody);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Absorption Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCAbsorption.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCAbsorption.mBody);
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Complement"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCCompAbsorption.mBody);
    pCurr->SetRange(0, 100, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCCompAbsorption.mBody);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Scattering Factor"));
    pCurr->CombinePointer(_GP_FLOAT_SLIDER, 
			  &mpCParent->mCData.mCScatter.mBody);
    pCurr->SetRange(0, 200, _MN_LASTITEM, 100);
    pCurr->CombinePointer(_GP_FLOAT, 
			  &mpCParent->mCData.mCScatter.mBody);

    pCurr->FinishSetting(pBoxSizer, FALSE, index);
    pCurr->UpdateValues();

    //mpObjectForCustomA = pCurr->mVCPointers[mIndexColorButton].mpPointerCtrl;

    pCurr->SetFuncHook(&FuncHookComboBox, _MN_HOOK_COMBOBOX_CHANGE);
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookRefreshTop, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FilterCtrl_cls::FuncHookChanged, _MN_HOOK_CHANGE);
    pCurr->SetFuncHook(&FuncHookPanelClick, _MN_HOOK_PANELCLICK);
    
    UpdateColorList();
    HideAndShowItems();
}

MACRO_SetCustomFunctions(PovVolumeGUI_cls,/*ChooseColor()*/,,)

void PovVolumeGUI_cls::
ChooseColor(){
    mpCParent->ChooseColor();
    FilterCtrl_cls::FuncHookRefreshTop(mpCMNumPanel, NULL);
}

void PovVolumeGUI_cls::
UpdateColorList(){
    if (GCColoring.mpCCustomColor == NULL) return; // ŏ̃R[łNULL

    wxString pColorList[_CLR_CUSTOM + _CLR_CUSTOM_NUM_MAX];
    for (int i = 0; i < _CLR_CUSTOM; i++) 
	pColorList[i] << mListType[i];
    for (int i = 0; i < GCColoring.mpCCustomColor->mCItemNum.mBody; i++)
	pColorList[_CLR_CUSTOM + i] << wxT("Color ") << i;
    mpCMNumPanel->mVCPointers[mIndexColorType].CombinePointer
	(_GP_UNSIGNED_INT_LIST, NULL, wxT(""), TRUE, _CLR_CUSTOM + GCColoring.mpCCustomColor->mCItemNum.mBody, pColorList);

}

#include "../MYAPI/MsgBox.h"

void PovVolumeGUI_cls::
FuncHookComboBox(void *pArg, void *pEvent){
    PovVolumeGUI_cls *pThis = 
	(PovVolumeGUI_cls*)DF3SeqCtrl_cls::GetPFilterGUI(pArg);
    pThis->HideAndShowItems();

    if (pThis->mpCParent->GetPCData()->mCColoring.mBody >= _CLR_CUSTOM &&
	FilterCtrl_cls::mspCPreviewPropertyLink->mCTextureType.mBody != _PREVIEW_RGBA &&
	bWarningForRGBA == FALSE){
	bWarningForRGBA = TRUE; // x͈x̂
	if (MsgBoxYesno(pThis->mpCMNumPanel, wxT("warning"), _("To use color curvem, preview mode must be RGBA")) == TRUE){
	    FilterCtrl_cls::mspCPreviewPropertyLink->mCTextureType.mBody = _PREVIEW_RGBA;
	    FilterCtrl_cls::mspCPreviewPropertyLink->UpdateItems(TRUE);
	}
    }

    MultiNumericalInputsPanel_cls *pMNumPanel = (MultiNumericalInputsPanel_cls*)pArg;
    if (FilterCtrl_cls::mspCPreviewPropertyLink->mCTextureType.mBody == _PREVIEW_RGBA){
	FilterCtrl_cls::FuncHookRetexture(pMNumPanel, NULL);
	return;
    }
    FilterCtrl_cls::FuncHookRefreshTop(pMNumPanel, NULL);
}

void PovVolumeGUI_cls::
HideAndShowItems(){
    wxString strType = ((AtoiComboCtrl_cls*)mpCMNumPanel->mVCPointers[mIndexColorType].mpPointerCtrl)->GetValue();

    int colorType = mpCParent->GetPCData()->mCColoring.mBody;
    mpCMNumPanel->ShowItems(mIndexDataType, 1, colorType != 0);
    if (colorType == 0)
	((PanelColorBar_cls*)mpCMNumPanel->mVCPointers[mIndexColorBar].GetPCtrlWindow())->mstrLabel = _("Mono Color");
    else if (colorType >= _CLR_CUSTOM)
	((PanelColorBar_cls*)mpCMNumPanel->mVCPointers[mIndexColorBar].GetPCtrlWindow())->mstrLabel = wxT("Color Curve");
    else 
	((PanelColorBar_cls*)mpCMNumPanel->mVCPointers[mIndexColorBar].GetPCtrlWindow())->mstrLabel = _("Pseudo Color");

    mpCPanelColorBar->Refresh(FALSE);
    mpCParent->mpCParentCtrl->mpCFilterPanelsLink->UpdatePanelLayout();
}

void PovVolumeGUI_cls::
FuncHookPanelClick(void *pArg, void *pEvent){
    PovVolumeGUI_cls *pCThis = ((PovVolumeGUI_cls*)DF3SeqCtrl_cls::GetPFilterGUI(pArg));
    int scrx, scry;
    pCThis->mpCMNumPanel->mVCPointers[pCThis->mIndexColorBar].GetPCtrlWindow()->GetScreenPosition(&scrx, &scry);
    int x = scrx + ((wxMouseEvent*)pEvent)->GetX();

    int coloring = pCThis->mpCParent->GetPCData()->mCColoring.mBody;
    if (coloring == 0)
	pCThis->mpCParent->ChooseColor(x-240, scry+36);
    else if (coloring >= _CLR_CUSTOM){
	wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuItemCustomColor);
	wxGetTopLevelParent(pCThis->mpCMNumPanel)->AddPendingEvent(com);
    }
}

void PovVolumeGUI_cls::
RefreshColorBar(){
    mpCPanelColorBar->Refresh();
}

double PovVolumeGUI_cls::
GetColor(double value, int rgbIndex){
    int coloring = mpCParent->GetPCData()->mCColoring.mBody;
    if (coloring == 0)
	return mpCParent->GetPCData()->mCMonoColor.mpBody[rgbIndex];
    if (coloring >= _CLR_CUSTOM)
	return GCColoring.mpCCustomColor->GetCustomColor(value, rgbIndex, TRUE, coloring - _CLR_CUSTOM);

    float pRGB[3];
    GCColoring.GetColor(pRGB, value, coloring, TRUE);
    return pRGB[rgbIndex];
}

double PovVolumeGUI_cls::
GetColorR(double value, void *pClue){
    PovVolumeGUI_cls *pCThis = (PovVolumeGUI_cls*)pClue;
    return pCThis->GetColor(value, 0);
}

double PovVolumeGUI_cls::
GetColorG(double value, void *pClue){
    PovVolumeGUI_cls *pCThis = (PovVolumeGUI_cls*)pClue;
    return pCThis->GetColor(value, 1);
}

double PovVolumeGUI_cls::
GetColorB(double value, void *pClue){
    PovVolumeGUI_cls *pCThis = (PovVolumeGUI_cls*)pClue;
    return pCThis->GetColor(value, 2);
}

bool PovVolumeGUI_cls::bWarningForRGBA = FALSE;

// ---------------------------------

PovVolumeData_cls::
PovVolumeData_cls(){
    ReDefineList();
    mCConfList.mbSetWithInitValue = FALSE;
}

void PovVolumeData_cls::
ReDefineList(){
    // ʂ̃Xgɒǉ
    FilterData_cls::ReDefineListBase();
    mCConfList.Add(&mCColoring  , wxT("VolumeColor="), L"0");
    mCConfList.Add(&mCEmission  , wxT("VolumeEmission="), L"1");
    mCConfList.Add(&mCAbsorption, wxT("VolumeAbsorption="), L"1");
    mCConfList.Add(&mCScatter   , wxT("VolumeScattering="), L"0");
    mCConfList.Add(&mCScatterType, wxT("VolumeScatterType="), L"0");
    mCConfList.Add(&mCNoiseType, wxT("VolumeNoise="), L"0");
    mCConfList.Add(&mCNoiseScale, wxT("VolumeNoiseScale="), L"0.01");
    mCConfList.Add(&mCCompAbsorption, wxT("VolumeComplementaryAbsorption="), L"0.5");
    mCConfList.Add(&mCMonoColor, wxT("VolumeMonoColor="), wxT("1 0 0"));
    mCConfList.Add(&mCScatterColor, wxT("VolumeScatterColor="), wxT("1 1 1"));
    mCConfList.Add(&mCbUseLocalSetting, wxT("VolumeUseLocalSet"), wxT("0"));
    mCConfList.Add(&mCSampleMin, wxT("VolumeSampleMin="), wxT("64"));
    mCConfList.Add(&mCSampleMax, wxT("VolumeSampleMax="), wxT("128"));

    mCConfList.Add(&mCMultiValueType, wxT("VolumeMultiValue="), wxT("0"));
    mCConfList.Add(&mCDataForColoring, wxT("VolumeDataForColoring="), wxT("0"));
}

MACRO_MakeInstanceAndAddToList(PovVolume);

void PovVolumeData_cls::
SetDefaultColour(int df3index){
    switch(df3index){
    case 0:
	mCMonoColor.mpBody[0] = 1; 
	mCMonoColor.mpBody[1] = 0; 
	mCMonoColor.mpBody[2] = 0; break;
    case 1:
	mCMonoColor.mpBody[0] = 0; 
	mCMonoColor.mpBody[1] = 1; 
	mCMonoColor.mpBody[2] = 0; break;
    case 2:
	mCMonoColor.mpBody[0] = 0; 
	mCMonoColor.mpBody[1] = 0; 
	mCMonoColor.mpBody[2] = 1; break;
    default:
	mCMonoColor.mpBody[0] = 1; 
	mCMonoColor.mpBody[1] = 1; 
	mCMonoColor.mpBody[2] = 1; break;
    }
}

// ---------------------------------

PovVolume_cls::
PovVolume_cls(FilterCtrl_cls *pCParentCtrl, PovVolumeData_cls *pCInitData) 
  : FilterBase_cls(pCParentCtrl) {

    bool bNULLInit = TRUE;
    if (pCInitData != NULL){
	mCData = *pCInitData;
	bNULLInit = FALSE;
    }

    ConstructorCommon(pCParentCtrl, _("Volume"), _FILTER_VOLUME);
    mpCGUIPanel = new PovVolumeGUI_cls(this);

    mbRefreshWhenComming = TRUE;
    if (bNULLInit == TRUE || msbConstructedFromMenu == TRUE)
	mCData.SetDefaultColour(mpCParentCtrl->mSelfDF3Index);
}

MACRO_GetPData(PovVolume);
MACRO_GetPCData(PovVolume);

MACRO_AddFilterItemPolymorph(PovVolume);

void PovVolume_cls::
ChooseColor(int x, int y){
    wxColour CColor(mCData.mCMonoColor.mpBody[0]*255,
		    mCData.mCMonoColor.mpBody[1]*255, 
		    mCData.mCMonoColor.mpBody[2]*255);
    GpCClrDialog->SetColour(CColor);
    GpCClrDialog->mpCDialog->Move(x, y);
    if (GpCClrDialog->ShowModal() == wxID_OK){
	mCData.mCMonoColor.mpBody[0] = (float)GpCClrDialog->GetR()/255;
	mCData.mCMonoColor.mpBody[1] = (float)GpCClrDialog->GetG()/255;
	mCData.mCMonoColor.mpBody[2] = (float)GpCClrDialog->GetB()/255;
	mpCGUIPanel->mpCMNumPanel->UpdateValues();
	((PovVolumeGUI_cls*)mpCGUIPanel)->HideAndShowItems();
    }
}

void PovVolume_cls::
UpdateColors(){
    ((PovVolumeGUI_cls*)mpCGUIPanel)->UpdateColorList();
}

void PovVolume_cls::
SetMNumAdvanced(wxWindow *pCParentWindow){
    MultiNumericalInputs_cls *pCDLG = 
	wxDialogEXKey::InitializeMNumDLG(&mpCMNumAdvanced, pCParentWindow, wxT("Advanced Setting"));

    pCDLG->mTextWidth = 275;
    pCDLG->mButtonWidth = 80;

    int boxWidth = 1;
    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Use followings instead of common settings"));
    pCDLG->CombinePointer(_GP_BOOL, &GetPCData()->mCbUseLocalSetting.mbBody);
    
    pCDLG->mTextWidth = 120;
    boxWidth = 2;
    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Samples(min,max)"));
    pCDLG->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCSampleMin.mBody);
    pCDLG->SetRange(1, 2048);
    pCDLG->CombinePointer(_GP_INTEGER_SPIN, &GetPCData()->mCSampleMax.mBody);
    pCDLG->SetRange(2, 2048);

    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Scattering Type"));
    wxString listTarget[5] = { _("Isotoropic"), _("Mie Haze"), _("Mie Murky"), _("Rayleigh"), wxT("Henyey-Greenstein") };
    pCDLG->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCScatterType.mBody,
	 wxT(""), TRUE, 5, listTarget, 0);

    mIndexScatterColor = pCDLG->CombinePointerButton
	(wxID_ANY, _("Color"));
    pCDLG->mVCPointers[mIndexScatterColor].mpConnectColorFloat 
	= GetPCData()->mCScatterColor.mpBody;

    pCDLG->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("AddNoise"));
    wxString listNoise[3] = { _("No noise"), _("noise"), _("agate")};
    pCDLG->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GetPCData()->mCNoiseType.mBody,
	 wxT(""), TRUE, 3, listNoise, 0);
    pCDLG->CombinePointer
	(_GP_FLOAT, &GetPCData()->mCNoiseScale.mBody);
    
    pCDLG->mMinimumWidth = 300;
    pCDLG->FinishSettingDLG(FALSE);

    pCDLG->SetFuncHook(&FuncHookScatterColor, _MN_HOOK_BUTTON);
}

// ^C}bvpɂQDF3`
wxString PovVolume_cls::
GetPovStrDensityFile(wxString strTag,
		     std::vector < wxString > VSFilePath, 
		     bool bTimeRemap, float remapFactor){
    strTag = mstrId + strTag;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTagDFile.size(); i++){
	if (strTag == mVSAlreadyOutputedTagDFile[i]) return L"";
    }
    mVSAlreadyOutputedTagDFile.push_back(strTag);
    
    wxString strDFile;

    int index = mpCParentCtrl->mSelfDF3Index;
    wxString strFilePathA = VSFilePath[index];

    strDFile << wxT("#declare DF3A_") << strTag << wxT("=function{") << ENDLWX
	     << wxT("\tpattern{") << ENDLWX;
    if (strFilePathA == wxT(""))
	strDFile << wxT("\t\tfunction { 0 }") << ENDLWX;
    else
	strDFile << wxT("\t\tdensity_file df3 \"") << strFilePathA << wxT("\"") << ENDLWX 
		 << wxT("\t\tinterpolate ") << msCInterpolate.mBody << ENDLWX ;
    strDFile << wxT("\t\t//scale <1.00392,1.00392,1.03226>") << ENDLWX
	     << wxT("\t} }") << ENDLWX;

    
    if (bTimeRemap == TRUE && remapFactor != 0){
	wxString strFilePathB = VSFilePath[index + _DF3_MAX_SLOT];
	strDFile << wxT("#declare DF3B_") << strTag << wxT("=function{") << ENDLWX
		 << wxT("\tpattern{") << ENDLWX;
	if (strFilePathB == wxT(""))
	    strDFile << wxT("\t\tfunction { 0 }") << ENDLWX;
	else
	    strDFile << wxT("\t\tdensity_file df3 \"") << strFilePathB << wxT("\"") << ENDLWX 
		     << wxT("\t\tinterpolate ") << msCInterpolate.mBody << ENDLWX ;
	strDFile << wxT("\t\t//scale <1.00392,1.00392,1.03226>") << ENDLWX
		 << wxT("\t} }") << ENDLWX;
    }

    return strDFile;
}

wxString PovVolume_cls::
GetPovStrDProfile(wxString strTag, std::vector < wxString > VSFilePath,
		  DrawingBlock_cls *pCDrawingBlock,
		  bool bTimeRemap, float remapFactor){

    strTag = mstrId + strTag;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTagDProfile.size(); i++){
	if (strTag == mVSAlreadyOutputedTagDProfile[i]) return L"";
    }
    mVSAlreadyOutputedTagDProfile.push_back(strTag);

    wxString strDProfile;

    wxString strFilePath = VSFilePath[mpCParentCtrl->mSelfDF3Index];
    
    strDProfile << GetPovStrDensity(strTag, strFilePath, pCDrawingBlock, bTimeRemap, remapFactor);
    
    if (GetPCData()->mCDataForColoring.mBody != 0){
	strFilePath = VSFilePath[GetPCData()->mCDataForColoring.mBody - 1];
	strDProfile << GetPovStrDensity(strTag + wxT("CLR"), strFilePath, pCDrawingBlock, bTimeRemap, remapFactor);
    }

    return strDProfile;
}

wxString PovVolume_cls::
GetPovStrVolumeBox(wxString strTag, wxString strGeometry, 
		   float treeDivideParam, unsigned int depth,
		   double baseAlpha, double doubleIndex, 
		   DrawingBlock_cls* pCDrawingBlock){
    
    strTag = mstrId + strTag;

    for (unsigned int i = 0; i < mVSAlreadyOutputedTag.size(); i++){
	if (strTag == mVSAlreadyOutputedTag[i]) return L"";
    }
    mVSAlreadyOutputedTag.push_back(strTag);
    
    wxString strVBox;
    
    double shadeOffset[3] = {0, 0, 0};
    mspCLightSRC->GetShadeOffset(shadeOffset, doubleIndex);
    PositionAndDirection_cls CShadeOffset;
    if (DF3GL_cls::mspCProperty->mCShadingType.mBody == 1)
	LOOP_XYZ CShadeOffset.mPos[xyz] = shadeOffset[xyz];
    CShadeOffset = pCDrawingBlock->
	GetInvertedPosDir(CShadeOffset, pCDrawingBlock->mbMirroring, FALSE, 
			  (FilterCtrl_cls::mspCPreviewPropertyLink->mCShadingCoord.mBody == 1));
    
    strVBox << PovOut3dS(L"#declare ShadeOffset_" + strTag + L"=", CShadeOffset.mPos) << L";" << ENDLWX;
    
    strVBox << GetPovStrInterior
	(strTag, strGeometry, treeDivideParam, depth, baseAlpha);
    
    return strVBox;
    
}

wxString PovVolume_cls::
GetPovStrDensity(wxString strTag, wxString strDensityFile, 
		 DrawingBlock_cls *pCDrawingBlock, 
		 bool bTimeRemap, float remapFactor){
    /*
    if (mCbWinePov.mbBody == TRUE) 
	strDensityFile.insert(1, L"z:"); 
    */
    
    bool bUse = 
	mpCParentCtrl->IsIsosurfTargetingMeExist(mstrLabel);
    
    if (mCData.mCbValid.mbBody == FALSE && bUse == FALSE) return L"";

    if (remapFactor > 0.9998) remapFactor = 0.9998; //ۂߌ덷ōvPȏɂȂȂ

    wxString strDensity;

    strDensity << ENDLWX;
    strDensity << L"#macro " << wxT("DensityProfile_") << strTag 
	       << L" ( VTranslate, VScale, colorMap)" << ENDLWX;
    strDensity << L"density{" << ENDLWX;
    bool bTimeRemapUse = bTimeRemap & (remapFactor != 0);
    if (strDensityFile != wxT("")){
	// FIX_ME!! gݍ킹܂
	if (bTimeRemapUse == TRUE && pCDrawingBlock->IsNestOnDensity() == FALSE){
	    strDensity << wxT("\tfunction { ") << GetShortString(0.9998 - remapFactor)
		       << wxT(" * DF3A_") << strTag << wxT("(x,y,z) + ") << GetShortString(remapFactor) 
		       << wxT(" * DF3B_") << strTag << wxT("(x,y,z) }") << ENDLWX;
	}
	else if (pCDrawingBlock->IsNestOnDensity() == TRUE){
	    float minPos[3], maxPos[3], coef1[3], coef0[3];
	    mspCTreeBoxRuleLink->GetChildNestPos(minPos, maxPos);
	    mspCTreeBoxRuleLink->GetCoordinateInChild(coef1, coef0);
	    wxString strFunctionA, strFunctionB;
	    strFunctionA 
		<< wxT("(x<=") << GetShortString(minPos[X]) << wxT(")|") 
		<< wxT("(x>=") << GetShortString(maxPos[X]) << wxT(")|")
		<< wxT("(y<=") << GetShortString(minPos[Y]) << wxT(")|")
		<< wxT("(y>=") << GetShortString(maxPos[Y]) << wxT(")|")
		<< wxT("(z<=") << GetShortString(minPos[Z]) << wxT(")|")
		<< wxT("(z>=") << GetShortString(maxPos[Z]) << wxT(")");
	    strFunctionB 
		<< wxT("(x>") << GetShortString(minPos[X]) << wxT(")&") 
		<< wxT("(x<") << GetShortString(maxPos[X]) << wxT(")&")
		<< wxT("(y>") << GetShortString(minPos[Y]) << wxT(")&")
		<< wxT("(y<") << GetShortString(maxPos[Y]) << wxT(")&")
		<< wxT("(z>") << GetShortString(minPos[Z]) << wxT(")&")
		<< wxT("(z<") << GetShortString(maxPos[Z]) << wxT(")");
	    wxString strArgument;
	    strArgument << GetShortString(coef1[X]) << wxT("*x");
	    if (coef0[X] > 0) strArgument << wxT("+");
	    if (coef0[X] != 0) strArgument << GetShortString(coef0[X]);
	    strArgument << wxT(",");
	    strArgument << GetShortString(coef1[Y]) << wxT("*y");
	    if (coef0[Y] > 0) strArgument << wxT("+");
	    if (coef0[Y] != 0) strArgument << GetShortString(coef0[Y]);
	    strArgument << wxT(",");
	    strArgument << GetShortString(coef1[Z]) << wxT("*z");
	    if (coef0[Z] > 0) strArgument << wxT("+");
	    if (coef0[Z] != 0) strArgument << GetShortString(coef0[Z]);
	    
	    strDensity << wxT("\tfunction {(") << ENDLWX
		       << wxT("\t\t(") << strFunctionA << wxT(") * DF3A_") << strTag << wxT("(x,y,z) +") << ENDLWX
		       << wxT("\t\t(") << strFunctionB << wxT(") * DF3A_") << strTag << wxT("0(") << strArgument << wxT("))") << ENDLWX;

	    int noiseCoord = 1.0f/GetPCData()->mCNoiseScale.mBody;
	    wxString strNoiseArg;
	    strNoiseArg << wxT("x*") << noiseCoord << wxT(",y*") << noiseCoord << wxT(",z*") << noiseCoord;
	    if (GetPCData()->mCNoiseType.mBody == 2)
		strDensity << wxT("\t\t* (1-f_agate(") << strNoiseArg << wxT(")*0.25)") << ENDLWX;
	    if (GetPCData()->mCNoiseType.mBody == 1)
		strDensity << wxT("\t\t* (1-f_noise3d(") << strNoiseArg << wxT(")*0.25)") << ENDLWX;
	    
	    strDensity << wxT("\t}") << ENDLWX;
	}
	else
	    strDensity << L"\tdensity_file df3 \"" << strDensityFile << L"\"" << ENDLWX
		       << L"\tinterpolate " << msCInterpolate.mBody << ENDLWX;
	
	strDensity << L"\tfrequency 0" << ENDLWX
		   << L"\tscale VScale" << ENDLWX
		   << L"\ttranslate VTranslate" << ENDLWX
		   << L"\tcolor_map { colorMap }" << ENDLWX;
    }
    strDensity << L"}" << ENDLWX
	       << L"#end" << ENDLWX;
    
    return strDensity;
}


wxString PovVolume_cls::
GetPovStrCommonDensityModifier(wxString strTab, float treeDivideParam){
    if (treeDivideParam > 1) treeDivideParam = 1.0;
    wxString strMod;
    strMod << strTab << wxT("intervals ") << msCInterval.mBody << ENDLWX
	   << strTab << wxT("samples ") << (int)(msCSampleMin.mBody * (1.0+treeDivideParam)) 
	   << wxT(",") << (int)(msCSampleMax.mBody * (1.0 + treeDivideParam)) << ENDLWX
	   << strTab << wxT("aa_level ") << msCAALevel.mBody << ENDLWX
	   << strTab << wxT("aa_threshold ") << GetShortString(msCAAThreshould.mBody) << ENDLWX
    	   << strTab << L"method " << msCMethod.mBody << ENDLWX;
    return strMod;
}

wxString PovVolume_cls::
GetPovStrDensityModifier(wxString strTab, float treeDivideParam,
			 unsigned int depth){
    if (treeDivideParam > 1) treeDivideParam = 1.0;
    wxString strMod;
    strMod << strTab << L"intervals " << msCInterval.mBody << ENDLWX;

    int sampleMin, sampleMax;
    
    if (mCData.mCbUseLocalSetting.mbBody == TRUE){
	sampleMin = (int)(mCData.mCSampleMin.mBody * (1.0+treeDivideParam));
	sampleMax = (int)(mCData.mCSampleMax.mBody * (1.0+treeDivideParam));
    }
    else{
	sampleMin = (int)(msCSampleMin.mBody * (1.0+treeDivideParam));
	sampleMax = (int)(msCSampleMax.mBody * (1.0+treeDivideParam));
    }

    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED){
	double factor;
	switch(mspCTreeBoxRuleLink->mCVolumeQuality.mBody){
	case 0: factor = 2; break;
	case 1: factor = 1.75; break;
	case 2: factor = 1.5; break;
	case 3: factor = 1.25; break;
	case 4: factor = 1; break;
	}
	for (int i = 0; i < depth; i++){
	    sampleMin /= factor;
	    sampleMax /= factor;
	}
    }
	
    if (sampleMin < 1) sampleMin = 1;
    if (sampleMax < 1) sampleMax = 1;

    strMod << strTab << L"samples " << sampleMin << L"," << sampleMax << ENDLWX
	   << strTab << wxT("aa_level ") << msCAALevel.mBody << ENDLWX
	   << strTab << wxT("aa_threshold ") << GetShortString(msCAAThreshould.mBody) << ENDLWX
	   << strTab << L"method " << msCMethod.mBody << ENDLWX;
    return strMod;
}

wxString PovVolume_cls::
GetPovStrInterior(wxString strTag, wxString strDensityGeometry,
		  float treeDivideParam, unsigned int depth,
		  double baseAlpha){
    
    int selfIndex = mpCParentCtrl->GetSelfIndex(this);
    bool bUse = mpCParentCtrl->IsIsosurfTargetingMeExist(mstrLabel/*selfIndex*/);
    if (mCData.mCbValid.mbBody == FALSE && bUse == FALSE) return L"";

    wxString strOut;
    
    strOut << L"#macro INTERIOR_" << strTag << L" ()" << ENDLWX
	   << L"interior {" << ENDLWX;
    
    float factor = 1;
    int loopmax = depth; if (loopmax > 5) loopmax = 5;
    for (int i = 0; i < loopmax; i++)
	//float factor = pow(mTreeDepthParam, (float)depth);
	factor *= mspCPov->mTreeDepthParam[i];
    
    float emission3f[3], absorb3f[3], scatter3f[3];
    LOOP_XYZ emission3f[xyz] = baseAlpha * mCData.mCEmission.mBody * factor;
    if (DF3GL_cls::mspCProperty->mCShadingType.mBody == TRUE ||
	mCData.mCAbsorption.mBody == 0 ||
	mCData.mCCompAbsorption.mBody != 0)
	LOOP_XYZ absorb3f[xyz] = 0;
    else
	LOOP_XYZ absorb3f[xyz] = baseAlpha * mCData.mCAbsorption.mBody * factor;
    LOOP_XYZ scatter3f[xyz] = baseAlpha * mCData.mCScatterColor.mpBody[xyz] * mCData.mCScatter.mBody * factor;
    
    strOut << 
	GetPovStrMedia(emission3f, absorb3f, scatter3f, 
		       L"DensityProfile_"+strTag, strDensityGeometry,
		       L"<0, 0, 0>", treeDivideParam, depth);
    
    // -- SHADE

    double compAbs = mCData.mCCompAbsorption.mBody;
    if (DF3GL_cls::mspCProperty->mCShadingType.mBody == TRUE ||
	(compAbs != 0 && mCData.mCAbsorption.mBody != 0)){
	factor = 1.0 + compAbs - 0.5 * compAbs * compAbs;
	
	loopmax = depth; if (loopmax > 5) loopmax = 5;
	for (int i = 0; i < loopmax; i++)
	    factor *= mspCPov->mTreeDepthParam[i];
	
	LOOP_XYZ emission3f[xyz] = 0;
	if (mCData.mCAbsorption.mBody != 0)
	    LOOP_XYZ absorb3f[xyz] = baseAlpha * mCData.mCAbsorption.mBody * factor;
	else
	    LOOP_XYZ absorb3f[xyz] = 0;
	LOOP_XYZ scatter3f[xyz] = 0;
	
	strOut << 
	    GetPovStrMedia(emission3f, absorb3f, scatter3f, 
			   L"DensityProfile_"+strTag, strDensityGeometry,
			   L"ShadeOffset_"+strTag, treeDivideParam, depth,
			   _CLR_SHADE);
    }
    
    /*
      strOut << mspCPov->GetStrFogList(); // IsosurfaceƓl
    */

    /*
    if (mspCPov->IsFogExist() == TRUE)
	strOut << L"\tmedia { MediaFog }" << ENDL;
    */

    strOut << DrawingBlock_cls::mspCTexture->GetPovStrInterior() << ENDLWX;

    strOut << L"}" << ENDLWX
	   << L"#end" << ENDLWX
	   << ENDLWX;
    
    return strOut;
}

wxString PovVolume_cls::
GetPovStrDProfileArgument(wxString strTranslate, wxString strDensityGeometry,
			  bool bShade, bool bForColor){
    int df3index = mpCParentCtrl->mpCDF3SeqParent->mSelfIndex;
    int coloring = mCData.mCColoring.mBody;

    wxString strColorMap = wxT("CLRMAP");
    strColorMap << df3index << coloring;

    if (bShade == TRUE) strColorMap << wxT("SHADE");
    if (bForColor == TRUE) strColorMap << wxT("CLR");

    wxString strArgment = wxT(" (");
    strArgment << strTranslate << wxT(", ")
	       << strDensityGeometry << wxT(", ")
	       << strColorMap
	       << wxT("())") << ENDLWX;

    return strArgment;
}

wxString PovVolume_cls::
GetPovStrMedia(float emission3f[3], float absorb3f[3], float scatter3f[3],
	       wxString strDensityProfile,
	       wxString strDensityGeometry,
	       wxString strTranslate,
	       float treeDivideParam, unsigned int depth,
	       bool bShade){
    wxString strMedia;

    float emissionTotal = 0, absorbTotal = 0, scatterTotal = 0;
    LOOP_RGB emissionTotal += emission3f[rgb];
    LOOP_RGB absorbTotal += absorb3f[rgb];
    LOOP_RGB scatterTotal += scatter3f[rgb];

    if (emissionTotal == 0 && absorbTotal == 0 && scatterTotal == 0)
	return wxT("");

    strMedia << L"\tmedia {" << ENDLWX;
    
    strMedia << L"\t\t" << strDensityProfile
	     << GetPovStrDProfileArgument(strTranslate, strDensityGeometry,
					  bShade);
    if (GetPCData()->mCDataForColoring.mBody != 0){
	strMedia << L"\t\t" << strDensityProfile << wxT("CLR")
		 << GetPovStrDProfileArgument(strTranslate, strDensityGeometry,
					      bShade, TRUE);
    }

    strMedia << GetPovStrDensityModifier(L"\t\t", treeDivideParam, depth);

    float factor = 1;
    int loopmax = depth; if (loopmax > 5) loopmax = 5;
    for (int i = 0; i < loopmax; i++)
	factor *= mspCPov->mTreeDepthParam[i];
    
    if (emissionTotal != 0)
	strMedia << PovOut3fS(wxT("\t\temission rgb"), emission3f, 4) << ENDLWX;
    if (absorbTotal != 0)
	strMedia << PovOut3fS(wxT("\t\tabsorption rgb"), absorb3f, 4) << ENDLWX;
    if (scatterTotal != 0)
	strMedia << wxT("\t\tscattering {") << GetPCData()->mCScatterType.mBody +1 << PovOut3fS(L",rgb", scatter3f, 4) << L" }" << ENDLWX;
    strMedia << L"\t}" << ENDLWX;

    return strMedia;
}

wxString PovVolume_cls::
GetPovStrBox(wxString strBox, double pMin[3], double pMax[3], bool bClipping,
	     std::vector < FilterBase_cls* > VCPFog,
	     std::vector < wxString > *pVstrInteriorTag,
	     wxString strDifference, DrawingBlock_cls *pCDrawingBlock){
    wxString strOut;
    /* 
    double sizeX = 1.0 - mCSplitBox.mBody * mCameraDirection[0]
	* ((mCameraDirection[0] < 0) ? -1 : 1);
    double sizeY = 1.0 - mCSplitBox.mBody * mCameraDirection[1]
	* ((mCameraDirection[1] < 0) ? -1 : 1);
    double sizeZ = 1.0 - mCSplitBox.mBody * mCameraDirection[2]
	* ((mCameraDirection[2] < 0) ? -1 : 1);
	*/

    double X0[3], X1[3];
    double X0org[3] = {mspCPov->mSplit, mspCPov->mSplit, mspCPov->mSplit};
    double X1org[3] = 
	{1.0-mspCPov->mSplit, 1.0-mspCPov->mSplit, 1.0-mspCPov->mSplit};
    LOOP_XYZ X0[xyz] = X0org[xyz];
    LOOP_XYZ X1[xyz] = X1org[xyz];

    if (bClipping == TRUE){
	LOOP_XYZ X0[xyz] = pMin[xyz] > X0[xyz] ? pMin[xyz] : X0[xyz];
	LOOP_XYZ X1[xyz] = pMax[xyz] < X1[xyz] ? pMax[xyz] : X1[xyz];
    }

    strOut << L"#declare " << strBox << L"=" << ENDLWX
	   << L"object {" << ENDLWX
	   << L"\tdifference {" << ENDLWX;

    wxString strRoundedBox;
    strRoundedBox << L"Round_Box_Merge( " << PovOut3d(wxT(""), X0org) << PovOut3d(wxT(","), X1org) << wxT(",") << mspCTreeBoxRuleLink->mCRoundedCurve.mBody << wxT(")");
    wxString strPerpendicularBox;
    strPerpendicularBox << L"box{" << PovOut3d(wxT(""), X0) << PovOut3d(wxT(","), X1) << wxT("}");
    
    if (mspCTreeBoxRuleLink->mCType.mBody == _TREE_NESTED &&
	mspCTreeBoxRuleLink->mCRoundedCurve.mBody != 0 &&
	pCDrawingBlock->mDepth >= 1){
	if (bClipping == TRUE)
	    strOut << wxT("\t\tintersection {") << ENDLWX
		   << wxT("\t\t\t") << strRoundedBox << ENDLWX
		   << wxT("\t\t\t") << strPerpendicularBox << ENDLWX
		   << wxT("\t\t}") << ENDLWX;
	else
	    strOut << wxT("\t\t") << strRoundedBox << ENDLWX;
    }
    else
	strOut << wxT("\t\t") << strPerpendicularBox << ENDLWX;
	/*
	strOut << L"\t\tbox{ " 
	       << L"<" << X0[X] << L", " << X0[Y] << L", " << X0[Z] << L">, " 
	       << L"<" << X1[X] << L", " << X1[Y] << L", " << X1[Z] << L">" << L" }" << ENDLWX;
	*/
    
    if (strDifference != wxT(""))
	strOut << strDifference ;

    if (LightSRC_cls::msstrLightSRCSphere != wxT(""))
	strOut << wxT("\t\tobject{") << ENDLWX
	       << wxT("\t\t\tLIGHTSRC_SPHERE()") << ENDLWX
	       << pCDrawingBlock->GetPovStrInverseCoordinate()
	       << wxT("\t\t\ttranslate <0.5, 0.5, 0.5>") << ENDLWX
	       << wxT("\t\t}") << ENDLWX;
    
    strOut << L"\t}" << ENDLWX 
	   << L"\thollow" << ENDLWX;
    //<< L"\tpigment { rgbt < 1, 1, 1, 1> }" << ENDLWX ;
    
    //<< wxConvFile.cWX2WC(strInterior);
    
    for (unsigned int i = 0; i < (*pVstrInteriorTag).size(); i++)
	strOut << (*pVstrInteriorTag)[i] << ENDLWX;

    if (TRUE){ // FIX_ME photon KvȂƂ
	strOut << wxT("\tphotons{ pass_through on reflection off refraction on collect off}") << ENDLWX;
    }

    for (unsigned int i = 0; i < VCPFog.size(); i++)
	if (VCPFog[i] != NULL)
	    strOut << L"\tinterior{" << ENDLWX
		   << L"\t" << ((PovFog_cls*)VCPFog[i])->GetPovStrFogForUse()
		   << L"\t}" << ENDLWX;

    strOut << L"\ttranslate < -0.5, -0.5, -0.5 >" << ENDLWX // [-0.5, 0.5]box
	  << L"}" << ENDLWX << ENDLWX;
    return strOut;
}

void PovVolume_cls::
FuncHookScatterColor(void *pArg, void *pEvent){
    PovVolume_cls *pCThis = mspCConnectedToDLG;

    wxColour CColor(pCThis->GetPCData()->mCScatterColor.mpBody[0]*255,
		    pCThis->GetPCData()->mCScatterColor.mpBody[1]*255,
		    pCThis->GetPCData()->mCScatterColor.mpBody[2]*255);

    GpCClrDialog->SetColour(CColor);
    if (GpCClrDialog->ShowModal() == wxID_OK){
	pCThis->GetPCData()->mCScatterColor.mpBody[0] = (float)GpCClrDialog->GetR()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[1] = (float)GpCClrDialog->GetG()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[2] = (float)GpCClrDialog->GetB()/255;
	pCThis->mpCMNumAdvanced->mVCPointers[pCThis->mIndexScatterColor].UpdateValue();
    }
    /*
    wxColourDialogEX CDialog(pCThis->mpCParentCtrl->mpCFilterPanelsLink,
			     CColor, TRUE);
    if (CDialog.ShowModal() == wxID_OK){
	pCThis->GetPCData()->mCScatterColor.mpBody[0] = (float)CDialog.GetR()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[1] = (float)CDialog.GetG()/255;
	pCThis->GetPCData()->mCScatterColor.mpBody[2] = (float)CDialog.GetB()/255;
	pCThis->mpCMNumAdvanced->mVCPointers[pCThis->mIndexScatterColor].UpdateValue();
    }
    */
}

// ----

wxString PovVolume_cls::
GetFilterStatusStr(){
    wxString strStatus;
    strStatus << wxT("PovVolume::");
    
    return strStatus;
}

MACRO_SerializeConf(PovVolume)
MACRO_ReadDataBlock(PovVolume)

wxString PovVolume_cls::
SerializeStaticConf(){
    return 
	msCInterpolate.Serialize() +
	msCMethod.Serialize() + 
	msCInterpolate.Serialize() + 
	msCSampleMin.Serialize() + 
	msCSampleMax.Serialize() + 
	msCAALevel.Serialize() + 
	msCAAThreshould.Serialize() + 
	msCSplitBox.Serialize();
}

void PovVolume_cls::
ReadStrLineStaticConf(std::wstring strLine){
    msCInterpolate.ReadStrLine(strLine);
    msCMethod.ReadStrLine(strLine);
    msCInterpolate.ReadStrLine(strLine);
    msCSampleMin.ReadStrLine(strLine);
    msCSampleMax.ReadStrLine(strLine);
    msCAALevel.ReadStrLine(strLine);
    msCAAThreshould.ReadStrLine(strLine);
    msCSplitBox.ReadStrLine(strLine);
}

// ---

void PovVolume_cls::
InitializeStaticMember(PovRay_cls* pCPov, LightSRC_cls* pCLightSRC){
    mspCPov = pCPov;
    mspCLightSRC = pCLightSRC;
}

PovRay_cls* PovVolume_cls::mspCPov;
LightSRC_cls* PovVolume_cls::mspCLightSRC;

PovVolume_cls *PovVolume_cls::mspCConnectedToDLG;

ConfItemUINT_cls PovVolume_cls::msCInterpolate(L"PovRenderInterpolate=",L"1");
ConfItemUINT_cls PovVolume_cls::msCMethod(L"PovRenderSamplingMethod=",L"3");
ConfItemINT_cls PovVolume_cls::msCInterval(L"PovRenderIntervals=", L"1");
ConfItemINT_cls PovVolume_cls::msCSampleMin(L"PovRenderSampleMin=", L"64");
ConfItemINT_cls PovVolume_cls::msCSampleMax(L"PovRenderSampleMax=", L"128");
ConfItemINT_cls PovVolume_cls::msCAALevel(L"PovRenderAALevel=", L"4");
ConfItemFLOAT_cls PovVolume_cls::msCAAThreshould(L"PovRenderAAThreshold=", L"0.05");
ConfItemFLOAT_cls PovVolume_cls::msCSplitBox(L"PovRenderSplitBox=", L"-3");
