// -*- c -*-
// "clsUndoManager.h"
// (setq c-basic-offset 4)

#ifndef _UNDO_MANAGER_H_
#define _UNDO_MANEGER_H_

#include "clsUndoManager.h"

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

UndoManager_cls GCUndoList;

CommandItem_cls::
CommandItem_cls(){
    mCommandType = _COMMAND_TYPE_NORMAL;
}

CommandItem_cls::
~CommandItem_cls(){
}

// --------------------------

ComBlockOpen::
ComBlockOpen(){
    mCommandType = _COMMAND_TYPE_BLOCK_OPEN;
}

bool ComBlockOpen::
Execute(bool bFlag){
    GCUndoList.mBlockDepth++;
    return FALSE;
}

ComBlockClose::
ComBlockClose(){
    mCommandType = _COMMAND_TYPE_BLOCK_CLOSE;
}

bool ComBlockClose::
Execute(bool bFlag){
    GCUndoList.mBlockDepth--;
    return FALSE;
}

// --------------------------

UndoManager_cls::
UndoManager_cls(){
    mCurrentIndex = 0;
    mBlockDepth = 0;
}

UndoManager_cls::
~UndoManager_cls(){
    for (unsigned int i = 0; i < mVPCRedoList.size(); i++){
	delete mVPCRedoList[i];
	delete mVPCUndoList[i];
    }
}

void UndoManager_cls::
Clear(){
    for (unsigned int i = 0; i < mVPCRedoList.size(); i++) RemoveItem(i);
    mVPCRedoList.clear();
    mVPCUndoList.clear();
    mCurrentIndex = 0;
}

void UndoManager_cls::
RemoveItem(unsigned int i){
    delete mVPCRedoList[i];
    delete mVPCUndoList[i];
}

void UndoManager_cls::
OpenBlock(){
    SetNewCommand
	(new ComBlockOpen, new ComBlockClose);
    ExecCurrent();
}

void UndoManager_cls::
CloseBlock(){
    int type = GetRedoCommandType();

    SetNewCommand
	(new ComBlockClose, new ComBlockOpen);
    ExecCurrent();

    // $B6u$C$]$N%V%m%C%/$J$i0UL#$,L5$$$N$G0l$DLa$k(B
    //UndoCurrent$B$GLa$k$K$O(BmBlockDepth$B$,(B0$B$G$J$$$H:$$k$N$GF~$l;RMQ$K2>$K(B0$B$K!#(B
    int blockDepthOrg = mBlockDepth;
    mBlockDepth = 0;
    if (type == _COMMAND_TYPE_BLOCK_OPEN)
	UndoCurrent(); 
    mBlockDepth = blockDepthOrg;
}

void UndoManager_cls::
SetNewCommand(CommandItem_cls* pComRedo, CommandItem_cls* pComUndo){
    for (unsigned int i = mCurrentIndex; i < mVPCRedoList.size(); i++)
	RemoveItem(i);
    
    mVPCRedoList.resize(mCurrentIndex);
    mVPCUndoList.resize(mCurrentIndex);
    mVPCRedoList.push_back(pComRedo);
    mVPCUndoList.push_back(pComUndo);
}

void UndoManager_cls::
PushNewCommand(CommandItem_cls* pComRedo, CommandItem_cls* pComUndo){
    mVPCRedoList.push_back(pComRedo);
    mVPCUndoList.push_back(pComUndo);
}

#include "../MYAPI/MsgBox.h"

int UndoManager_cls::
GetRedoCommandType(int level){
    if (mCurrentIndex <= level) return _COMMAND_TYPE_NORMAL;
    return mVPCRedoList[mCurrentIndex - level]->mCommandType;
}

int UndoManager_cls::
GetUndoCommandType(int level){
    if (mCurrentIndex <= level) return _COMMAND_TYPE_NORMAL;
    return mVPCUndoList[mCurrentIndex - level]->mCommandType;
}

void UndoManager_cls::
SkipCurrent(){
    if (mCurrentIndex < mVPCRedoList.size()){
	mCurrentIndex++;
	if (mBlockDepth > 0) SkipCurrent();
    }
}

void UndoManager_cls::
ExecCurrent(){

    // $B$3$N(Bif$B$K$h$C$F!"<g=>4X78$+$i$NM>J,$J<B9T$rM^@)(B
    if (mCurrentIndex < mVPCRedoList.size() ){
	bool bRecursiveMasterSlave = 
	    mVPCRedoList[mCurrentIndex]->Execute(_COMMAND_REDO);
	//MsgBoxf(NULL, "Call In Exec", "MasterSlave:%d, Chain:%d", (int)bRecursiveMasterSlave, (int)mVPCRedoList[mCurrentIndex]->mbDoNext);
	mCurrentIndex++;
	if (bRecursiveMasterSlave) ExecCurrent();
	if (mBlockDepth > 0) ExecCurrent();
    }
}

bool UndoManager_cls::
UndoCurrent(){
    mCurrentIndex--;
    //MsgBoxf(NULL, "UndoCurrent", "%d size:%d", mCurrentIndex, mVPCRedoList.size());
    if (mCurrentIndex < 65535){ 
	bool bRecursiveMasterSlave = 
	    mVPCUndoList[mCurrentIndex]->Execute(_COMMAND_UNDO);
	//MsgBoxf(NULL, "Call In Undo", "MasterSlave:%d, depth:%d", (int)bRecursiveMasterSlave, mBlockDepth);
	if (bRecursiveMasterSlave) UndoCurrent();
	if (mBlockDepth > 0) UndoCurrent();
	return TRUE; }
    else 
	mCurrentIndex = 0;
    return FALSE;
}

#endif
