// -*- c -*- 
// "clsGraphPanel.h"

#include "clsGraphPanel.h"

BEGIN_EVENT_TABLE(GraphPanel_cls, wxPanel)
     EVT_PAINT       (GraphPanel_cls::OnPaint)
     EVT_CHAR        (GraphPanel_cls::OnChar)
     EVT_MOUSE_EVENTS(GraphPanel_cls::OnMouse)
END_EVENT_TABLE()

#define X 0
#define Y 1

GraphPanel_cls::
GraphPanel_cls(wxWindow *parent, wxWindowID id, 
	       wxPoint pos, wxSize size) :
     wxPanel(parent, id, pos, size)
{
    mOffset[0] = 2; mOffset[1] = 2;
    mBorderSizeOffset[0] = 0; mBorderSizeOffset[1] = 0;
    mIni[0] = 0; mIni[1] = 0;
    mFin[0] = 1; mFin[1] = 1;
    mCurrentDoubleIndex = 0;
    mbDisplayCurrent = FALSE;
    mbDisplayKeyPoint = FALSE;
    mbDisplayBorder = TRUE;
    mpClue = NULL;
}

void GraphPanel_cls::
SetFunction(double (*pFuncGetValue)(double, void*), unsigned int resolution){
    structFuncPointer structPtr;
    structPtr.mpFuncGetValue = pFuncGetValue;
    mVGraph.push_back(structPtr);
    mVResolution.push_back(resolution);
    mVColor.push_back(wxColour(0, 0, 0));
}

void GraphPanel_cls::
SetColorLastItem(wxColour color){
    mVColor[mVColor.size() - 1] = color;
}

void GraphPanel_cls::
SetMemberSize(wxSize size){
    mPanelSize[X] = size.GetWidth();
    mPanelSize[Y] = size.GetHeight();
}

void GraphPanel_cls::
DrawCurrentIndex(wxPaintDC &dc){
    dc.SetPen(*wxBLACK_PEN);
    dc.SetLogicalFunction(wxINVERT);
    CalcCoordinates(mCurrentDoubleIndex, 0, mCurrentDoubleIndex, 1);
    dc.DrawLine(mX0, mY0, mX1, mY1);
    dc.SetLogicalFunction(wxCOPY);
}

void GraphPanel_cls::
ClearKeyPoint(){
    mVVKeyPoint.clear();
    mVVKeyPoint.resize(mVGraph.size());
}

void GraphPanel_cls::
SetKeyPoint(unsigned int index, double x, double y){
    if (index > mVVKeyPoint.size()) return;
    structKeyPoint keyPos = {x, y};
    mVVKeyPoint[index].push_back(keyPos);
}

void GraphPanel_cls::
DrawKeyPoint(wxPaintDC &dc){
    for (unsigned int graph = 0; graph < mVVKeyPoint.size(); graph++){
	dc.SetPen(mVColor[graph]);
	for (unsigned int index = 0; index < mVVKeyPoint[graph].size(); index ++){
	    double dindex = mVVKeyPoint[graph][index].mX;
	    double dvalue = mVVKeyPoint[graph][index].mY;
	    CalcCoordinates(dindex, dvalue, 0, 0);
	    dc.DrawRectangle(mX0-2, mPanelSize[Y] - mY0 - 3, 5, 5);
	}
    }
}

void GraphPanel_cls::
UpdateDisplay(){
    wxPaintDC dc(this);
    
    SetMemberSize(dc.GetSize());
    
    dc.SetBrush(*wxTRANSPARENT_BRUSH);

    CalcCoordinates(0, 0, 1, 1);
    dc.DrawRectangle(mX0-1, mY0-1, 
		     mX1-mX0+2+mBorderSizeOffset[0],
		     mY1-mY0+2+mBorderSizeOffset[1]);

    for (unsigned int graph = 0; graph < mVGraph.size(); graph++){
	dc.SetPen(mVColor[graph]);
	wxPoint *pointList;
	unsigned int resolution = mVResolution[graph];
	pointList = new wxPoint [resolution];
	for (unsigned int index = 0; index < resolution; index ++){
	    double dindex = (double)index / (resolution - 1);
	    CalcCoordinates(dindex, (*mVGraph[graph].mpFuncGetValue)(dindex, mpClue), 0, 0);
	    pointList[index].x = mX0;
	    pointList[index].y = mPanelSize[Y] - mY0 - 1;
	    if (pointList[index].y < mOffset[Y]) pointList[index].y = mOffset[Y];
	}
	dc.DrawLines(resolution, pointList);
	delete [] pointList;
    }

    if (mbDisplayKeyPoint == TRUE) DrawKeyPoint(dc);
    if (mbDisplayCurrent == TRUE) DrawCurrentIndex(dc);
}

void GraphPanel_cls::
OnPaint(wxPaintEvent& event){
    event.Skip();
    UpdateDisplay();
}

void GraphPanel_cls::
OnChar(wxKeyEvent& event){
    GetParent()->AddPendingEvent(event);
}

void GraphPanel_cls::
OnMouse(wxMouseEvent& event){
    GetParent()->AddPendingEvent(event);
}

void GraphPanel_cls::
CalcCoordinates(double valuex0, double valuey0,
		double valuex1, double valuey1){
    double normx0 = (valuex0 - mIni[X]) / ( mFin[X] - mIni[X]);
    double normy0 = (valuey0 - mIni[Y]) / ( mFin[Y] - mIni[Y]);
    mX0 = GetCoordinate(normx0, X);
    mY0 = GetCoordinate(normy0, Y);
    double normx1 = (valuex1 - mIni[X]) / ( mFin[X] - mIni[X]);
    double normy1 = (valuey1 - mIni[Y]) / ( mFin[Y] - mIni[Y]);
    mX1 = GetCoordinate(normx1, X);
    mY1 = GetCoordinate(normy1, Y);
}

int  GraphPanel_cls::
GetCoordinate(double norm, unsigned int index){
    return norm * (mPanelSize[index] - 2 * mOffset[index]) + mOffset[index];
}

// --------------------

PanelColorBar_cls::
PanelColorBar_cls(wxWindow *parent, wxWindowID id, 
		  wxPoint pos, wxSize size) :
    GraphPanel_cls(parent, id, pos, size)
{
    mstrLabel = wxT("");
}


void PanelColorBar_cls::
UpdateDisplay(){
    wxPaintDC dc(this);
    
    SetMemberSize(dc.GetSize());
    
    dc.SetBrush(*wxTRANSPARENT_BRUSH);

    CalcCoordinates(0, 0, 1, 1);
    if (mbDisplayBorder == TRUE) 
	dc.DrawRectangle(mX0-1, mY0-1, 
			 mX1-mX0+2+mBorderSizeOffset[0],
			 mY1-mY0+2+mBorderSizeOffset[1]);

    int left = mX0;

    if (mVGraph.size() < 3) return;
        
    unsigned int resolution = mVResolution[0];

    dc.SetPen(*wxTRANSPARENT_PEN);

    for (unsigned int index = 0; index < resolution ; index ++){
	double dindexA = (double)index / resolution;
	double dindexB = (double)(index + 1) / resolution;
	unsigned int r = 255*(*mVGraph[0].mpFuncGetValue)(dindexA, mpClue);
	unsigned int g = 255*(*mVGraph[1].mpFuncGetValue)(dindexA, mpClue);
	unsigned int b = 255*(*mVGraph[2].mpFuncGetValue)(dindexA, mpClue);
	CalcCoordinates(dindexA, 0, dindexB, 1);
	dc.SetBrush(wxColour(r, g, b));
	dc.DrawRectangle(mX0, mY0, mX1-mX0, mY1-mY0);
    }
	

    if (mbDisplayCurrent == TRUE) DrawCurrentIndex(dc);

    if (mstrLabel != wxT("")){
	dc.SetFont(*wxNORMAL_FONT);
	int charHeight = dc.GetCharHeight();
	int charWidth = dc.GetCharWidth();
	int charY = (mY0 + mY1 - charHeight) / 2;
	dc.SetTextForeground(wxColour(64, 64, 64));
	dc.DrawText(mstrLabel, left + 23, charY); 
	dc.SetTextForeground(wxColour(216, 216, 216));
	dc.DrawText(mstrLabel, left + 22, charY-1); // FIX_MEZ^Oǂ́H
    }
}
