// -*- c -*- 
// "clsMultiNumericalInputs.cpp"

#ifndef _MULTI_NUMERICAL_INPUTS_H_
#define _MULTI_NUMERICAL_INPUTS_H_

#include "EX/clsButtonEX.h"
#include "clsDialogEXKey.h"

#include "clsGeneralPointers.h"

#include <vector>

#define _MN_FORWARD 1
#define _MN_BACKWARD -1

#define _MN_LASTITEM -1
#define _MN_ALLITEM  -2
#define _MN_NOITEM   -3
#define _MN_ASSIGN_ALL -2

#define _MN_NO_BUTTONS 0
#define _MN_ASSIGN_UPDATE_HIDE 1
#define _MN_CANCEL_OK 2
#define _MN_OK_CANCEL 3
#define _MN_HELP      4
#define _MN_ADVANCED  8

#define _MN_PANEL  0
#define _MN_DIALOG 1

#define _MN_HOOK_ID_ANY -1
#define _MN_HOOK_COMBOBOX 0
#define _MN_HOOK_MOUSE    1
#define _MN_HOOK_ASSIGN   2
#define _MN_HOOK_SHOW     3
#define _MN_HOOK_CHECKBOX 4
#define _MN_HOOK_COMBOBOX_CHANGE 5 // $BF1$8J8;z$r(BSelect$B$7$?>l9g$r=|30(B
#define _MN_HOOK_BUTTON 6
#define _MN_HOOK_OPEN 7
#define _MN_HOOK_CHANGE 8
#define _MN_HOOK_PANELCLICK 9

#define _MN_NOFORCE FALSE

#define _MN_DLG_MAKEGEOM TRUE
#define _MN_DLG_NOMAKEGEOM FALSE
#define _MN_DLG_MODAL TRUE

#ifdef WIN32
#define _MN_OK_CANCEL_DEFAULT _MN_OK_CANCEL
#else
#define _MN_OK_CANCEL_DEFAULT _MN_CANCEL_OK
#endif

class wxCheckBoxEX;
class MultiNumericalInputs_cls;

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

// $B4X?t%]%$%s%?$r(Bvetctor$B$KF~$l$k$?$a$N9=B$BN(B
struct structFuncPointerVoidVoid{
    void (*mpFunc)(void*, void*);
    int mId;
    int mType;
    void *mpEventOvwt; //$BDL>o$O(BpEvent$B$@$1$I>l9g$K$h$C$F(B
};

class MultiNumericalInputsBase_cls {
 public:
    int mType;
 public:
    MultiNumericalInputsBase_cls(void **ppToSub);
    void **mppToSub;

    int CombinePointerButton
	(int item, wxString title = wxT(""), bool enabled = TRUE);
    int  CombinePointer
	(int type, void* pPointerVal = NULL, wxString title = wxT(""), 
	 bool enabled = TRUE, 
	 unsigned int listMax = 0, wxString *pChoiceList = NULL,
	 unsigned int init = 0, unsigned int offset = 0);
    
    void SetRange(double min, double max, 
		  int index = _MN_LASTITEM, int scale = 1);
    void DisableHookSpace(int index = _MN_LASTITEM);
    void AlignRight(int index = _MN_LASTITEM);

    void MakeGeometryBase(wxBoxSizer *pBoxSizerSrc, wxWindow *pWindow);
    void AdjustTabNavigationOrder();

    virtual void ConnectButton(wxWindow *pWindow, int id) = 0;
    virtual void ConnectComboBox(wxWindow *pWindow, int id) = 0;
    virtual void ConnectCheckBox(wxWindow *pWindow, int id) = 0;
    virtual void ConnectScroll(wxWindow *pWindow, int id) = 0;
    virtual void ConnectClick(wxWindow *pWindow, int id) = 0;

    wxBoxSizer *mpBoxSizerSrc;
    wxSizerItem *mpPadding;

    void SetDefaultButtons(wxWindow* pC, wxBoxSizer *pSiczer);
    wxButtonEX* mpBTNAssign;
    wxButtonEX* mpBTNUpDate;
    wxButtonEX* mpBTNHide; // = CANCEL
    wxButtonEX* mpBTNOK;
    wxButtonEX* mpBTNHelp;
    wxButtonEX* mpBTNAdvanced;
    MultiNumericalInputs_cls* mpDialogAdvanced;
    wxString mstrHelp;

    int  SearchNextTabstop(int index, int direction);
    static void SetTabRing
	(std::vector < MultiNumericalInputsBase_cls* > VpList);
    
    void UpdateValues(int index = _MN_ALLITEM, int exceptionID = _MN_NOITEM, bool bForce = TRUE);
    void UpdateValuesAll(bool bForce = FALSE);
    bool CheckDirty();

    void UpdateFromText(unsigned int index = _MN_ASSIGN_ALL);
    void AssignBase(unsigned int index = _MN_ASSIGN_ALL);
    void AssignBaseById(int id);
    void ShowItem(unsigned int index);
    void HideItem(unsigned int index);
    void ColoringItem(int index, wxColour color);
    void OnScrollBase(wxCommandEvent& event);
    void OnScrollThumbBase(wxCommandEvent& event);
    void OnComboBoxBase(wxCommandEvent& event, void *pThis);

    std::vector < GeneralPointer_cls > mVCPointers;
    GeneralPointer_cls* GetPLastVCPointer();
    unsigned int GetLastVCIndex();
    void SetItemTitle(unsigned int index, wxString title);
    void SetColorForXYZ(unsigned int index);

    unsigned int mSizerIndex;
    unsigned int mSizerCounter;
    bool     mbDirty;
    bool     mbHideButton;

    void     SetDefaultButtonType(int type);
    int      mDefaultButton;
    int      mTextWidth, mButtonWidth, mComboWidth, mCheckBoxWidth;
    int      mItemWidth;
    int      mMinimumWidth;
    int      mProportion;

    float    mCachedFloat;
    int      mCachedInt;
    bool     mCachedBool;
    void CacheOriginalVaues(int index);

    bool     mbAssignHook;
    std::vector < structFuncPointerVoidVoid > mVpStructFuncHook;

    void     SetFuncHook(void (*pFunc)(void*, void*), int type,
			 int ID = wxID_ANY, void *pEventOvwr = NULL);
};

class MultiNumericalInputsPanel_cls : 
public wxPanel, public MultiNumericalInputsBase_cls {
 public:
    MultiNumericalInputsPanel_cls(MultiNumericalInputsPanel_cls **ppToThis,
				  wxWindow *parent, wxWindowID id,
				  const wxPoint& pos  = wxDefaultPosition,
				  const wxSize&  size = wxDefaultSize,
				  long style = wxBORDER_SIMPLE,
				  const wxString& name = wxT("window"));
    
    ~MultiNumericalInputsPanel_cls();
    int mIndexMinimize;
    
    void MakeGeometry();
    int GetPaddingWidth(int minimumWidth);
    void SetPaddingWidth(int paddingWidth);
    
    void ConnectButton(wxWindow *pWindow, int id);
    void ConnectComboBox(wxWindow *pWindow, int id);
    void ConnectCheckBox(wxWindow *pWindow, int id);
    void ConnectScroll(wxWindow *pWindow, int id);
    void ConnectClick(wxWindow *pWindow, int id);

    void OnAssign(wxCommandEvent& event);
    void OnHide(wxCommandEvent& event);
    void OnShow(wxCommandEvent& event);
    void OnUpdate(wxCommandEvent& event);
    void OnOK(wxCommandEvent& event);
    void OnButton(wxCommandEvent& event);
    void OnCheckBox(wxCommandEvent& event);
    void OnComboBox(wxCommandEvent& event);
    void OnScroll(wxCommandEvent& event);
    void OnScrollThumb(wxCommandEvent& event);
    void OnChar(wxKeyEvent& event);
    void OnKeyDown(wxKeyEvent& event);
    void OnMouse(wxMouseEvent& event);
    void OnMouseOnHide(wxMouseEvent& event);
    void OnMouseOnItem(wxMouseEvent& event);
    void OnRgtDownForButton(wxMouseEvent& event);
    void OnChildFocusEventCustum(wxChildFocusEvent& evevt);
    void OnSize(wxSizeEvent& event);
    void OnHelp(wxCommandEvent& event);
    void OnAdvanced(wxCommandEvent& event);

    wxButtonEX* mpBTNMinimize;
    wxCheckBoxEX* mpCheckBoxValid;
    wxBoxSizer* mpHorSizer;
    void AssignMinimizeButtonBoth(wxButtonEX* &pBTNMinimize,
				  int ID, wxString title);
    void SetMinimizeButtonLabel(wxString strLabel);
    void AssignValidCheckBoxBoth(wxCheckBoxEX* &pCheckBoxValid,
				 int ID, bool *pBool);
    void RefreshBoth(bool mode);
    void ShowPanel(bool mode);
    void ShowItems(int index, int num, bool bShow);
    void EnableItems(int index, int num, bool bEnable);
    void SetHilight(bool bHilight = TRUE);
    void AddToSizer(wxBoxSizer *pBoxSizer, int index = -1);
    int  FinishSetting(wxBoxSizer *pBoxSizer, bool vis, int index = -1);
    
    MultiNumericalInputsPanel_cls **mppToThis;

    DECLARE_EVENT_TABLE()
};

// wxDialog $B%/%i%9$N7Q>5$@$H!"B>$N%&%#%s%I%&$GI=<($,1#$l$J$$(B
// wxFrame $B%/%i%9$N7Q>5$@$H!"B>$N%&%#%s%I%&$GI=<($,1#$l$k(B

class MultiNumericalInputs_cls : 
public wxDialogEXKey, public MultiNumericalInputsBase_cls {
 public:
    MultiNumericalInputs_cls(MultiNumericalInputs_cls **ppToThis,
			     wxWindow *parent, wxWindowID id, wxString title,
			     const wxPoint& pos  = wxDefaultPosition,
			     const wxSize&  size = wxDefaultSize,
			     long style = wxDEFAULT_DIALOG_STYLE,
			     const wxString& name = wxT("dialogBox"));
    ~MultiNumericalInputs_cls();
    
    void MakeGeometry();
    void OpenDialogEX();

    void ConnectButton(wxWindow *pWindow, int id);
    void ConnectComboBox(wxWindow *pWindow, int id);
    void ConnectCheckBox(wxWindow *pWindow, int id);
    void ConnectScroll(wxWindow *pWindow, int id);
    void ConnectClick(wxWindow *pWindow, int id);

    void OnAssign(wxCommandEvent& event);
    void OnHide(wxCommandEvent& event);
    void OnUpdate(wxCommandEvent& event);
    void OnOK(wxCommandEvent& event);
    void OnButton(wxCommandEvent& event);
    void OnCheckBox(wxCommandEvent& event);
    void OnComboBox(wxCommandEvent& event);
    void OnScroll(wxCommandEvent& event);
    void OnScrollThumb(wxCommandEvent& event);
    void OnChar(wxKeyEvent& event);
    void OnMouseOnItem(wxMouseEvent& event);
    void OnChildFocusEventCustum(wxChildFocusEvent& evevt);
    void OnSize(wxSizeEvent& event);
    void OnHelp(wxCommandEvent& event);
    void OnAdvanced(wxCommandEvent& event);


    int  FinishSettingDLG(bool vis, bool modal = FALSE, bool bMakeGeom = TRUE);
    
    MultiNumericalInputs_cls **mppToThis;

    wxString mTitle;
    wxPoint  mCPosition;
 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

#endif
