// "clsCameraPath.h"

#include "clsCameraPath.h"

#include "wxWidgetEnum.h"
#include "wxWidgetEnumMenu.h"

#include "clsFrameMain.h"
#include "clsMain.h"
#include "clsTimeCtrl.h"
#include "SRC/COMMON/MyVector.h"
#include "SRC/DF3/clsDF3Sequence.h"
#include "SRC/GLCANVAS/clsMainGLCanvas.h"
extern MainGLCanvas_cls* GpCurrentGLPanel;
#include "SRC/GLCANVAS/clsHitTest.h"
#include "SRC/MYAPI/MsgBox.h"
#include "SRC/MYGL/clsGLGuides.h"
#include "SRC/POVRAY/clsStereoGram.h"
#include "SRC/WIDGETCTRL/clsMultiPointInput.h"
#include "SRC/WIDGETCTRL/clsMultiNumericalInputs.h"
#include "SRC/WIDGETCTRL/EX/clsColourDialogEX.h"
#include "SRC/MYGL/clsFBO.h"
#include "SRC/UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include <fstream>
#include <sstream>
extern wxColourDialogEX* GpCClrDialog;

extern FrameMain_cls* GpCMainFrameCLS;

#define _POV_PERSPECTIVE  0
#define _POV_ORTHOGRAPHIC 1
#define _POV_FISHEYE      2
#define _POV_PANORAMIC    3

#define X 0
#define Y 1
#define Z 2
#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)
#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)

#define ENDLWX wxT("\n");

// $B@EE*$K%$%Y%s%H$r7k$S$D$1$k(B
BEGIN_EVENT_TABLE(CameraPathGUI_cls, wxDialog)
     EVT_SIZE(CameraPathGUI_cls::OnSize)
     EVT_MOUSE_EVENTS(CameraPathGUI_cls::OnMouse)
     EVT_BUTTON    (btnView,    CameraPathGUI_cls::OnButtonShowItemView)
     EVT_BUTTON    (btnViewAdv,    CameraPathGUI_cls::OnButtonShowItemViewAdv)
     EVT_BUTTON    (btnViewPath,    CameraPathGUI_cls::OnButtonShowItemPath)
     EVT_BUTTON    (btnGLPanelBG0,  CameraPathGUI_cls::OnButtonBGColor0)
     EVT_BUTTON    (btnGLPanelBG1,  CameraPathGUI_cls::OnButtonBGColor1)
     EVT_BUTTON    (btnGLPanelAll, CameraPathGUI_cls::OnButtonApplyAll)
     EVT_MENU(menuItemMinmizeButton, CameraPathGUI_cls::OnMenuMinimize)
END_EVENT_TABLE()

CameraPathGUI_cls::
CameraPathGUI_cls(CameraPathCtrl_cls* pCParent,
		  wxWindow *pC, wxPoint pos, wxSize size) :
    wxDialogEXKey(pC, wxID_ANY, _("View and Camera Configure"), 
		  pos, size, pCParent->mpBitmapLink) {
    mspCParent = pCParent;
    mspCThis = this;
    //SetScreenRectPos(&pCParent->mCScreenRect);
    
    wxBoxSizer *pBoxSizerSrc     = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerView    = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerViewAdv = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerPath    = new wxBoxSizer(wxVERTICAL);

    SetMNumView(pBoxSizerView);
    SetMNumViewAdv(pBoxSizerViewAdv);
    SetMNumPath(pBoxSizerPath);

    pBoxSizerSrc->Add(pBoxSizerView, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerViewAdv, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerPath, 0, wxEXPAND);
    
    SetSizerAndFit(pBoxSizerSrc);

    int minimumWidth = GetClientSize().GetWidth() - 8;
    minimumWidth = mpCMNumView->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumViewAdv->GetPaddingWidth(minimumWidth);
    minimumWidth = mpCMNumPath->GetPaddingWidth(minimumWidth);
    mpCMNumView->SetPaddingWidth(minimumWidth);
    mpCMNumViewAdv->SetPaddingWidth(minimumWidth);
    mpCMNumPath->SetPaddingWidth(minimumWidth);

    mpCMNumViewAdv->ShowPanel(FALSE);
    mpCMNumPath->ShowPanel(FALSE);
}

CameraPathGUI_cls::
~CameraPathGUI_cls(){
}

void CameraPathGUI_cls::
SetMNumView(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr = 
	InitializeMNumPanel(&mpCMNumView, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNView, btnView,
				    _("[+]View Setting"));

    int boxWidth = 3;
#ifdef WIN32
    pCurr->mTextWidth = 90;
#else
    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 108;
#endif
    
    mIndexCamPos = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Geometry"));

    wxString strListView[4] = {
	wxT("View0"),wxT("View1"),wxT("View2"), wxT("View3")};
    pCurr->CombinePointer(_GP_UNSIGNED_INT_LIST, 
			  &mspCParent->mActiveView,
			  wxT(""), TRUE, 4, strListView, 1);

    wxString strGeometry[4] = {
	_("Perspective"), wxT("Orthographic"), _("FishEye"), _("Panoramic")};
    int typeSelect = 4 - (FBO_cls::msbSupported == FALSE) * 2;
    mIndexViewGeom = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCGLGeometry.mBody,
	 wxT(""), _GP_ENABLE, typeSelect, strGeometry);

    mIndexCamPos = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Position"));
    pCurr->CombinePointer(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[X]);
    pCurr->CombinePointer(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[Y]);
    pCurr->CombinePointer(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[Z]);
    mIndexAncPos = pCurr->CombinePointer
	(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Anchor"));
    pCurr->CombinePointer(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[X]);
    pCurr->CombinePointer(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Y]);
    pCurr->CombinePointer(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Z]);

    boxWidth = 2;
    mIndexPitchAngle = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
					     _("Pitch Angle"));
    pCurr->CombinePointer
	(_GP_DOUBLE_SLIDER, &mspCParent->GetActivePath()->mpCStereoGramLinked->mCPitchAngle.mBody);
    pCurr->SetRange(-180, 180, _MN_LASTITEM, 1);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mpCStereoGramLinked->mCPitchAngle.mBody);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("View Angle"));
    mIndexAngleView = pCurr->CombinePointer
	(_GP_DOUBLE_SLIDER, &GpCurrentGLPanel->mCGLPerspective.mDeg);
    pCurr->SetRange(1, 170, _MN_LASTITEM, 1);
    pCurr->CombinePointer(_GP_DOUBLE, &GpCurrentGLPanel->mCGLPerspective.mDeg);

#ifdef WIN32
    pCurr->mMinimumWidth = 280;
#else
    pCurr->mMinimumWidth = 320;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetFuncHook(FrameMain_cls::FuncHookParamChanged, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);

    pCurr->SetColorForXYZ(mIndexCamPos+1);
    pCurr->SetColorForXYZ(mIndexAncPos+1);
    
    HideAndShowView();
}

void CameraPathGUI_cls::
SetMNumViewAdv(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr =
	InitializeMNumPanel(&mpCMNumViewAdv, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNViewAdv, btnViewAdv,
				    _("[+]View Advanced Setting"));
#ifdef WIN32
    pCurr->mTextWidth = 90;
    pCurr->mButtonWidth = 80;
#else
    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 108;
    pCurr->mButtonWidth = 90;
#endif

    int boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexBG0 = pCurr->CombinePointerButton(btnGLPanelBG0, _("BG Color 0"));
    pCurr->mVCPointers[mIndexBG0].mpConnectColorFloat
	= GpCurrentGLPanel->mCColorBG0.mpBody;
    mIndexBG1 = pCurr->CombinePointerButton(btnGLPanelBG1, _("Outside Color"));
    pCurr->mVCPointers[mIndexBG1].mpConnectColorFloat
	= GpCurrentGLPanel->mCColorBG1.mpBody;

    pCurr->mButtonWidth = -1;
    pCurr->CombinePointerButton(btnGLPanelAll, _("Apply to All"));
    
    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Grid and Width"));
    wxString strListGrid[3] = {
	_("Grid"), _("Grid+Circle"), _("Polar")};
    mIndexGrid = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCGridType.mBody, 
	 _("Grid Type"), TRUE, 3, strListGrid, 
	 GpCurrentGLPanel->mCGridType.mBody);

    pCurr->CombinePointer(_GP_FLOAT, &MainGLCanvas_cls::msLineWidth, _("Line Width"));

    wxString listFBOSize[5] = {wxT("256x256"), wxT("512x512"), wxT("1024x1024"), wxT("2048x2048"), wxT("4096x4096")};
    mIndexFBOSize = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCFBOSizeIndex.mBody,
	 _("FBO Size"), _GP_ENABLE, 5, listFBOSize);

    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexAntiAlias = pCurr->CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbGridAnti.mbBody, _("AntiAlias"));
    mIndexCoordinate = pCurr->CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbShowGridTag.mbBody, _("Show Coordinate"));

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexShowOtherCam = pCurr->CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbShowOtherCam.mbBody, _("view point"));
    mIndexXOR = 
	pCurr->CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbXOR.mbBody, _("Draw item in XOR"));
    
    
#ifdef WIN32
    pCurr->mMinimumWidth = 280;
#else
    pCurr->mMinimumWidth = 320;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);
}

void CameraPathGUI_cls::
ReconnectView(){
    mspCParent->mActiveView = GpCurrentGLPanel->mSelfIndex; 
    
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumView;

    pCurr->mVCPointers[mIndexCamPos+1].CombinePointer
	(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[X]);
    pCurr->mVCPointers[mIndexCamPos+2].CombinePointer
	(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[Y]);
    pCurr->mVCPointers[mIndexCamPos+3].CombinePointer
	(_GP_DOUBLE, &GpCurrentGLPanel->mCCameraPos.mPos[Z]);

    pCurr->mVCPointers[mIndexAncPos+1].CombinePointer
	(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[X]);
    pCurr->mVCPointers[mIndexAncPos+2].CombinePointer
	(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Y]);
    pCurr->mVCPointers[mIndexAncPos+3].CombinePointer
	(_GP_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Z]);

    pCurr->mVCPointers[mIndexViewGeom].CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCGLGeometry.mBody,
	 _("Preview Setting"));

    int angleMax = 170;
    if (GpCurrentGLPanel->mCGLGeometry.mBody != 0) angleMax = 180;
    pCurr->mVCPointers[mIndexAngleView].CombinePointer
	(_GP_DOUBLE_SLIDER, &GpCurrentGLPanel->mCGLPerspective.mDeg);
    pCurr->mVCPointers[mIndexAngleView+1].CombinePointer
	(_GP_DOUBLE, &GpCurrentGLPanel->mCGLPerspective.mDeg);
    pCurr->SetRange(1, angleMax, mIndexAngleView, 1);
    pCurr->SetRange(1, angleMax, mIndexAngleView+1, 1);
    pCurr->mVCPointers[mIndexAngleView].ApplyRange();
    pCurr->mVCPointers[mIndexAngleView+1].ApplyRange();

    if (GpCurrentGLPanel->mCGLPerspective.mDeg > angleMax){
	GpCurrentGLPanel->mCGLPerspective.mDeg = angleMax;
    }

    pCurr->UpdateValues();

    HideAndShowAll();
}

void CameraPathGUI_cls::
ReconnectViewAdv(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumViewAdv;

    pCurr->mVCPointers[mIndexBG0].mpConnectColorFloat
	= GpCurrentGLPanel->mCColorBG0.mpBody;
    pCurr->mVCPointers[mIndexBG1].mpConnectColorFloat
	= GpCurrentGLPanel->mCColorBG1.mpBody;
    
    pCurr->mVCPointers[mIndexGrid].CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCGridType.mBody,
	 _("Grid Type"));

    pCurr->mVCPointers[mIndexAntiAlias].CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbGridAnti.mbBody);
    pCurr->mVCPointers[mIndexCoordinate].CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbShowGridTag.mbBody);
    pCurr->mVCPointers[mIndexShowOtherCam].CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbShowOtherCam.mbBody);
    pCurr->mVCPointers[mIndexXOR].CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbXOR.mbBody);

    pCurr->mVCPointers[mIndexFBOSize].CombinePointer
	(_GP_UNSIGNED_INT_LIST, &GpCurrentGLPanel->mCFBOSizeIndex.mBody,
	 _("FBO Size"));
    
    pCurr->UpdateValues();
}

void CameraPathGUI_cls::
FuncHookRefresh(void *pArg, void *pEvent){
    if ((MultiNumericalInputsPanel_cls*)pArg == mspCThis->mpCMNumView){
	mspCParent->mpCGUI->ReconnectView();
	mspCParent->mpCGUI->ReconnectViewAdv(); 
    }
    if ((MultiNumericalInputsPanel_cls*)pArg == mspCThis->mpCMNumPath)
	mspCParent->mpCGUI->ReconnectPath();

    mspCThis->HideAndShowAll();

    wxCommandEvent comRefreshTL(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshTL);
    mspCThis->GetParent()->AddPendingEvent(comRefreshTL);

    wxCommandEvent comRefresh(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mspCThis->GetParent()->AddPendingEvent(comRefresh);
}

void CameraPathGUI_cls::
FuncHookButton(void *pArg, void *pEvent){
    if (((wxCommandEvent*)pEvent)->GetId() == buttonCtrlAddPath)
	mspCParent->AddPath();
    if (((wxCommandEvent*)pEvent)->GetId() == buttonCtrlDeletePath)
	mspCParent->DeletePath();
}

void CameraPathGUI_cls::
HideAndShowView(){
    MultiNumericalInputsPanel_cls* pCurr = (MultiNumericalInputsPanel_cls*)mpCMNumView;

    bool bPitchAngle = (GpCurrentGLPanel->mCGLGeometry.mBody == _POV_FISHEYE);
    pCurr->ShowItems(mIndexPitchAngle, 3, bPitchAngle);

    pCurr->GetSizer()->Layout(); 
}

void CameraPathGUI_cls::
HideAndShowPath(){
    MultiNumericalInputsPanel_cls* pCurr = (MultiNumericalInputsPanel_cls*)mpCMNumPath;
    bool bPointSky = mspCParent->GetActivePath()->mCbSky.mbBody;
    pCurr->ShowItems(mIndexSky-1, 4, bPointSky);

    pCurr->GetSizer()->Layout();	
}

void CameraPathGUI_cls::
HideAndShowAll(){
    HideAndShowView();
    HideAndShowPath();

    mspCThis->ResizeDialog();
}

void CameraPathGUI_cls::
SetMNumPath(wxBoxSizer *pBoxSizer){
    MultiNumericalInputsPanel_cls* pCurr =
	InitializeMNumPanel(&mpCMNumPath, this);
    pCurr->AssignMinimizeButtonBoth(mpCBTNPath, btnViewPath,
				    _("[+]Camera Path Setting"));

    int boxWidth = 3;
    pCurr->mDefaultButton = _MN_NO_BUTTONS;
#ifdef WIN32
    pCurr->mTextWidth = 90;
#else
    pCurr->mTextWidth = 100;
    pCurr->mComboWidth = 108;
#endif
    pCurr->mButtonWidth = 60;

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Active Cam"));
    wxString strListView[1] = {wxT("Cam0")};
    mIndexActivePath = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mActivePath,
	 _("Active Path"), TRUE, 1, strListView, 0);
    pCurr->CombinePointerButton(buttonCtrlAddPath, wxT("Add"));
    pCurr->CombinePointerButton(buttonCtrlDeletePath, wxT("Delete"));

    wxString strGeometry[4] = {
	_("Perspective"), wxT("Orthographic"), _("FishEye"), _("Panoramic")};
    int typeSelect = 4 - (FBO_cls::msbSupported == FALSE) * 2;
    mIndexGeom = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, 
	 &mspCParent->GetActivePath()->mCCamGeometry.mBody,
	 _("Camera Geom."), _GP_ENABLE, typeSelect, strGeometry);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("View Angle"));
    mIndexAnglePath = pCurr->CombinePointer
	(_GP_DOUBLE_SLIDER, &mspCParent->GetActivePath()->mCVAngle.mBody);
    pCurr->SetRange(1, 170, _MN_LASTITEM, 1);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCVAngle.mBody);
    
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Position"));
    mIndexPosition = pCurr->CombinePointer
	(_GP_DOUBLE, 
	 &mspCParent->GetActivePath()->mCPosition.mpInterpolatedValue[X]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCPosition.mpInterpolatedValue[Y]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCPosition.mpInterpolatedValue[Z]);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Anchor"));
    mIndexAnchor = pCurr->CombinePointer
	(_GP_DOUBLE, 
	 &mspCParent->GetActivePath()->mCAncPos.mpInterpolatedValue[X]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCAncPos.mpInterpolatedValue[Y]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCAncPos.mpInterpolatedValue[Z]);

    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Y/P/R"));
    mIndexYPR = pCurr->CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCRotYPR.mpInterpolatedValue[0]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCRotYPR.mpInterpolatedValue[1]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCRotYPR.mpInterpolatedValue[2]);

    boxWidth = 1;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Point Sky"));
    mIndexPointSky = pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->GetActivePath()->mCbSky.mbBody);
    //pCurr->CombinePointer(_GP_STATIC_TEXT, NULL);
    //pCurr->CombinePointer(_GP_STATIC_TEXT, NULL);
    
    boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Sky Vector"));
    mIndexSky = pCurr->CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[0]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[1]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[2]);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 280;
#else
    pCurr->mMinimumWidth = 320;
#endif
    pCurr->FinishSetting(pBoxSizer, TRUE);

    pCurr->SetColorForXYZ(mIndexPosition);
    pCurr->SetColorForXYZ(mIndexAnchor);

    pCurr->SetFuncHook(FrameMain_cls::FuncHookParamChanged, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookButton, _MN_HOOK_BUTTON, _MN_HOOK_ID_ANY);
}

void CameraPathGUI_cls::
OnMouse(wxMouseEvent &event){
    if (event.RightDown() == TRUE){
	MACRO_PopupMinimizeButton(this,View);
	MACRO_PopupMinimizeButton(this,ViewAdv);
	MACRO_PopupMinimizeButton(this,Path);
    }
    event.Skip();
}

void CameraPathGUI_cls::
UpdateComboboxList(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumPath;

    wxString pStrListPath[_PATH_NUM_MAX];
    for (int i = 0; i < mspCParent->mCCameraNum.mBody; i++){
	pStrListPath[i] << wxT("Cam") << i;
    }
    pCurr->mVCPointers[mIndexActivePath].CombinePointer
	(_GP_UNSIGNED_INT_LIST, NULL,
	 _("Active Path"), TRUE, mspCParent->mCCameraNum.mBody, pStrListPath);
}

void CameraPathGUI_cls::
OnMenuMinimize(wxCommandEvent &event){
    MACRO_CallMinimizeButtonPath(this, View, _("[+]View Setting"));
    MACRO_CallMinimizeButtonPath(this, ViewAdv, _("[+]View Advanced Setting"));
    MACRO_CallMinimizeButtonPath(this, Path, _("[+]Camera Path Setting"));
    ResizeDialog();
    Refresh(FALSE);
}

#define MACRO_UpdateBTNMinimize(BTNName, LABEL)\
    if (mspCParent->mCConfBTN##BTNName.mbBody == TRUE){	\
    mspCParent->mCConfBTN##BTNName.mbBody = FALSE;	\
    MinimizeButton(mpCBTN##BTNName, \
		   &mspCParent->mCConfBTN##BTNName.mbBody, LABEL,\
		   (wxButtonEX*)mpCMNum##BTNName->mVCPointers[0].mpPointerCtrl); }

void CameraPathGUI_cls::
UpdateButtonMinimize(){
    MACRO_UpdateBTNMinimize(View, _("[+]View Setting"));
    MACRO_UpdateBTNMinimize(ViewAdv, _("[+]View Setting"));
    MACRO_UpdateBTNMinimize(Path, _("[+]View Setting"));
}

void CameraPathGUI_cls::
ReconnectPath(){
    MultiNumericalInputsPanel_cls* pCurr = mpCMNumPath;
    
    pCurr->mbAssignHook = FALSE;
    mspCParent->GetActivePath()->
	GetSetCurrentPos(GpCMainFrameCLS->mpCParent->mpCTime->GetDoubleIndex());
    
    wxString strDomeType[4] = {
	_("Perspective"), wxT("Orthographic"), _("FishEye"), _("Panoramic")};
    int typeSelect = 4 - (FBO_cls::msbSupported == FALSE) * 2;
    pCurr->mVCPointers[mIndexGeom].CombinePointer
	(_GP_UNSIGNED_INT_LIST, 
	 &mspCParent->GetActivePath()->mCCamGeometry.mBody,
	 _("Camera Geom."), _GP_ENABLE, typeSelect, strDomeType);

    int angleMax = 170;
    if (mspCParent->GetActivePath()->mCCamGeometry.mBody != 0) angleMax = 180;
    pCurr->mVCPointers[mIndexAnglePath].CombinePointer
	(_GP_DOUBLE_SLIDER, &mspCParent->GetActivePath()->mCVAngle.mBody);
    pCurr->mVCPointers[mIndexAnglePath+1].CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCVAngle.mBody);
    pCurr->SetRange(1, angleMax, mIndexAnglePath, 1);
    pCurr->SetRange(1, angleMax, mIndexAnglePath+1, 1);
    pCurr->mVCPointers[mIndexAnglePath].ApplyRange();
    pCurr->mVCPointers[mIndexAnglePath+1].ApplyRange();

    if (mspCParent->GetActivePath()->mCVAngle.mBody > angleMax){
	mspCParent->GetActivePath()->mCVAngle.mBody = angleMax;
    }
    
    LOOP_XYZ
	pCurr->mVCPointers[mIndexPosition + xyz].CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCPosition.mpInterpolatedValue[xyz]);
    LOOP_XYZ
	pCurr->mVCPointers[mIndexAnchor + xyz].CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCAncPos.mpInterpolatedValue[xyz]);
    LOOP_XYZ
	pCurr->mVCPointers[mIndexYPR + xyz].CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCRotYPR.mpInterpolatedValue[xyz]);
    LOOP_XYZ
	pCurr->mVCPointers[mIndexSky + xyz].CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[xyz]);
    
    pCurr->mVCPointers[mIndexPointSky].CombinePointer
	(_GP_BOOL, &mspCParent->GetActivePath()->mCbSky.mbBody);

    pCurr->UpdateValues(_MN_ALLITEM, 2);
    pCurr->mbAssignHook = TRUE;
}

void CameraPathGUI_cls::
OnSize(wxSizeEvent &event){
    event.Skip();
    SetSizerAndFit(GetSizer());
}

void CameraPathGUI_cls::
OnButtonShowItemView(wxCommandEvent &event){
    mpCMNumView->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void CameraPathGUI_cls::
OnButtonShowItemViewAdv(wxCommandEvent &event){
    mpCMNumViewAdv->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void CameraPathGUI_cls::
OnButtonShowItemPath(wxCommandEvent &event){
    mpCMNumPath->ShowPanel(TRUE);
    ResizeDialog();
    Refresh(FALSE);
}

void CameraPathGUI_cls::
OnButtonBGColor0(wxCommandEvent& event){
    GpCClrDialog->SetColour(GpCurrentGLPanel->GetColourBG(0));
    if (GpCClrDialog->ShowModal() == wxID_OK)
	GpCurrentGLPanel->
	    ChangeBGColor(GpCClrDialog->GetPColorFloat(), 0);
    
    mpCMNumViewAdv->UpdateValues();
    GpCurrentGLPanel->mCDrawInfo.mbForceNewDraw = TRUE;
}

void CameraPathGUI_cls::
OnButtonBGColor1(wxCommandEvent& event){
    GpCClrDialog->SetColour(GpCurrentGLPanel->GetColourBG(1));
    if (GpCClrDialog->ShowModal() == wxID_OK)
	GpCurrentGLPanel->
	    ChangeBGColor(GpCClrDialog->GetPColorFloat(), 1);
    
    mpCMNumViewAdv->UpdateValues();
    GpCurrentGLPanel->mCDrawInfo.mbForceNewDraw = TRUE;
}

void CameraPathGUI_cls::
OnButtonApplyAll(wxCommandEvent& event){
    GpCurrentGLPanel->CopySettingToAll();
}

void CameraPathGUI_cls::
FuncHookAssign(void *pArg, void *pEvent){
    MultiNumericalInputsPanel_cls *pCurr = (MultiNumericalInputsPanel_cls*)pArg;
    GeneralPointer_cls *pPoint = (GeneralPointer_cls*)pEvent;
    CameraPath_cls *pCActivePath = mspCParent->GetActivePath();
    
    // AD_HOC!! $BHs>o$K2s$j$/$I$$=hM}(B
    // $B$J$<$+F~NOMs$NCM$,$:$l$k>I>u$,$"$k$N$r2sHr$9$k$?$a$K!"(B
    // $BF~NO$N$"$C$?Ms$@$1$r=hM}!#2?=h$+$K%P%0$,@x$s$G$$$k!D(B
    bool bChanged = FALSE;
    double doubleIndex = GpCMainFrameCLS->mpCParent->mpCTime->GetDoubleIndex();
    double pos[3], posInput[3];
    double anc[3], ancInput[3];
    double YPR[3], YPRInput[3];
    double Sky[3], SkyInput[3];

    LOOP_XYZ
	posInput[xyz] = pCurr->mVCPointers[mspCThis->mIndexPosition + xyz].GetValueDouble();
    LOOP_XYZ
	ancInput[xyz] = pCurr->mVCPointers[mspCThis->mIndexAnchor + xyz].GetValueDouble();
    LOOP_XYZ
	YPRInput[xyz] = pCurr->mVCPointers[mspCThis->mIndexYPR + xyz].GetValueDouble();
    LOOP_XYZ
	SkyInput[xyz] = pCurr->mVCPointers[mspCThis->mIndexSky + xyz].GetValueDouble();
    
    pCActivePath->mCPosition.GetPInterpolatedValue(doubleIndex);
    pCActivePath->mCAncPos.GetPInterpolatedValue(doubleIndex);
    pCActivePath->mCRotYPR.GetPInterpolatedValue(doubleIndex);
    pCActivePath->mCSky.GetPInterpolatedValue(doubleIndex);
    
    LOOP_XYZ
	if (pPoint == &pCurr->mVCPointers[mspCThis->mIndexPosition + xyz])
	    pos[xyz] = posInput[xyz];
	else
	    pos[xyz] = pCActivePath->mCPosition.mpInterpolatedValue[xyz];
    LOOP_XYZ
	if (pPoint == &pCurr->mVCPointers[mspCThis->mIndexAnchor + xyz])
	    anc[xyz] = ancInput[xyz];
	else
	    anc[xyz] = pCActivePath->mCAncPos.mpInterpolatedValue[xyz];
    LOOP_XYZ
	if (pPoint == &pCurr->mVCPointers[mspCThis->mIndexYPR + xyz])
	    YPR[xyz] = YPRInput[xyz];
	else
	    YPR[xyz] = pCActivePath->mCRotYPR.mpInterpolatedValue[xyz];
    LOOP_XYZ
	if (pPoint == &pCurr->mVCPointers[mspCThis->mIndexSky + xyz])
	    Sky[xyz] = SkyInput[xyz];
	else
	    Sky[xyz] = pCActivePath->mCSky.mpInterpolatedValue[xyz];
    
    LOOP_XYZ pCActivePath->mCPosition.mpInterpolatedValue[xyz] = pos[xyz];
    LOOP_XYZ pCActivePath->mCAncPos.mpInterpolatedValue[xyz] = anc[xyz];
    LOOP_XYZ pCActivePath->mCRotYPR.mpInterpolatedValue[xyz] = YPR[xyz];
    LOOP_XYZ pCActivePath->mCSky.mpInterpolatedValue[xyz] = Sky[xyz];
    bChanged |= pCActivePath->mCPosition.SetKeyIfChanged(doubleIndex);
    bChanged |= pCActivePath->mCAncPos.SetKeyIfChanged(doubleIndex);
    bChanged |= pCActivePath->mCRotYPR.SetKeyIfChanged(doubleIndex);
    bChanged |= pCActivePath->mCSky.SetKeyIfChanged(doubleIndex);

    if (bChanged == TRUE){
	pCActivePath->
	    GetSetCurrentPos(GpCMainFrameCLS->mpCParent->mpCTime->GetDoubleIndex());
	GpCMainFrameCLS->mbPathChanged = TRUE;
	GpCMainFrameCLS->Refresh(FALSE);
	
	mspCParent->UpdateItems();
    }
}

CameraPathCtrl_cls *CameraPathGUI_cls::mspCParent;
CameraPathGUI_cls *CameraPathGUI_cls::mspCThis;

// -------------------------------------

CameraPathCtrl_cls::
CameraPathCtrl_cls(FrameMain_cls* pCParentFrame, StereoGram_cls* pCStereoGram,
		   wxBitmap *pBitmap) : 
mVpCPath(_PATH_NUM_MAX)
{
    mpCGUI  = NULL;
    mpCParentFrame = pCParentFrame;
    mpBitmapLink = pBitmap;
    // Undo$B$,$i$_$G%"%I%l%9$NJQ2=$7$J$$(BParameter_cls$B$,:G=i$+$i$$$J$$$H$$$1$J$$!#(B
    for (unsigned int i = 0; i < _PATH_NUM_MAX; i++){
	mVpCPath[i] = new CameraPathGLGUI_cls();
	mVpCPath[i]->LinkStereoGramCLS(pCStereoGram);
    }
    mActiveView = 0;
    mActivePath = 0;
    mCConfList.Clear();
    mCConfList.Add(&mCCameraNum, wxT("CameraNum="), L"1");

    mCConfListLayout.Clear();
    mCConfListLayout.Add(&mCConfBTNView, wxT("CamMinBTNView="), L"0");
    mCConfListLayout.Add(&mCConfBTNViewAdv, wxT("CamMinBTNViewAdv="), L"0");
    mCConfListLayout.Add(&mCConfBTNPath, wxT("CamMinBTNViewPath="), L"0");

    // mCCameraNum.mBody = 1;

    // mspCStereoLinked = pCStereoGram;
}

CameraPathCtrl_cls::
~CameraPathCtrl_cls(){
    if (mpCGUI != NULL)
	delete mpCGUI;
    for (unsigned int i = 0; i < _PATH_NUM_MAX; i++)
	delete mVpCPath[i];
}

CameraPath_cls* CameraPathCtrl_cls::
GetActivePath(){
    return mVpCPath[mActivePath];
}

CameraPathGLGUI_cls* CameraPathCtrl_cls::
GetActivePathGLGUI(){
    return mVpCPath[mActivePath];
}

void CameraPathCtrl_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new CameraPathGUI_cls(this, pC,
				   mCScreenRect.GetPosition(),
				   mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void CameraPathCtrl_cls::
UpdateItems(bool bForce){
    if (mpCGUI == NULL) return;

    double doubleIndex = GpCMainFrameCLS->mpCParent->mpCTime->GetDoubleIndex();
    GetActivePath()->mCPosition.GetPInterpolatedValue(doubleIndex);
    mpCGUI->mpCMNumPath->UpdateValues();
    mpCGUI->mpCMNumView->UpdateValues();
    mpCGUI->mpCMNumViewAdv->UpdateValues();
}

void CameraPathCtrl_cls::
ClearPath(){
    mCCameraNum.mBody = 0;
    mActivePath = 0;
}

void CameraPathCtrl_cls::
AddPath(){
    GCUndoList.SetNewCommand
	(new ComCameraAdd_cls(this, mCCameraNum.mBody, NULL),
	 new ComCameraDelete_cls(this, mCCameraNum.mBody));
    GCUndoList.ExecCurrent();
}

void CameraPathCtrl_cls::
AddPathExecute(int index, CameraPathGLGUI_cls *pCPathGLGUI, bool bRefresh){
    for (int i = _PATH_NUM_MAX - 1; index > index; i--)
	mVpCPath[i]->CopyFromP(mVpCPath[i-1]);
    if (pCPathGLGUI == NULL)
	mVpCPath[index]->InitializeCamera();
    else
	mVpCPath[index]->CopyFromP(pCPathGLGUI);
    mCCameraNum.mBody++;

    if (bRefresh == TRUE){
	mpCGUI->UpdateComboboxList();
    }
    
    wxCommandEvent comRefreshVL(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshVList);
    mpCParentFrame->AddPendingEvent(comRefreshVL);
    wxCommandEvent comRefresh(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mpCParentFrame->AddPendingEvent(comRefresh);

}

void CameraPathCtrl_cls::
DeletePath(){
    if (mCCameraNum.mBody == 1) return;
    GCUndoList.SetNewCommand
	(new ComCameraDelete_cls(this, mActivePath),
	 new ComCameraAdd_cls(this, mActivePath, mVpCPath[mActivePath]));
    GCUndoList.ExecCurrent();
}

void CameraPathCtrl_cls::
DeletePathExecute(int index){
    mpCParentFrame->DeselectAllIf(wxT("CamPos"));
    mpCParentFrame->DeselectAllIf(wxT("AncPos"));

    for (int i = index; i < _PATH_NUM_MAX - 2; i++)
	mVpCPath[i]->CopyFromP(mVpCPath[i+1]);
    mCCameraNum.mBody--;
    
    if (mActivePath >= index) mActivePath--;
    mpCGUI->ReconnectPath();
   
    mpCGUI->UpdateComboboxList();
    wxCommandEvent comRefreshVL(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshVList);
    mpCGUI->GetParent()->AddPendingEvent(comRefreshVL);
    wxCommandEvent comRefreshTL(wxEVT_COMMAND_MENU_SELECTED, menuItemRefreshTL);
    mpCGUI->GetParent()->AddPendingEvent(comRefreshTL);
    wxCommandEvent comRefresh(wxEVT_COMMAND_MENU_SELECTED, menuItemRefresh);
    mpCGUI->GetParent()->AddPendingEvent(comRefresh);
}

void CameraPathCtrl_cls::
DrawPath(PositionAndDirection_cls CViewPos){
    for (int i = 0; i < mCCameraNum.mBody; i++)
	if (CViewPos == mVpCPath[i]->mCCurrentCam)
	    continue;
	else
	    mVpCPath[i]->DrawPath();
}

wxString CameraPathCtrl_cls::
Serialize(bool flag){
    if (flag == _FM_DEFAULT_CONF)
	return mCConfListLayout.SerializeAll();
    
    wxString strOut;
    for (int i = 0; i < mCCameraNum.mBody; i++)
	strOut << mVpCPath[i]->Serialize();
    strOut << mCConfList.SerializeAll();
    return strOut;
}

void CameraPathCtrl_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"--CameraPath");
    if (loc != std::string::npos){
	AddPathExecute(mCCameraNum.mBody, NULL, FALSE);
	mVpCPath[mCCameraNum.mBody - 1]->ReadStrLine(istream, strLine);
    }
    mCConfList.ReadStrLineAll(strLine);
    mCConfListLayout.ReadStrLineAll(strLine);
}

//StereoGram_cls* CameraPathCtrl_cls::mspCStereoLinked;

// -------------------------------------

CameraPath_cls::
CameraPath_cls():
    mCPosition(3, L"CamPos"), mCTop(3, L"CamTop"), mCSky(3, L"CamSky"),
    mCAncPos(3, L"AncPos"), mCRotYPR(3, L"CamRotYPR")
{
    InitializeCamera();
}

void CameraPath_cls::
InitializeCamera(){
    mCPosition.mbUpdateCamera = TRUE;
    mCTop.mbUpdateCamera = TRUE;
    mCAncPos.mbUpdateCamera = TRUE;
    mCPosition.mCbTimeDependent.mbBody = mCTop.mCbTimeDependent.mbBody = TRUE;
    double initPos[3] = {0, 0, 1.5};
    mCPosition.ExecCommandInitSet(initPos);
    double initTop[3] = {0, 1, 0};
    mCTop     .ExecCommandInitSet(initTop);
    double initZero[3] = {0, 0, 0};
    mCAncPos  .ExecCommandInitSet(initZero);
    mCPosition.SetSlaveParam(&mCTop);
    double initSky[3] = {0, 0, 100};
    mCSky     .ExecCommandInitSet(initSky);
    mCRotYPR  .ExecCommandInitSet(initZero);
    mCurrentAnc[0] = mCurrentAnc[1] = mCurrentAnc[2] = 0;
    mCConfList.Clear();
    mCConfList.Add(&mCbSky, wxT("CameraUseSky="), L"0");
    mCConfList.Add(&mCCamGeometry, wxT("CameraGeometry="), L"0");
    mCConfList.Add(&mCVAngle, wxT("CameraVAngle="), L"45");
    //mCbSky.SetItem(wxT("CameraUseSky="), L"0");
}

CameraPath_cls::
CameraPath_cls(const CameraPath_cls& obj) : 
    mCPosition(obj.mCPosition),
    mCTop(obj.mCTop),
    mCSky(obj.mCSky),
    mCAncPos(obj.mCAncPos),
    mCRotYPR(obj.mCRotYPR)
{
    mpCStereoGramLinked = obj.mpCStereoGramLinked;
    mCCurrentCam = obj.mCCurrentCam;
    LOOP_XYZ{
	mCurrentAnc[xyz] = obj.mCurrentAnc[xyz];
	mCurrentFocus[xyz] = obj.mCurrentFocus[xyz];
    }
    mDistanceToFocus = obj.mDistanceToFocus;

    mCbSky = obj.mCbSky;
    mCCamGeometry = obj.mCCamGeometry;

    ReDefineList();
}

void CameraPath_cls::
ReDefineList(){
    mCConfList.Clear();
    mCConfList.Add(&mCbSky, wxT("CameraUseSky="));
    mCConfList.Add(&mCCamGeometry, wxT("CameraGeometry="));
    mCConfList.Add(&mCVAngle, wxT("CameraVAngle="));
}

CameraPath_cls::
~CameraPath_cls(){
    //if (mpCMNumCamConfigure != NULL) delete mpCMNumCamConfigure;
}

void CameraPath_cls::
CopyFromP(CameraPath_cls *pCObj){
    mCPosition.CopyFromP(&pCObj->mCPosition);
    mCTop     .CopyFromP(&pCObj->mCTop);
    mCAncPos  .CopyFromP(&pCObj->mCAncPos);
    mCRotYPR  .CopyFromP(&pCObj->mCRotYPR);
    mCSky     .CopyFromP(&pCObj->mCSky);
    
    mCbSky.mbBody = pCObj->mCbSky.mbBody;
    mCCamGeometry.mBody = pCObj->mCCamGeometry.mBody;
    mCVAngle.mBody = pCObj->mCVAngle.mBody;
}

void CameraPath_cls::
LinkStereoGramCLS(StereoGram_cls* pCStereoGram){
    mpCStereoGramLinked = pCStereoGram;
}

void CameraPath_cls::
GetSetCurrentPos(double doubleIndex){
    mCurrentDoubleIndex = doubleIndex;
    double *pPos = mCPosition.GetPInterpolatedValue(doubleIndex);
    double *pTop = mCTop.GetPInterpolatedValue(doubleIndex);
    double *pSky = mCSky.GetPInterpolatedValue(doubleIndex);
    double *pAnc = mCAncPos.GetPInterpolatedValue(doubleIndex);
    double *pYPR = mCRotYPR.GetPInterpolatedValue(doubleIndex);
    mDistanceToFocus = mpCStereoGramLinked->GetDistanceToFocus(doubleIndex);
    LOOP_XYZ{
	mCCurrentCam.mPos[xyz] = pPos[xyz];
	if (mCbSky.mbBody == TRUE)
	    mCCurrentCam.mTop[xyz] = pSky[xyz] - pPos[xyz];
	else
	    mCCurrentCam.mTop[xyz] = pTop[xyz];
	mCurrentAnc[xyz] = pAnc[xyz];
    }
    mCCurrentCam.PointAt(pAnc);

    double relativeNorm = MyVector::GetNormRelative(pPos, pAnc);
    LOOP_XYZ{
	mCurrentFocus[xyz] = pPos[xyz] + 
	    mDistanceToFocus * (pAnc[xyz] - pPos[xyz]) / relativeNorm;
    }
    
    mCCurrentCam.RotateYPR3d(pYPR);
    return;
}

void CameraPath_cls::
ModifyStereogramPos(PositionAndDirection_cls &CPOS){
    if (mpCStereoGramLinked->mCType.mBody == _STEREO_NONE) return;
    LOOP_XYZ CPOS.mPos[xyz] +=
	0.5 * mpCStereoGramLinked->GetOffsetRight() * CPOS.mRgt[xyz];
};

double CameraPath_cls::
GetStereogramYawAngle(){
    return mpCStereoGramLinked->GetStereogramYawAngle();
}

double CameraPath_cls::
GetDomePitchAngle(){
    return mpCStereoGramLinked->GetDomePitchAngle();
}

void CameraPath_cls::
RotateAroundSelectedKey(double dRot){
    GCUndoList.OpenBlock();
    size_t keySize = mCPosition.mVCKeys.size();

    PositionAndDirection_cls CTmp;
    for (size_t i = 0; i < keySize; i++){
	if (mCPosition.mVCKeys[i].mbSelected == FALSE) continue;

	double doubleIndex = mCPosition.GetKeyFrameDoubleIndex(i);
	double axis[3], top[3];
	LOOP_XYZ axis[xyz] = 
	    mCAncPos.GetPInterpolatedValue(doubleIndex)[xyz] -
	    mCPosition.GetPInterpolatedValue(doubleIndex)[xyz];
	MyVector::Normalize(axis, 1);
	
	LOOP_XYZ top[xyz] = mCTop.GetPKeyFrameValue(i)[xyz];
	MyVector::RotateAround(top, axis, dRot);

	/*
	top[X] = 0;
	top[Y] = 1;
	top[Z] = 0;
	*/

	KeyFrame_cls CKey(mCTop.mVCKeys[i]);
	CKey.SetValue(top);
	mCTop.SetKey(CKey);
	mCTop.mVCKeys[i].mbSelected = TRUE;
    }
    
    GCUndoList.CloseBlock();
}

void CameraPath_cls::
DeleteKeyPos(double doubleIndex){
    mCPosition.DeleteKey(doubleIndex);
    mCTop.DeleteKey(doubleIndex);
}

void CameraPath_cls::
InsertKeyPos(double doubleIndex){
    mCPosition.InsertKey(doubleIndex);
    mCTop.InsertKey(doubleIndex);
}

void CameraPath_cls::
SetKeyPos(double doubleIndex, PositionAndDirection_cls CPOS){

    KeyFrame_cls 
	CKeyPos(mCPosition.mCDimension.mBody),
	CKeyTop(mCTop.mCDimension.mBody);
    mCPosition.SetKeyToValueAt(&CKeyPos, doubleIndex);
    mCTop.SetKeyToValueAt(&CKeyTop, doubleIndex);

    CKeyPos.SetValue(CPOS.mPos);
    CKeyTop.SetValue(CPOS.mTop);

    mCPosition.SetKey(CKeyPos, CKeyTop);
}

void CameraPath_cls::
SetKeyAnc(double doubleIndex, Position_cls CAnc){
    KeyFrame_cls CKeyAnc(mCAncPos.mCDimension.mBody);//, CKeySlave(0);
    mCAncPos.SetKeyToValueAt(&CKeyAnc, doubleIndex);
    CKeyAnc.SetValue(CAnc.mPos);

    mCAncPos.SetKey(CKeyAnc);
}

void CameraPath_cls::
NormalizeTop(){
    GCUndoList.OpenBlock();
    for (unsigned int i = 0; i < mCPosition.GetKeyNumber(); i++){
	Direction_cls CDir;
	double doubleIndex = mCPosition.mVCKeys[i].mDoubleIndex;
	LOOP_XYZ CDir.mTgt[xyz] = 
	    mCPosition.GetPKeyFrameValue(i)[xyz] - mCAncPos.GetPInterpolatedValue(doubleIndex)[xyz];
	LOOP_XYZ CDir.mTop[xyz] = 
	    mCTop.GetPKeyFrameValue(i)[xyz];
	CDir.NormalizeAndOrthogonalize();

	bool bChangeValue = FALSE;
	LOOP_XYZ if (mCTop.GetPKeyFrameValue(i)[xyz] != CDir.mTop[xyz])
	    bChangeValue = TRUE;

	if (bChangeValue == TRUE)
	    mCTop.InsertKey(doubleIndex, CDir.mTop);
    }
    GCUndoList.CloseBlock();
}

void CameraPath_cls::
ApplySky(PositionAndDirection_cls* pCPos){
    if (mCbSky.mbBody == FALSE) return;
    LOOP_XYZ pCPos->mTop[xyz] = 
	mCSky.mpInterpolatedValue[xyz] - pCPos->mPos[xyz];
    pCPos->NormalizeAndOrthogonalize();
}

void CameraPath_cls::
ApplySky(){
    if (mCbSky.mbBody == FALSE) return;
    LOOP_XYZ mCTop.mpInterpolatedValue[xyz] = 
	mCSky.mpInterpolatedValue[xyz] - mCPosition.mpInterpolatedValue[xyz];
    MyVector::Normalize(mCTop.mpInterpolatedValue, 1.0);
}

void CameraPath_cls::
ApplySkyToSelectedKey(){
    if (mCbSky.mbBody == FALSE) return;
    size_t keySize = mCPosition.mVCKeys.size();
    for (size_t i = 0; i < keySize; i++){
	if (mCPosition.mVCKeys[i].mbSelected == TRUE){
	    double doubleIndex = mCPosition.mVCKeys[i].mDoubleIndex;
	    double top[3];
	    LOOP_XYZ top[xyz] = 
		mCSky.GetPInterpolatedValue(doubleIndex)[xyz] - mCPosition.GetPInterpolatedValue(doubleIndex)[xyz];
	    
	    KeyFrame_cls CKey(mCTop.mVCKeys[i]);
	    CKey.SetValue(top);
	    mCTop.SetKey(CKey);
	}
    }
}

wxString CameraPath_cls::
Serialize(){
    wxString strOut;
    strOut << wxT("--CameraPath") << ENDLWX;
    strOut << mCPosition.Serialize();
    strOut << mCTop.Serialize();
    strOut << mCAncPos.Serialize();
    strOut << mCRotYPR.Serialize();
    strOut << mCConfList.SerializeAll();
    strOut << wxT("--END") << ENDLWX;
    return strOut;
}

void CameraPath_cls::
ReadStrLine(std::wistream &istream, std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"--CameraPath");
    //std::wstring strLine;
    if (loc != std::string::npos){
	std::getline(istream, strLine);
	mCPosition.ReadStrLine(istream, strLine);
	std::getline(istream, strLine);
	mCTop.ReadStrLine(istream, strLine);
	std::getline(istream, strLine);
	mCAncPos.ReadStrLine(istream, strLine);
	std::getline(istream, strLine);
	mCRotYPR.ReadStrLine(istream, strLine);
    }
    //mCbSky.ReadStrLine(strLine);
    std::getline(istream, strLine);
    while (strLine.find(L"--END") == std::string::npos){
	mCConfList.ReadStrLineAll(strLine);
	std::getline(istream, strLine);
	// $B8_49@-$N$?$a$KNc30E*$K=hM}(B
	if (strLine.find(L"--AnchorPos") != std::string::npos) break;
    }
}

//CameraPath_cls* CameraPath_cls::mpCThisCurrent;

// ---------------

#include <GL/glut.h>
// dll$B$N%X%C%@(B
#include "SRC/GLOBJ/GLOBJECTS.h"
#include "SRC/COMMON/MyVector.h"

CameraPathGLGUI_cls::CameraPathGLGUI_cls(){
    mVDisplayListCameraCurr.push_back(_CAMERA_1);
    mVDisplayListCameraCurr.push_back(_CAMERA_2);
    mVDisplayListAnchorCurr.push_back(_CAMERA_CROSSHAIR);
};

CameraPathGLGUI_cls::CameraPathGLGUI_cls(const CameraPathGLGUI_cls& obj) :
    CameraPath_cls(obj)
{};

void CameraPathGLGUI_cls::
CopyFromP(CameraPathGLGUI_cls *pCObj){
    CameraPath_cls::CopyFromP(pCObj);
}

void CameraPathGLGUI_cls::
SetMatrix(Parameters_cls* pParam, int i, double dpos[3], int flag){
    double pos[3] = {
	pParam->GetPKeyFrameValue(i)[0] + dpos[0],
	pParam->GetPKeyFrameValue(i)[1] + dpos[1],
	pParam->GetPKeyFrameValue(i)[2] + dpos[2]};
    Direction_cls CDir;

    if (flag == _CAMERA_PATH_CAMERA){
	double dobIndexAtKey = pParam->mVCKeys[i].mDoubleIndex;
	LOOP_XYZ CDir.mTgt[xyz] = 
	    mCAncPos.GetPInterpolatedValue(dobIndexAtKey)[xyz] - pos[xyz];
	LOOP_XYZ CDir.mTop[xyz] = mCTop.GetPKeyFrameValue(i)[xyz];
	CDir.NormalizeAndOrthogonalize();
	CDir.RotateYPR3d(mCRotYPR.GetPInterpolatedValue(dobIndexAtKey));
    }
    
    GLOBJ::SetMatrix3dv3(pos, CDir.mTgt, CDir.mTop);
    glScalef(mScale, mScale, mScale);
}

void CameraPathGLGUI_cls::
DrawCamPath(GLGuides_cls* pCGLGuide, double dpos[3], int manipMode,
	    double dScale[3], double dRot[4]){
    if (DrawSelectedKey(pCGLGuide, &mCPosition, &mCAncPos, &mCTop,
			_CAMERA_1, dpos, mScale, _CAMERA_PATH_CAMERA, 
			manipMode, dRot) == FALSE){
	PositionAndDirection_cls CTmp;
	DrawCurrent(pCGLGuide, mCCurrentCam, mVDisplayListCameraCurr, 
		    dpos, mScale, mCurrentAnc, CTmp, manipMode,
		    dScale, dRot);
    }
}

void CameraPathGLGUI_cls::
DrawAncPath(GLGuides_cls* pCGLGuide, double dpos[3], int manipMode){
    double dRotDummy[4] = {0, 0, 0, 0};
    if (DrawSelectedKey(pCGLGuide, &mCAncPos, NULL, NULL,
			_CAMERA_CROSSHAIR, dpos, mScale, _CAMERA_PATH_ANCHOR, manipMode, dRotDummy) == FALSE){
	PositionAndDirection_cls CPosDir;
	LOOP_XYZ CPosDir.mPos[xyz] = mCurrentAnc[xyz];
	PositionAndDirection_cls CTmp;
	DrawCurrent(pCGLGuide, CPosDir, mVDisplayListAnchorCurr,
		    dpos, mScale, NULL, CTmp);
    }
}

void CameraPathGLGUI_cls::
DrawPath(){

    bool bSelectedCam = 
	GpCMainFrameCLS->IsPanelActive(wxT("[Cam XYZ]"));
    bool bSelectedAnc = 
	GpCMainFrameCLS->IsPanelActive(wxT("[Anc XYZ]"));

    double top[3], pZero[3] = {0, 0, 0};
    Direction_cls CDir;

    if (mCPosition.mCbVisible.mbBody == TRUE){
	for (unsigned int i = 0; i < mCPosition.GetKeyNumber(); i++){
	    double dobIndexAtKey = mCPosition.mVCKeys[i].mDoubleIndex;
	    if (dobIndexAtKey >= mDobWAMin && dobIndexAtKey <= mDobWAMax + 1e-5)
		glColor4fv(mCPosition.GetPColor(bSelectedCam, TRUE));
	    else
		glColor4fv(mCPosition.GetPColor(bSelectedCam, FALSE));
	
	    if (mCPosition.mVCKeys[i].mbSelected == TRUE &&
		bSelectedCam == TRUE)
		glColor4f(1.0f, 1.0f, 0.2f, 1.0f);
	    
	    //$B%+%a%i(BKey
	    glPushMatrix();
	    SetMatrix(&mCPosition, i, pZero, _CAMERA_PATH_CAMERA);
	    GLOBJ::Object::Draw(_CAMERA_1);
	    glPopMatrix();

	    // TOP$BJ}8~(B
	    glBegin(GL_LINES);
	    glVertex3dv(mCPosition.GetPKeyFrameValue(i));
	    LOOP_XYZ top[xyz] = mCPosition.GetPKeyFrameValue(i)[xyz] + mCTop.GetPKeyFrameValue(i)[xyz] * mScale * 5;
	    glVertex3dv(top);
	    glEnd();
	    
	    // Anc$BJ}8~(B
	    if (mCAncPos.mCbVisible.mbBody == TRUE &&
		dobIndexAtKey >= mDobWAMin && dobIndexAtKey <= mDobWAMax + 1e-5) {
		glColor4f(0.1f, 0.5f, 0.1f, 0.5f);
		glEnable(GL_LINE_STIPPLE);
		glBegin(GL_LINES);
		glVertex3dv(mCPosition.GetPKeyFrameValue(i));
		glVertex3dv(mCAncPos.GetPInterpolatedValue(dobIndexAtKey));
		glEnd();
		glDisable(GL_LINE_STIPPLE);
	    }

	}

	mCPosition.DrawGLPath3dv(mStepNum, mDobWAMin, mDobWAMax,
				 mScale, mInterval, bSelectedCam, FALSE);
	
    	// $B8=:_$N%+%a%i(B
	if (bSelectedCam == TRUE)
	    glColor4f(1.f, 1.f, 1.f, 1.f);
	else
	    glColor4f(0.8f, 0.8f, 0.8f, 1.f);
	
	glPushName(_HITTEST_CAMERA);
	glPushName(0); //FIX_ME!! SelfIndex!!

	glPushMatrix();
	PositionAndDirection_cls CPosDir(mCCurrentCam);
	GLOBJ::SetMatrix3dv3(CPosDir.mPos, CPosDir.mTgt, CPosDir.mTop);
	glScalef(mScale, mScale, mScale);
	GLOBJ::Object::Draw(_CAMERA_1);
	GLOBJ::Object::Draw(_CAMERA_2);
	glPopMatrix();

	glInitNames();
    }

    // $B%"%s%+!<(B
    if (mCAncPos.mCbVisible.mbBody == TRUE){
	mCAncPos.DrawGLPath3dv(mStepNum, mDobWAMin, mDobWAMax,
			       mScale, mInterval, bSelectedAnc, TRUE);
	
	// $B8=:_$N%"%s%+!<(B
	if (bSelectedAnc == TRUE)
	    glColor4f(1.f, 1.f, 1.f, 1.f);
	else
	    glColor4f(0.8f, 0.8f, 0.8f, 1.f);

	glPushName(_HITTEST_ANCHOR);
	glPushName(0); //FIX_ME!! SelfIndex!!

	glPushMatrix();
	glTranslated(mCurrentAnc[0], mCurrentAnc[1], mCurrentAnc[2]);
	glScalef(mScale*3, mScale*3, mScale*3);
	GLOBJ::Object::Draw(_CAMERA_CROSSHAIR);
	glPopMatrix();
	
	glInitNames();
    }
}

// ----------------------

ComCameraBase_cls::
ComCameraBase_cls(CameraPathCtrl_cls *pCTgt, int index){
    mpCTgt = pCTgt;
    mIndex = index;
}

ComCameraAdd_cls::
ComCameraAdd_cls(CameraPathCtrl_cls *pCTgt, int index, CameraPathGLGUI_cls *pCCameraPathGLGUI):
    ComCameraBase_cls(pCTgt, index){
    if (pCCameraPathGLGUI == NULL)
	mpCPathGLGUI = NULL; // NULL$B$J$i%$%K%7%c%i%$%:$9$k$H$9$l$P$$$$(B
    else
	mpCPathGLGUI = new CameraPathGLGUI_cls(*pCCameraPathGLGUI);
}

ComCameraAdd_cls::
~ComCameraAdd_cls(){
    if (mpCPathGLGUI != NULL) delete mpCPathGLGUI;
}

bool ComCameraAdd_cls::
Execute(bool bFlag){
    mpCTgt->AddPathExecute(mIndex, mpCPathGLGUI);
    return _COMMAND_NODEPENDENT;
}

ComCameraDelete_cls::
ComCameraDelete_cls(CameraPathCtrl_cls *pCTgt, int index) : 
    ComCameraBase_cls(pCTgt, index){
}

bool ComCameraDelete_cls::
Execute(bool bFlag){
    mpCTgt->DeletePathExecute(mIndex);
    return _COMMAND_NODEPENDENT;
}
