// "clsCameraPath.cpp"

#ifndef _CAMERA_PATH_H
#define _CAMERA_PATH_H

#include "SRC/CONFITEM/clsConfItem.h"
#include "SRC/PARAMETER/clsParameters.h"
#include "SRC/MYGL/clsDrawPath.h"
#include "SRC/MYGL/clsPositionAndDirection.h"
#include "SRC/MAINCANVAS/clsDrawItem.h"

#include "SRC/WIDGETCTRL/clsDialogEXKey.h"
#include "SRC/WIDGETCTRL/clsDialogController.h"
#include "SRC/WIDGETCTRL/clsMinimizeButton.h"

class CameraPath_cls; 
class CameraPathCtrl_cls;
class CameraPathGUI_cls; 
class CameraPathGLGUI_cls; 
class DF3Sequence_cls; 
class FrameMain_cls;
class GLGuides_cls;
class MultiNumericalInputs_cls; 
class MultiNumericalInputsPanel_cls;
class StereoGram_cls; 

class wxButtonEX;

#define _CAMERA_PATH_CAMERA 0
#define _CAMERA_PATH_ANCHOR 1
#define _PATH_NUM_MAX 16

#ifdef _DEBUG
#include <stdlib.h>
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

class CameraPathGUI_cls : public wxDialogEXKey, public MinimizeButton_cls {
 public:
    CameraPathGUI_cls(CameraPathCtrl_cls* pCParent, 
		      wxWindow *pC, wxPoint pos, wxSize size);
    ~CameraPathGUI_cls();

    static CameraPathGUI_cls *mspCThis;
    static CameraPathCtrl_cls *mspCParent;
    
    wxButtonEX *mpCBTNView, *mpCBTNViewAdv, *mpCBTNPath;

    MultiNumericalInputsPanel_cls* mpCMNumView;
    MultiNumericalInputsPanel_cls* mpCMNumViewAdv;
    MultiNumericalInputsPanel_cls* mpCMNumPath;

    int mIndexCamPos, mIndexAncPos, mIndexViewGeom;
    int mIndexPitchAngle;
    int mIndexAngleView, mIndexAnglePath;
    int mIndexActivePath, mIndexGeom;
    int mIndexPosition, mIndexAnchor, mIndexYPR, mIndexSky;
    int mIndexPointSky, mIndexFBOSize;
    int mIndexGrid, mIndexAntiAlias, mIndexCoordinate, mIndexShowOtherCam, mIndexXOR;
    int mIndexBG0, mIndexBG1;

    void HideAndShowView();
    void HideAndShowPath();
    void HideAndShowAll();

    void SetMNumView(wxBoxSizer *pBoxSizer);
    void SetMNumViewAdv(wxBoxSizer *pBoxSizer);
    void SetMNumPath(wxBoxSizer *pBoxSizer);
    void ReconnectView();
    void ReconnectViewAdv();
    void ReconnectPath();
    void UpdateComboboxList();

    void OnSize(wxSizeEvent &event);
    void OnMouse(wxMouseEvent &event);
    void OnMenuMinimize(wxCommandEvent &event);
    void UpdateButtonMinimize();
    void OnButtonShowItemView(wxCommandEvent &event);
    void OnButtonShowItemViewAdv(wxCommandEvent &event);
    void OnButtonShowItemPath(wxCommandEvent &event);
    void OnButtonBGColor0(wxCommandEvent &event);
    void OnButtonBGColor1(wxCommandEvent &event);
    void OnButtonOutsideColor(wxCommandEvent &event);
    void OnButtonApplyAll(wxCommandEvent &event);

    static void FuncHookAssign(void *pArg, void *pEvent);
    static void FuncHookRefresh(void *pArg, void *pEvent);
    static void FuncHookButton(void *pArg, void *pEvent);

     private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class CameraPathCtrl_cls : public DialogControl_cls {
 public:
    CameraPathCtrl_cls(FrameMain_cls *pCParentFrame, StereoGram_cls *pCStereoGram,
		       wxBitmap *pBitmap);
    ~CameraPathCtrl_cls();

    FrameMain_cls *mpCParentFrame;
    wxBitmap *mpBitmapLink;

    std::vector < CameraPathGLGUI_cls* > mVpCPath;
    CameraPath_cls* GetActivePath();
    CameraPathGLGUI_cls* GetActivePathGLGUI();

    CameraPathGUI_cls *mpCGUI;

    void InitGUI(wxWindow *pC);
    void UpdateItems(bool bForce = FALSE);
    
    ConfItemList_cls mCConfList;
    ConfItemINT_cls mCCameraNum;
    ConfItemList_cls mCConfListLayout;
    ConfItemBOOL_cls mCConfBTNView;
    ConfItemBOOL_cls mCConfBTNViewAdv;
    ConfItemBOOL_cls mCConfBTNPath;
    int mActiveView;
    //int mPathNum;
    int mActivePath;

    void ClearPath();
    void AddPath();
    void AddPathExecute(int index, CameraPathGLGUI_cls *pCPathGLGLUI,
			bool bRefresh = TRUE);
    void DeletePath();
    void DeletePathExecute(int index);
    void DrawPath(PositionAndDirection_cls CViewPos);

    wxString Serialize(bool flag);
    void ReadStrLine(std::wistream &istream, std::wstring strLine);
};

class CameraPath_cls : public DrawPath_cls {
    friend class DF3SeqGUI_cls; 
    friend class FrameMain_cls; 
    friend class MainGLCanvas_cls; 
    
 public:
    CameraPath_cls();
    void InitializeCamera();
    CameraPath_cls(const CameraPath_cls &obj);
    ~CameraPath_cls();
    void CopyFromP(CameraPath_cls *pCObj);
    void ReDefineList();
    
 public:
    Parameters_cls mCPosition;
    Parameters_cls mCTop;
    Parameters_cls mCAncPos;
    Parameters_cls mCRotYPR;
    Parameters_cls mCSky;
    ConfItemList_cls mCConfList;
    ConfItemBOOL_cls mCbSky;
    ConfItemINT_cls  mCCamGeometry;
    ConfItemDOUBLE_cls mCVAngle;
   
    StereoGram_cls *mpCStereoGramLinked;
    void LinkStereoGramCLS(StereoGram_cls* pCStereoGram);
    
 public:
    double mCurrentDoubleIndex;
    PositionAndDirection_cls mCCurrentCam;
    double mCurrentAnc[3];
    
    double mDistanceToFocus;
    double mCurrentFocus[3];
    void   GetSetCurrentPos(double doubleIndex);
    void   ModifyStereogramPos(PositionAndDirection_cls &CPos);
    double GetStereogramYawAngle();
    double GetDomePitchAngle();
    void RotateAroundSelectedKey(double dRot);
    
    void DeleteKeyPos(double doubleIndex);
    void SetKeyPos(double doubleIndex, PositionAndDirection_cls CPOS);
    void SetKeyAnc(double doubleIndex, Position_cls CAnc);
    void NormalizeTop();
    void ApplySky(PositionAndDirection_cls* pCPos);
    void ApplySky();
    void ApplySkyToSelectedKey();
    
    void InsertKeyPos(double doubleIndex);

    wxString Serialize();
    void ReadStrLine(std::wistream &istream, std::wstring strLine);
};

// GL$BI=<($N(BGUI
class CameraPathGLGUI_cls : public CameraPath_cls, DrawItem_cls {
 
 public:
    CameraPathGLGUI_cls();
    CameraPathGLGUI_cls(const CameraPathGLGUI_cls& obj);
    void CopyFromP(CameraPathGLGUI_cls *pCObj);
    
    std::vector < int > mVDisplayListCameraCurr;
    std::vector < int > mVDisplayListAnchorCurr;

    void SetMatrix(Parameters_cls* pParam, int i, double dpos[3], 
		   int flag = _CAMERA_PATH_ANCHOR);
    
    void DrawCamPath(GLGuides_cls* pCGLGuide, double dpos[3], int manipMode,
		     double dscale[3], double drot[4]);
    void DrawAncPath(GLGuides_cls* pCGLGuide, double dpos[3], int manipMode);
    void DrawPath();
};

#include "SRC/UNDO/clsUndoManager.h"

class ComCameraBase_cls : public CommandItem_cls {
public:
    ComCameraBase_cls(CameraPathCtrl_cls *pCTgt, int index);
    //~ComCameraBase_cls();
    CameraPathCtrl_cls *mpCTgt;
    int mIndex;

    //bool Execute(bool bFlag = _COMMAND_REDO);
};

class ComCameraDelete_cls : public ComCameraBase_cls {
public:
    ComCameraDelete_cls(CameraPathCtrl_cls *pCTgt, int index);
    bool Execute(bool bFlag = _COMMAND_REDO);
};

class ComCameraAdd_cls : public ComCameraBase_cls {
public:
    ComCameraAdd_cls(CameraPathCtrl_cls *pCTgt, int index, CameraPathGLGUI_cls *pCCameraPathGLGUI);
    ~ComCameraAdd_cls();
    
    CameraPathGLGUI_cls *mpCPathGLGUI;

    bool Execute(bool bFlag = _COMMAND_REDO);
};

#endif
