// -*- c++ -*-
// (setq c-basic-offset 4)
// "clsMain.h"

#include "clsMain.h"

#include "clsCameraPath.h"
#include "clsFrameMain.h"
#include "clsTimeCtrl.h"

#include "SRC/GLCANVAS/clsGLCanvasCtrl.h"

#include "clsDrawingBox.h"
#include "SRC/CONFITEM/clsImportList.h"
#include "SRC/DF3/clsDF3SeqCtrl.h"
#include "SRC/DF3/clsDF3Sequence.h"
#include "SRC/DF3/clsDF3SeqGUI.h"
#include "SRC/DF3/GL/clsColoring.h"
extern Coloring_cls GCColoring;

#include "SRC/DRAWBOX/clsBoxCopy.h"
#include "SRC/FILTER/clsFilterCtrl.h"
#include "SRC/FILTER/clsClipping.h"
#include "SRC/GLCANVAS/clsMainGLCanvas.h"
#include "SRC/GRAPHIC/clsImageIO.h"
extern ImageIO_cls GCImgIO;
#include "SRC/GRAPHIC/clsImgSequence.h"
#include "SRC/MAINCANVAS/clsScene.h"
#include "SRC/MYGL/clsTextureProperty.h"

#include "SRC/PARAMETER/clsTimeLinePanel.h"
#include "SRC/POVRAY/clsLightSource.h"
#include "SRC/POVRAY/clsPovRay.h"
#include "SRC/POVRAY/clsPovIsosurface.h"
#include "SRC/POVRAY/clsPovCrossSection.h"
#include "SRC/POVRAY/clsPovObject.h"
#include "SRC/POVRAY/clsStereoGram.h"
#include "SRC/POVRAY/clsTreeBoxRule.h"
#include "SRC/WIDGETCTRL/EX/clsFileDialogEX.h"
extern wxFileDialogEX*   GpCFileDialog;

//#include "SRC/POVRAY/POVTEXTURE/texture.h"
extern void InitTextureTable();
extern void InitSolidNoise();

#include "SRC/UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include "wxWidgetEnum.h"
#include "wxWidgetString.h"

int GClientDisplayRightPos, GClientDisplayBottomPos;

#include "SRC/MYAPI/MsgBox.h"
#include <wx/tooltip.h>
#include <wx/file.h>
#include <wx/gdicmn.h>
#include <wx/progdlg.h>

//#include <wx/app.h>
//#include "Main.h"
//DECLARE_APP(MyApp)

#define LOOP_XYZ for(int xyz = 0; xyz < 3; xyz++)

#ifdef WIN32
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

Main_cls::
Main_cls() : mpCTexPropertyCtrl(NULL) 
{
#ifdef WIN32
    mPID = _getpid();
#else
    mPID = getpid();
#endif

    Parameters_cls::mspVPInstancePrim = &mVPParamInstancePrim;
    mpCBoxSize = new Parameters_cls(6, L"BoxSize");
    DrawingBlock_cls::mspCBoxSize = mpCBoxSize;
    mpCBoxRotate = new Parameters_cls(4, L"BoxRotate");
    DrawingBlock_cls::mspCRotate = mpCBoxRotate;

    int posx = 20, posy = 10;
    int width = 1024, height = 900;
    wxRect rect = wxGetClientDisplayRect();
    GClientDisplayRightPos = rect.x + rect.width;
    GClientDisplayBottomPos = rect.y + rect.height;
    //wxCDisplaySize(&GDisplayWidth, &GDisplayHeight);
    if (posx + width > GClientDisplayRightPos) { 
	posx = rect.x; width = rect.width; }
    if (posy + height > GClientDisplayBottomPos) {
	posy = rect.y; height = rect.height; }
    
    mpBitmapColor = 
	new wxBitmap(wxT("ICONS/Color.png"), wxBITMAP_TYPE_PNG);
    mpBitmapLight = 
	new wxBitmap(wxT("ICONS/LightSource.png"), wxBITMAP_TYPE_PNG);
    
    mpCTime  = new TimeCtrl_cls;
    mpCFrame = new FrameMain_cls(_FRAME_MAIN_APPNAME, this, 
				 posx, posy, width, height);
    mpCDF3SeqCtrlLink = mpCFrame->mpCDF3SeqCtrl;
    mpCScene = new Scene_cls(mpCFrame, mpCTime, mpCFrame->mpBitmapScene);

    mCbToolTip.SetItem(L"ToolTip=", L"1");
    mCFPathEditDLG.SetMNumEditor(mpCFrame);
    
    //--- Class with Dialogs (݂Ɉˑ̂ŏ̏ԏdv)
    mpCDBoxMain    = new DrawingBlock_cls();
    mpCTreeBoxRule = new TreeBoxRule_cls(mpCFrame, mpCFrame->mpBitmapTree);
    mpCFrame->mpCDF3SeqCtrl->mspCTreeBoxRuleLink = mpCTreeBoxRule;
    //---

    mpCLightSRC    = new LightSRC_cls(mpCFrame, mpBitmapLight);

    mpCImportList  = new ImportList_cls(mpCFrame);
    mpCCustomColor = new CustomColor_cls(mpCFrame, &GCColoring, mpBitmapColor);

    mpCTexPropertyCtrl = new TexPropertyCtrl_cls(mpCFrame, mpCFrame->mpBitmapTexture);
    FilterWithTextureGUI_cls::LinkTexturePropertyCtrl(mpCTexPropertyCtrl);

    mpCImgSequence = new ImgSequence_cls(mpCFrame);

    GCColoring.mpCCustomColor = mpCCustomColor;

    FilterCtrl_cls::InitializeStaticMembers
	(MainGLCanvas_cls::mspCPreference, mpCFrame->mpCDF3SeqCtrl, 
	 mpCFrame->mpCPovRay, mpCTreeBoxRule, DF3_cls::mspCProperty);

    DrawingBlock_cls::InitializeStaticMembers
	(mpCFrame, mpCFrame->mpCDF3SeqCtrl, 
	 mpCFrame->mpCDF3SeqCtrl->GetPFilterCtrl(0/*CHECK_ME!!*/), mpCTreeBoxRule); 
    
    TimeLinePanel_cls::InitializeStaticMembers
	(mpCLightSRC, &DrawingBlock_cls::msCHotSpot,
	 mpCFrame->mpCDF3SeqCtrl->mpCGUI);

    InitTextureTable();
    InitSolidNoise();
    
    wxToolTip::SetDelay(400);//ms

    // ŏɕ\ĂTLACe
    mpCFrame->mpCDF3SeqCtrl->mpCGUI->ShowItem(wxT("[Cam XYZ]"));
    mpCFrame->mpCDF3SeqCtrl->mpCGUI->ShowItem(wxT("[Anc XYZ]"));

    // ftHgCg
    wxCommandEvent com(wxEVT_COMMAND_BUTTON_CLICKED, btnLightAdd);
    mpCLightSRC->mpCGUI->AddPendingEvent(com);

    // ftHgJ[
    wxCommandEvent com2(wxEVT_COMMAND_BUTTON_CLICKED, btnCtrlCustomClrAdd);
    mpCCustomColor->mpCGUI->AddPendingEvent(com2);

    bOneTimeWarningShowed_SceneLength = FALSE;
}

Main_cls::
~Main_cls(){
    FilterClipping_cls::mspCDF3SeqGUI = NULL; //폜ς݂Ȃ̂NULL
    //delete mpCFilterCtrl;
    delete mpCDBoxMain;
    delete mpCTreeBoxRule;
    delete mpCLightSRC;
    delete mpCImportList;
    delete mpCTexPropertyCtrl;
    delete mpCImgSequence;
    delete mpCBoxSize;
    delete mpCBoxRotate;
    delete mpCScene;
    delete mpCTime;

    delete mpBitmapColor;
    delete mpBitmapLight;
}

void Main_cls::
UpdateParamInstancePrim(){
    mVPParamInstancePrim.clear();
    mVPParamInstancePrim.push_back(&mpCTime->mCTimeRemap);
    mVPParamInstancePrim.push_back(&mpCFrame->mpCPathCtrl->mVpCPath[0]->mCPosition);
    mVPParamInstancePrim.push_back(&mpCFrame->mpCPathCtrl->mVpCPath[0]->mCTop);
    mVPParamInstancePrim.push_back(&mpCFrame->mpCPathCtrl->mVpCPath[0]->mCRotYPR);
    mVPParamInstancePrim.push_back(&mpCFrame->mpCPathCtrl->mVpCPath[0]->mCAncPos);
    mVPParamInstancePrim.push_back(&mpCFrame->mpCStereo->mCDistanceOfEyes);

    for (unsigned int i = 0; i < mpCLightSRC->mpCGUI->mVCParam.size(); i++)
	mVPParamInstancePrim.push_back(&mpCLightSRC->mpCGUI->mVCParam[i]);
    for (unsigned int i = 0; i < DrawingBlock_cls::msCHotSpot.mpCGUI->mVCParam.size(); i++)
	mVPParamInstancePrim.push_back(&DrawingBlock_cls::msCHotSpot.mpCGUI->mVCParam[i]);

    // Isosurfacẽp[^ǉ
    for (unsigned df3 = 0; df3 < _DF3_MAX_SLOT; df3++){
	std::vector < Parameters_cls* > VpCPrimParam = 
	    mpCDF3SeqCtrlLink->GetPFilterCtrl(df3)->GetVpCPrimParam();
	for (unsigned int i = 0; i < VpCPrimParam.size(); i++)
	    mVPParamInstancePrim.push_back(VpCPrimParam[i]);
    }
    
    mVPParamInstancePrim.push_back(mpCBoxSize);
    mVPParamInstancePrim.push_back(&mpCDF3SeqCtrlLink->mCBaseAlpha);
}

DF3Sequence_cls* Main_cls::
GetPDF3SeqActive(){
    return mpCDF3SeqCtrlLink->GetPDF3SeqActive();
}

DF3_cls* Main_cls::
GetPDF3Active(){
    return mpCDF3SeqCtrlLink->GetPDF3Active();
}

// ----

void Main_cls::
RefreshDialogs(){
    /*
    if (mpCCustomColor->IsGUIShownOnScreen() == TRUE)
	mpCCustomColor->mpCGUI->UpdateValues();
    */

    if (mpCTreeBoxRule->mpCGUI->IsShownOnScreen() == TRUE)
	mpCTreeBoxRule->mpCGUI->UpdateDisplay();
}

// ----

void Main_cls::
OpenFileSequence(wxString strFilename){
    unsigned int maxFrame = mpCTime->mCMaxFrame.mBody;
    unsigned int WAMin = mpCTime->mCWAMin.mBody;
    unsigned int WAMax = mpCTime->mCWAMax.mBody;
    GetPDF3SeqActive()->ClearFileSequence();
    AddFileSequence(strFilename, FALSE);
    GCUndoList.Clear();
    if (maxFrame != 0 && mpCTime->mCMaxFrame.mBody == 0)
	if (bOneTimeWarningShowed_SceneLength == TRUE || 
	    MsgBoxYesno(mpCFrame, _("Query"), _("Snap Shot is loaded. Set Scene Length to 0?")) == FALSE){
	    mpCTime->mCMaxFrame.mBody = maxFrame;
	    mpCTime->mCWAMin.mBody = WAMin;
	    mpCTime->mCWAMax.mBody = WAMax;
	    mpCTime->mCAAMin.mBody = WAMin;
	    mpCTime->mCAAMax.mBody = WAMax;
	    mpCTime->UpdateDF3SeqGUI();
	    bOneTimeWarningShowed_SceneLength = TRUE;
	}
    mpCDF3SeqCtrlLink->mpCGUI->mpCPanelList->UpdateBitmapPanels();
}

void Main_cls::
AddFileSequence(wxString strFilename, bool bImageTimeSequence){
    UpdateParamInstancePrim();
    GetPDF3SeqActive()->AppendFileSequence(strFilename, TRUE, 
					   bImageTimeSequence);
    GetPDF3SeqActive()->SetIndividualFileNames();
    mpCFrame->mCLastOpenedFileDF3.msBody = strFilename;

    if (GetPDF3SeqActive()->mVCFileSequence.size() > 0){
	wxString strInfo;
	strInfo << GetPDF3SeqActive()->mVCFileSequence[0].GetInfoStr() << wxT(" - ") << _APPNAME_SHORT;
	mpCFrame->SetTitle(strInfo);
    }
    
    mpCFrame->mpCMNumPlayOption->UpdateValues();
    mpCFrame->Refresh(false);
}

void Main_cls::
ImportParamInitialize(){
    if (mpCImportList->mCbPathLine.mbBody == TRUE)
	mpCFrame->mpCPathCtrl->ClearPath();
}

void Main_cls::
ImportParamMain(std::wistream &istream, std::wstring strLine){
    if (mpCImportList->mCbPathLine.mbBody == TRUE)
	mpCFrame->mpCPathCtrl->ReadStrLine(istream, strLine);
    if (mpCImportList->mCbStereo.mbBody == TRUE)
	mpCFrame->mpCStereo->ReadStrLine(istream, strLine);
    if (mpCImportList->mCbHotSpot.mbBody == TRUE)
	DrawingBlock_cls::msCHotSpot.mpCGUI->ReadStrLine(istream, strLine);
    if (mpCImportList->mCbLight.mbBody == TRUE)
	mpCLightSRC->ReadStrLine(istream, strLine);
    if (mpCImportList->mCbBoxSize.mbBody == TRUE){
	mpCBoxSize->ReadStrLine(istream, strLine);
	mpCBoxRotate->ReadStrLine(istream, strLine);
	/*
	DrawingBlock_cls::mspCBoxSize->ReadStrLine(istream, strLine);
	DrawingBlock_cls::mspCRotate->ReadStrLine(istream, strLine);
	*/
    }	
    if (mpCImportList->mCbAlpha.mbBody == TRUE)
	mpCDF3SeqCtrlLink->mCBaseAlpha.ReadStrLine(istream, strLine);
    if (mpCImportList->mCbSceneInfo.mbBody == TRUE){
	mpCDF3SeqCtrlLink->ReadStrLineSceneConf(strLine);
	mpCTime->ReadStrLine(istream, strLine);
    }

    if (mpCImportList->mCbDF3Seq.mbBody == TRUE){
	mpCDF3SeqCtrlLink->ReadStrLineRenderParam(strLine);
	mpCTreeBoxRule->ReadStrLine(strLine);
	mpCDBoxMain->ReadStrLine(strLine);
	//DrawingBlock_cls::msCConfsCopy.ReadStrLineAll(strLine);
	//GCColoring.ReadStrLine(strLine);
    }

    if (mpCImportList->mCbFilters.mbBody == TRUE){
	GetPDF3SeqActive()->mpCFilterCtrl->ReadStrLine(istream, strLine);
	DrawingBlock_cls::ReadStrLineTex(istream, strLine);
    }
    if (mpCImportList->mCbPovRay.mbBody == TRUE)
	mpCFrame->mpCPovRay->ReadStrLine(istream, strLine);
    if (mpCImportList->mCbCustomColor.mbBody == TRUE)
	mpCCustomColor->ReadStrLine(istream, strLine);
}

void Main_cls::
ImportParam(unsigned int type){
    
    mpCImportList->SetToDefault(type);
    if (mpCImportList->OpenWindowModal(mpCFrame) == FALSE) {
	return;
    }

    GpCFileDialog->SetMessage(wxT("Select a file"));
    GpCFileDialog->SetWildcard(wxT("Project file (*.osp)|*.osp"));
        
    if (GpCFileDialog->ShowModal() == wxID_OK){
	std::wifstream ifstream(GpCFileDialog->GetPath().ToAscii(), std::ios::in);
	ImportParamInitialize();
	if (!ifstream.fail()){
	    FilterCtrl_cls::PushListOmit();
	    std::wstring strLine;
	    while ( std::getline (ifstream, strLine))
		ImportParamMain(ifstream, strLine);
	    FilterCtrl_cls::PopListOmit();
	}
    }
    //delete pDialog;

    mpCFrame->UpdateMenu();
    mpCFrame->RefreshCameraCombobox();
    mpCDF3SeqCtrlLink->UpdateFilterPanelListAll();
    
}
 
void Main_cls::
OpenFileEdit(wxString strTitle, bool mbAddMode){
    if (GetPDF3SeqActive()->mVCFileSequence.size() > 0)
	mCFPathEditDLG.mstrBuffer = GetPDF3SeqActive()->mVCFileSequence[0].mwsFormat;
    /*
    else 
	mCFPathEditDLG.mstrBuffer = wxT("");
    */
    
    mCFPathEditDLG.mIntegerA = GetPDF3SeqActive()->mDefaultDigit;
    mCFPathEditDLG.mIntegerB = GetPDF3SeqActive()->mDefaultZSliceInit;
    mCFPathEditDLG.mbEnableA = TRUE; mCFPathEditDLG.mbEnableB = TRUE;
    mCFPathEditDLG.mbSave = FALSE;
    mCFPathEditDLG.SetDisable();  
    mCFPathEditDLG.mpCMNumEditDialog->mstrHelp = _("Tooltip for Open Data");
    if (mCFPathEditDLG.OpenModalDialog(strTitle) == TRUE){
	if (mbAddMode == TRUE) GetPDF3SeqActive()->ClearFileSequence();
	GetPDF3SeqActive()->mDefaultDigit = mCFPathEditDLG.mIntegerA;
	GetPDF3SeqActive()->mDefaultZSliceInit = mCFPathEditDLG.mIntegerB;
	AddFileSequence(mCFPathEditDLG.mstrBuffer, TRUE);
	GCUndoList.Clear();
    }
}

bool Main_cls::
IsPovrayCondistionSatisfiled(){
    if (mpCDF3SeqCtrlLink->IsAllDataDF3() == FALSE){
	MsgBox(mpCFrame, _("Please Use DF3 File"), _("Can not Render"));
	return FALSE; }
    
    if (mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_VOLUME) == FALSE && 
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_ISOSURFACE) == FALSE &&
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_CROSS_SECTION) == FALSE&&
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_POV_OBJECT) == FALSE &&
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_THREEVIEW) == FALSE &&
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_STREAMLINE) == FALSE){
	if (MsgBoxYesno(mpCFrame, _("Output .pov file"), _("No Object")) == FALSE)
	    return FALSE; }
    if ((mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_ISOSURFACE) == TRUE ||
	 mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_CROSS_SECTION) == TRUE||
	 mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_POV_OBJECT)) &&
	mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_VOLUME) == FALSE &&
	mpCLightSRC->mpCGUI->mCItemNum.mBody == 0){
	if (MsgBoxYesno(mpCFrame, _("Output .pov file"), _("No Light")) == FALSE)
	    return FALSE; }
    return TRUE;
}


bool Main_cls::
ProcOutputPovRay(int stereo){
    PovIsosurface_cls::msstrWorkingDir = 
	mpCFrame->mpCPovRay->GetWorkingDirWithSep();

    if (IsPovrayCondistionSatisfiled() == FALSE) return FALSE;

    if (mpCFrame->mpCPovRay->mCCameraType.mBody == _POV_SAME_AS_PREVIEW0)
	mpCFrame->mpCCanvas[0]->mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_FREE);
    if (mpCFrame->mpCPovRay->mCCameraType.mBody == _POV_SAME_AS_PREVIEW1)
	mpCFrame->mpCCanvas[1]->mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_FREE);
    if (mpCFrame->mpCPovRay->mCCameraType.mBody == _POV_SAME_AS_PREVIEW2)
	mpCFrame->mpCCanvas[2]->mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_FREE);
    if (mpCFrame->mpCPovRay->mCCameraType.mBody == _POV_SAME_AS_PREVIEW3)
	mpCFrame->mpCCanvas[3]->mpCGLCanvas->SetDisplayMode(_GL_CANVAS_DISPLAY_FREE);

    wchar_t filename[64];
    swprintf(filename, 64, 
	     mpCFrame->mpCPovRay->GetFileStr(_POV_SNAP, stereo).c_str(),
	     mpCFrame->mpCPovRay->mSnapNumber);
    
    if (OutputPovRayIO(filename, mpCTime->mFrame) == FALSE){
	mpCFrame->PushStatusTextMain(_("Disk I/O Error"));
	MsgBox(mpCFrame, _("File output failed. Please check the working directory and its permission."), _("Disk I/O Error"));
		
	return FALSE;
    }
    
    wxString strStatus;
    strStatus << _("Disk Output Suceed.") << wxT(" ") 
	      << filename << wxT(".ini ") 
	      << filename << wxT(".pov");
    mpCFrame->PushStatusTextMain(strStatus);
	
    return TRUE;
}

void Main_cls::
OutputPovStrObjDensity(std::wfstream &ofstream,
		       std::vector < DrawingBlock_cls > &VCDrawingCls, 
		       PositionAndDirection_cls CCamGlobalCoord, int frame){
    
    mpCDF3SeqCtrlLink->ClearAlreadyOutputtedTagAll();
    LightSRC_cls::SetParentObjPos(&CCamGlobalCoord);
    
    ofstream << wxConvFile.cWX2WC(VCDrawingCls[0].GetPovBoxTex());

    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
	VCDrawingCls[i].OutputPovDataBlockDensity
	    (ofstream, mpCFrame->mpCPovRay, CCamGlobalCoord, frame);
}

void Main_cls::
OutputPovRayMain(std::wfstream &ofstream, int frame){
    
    mpCDF3SeqCtrlLink->UpdateFilterParametersAll(mpCTime->GetDoubleIndex());
    
    /*
    mpCFrame->mpCPovRay->SetBGColor
	(mpCFrame->mpCCanvas[0]->mpCGLCanvas->mCColorBG0.mpBody,
	 mpCFrame->mpCCanvas[0]->mpCGLCanvas->mCColorBG1.mpBody);
    */
    
    std::vector < DrawingBlock_cls > VCDrawingCls;
    mpCFrame->GainDrawingBlockCopies(VCDrawingCls,
				     _CHECK_TREE_FORCE_FILE_CHECK,
				     _CHECK_TREE_STOP_IF_FILE_NOT_EXIST);
    for (int i = 0; i < 5; i++)
	mpCFrame->mpCPovRay->mTreeDepthParam[i] = 
	    mpCTreeBoxRule->mCTreeDepthParam[i].mBody;
    
    //wBoxɌȂtO
    for (unsigned int i = 0; i < VCDrawingCls.size(); i++)
	VCDrawingCls[i].CheckVis
	    (mpCFrame->mpCPathCtrl->mVpCPath[0]->mCCurrentCam, 
	     mpCFrame->mpCPathCtrl->mVpCPath[0]->mCurrentAnc,
	     mpCFrame->mpCPovRay->GetCameraGeometry(),
	     mpCFrame->mpCPovRay->GetRadianForVisCheck());

    int glCanvasID;
    switch(mpCFrame->mpCPovRay->mCCameraType.mBody){
    case _POV_FOLLOW_CAMERA:
    case _POV_SAME_AS_PREVIEW0:	glCanvasID = 0; break;
    case _POV_SAME_AS_PREVIEW1:	glCanvasID = 1; break;
    case _POV_SAME_AS_PREVIEW2:	glCanvasID = 2; break;
    case _POV_SAME_AS_PREVIEW3:	glCanvasID = 3; break;
    default: glCanvasID = 0;
    }
    
    PositionAndDirection_cls CCamGlobalCoord = 
	mpCFrame->mpCCanvas[glCanvasID]->mpCGLCanvas->mCCameraPos;
    
    if (mpCFrame->mpCPovRay->mCCameraType.mBody == _POV_FOLLOW_CAMERA)
	LightSRC_cls::SetParentObjPos
	    (&mpCFrame->mpCPathCtrl->GetActivePath()->mCCurrentCam);
    else
        LightSRC_cls::SetParentObjPos(&CCamGlobalCoord);

    wxString strHead1;

    strHead1 << mpCFrame->mpCPovRay->GetPovStrCommon()
	     << mpCLightSRC->GetPovStrLightSource(mpCTime->GetDoubleIndex())
	     << mpCDF3SeqCtrlLink->GetPovStrFog()
	     << mpCFrame->mpCPovRay->GetPovStrCamInfo(CCamGlobalCoord, mpCFrame->mpCCanvas[glCanvasID]->GetPovStrCamPos(), mpCTime->GetDoubleIndex())
	     << mpCFrame->mpCCanvas[glCanvasID]->GetPovStrCamMatrix();
    
    ofstream << wxConvFile.cWX2WC(strHead1);
    
    if (mpCFrame->mpCPovRay->mCbSubsetRow.mbBody == TRUE){
	ofstream << wxConvFile.cWX2WC
	    (mpCFrame->mpCCanvas[glCanvasID]->GetPovStrMask
	     (mpCFrame->mpCPovRay->mCSubsetRowStart.mBody,
	      mpCFrame->mpCPovRay->mCSubsetRowEnd.mBody,
	      mpCFrame->mpCPovRay->mCHeight.mBody,
	      mpCFrame->mpCPovRay->GetVAngle()));
    }
    
    ofstream << (const char*)(mpCDF3SeqCtrlLink->GetPovStrColorMapsAllDF3()).ToAscii();
    
    DrawingBlock_cls::msCCamGlobalCoord = CCamGlobalCoord;
    //mpCFrame->OutputPovStrObjectsDensity(ofstream, VCDrawingCls, CCamGlobalCoord, frame);
    OutputPovStrObjDensity(ofstream, VCDrawingCls, CCamGlobalCoord, frame);
    
    ofstream << (const char*)(mpCFrame->GetPovStrObjects(VCDrawingCls)).ToAscii();

}

bool Main_cls::
OutputPovRayIO(std::wstring strFile, int frame){
    mpCFrame->mpCPovRay->mstrPostSceneCommand = wxT("");

    wxString strFileOut = mpCFrame->mpCPovRay->mCConfWorkingDir.msBody + wxT("/") + strFile + wxT(".pov");

#ifdef WIN32
    std::wfstream ofstreamPov(strFileOut.c_str(), std::ios::out);
#else
    std::wfstream ofstreamPov(strFileOut.ToAscii(), std::ios::out);
#endif

    mpCFrame->GetStatusBar()->SetStatusText(wxString::Format(wxT("making .pov %d/%d"), frame, mpCTime->mCMaxFrame.mBody), 2);
    while (wxTheApp->Pending())
	wxTheApp->Dispatch();
    
    if (ofstreamPov.fail()) return FALSE;

    mpCFrame->UpdateDrawingData();
    mpCTime->UpdateDF3SeqGUI();
    mpCFrame->mpCCanvas[0]->mpCGLCanvas->UpdateCameraPos();
    PositionAndDirection_cls CCamPosOrg = mpCFrame->mpCCanvas[0]->mpCGLCanvas->mCCameraPos; 

    mpCFrame->mpCCanvas[0]->mpCGLCanvas->ModifyByStereogram();
    if (mpCFrame->mpCPovRay->GetCameraGeometry() == _POV_FISHEYE)
	mpCFrame->mpCCanvas[0]->mpCGLCanvas->ModifyByDomeTilt();

    // -- Isosurface ȂǏKvȂ̂ŁAFilterPorcesss 
    
    mpCDF3SeqCtrlLink->CheckUpdate();
    for (int df3index = 0; df3index < _DF3_MAX_SLOT; df3index++){
	if (mpCDF3SeqCtrlLink->IsReprocessRequired(df3index) == TRUE){
	    mpCDF3SeqCtrlLink->
		ReleaseStepDF3(df3index, 
			       mpCTime->GetRemappedIndex(df3index),
			       _DF3SEQ_TEXTURE);
	    mpCDF3SeqCtrlLink->
		ClearTreeInfo(-1);
	}
    }
    mpCDF3SeqCtrlLink->ProcessXYZDataAllChannel();
    // --

    OutputPovRayMain(ofstreamPov, frame);
    ofstreamPov.close();

    mpCFrame->mpCCanvas[0]->mpCGLCanvas->mCCameraPos = CCamPosOrg;
    
    strFileOut = mpCFrame->mpCPovRay->mCConfWorkingDir.msBody + L"/" + strFile + L".ini";

    wxFile CFile(strFileOut, wxFile::write);
    if (CFile.IsOpened() == FALSE) return FALSE;

    CFile.Write(mpCFrame->mpCPovRay->GetPovStrIniFile
		(strFile, GCImgIO.mJPEGQuality), wxConvFile);
    
    return TRUE;
}

bool Main_cls::
OutputPovRayIOAnim(unsigned int *pInit, unsigned int *pFin, bool bQuery){

    if (IsPovrayCondistionSatisfiled() == FALSE) return FALSE;
    
    bool bSuccess = TRUE;
    bool bCachePause = mpCFrame->mpCPovRay->mCbPause.mbBody;
    mpCFrame->mpCPovRay->mCbPause.mbBody = FALSE;
   
    //unsigned int init, fin;
    switch (mpCFrame->mpCPovRay->mCRegionType.mBody){
    case _POV_RENDER_WORKAREA:
	*pInit = mpCTime->mCWAMin.mBody;
	*pFin  = mpCTime->mCWAMax.mBody;
	break;
    case _POV_RENDER_FULLSCENE:
	*pInit = 0;
	*pFin  = mpCTime->mCMaxFrame.mBody;
	break;
    case _POV_RENDER_CUSTOM: 
	if (mpCTime->OpenCustomRegionDialog(mpCFrame) == FALSE)
	    return FALSE;
	
	*pInit = mpCTime->mCustomRegionMin;
	*pFin  = mpCTime->mCustomRegionMax;

	break;
    }

    if (bQuery && (*pFin - *pInit) > 100)
	if (MsgBoxYesno(mpCFrame, _("Exporting many files."), _("You are goint to export many files.\nExecute it now?")) == FALSE)
	    return FALSE;  
    
    if (mpCFrame->mpCCanvas[0]->GetDisplayMode() < _GL_CANVAS_LINKED_CAM0) 
	if (MsgBoxYesno(mpCFrame, _("Output .pov file"), _("Panel 0 is not linked to camera")) == TRUE){
	    int camIndex = mpCFrame->mpCPovRay->GetCameraIndex();
	    if (camIndex != -2)
		mpCFrame->mpCCanvas[0]->mpCGLCanvas->LinkageConnect(camIndex);
	}
    
    if (mpCDF3SeqCtrlLink->DoesFilterExistInAllDF3(_FILTER_ISOSURFACE) == TRUE &&
	PovIsosurface_cls::msCOutputType.mBody == _ISO_PACK_IN_POV)
	if (MsgBoxYesno(mpCFrame, _("Output .pov file"), _("You are making .pov files with isosurface data.\nThese file may be very large ones.\nContinue?")) == FALSE)
	    return FALSE;

    wxBeginBusyCursor();
    mpCFrame->mpCPovRay->ExportPovAnim();
    mpCFrame->Disable();
    
    mpCFrame->mpCPovRay->mVstrAnimCommand.clear();
    mpCFrame->mpCPovRay->mVstrAnimPostSceneCommand.clear();
    
    int stereoEyeOrg = mpCFrame->mpCStereo->mCEye.mBody;
    for (int stereoIndex = 0; stereoIndex < 2; stereoIndex++){
	int stereo = mpCFrame->mpCPovRay->GetStereoId(stereoIndex);
	if (stereo == -1) continue;
	
	FilterCtrl_cls::msStereoTmpFileIOMode = 
	    mpCFrame->mpCPovRay->GetStereoTmpFileIOMode(stereoIndex);

	mpCFrame->mpCStereo->mCEye.mBody = stereo;
	for (unsigned int frame = *pInit; frame <= *pFin; frame++){
	    if (mpCFrame->mpCPovRay->mbExportingAnim == FALSE){
		wxEndBusyCursor();
		mpCFrame->Enable();
		return FALSE;
	    }

	    /*
	    while (wxGetApp().Pending())
		wxGetApp().Dispatch();
	    */
	    
	    //mpCDF3SeqCtrlLink->mbQuickShow = TRUE; QuickShowƉH
	    mpCTime->mFrame = frame;

	    mpCDF3SeqCtrlLink->ProcessXYZDataAllChannel();
	    
	    wchar_t filename[64];
	    swprintf(filename, 64,
		     mpCFrame->mpCPovRay->GetFileStr(_POV_ANIM, stereo).c_str(),
		     frame);
	    
	    if (OutputPovRayIO(filename, frame) == FALSE){
		mpCFrame->PushStatusTextMain(_("Disk I/O Error"));
		MsgBox(mpCFrame, _("File output failed. Please check the working directory and its permission."), _("Disk I/O Error")); 
		bSuccess = FALSE;
		frame = mpCTime->mCWAMax.mBody;
	    }

	    mpCFrame->mpCPovRay->PushBackCommands(stereo, frame);
	}
    }
    mpCFrame->mpCStereo->mCEye.mBody = stereoEyeOrg;
    FilterCtrl_cls::msStereoTmpFileIOMode = 3; // ftHg
    
    wxEndBusyCursor();
    mpCFrame->Enable();
    mpCFrame->Refresh(TRUE);
    mpCFrame->mpCPovRay->mCbPause.mbBody = bCachePause;

    if (bSuccess == TRUE){
	wxString strStatus;
	strStatus << _("Disk Output Suceed.") << wxT(" ") 
		  << mpCFrame->mpCPovRay->GetFileStr(_POV_ANIM, _STEREO_NONE) << wxT(".ini ") 
		  << mpCFrame->mpCPovRay->GetFileStr(_POV_ANIM, _STEREO_NONE) << wxT(".pov");
	mpCFrame->PushStatusTextMain(strStatus);
    }

    mpCFrame->GetStatusBar()->SetStatusText(wxT("field_2"), 2);
    return bSuccess;
}

void Main_cls::
ExportDF3Seq(){
    if (mpCFrame->mCLastExportedFileDF3.msBody != wxT(""))
	mCFPathEditDLG.mstrBuffer = mpCFrame->mCLastExportedFileDF3.msBody;
    else
	mCFPathEditDLG.mstrBuffer = wxT("");
    
    mCFPathEditDLG.mIntegerA = 0;    mCFPathEditDLG.mIntegerB = 0;
    mCFPathEditDLG.mbEnableA = TRUE; mCFPathEditDLG.mbEnableB = FALSE;
    mCFPathEditDLG.mbSave = TRUE;
    mCFPathEditDLG.SetDisable();
    mCFPathEditDLG.mpCMNumEditDialog->mstrHelp = _("Tooltip for Export DF3 Sequence");
    if (mCFPathEditDLG.OpenModalDialog(wxT("Exoprt DF3")) == FALSE) return;
 
    mpCFrame->mCLastExportedFileDF3.msBody = mCFPathEditDLG.mstrBuffer;
    FileSequenceWX_cls CFileSequence;
    CFileSequence.Set(mCFPathEditDLG.mstrBuffer);
    if (mCFPathEditDLG.mIntegerA != 0) 
	CFileSequence.mDigit = mCFPathEditDLG.mIntegerA;
    CFileSequence.mSeqLength = GetPDF3SeqActive()->mMaxIndex + 1;
    
    Filename_cls::DoubleSlash(CFileSequence.mwsFileHead);
    if (CFileSequence.mSeqLength == 0) return;

    bool bMirroring[3] = {FALSE, FALSE, FALSE};
    BoxCopyData_cls* pCBoxData =
	(BoxCopyData_cls*)mpCFrame->mpCDF3SeqCtrl->GetFirstPFilterDataInAllDF3(_FILTER_BOXCOPY);
    if (pCBoxData != NULL)
	LOOP_XYZ bMirroring[xyz] = pCBoxData->mCbValid.mbBody & pCBoxData->mCbMirroring.mpbBody[xyz];
    
    wxBeginBusyCursor();

    wxProgressDialog PCDialog(wxT("Exporting DF3"), wxT("Exporting..."), 100, mpCFrame, wxPD_APP_MODAL | wxPD_CAN_ABORT);
    
    int length = mpCTime->mCWAMax.mBody - mpCTime->mCWAMin.mBody + 1;
    for (int offset = mpCTime->mCWAMin.mBody; offset <= mpCTime->mCWAMax.mBody; offset++){
	double localDoubleIndex = mpCTime->GetDoubleIndex(offset);
	mpCDF3SeqCtrlLink->GetPDF3SeqActive()->mpCFilterCtrl->UpdateParameters(localDoubleIndex);
	int percent = 100 * (offset - mpCTime->mCWAMin.mBody) / length;
	wxString strMessage = wxString::Format(wxT("frame %d (%d%%)"), offset, percent);
	mpCFrame->GetStatusBar()->SetStatusText(strMessage, 2);
	if (mpCDF3SeqCtrlLink->GetPDF3SeqActive()->mVDF3GL[offset].ExportDF3(CFileSequence.GetModifiedWXStr(offset), bMirroring) == FALSE)
	    offset = CFileSequence.mSeqLength;
	if (PCDialog.Update(percent, strMessage) == FALSE)
	    offset = CFileSequence.mSeqLength;
    }
    mpCFrame->GetStatusBar()->SetStatusText(wxT("field_2"), 2);
    
    wxEndBusyCursor();
}

void Main_cls::
ExportImgSeq(){
    if (mpCFrame->mCLastExportedFileIMG.msBody != wxT(""))
	mCFPathEditDLG.mstrBuffer = mpCFrame->mCLastExportedFileIMG.msBody;
    else
	mCFPathEditDLG.mstrBuffer = wxT("");

    mCFPathEditDLG.mIntegerA = 0;    mCFPathEditDLG.mIntegerB = 0;
    mCFPathEditDLG.mbEnableA = TRUE; mCFPathEditDLG.mbEnableB = FALSE;
    mCFPathEditDLG.mbSave = TRUE;
    mCFPathEditDLG.SetDisable();
    mCFPathEditDLG.mpCMNumEditDialog->mstrHelp = _("Tooltip for Export IMG Sequence");
    if (mCFPathEditDLG.OpenModalDialog(wxT("Exoprt IMG")) == FALSE) return;

    mpCFrame->mCLastExportedFileIMG.msBody = mCFPathEditDLG.mstrBuffer;
    
    FileSequenceWX_cls CFileSequence;
    CFileSequence.Set(mCFPathEditDLG.mstrBuffer);
    if (mCFPathEditDLG.mIntegerA != 0) 
	CFileSequence.mDigit = mCFPathEditDLG.mIntegerA;
    FilenameWX_cls::DoubleSlash(CFileSequence.mwsFileHead);
    if (CFileSequence.mSeqLength == 0) return;

    wxBeginBusyCursor();
    mpCFrame->Disable();
    GetPDF3SeqActive()->GetPDF3()->ExportIMGSeq(CFileSequence, 
						mpCFrame->GetStatusBar());
    mpCFrame->GetStatusBar()->SetStatusText(wxT("field_2"), 2);
    mpCFrame->Enable();
    wxEndBusyCursor();
}

//----

void Main_cls::
ReadStrLine(std::wstring strLine){
    if (mCbToolTip.ReadStrLine(strLine) == TRUE)
	wxToolTip::Enable(mCbToolTip.mbBody);
}

wxString Main_cls::
Serialize(){
    return mCbToolTip.Serialize();
}
